// Binarize - Binarize (outputs either 0 or 255) an video plane
// Binarize (C) 2003 Kurosu (kurosu@inforezo.org)

// See legal notice in Interface.cpp and Readme.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

void BinUpper_MMX(unsigned char *srcp, const int src_pitch, const int row_size,
				  const int height, const unsigned char threshold)
{
	unsigned int thres = threshold;
	thres |= (thres << 8);
	thres |= (thres << 16);
	int modulo = src_pitch - row_size;
	int sizediv8 = row_size >> 3;
	__asm {
		mov			ebx, srcp
		mov			eax, [height]
		shr			eax, 1
		mov			edi, [modulo]
		mov			edx, [src_pitch]
		add			edi, edx
		mov			esi, [sizediv8]
		movd		mm7, [thres]
		movd		mm6, [thres]
		punpcklbw	mm7, mm6
		movq		mm6, mm7
		pxor		mm4, mm4
		pxor		mm5, mm5

		align 16

yloop :

		mov			ecx, esi

xloop : 

		movq		mm0, [ebx]
		 movq		 mm1, [ebx+edx]
		psubusb		mm0, mm6
		 psubusb	 mm1, mm7
		pcmpeqb		mm0, mm4
		 pcmpeqb	 mm1, mm5
		movq		[ebx], mm0
		 movq		 [ebx+edx], mm1

		add			ebx, 8
		dec			ecx
		jnz			xloop

		add			ebx, edi
		dec			eax
		jnz			yloop

		emms
	}
}

void BinUpper_C(unsigned char *srcpinit, const int src_pitch, const int row_size,
				const int height, const unsigned char threshold)
{
	unsigned char *srcp = srcpinit;
	int modulo = src_pitch - row_size;
	for (int y = 0; y < height; y++) 
	{
		for (int x = 0; x < row_size; x++)
		{
            if (*srcp > threshold) *srcp = 0;
			else *srcp = 255;
			srcp++;
		}
		srcp += modulo;
	}
}

void BinLower_MMX(unsigned char *srcp, const int src_pitch, const int row_size,
				  const int height, const unsigned char threshold)
{
	int  thres = threshold;
	thres |= (thres << 8);
	thres |= (thres << 16);
	int modulo = src_pitch - row_size;
	int sizediv8 = row_size >> 3;
	unsigned int complement = 0xffffffff;
	__asm {
		mov			ebx, srcp
		mov			eax, [height]
		shr			eax, 1
		mov			edi, [modulo]
		mov			edx, [src_pitch]
		add			edi, edx
		mov			esi, [sizediv8]
		movd		mm7, [thres]
		movd		mm6, [thres]
		punpcklbw	mm7, mm6
		movd		mm5, [complement]
		movd		mm4, [complement]
		punpcklbw	mm5, mm4
		movq		mm6, mm7
		movq		mm4, mm5
		pxor		mm3, mm3
		pxor		mm2, mm2

		align 16

yloop :

		mov			ecx, esi

xloop : 

		movq		mm0, [ebx]
		 movq		 mm1, [ebx+edx]
		psubusb		mm0, mm6
		 psubusb	 mm1, mm7
		pcmpeqb		mm0, mm2
		 pcmpeqb	 mm1, mm3
		pxor		mm0, mm4
		 pxor		 mm1, mm5
		movq		[ebx], mm0
		 movq		 [ebx+edx], mm1

		add			ebx, 8
		dec			ecx
		jnz			xloop

		add			ebx, edi
		dec			eax
		jnz			yloop

		emms
	}
}

void BinLower_C(unsigned char *srcpinit, const int src_pitch, const int row_size,
				const int height, const unsigned char threshold)
{
	unsigned char *srcp = srcpinit;
	int modulo = src_pitch - row_size;
	for (int y = 0; y < height; y++) 
	{
		for (int x = 0; x < row_size; x++)
		{
            if (*srcp <= threshold) *srcp = 0;
			else *srcp = 255;
			srcp++;
		}
		srcp += modulo;
	}
}