// Expand - set of tools for region growth of at best one pixel
// Expand (C) 2003 Kurosu (kurosu@inforezo.org)

// This work can also be referred to by the word AreaModifier

// See legal notice in Interface.cpp and Readme.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

void Expand_C(const unsigned char *srcpinit, unsigned char *dstpinit,
			const int src_pitch, const int dst_pitch,
			int row_size, const int height)
{
	const unsigned char *srcpn, *srcpp;
	int x, y;
	int max, temp;
	unsigned char *dstp = dstpinit;
	const unsigned char *srcp = srcpinit;

	srcpp = srcp;
	srcp += src_pitch;
	srcpn = srcp + src_pitch;

	//Top left corner
	max = srcpp[0];
	temp = srcpp[1];
	if (temp > max) max = temp;
	temp = srcp[0];
	if (temp > max) max = temp;
	temp = srcp[1];
	if (temp > max) max = temp;
	dstp[0] = max;

	//Topline
	for (x = 1; x < row_size-1; x++)
	{
		max = srcpp[x-1];
		temp = srcpp[x];
		if (temp > max) max = temp;
		temp = srcpp[x+1];
		if (temp > max) max = temp;

		temp = srcp[x-1];
		if (temp > max) max = temp;
		temp = srcp[x];
		if (temp > max) max = temp;
		temp = srcp[x+1];
		if (temp > max) max = temp;

		dstp[x] = max;
	}		

	//Top right corner
	max = srcpp[row_size-2];
	temp = srcpp[row_size-1];
	if (temp > max) max = temp;
	temp = srcp[row_size-2];
	if (temp > max) max = temp;
	temp = srcp[row_size-1];
	if (temp > max) max = temp;
	dstp[row_size-1] = max;

	dstp += dst_pitch;
	for (y = 1; y < height-1; y++) 
	{
		//Line's 1st pixel
		max = srcpp[0];
		temp = srcpp[1];
		if (temp > max) max = temp;

		temp = srcp[0];
		if (temp > max) max = temp;
		temp = srcp[1];
		if (temp > max) max = temp;

		temp = srcpn[0];
		if (temp > max) max = temp;
		temp = srcpn[1];
		if (temp > max) max = temp;
		dstp[0] = max;

		//Line's pixels
		for (x = 1; x < row_size-1; x++)
		{
			max = srcpp[x-1];
			temp = srcpp[x];
			if (temp > max) max = temp;
			temp = srcpp[x+1];
			if (temp > max) max = temp;

			temp = srcp[x-1];
			if (temp > max) max = temp;
			temp = srcp[x];
			if (temp > max) max = temp;
			temp = srcp[x+1];
			if (temp > max) max = temp;

			temp = srcpn[x-1];
			if (temp > max) max = temp;
			temp = srcpn[x];
			if (temp > max) max = temp;
			temp = srcpn[x+1];
			if (temp > max) max = temp;

			dstp[x] = max;
		}

		//Last line's pixel
		max = srcpp[row_size-2];
		temp = srcpp[row_size-1];
		if (temp > max) max = temp;

		temp = srcp[row_size-2];
		if (temp > max) max = temp;
		temp = srcp[row_size-1];
		if (temp > max) max = temp;

		temp = srcpn[row_size-2];
		if (temp > max) max = temp;
		temp = srcpn[row_size-1];
		if (temp > max) max = temp;
		dstp[row_size-1] = max;

		dstp += dst_pitch;
		srcpp = srcp;
		srcp = srcpn;
		srcpn += src_pitch;
	}

	//Bottom left corner
	max = srcpp[0];
	temp = srcpp[1];
	if (temp > max) max = temp;
	temp = srcp[0];
	if (temp > max) max = temp;
	temp = srcp[1];
	if (temp > max) max = temp;
	dstp[0] = max;

	//Bottom line
	for (x = 1; x < row_size-1; x++)
	{
		max = srcpp[x-1];
		temp = srcpp[x];
		if (temp > max) max = temp;
		temp = srcpp[x+1];
		if (temp > max) max = temp;

		temp = srcp[x-1];
		if (temp > max) max = temp;
		temp = srcp[x];
		if (temp > max) max = temp;
		temp = srcp[x+1];
		if (temp > max) max = temp;

		dstp[x] = max;
	}		

	//Bottom right corner
	max = srcpp[row_size-2];
	temp = srcpp[row_size-1];
	if (temp > max) max = temp;
	temp = srcp[row_size-2];
	if (temp > max) max = temp;
	temp = srcp[row_size-1];
	if (temp > max) max = temp;
	dstp[row_size-1] = max;
}

void Expand_MMX(const unsigned char *srcpinit, unsigned char *dstpinit,
			const int src_pitch, const int dst_pitch,
			int row_size, const int height)
{
	static const __int64	zero = 0x0000000000000000i64;
	const unsigned char		*srcpp = srcpinit;
	unsigned char			max, temp;
	unsigned char *dstp = dstpinit;
	const unsigned char *srcp = srcpinit;

	srcp += src_pitch;

	row_size -= 2;

	//Top left corner
	max = srcpp[0];
	temp = srcpp[1];
	if (temp > max) max = temp;
	temp = srcp[0];
	if (temp > max) max = temp;
	temp = srcp[1];
	if (temp > max) max = temp;
	dstp[0] = max;

	//Top line
	__asm
	{
		mov			eax,[srcp]
		mov			edx,[dstp]
		mov			ebx,eax
		add			edx,[dst_pitch]
		add			ebx,[src_pitch]
		mov			edi,0
		pxor		mm7,mm7
looptop:
		pxor		mm5,mm5
		pxor		mm6,mm6

		//C1
			//L1
		movd		mm0,[eax+edi]
		movq		mm6,mm0
		punpckhbw	mm6,mm7

			//L2
		movd		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4

		//C2
		inc			edi
			//L1
		movq		mm0,[eax+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4
			//L2
		movq		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4

		//C3
		inc			edi
			//L1
		movq		mm0,[eax+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4
			//L2
		movq		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4

		//Move back to C2 and store result
		dec			edi
		packuswb	mm6,mm7
		movd		[edx+edi],mm6

		add			edi,3
		cmp			edi,[row_size]
		jl			looptop
	}

	//Top right corner
	max = srcpp[row_size-2];
	temp = srcpp[row_size-1];
	if (temp > max) max = temp;
	temp = srcp[row_size-2];
	if (temp > max) max = temp;
	temp = srcp[row_size-1];
	if (temp > max) max = temp;
	dstp[row_size-1] = max;

	__asm
	{
		mov			eax,[srcpp]
		mov			edx,[dstp]
		mov			ebx,[srcp]
		add			edx,[dst_pitch]
		mov			esi,[height]
		mov			ecx,ebx		
		mov			edi,0
		add			ecx,[src_pitch]
		sub			esi,2
		pxor		mm7,mm7
		mov			[edx],di			//set to 0 1st pixel by default
loophere:
		//C1
			//L1
		movd		mm0,[eax+edi]
		movq		mm6,mm0
		punpckhbw	mm6,mm7

			//L2
		movd		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4
			//L3
		movq		mm0,[ecx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4

		//C2
		inc			edi
			//L1
		movq		mm0,[eax+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4
			//L2
		movq		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4
			//L3
		movq		mm0,[ecx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4

		//C3
		inc			edi
			//L1
		movq		mm0,[eax+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4
			//L2
		movq		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4
			//L3
		movq		mm0,[ecx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4

		//Move back to C2 and store result
		dec			edi
		packuswb	mm6,mm7
		movd		[edx+edi],mm6

		add			edi,3
		cmp			edi,[row_size]
		jl			loophere

		mov			edi,0
		add			edx,row_size
		mov			[edx+1],di
		mov			eax,ebx
		mov			ebx,ecx
		sub			edx,row_size
		add			ecx,[src_pitch]
		add			edx,[dst_pitch]

		dec			esi
		jnz			loophere
	}
/*
	srcp += height*src_pitch;
	dstp += height*dst_pitch;
	srcpp -= src_pitch;
*/
	//Bottom left corner
	max = srcpp[0];
	temp = srcpp[1];
	if (temp > max) max = temp;
	temp = srcp[0];
	if (temp > max) max = temp;
	temp = srcp[1];
	if (temp > max) max = temp;
	dstp[0] = max;

	//Bottom line
	__asm {
		mov			eax,[srcpp]
		mov			edx,[dstp]
		mov			ebx,[srcp]
		mov			edi,0
		pxor		mm7,mm7
loopbot:
		pxor		mm5,mm5
		pxor		mm6,mm6

		//C1
			//L1
		movd		mm0,[eax+edi]
		movq		mm6,mm0
		punpckhbw	mm6,mm7

			//L2
		movd		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4

		//C2
		inc			edi
			//L1
		movq		mm0,[eax+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4
			//L2
		movq		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4

		//C3
		inc			edi
			//L1
		movq		mm0,[eax+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4
			//L2
		movq		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4

		//Move back to C2 and store result
		dec			edi
		packuswb	mm6,mm7
		movd		[edx+edi],mm6

		add			edi,3
		cmp			edi,[row_size]
		jl			loopbot

		emms
	}

	//Bottom right corner
	max = srcpp[row_size-2];
	temp = srcpp[row_size-1];
	if (temp > max) max = temp;
	temp = srcp[row_size-2];
	if (temp > max) max = temp;
	temp = srcp[row_size-1];
	if (temp > max) max = temp;
	dstp[row_size-1] = max;

}

void Inpand_C(const unsigned char *srcpinit, unsigned char *dstpinit,
			const int src_pitch, const int dst_pitch,
			int row_size, const int height)
{
	const unsigned char *srcpn, *srcpp;
	int x, y;
	unsigned char min, temp;
	unsigned char *dstp = dstpinit;
	const unsigned char *srcp = srcpinit;

	srcpp = srcp;
	srcp += src_pitch;
	srcpn = srcp + src_pitch;

	//Top left corner
	min = srcpp[0];
	temp = srcpp[1];
	if (temp < min) min = temp;
	temp = srcp[0];
	if (temp < min) min = temp;
	temp = srcp[1];
	if (temp < min) min = temp;
	dstp[0] = min;

	//Topline
	for (x = 1; x < row_size-1; x++)
	{
		min = srcpp[x-1];
		temp = srcpp[x];
		if (temp < min) min = temp;
		temp = srcpp[x+1];
		if (temp < min) min = temp;

		temp = srcp[x-1];
		if (temp < min) min = temp;
		temp = srcp[x];
		if (temp < min) min = temp;
		temp = srcp[x+1];
		if (temp < min) min = temp;

		dstp[x] = min;
	}		

	//Top right corner
	min = srcpp[row_size-2];
	temp = srcpp[row_size-1];
	if (temp < min) min = temp;
	temp = srcp[row_size-2];
	if (temp < min) min = temp;
	temp = srcp[row_size-1];
	if (temp < min) min = temp;
	dstp[row_size-1] = min;

	dstp += dst_pitch;
	for (y = 1; y < height-1; y++) 
	{
		//Line's 1st pixel
		min = srcpp[0];
		temp = srcpp[1];
		if (temp < min) min = temp;

		temp = srcp[0];
		if (temp < min) min = temp;
		temp = srcp[1];
		if (temp < min) min = temp;

		temp = srcpn[0];
		if (temp < min) min = temp;
		temp = srcpn[1];
		if (temp < min) min = temp;
		dstp[0] = min;

		//Line's pixels
		for (x = 1; x < row_size-1; x++)
		{
			min = srcpp[x-1];
			temp = srcpp[x];
			if (temp < min) min = temp;
			temp = srcpp[x+1];
			if (temp < min) min = temp;

			temp = srcp[x-1];
			if (temp < min) min = temp;
			temp = srcp[x];
			if (temp < min) min = temp;
			temp = srcp[x+1];
			if (temp < min) min = temp;

			temp = srcpn[x-1];
			if (temp < min) min = temp;
			temp = srcpn[x];
			if (temp < min) min = temp;
			temp = srcpn[x+1];
			if (temp < min) min = temp;

			dstp[x] = min;
		}

		//Last line's pixel
		min = srcpp[row_size-2];
		temp = srcpp[row_size-1];
		if (temp < min) min = temp;

		temp = srcp[row_size-2];
		if (temp < min) min = temp;
		temp = srcp[row_size-1];
		if (temp < min) min = temp;

		temp = srcpn[row_size-2];
		if (temp < min) min = temp;
		temp = srcpn[row_size-1];
		if (temp < min) min = temp;
		dstp[row_size-1] = min;

		dstp += dst_pitch;
		srcpp = srcp;
		srcp = srcpn;
		srcpn += src_pitch;
	}

	//Bottom left corner
	min = srcpp[0];
	temp = srcpp[1];
	if (temp < min) min = temp;
	temp = srcp[0];
	if (temp < min) min = temp;
	temp = srcp[1];
	if (temp < min) min = temp;
	dstp[0] = min;

	//Bottom line
	for (x = 1; x < row_size-1; x++)
	{
		min = srcpp[x-1];
		temp = srcpp[x];
		if (temp < min) min = temp;
		temp = srcpp[x+1];
		if (temp < min) min = temp;

		temp = srcp[x-1];
		if (temp < min) min = temp;
		temp = srcp[x];
		if (temp < min) min = temp;
		temp = srcp[x+1];
		if (temp < min) min = temp;

		dstp[x] = min;
	}		

	//Bottom right corner
	min = srcpp[row_size-2];
	temp = srcpp[row_size-1];
	if (temp < min) min = temp;
	temp = srcp[row_size-2];
	if (temp < min) min = temp;
	temp = srcp[row_size-1];
	if (temp < min) min = temp;
	dstp[row_size-1] = min;
}

void Inpand_MMX(const unsigned char *srcpinit, unsigned char *dstpinit,
			const int src_pitch, const int dst_pitch,
			int row_size, const int height)
{
	static const __int64	zero = 0x0000000000000000i64;
	const unsigned char		*srcpp = srcpinit;
	unsigned char			min, temp;
	unsigned char *dstp = dstpinit;
	const unsigned char *srcp = srcpinit;

	srcp += src_pitch;

	row_size -= 2;

	//Top left corner
	min = srcpp[0];
	temp = srcpp[1];
	if (temp < min) min = temp;
	temp = srcp[0];
	if (temp < min) min = temp;
	temp = srcp[1];
	if (temp < min) min = temp;
	dstp[0] = min;

	//Top line
	__asm
	{
		mov			eax,[srcp]
		mov			edx,[dstp]
		mov			ebx,eax
		add			edx,[dst_pitch]
		add			ebx,[src_pitch]
		mov			edi,0
		pxor		mm7,mm7
looptop:
		//C1
			//L1
		movd		mm0,[eax+edi]
		movq		mm6,mm0
		punpckhbw	mm6,mm7

			//L2
		movd		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm3,mm6
		movq		mm4,mm5
		pcmpgtw		mm3,mm4
		pand		mm5,mm3
		pandn		mm3,mm6
		por			mm3,mm5
		movq		mm6,mm3

		//C2
		inc			edi
			//L1
		movq		mm0,[eax+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm3,mm6
		movq		mm4,mm5
		pcmpgtw		mm3,mm4
		pand		mm5,mm3
		pandn		mm3,mm6
		por			mm3,mm5
		movq		mm6,mm3
			//L2
		movq		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm3,mm6
		movq		mm4,mm5
		pcmpgtw		mm3,mm4
		pand		mm5,mm3
		pandn		mm3,mm6
		por			mm3,mm5
		movq		mm6,mm3

		//C3
		inc			edi
			//L1
		movq		mm0,[eax+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm3,mm6
		movq		mm4,mm5
		pcmpgtw		mm3,mm4
		pand		mm5,mm3
		pandn		mm3,mm6
		por			mm3,mm5
		movq		mm6,mm3
			//L2
		movq		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm3,mm6
		movq		mm4,mm5
		pcmpgtw		mm3,mm4
		pand		mm5,mm3
		pandn		mm3,mm6
		por			mm3,mm5
		movq		mm6,mm3

		//Move back to C2 and store result
		dec			edi
		packuswb	mm6,mm7
		movd		[edx+edi],mm6

		add			edi,3
		cmp			edi,[row_size]
		jl			looptop
	}

	//Top right corner
	min = srcpp[row_size-2];
	temp = srcpp[row_size-1];
	if (temp < min) min = temp;
	temp = srcp[row_size-2];
	if (temp < min) min = temp;
	temp = srcp[row_size-1];
	if (temp < min) min = temp;
	dstp[row_size-1] = min;

	__asm
	{
		mov			eax,[srcpp]
		mov			edx,[dstp]
		mov			ebx,[srcp]
		add			edx,[dst_pitch]
		mov			esi,[height]
		mov			ecx,ebx		
		mov			edi,0
		add			ecx,[src_pitch]
		sub			esi,2
		pxor		mm7,mm7
		mov			[edx],di			//set to 0 1st pixel by default
loophere:
		//C1
			//L1
		movd		mm0,[eax+edi]
		movq		mm6,mm0
		punpckhbw	mm6,mm7

			//L2
		movd		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm3,mm6
		movq		mm4,mm5
		pcmpgtw		mm3,mm4
		pand		mm5,mm3
		pandn		mm3,mm6
		por			mm3,mm5
		movq		mm6,mm3
			//L3
		movq		mm0,[ecx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm3,mm6
		movq		mm4,mm5
		pcmpgtw		mm3,mm4
		pand		mm5,mm3
		pandn		mm3,mm6
		por			mm3,mm5
		movq		mm6,mm3

		//C2
		inc			edi
			//L1
		movq		mm0,[eax+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm3,mm6
		movq		mm4,mm5
		pcmpgtw		mm3,mm4
		pand		mm5,mm3
		pandn		mm3,mm6
		por			mm3,mm5
		movq		mm6,mm3
			//L2
		movq		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm3,mm6
		movq		mm4,mm5
		pcmpgtw		mm3,mm4
		pand		mm5,mm3
		pandn		mm3,mm6
		por			mm3,mm5
		movq		mm6,mm3
			//L3
		movq		mm0,[ecx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm3,mm6
		movq		mm4,mm5
		pcmpgtw		mm3,mm4
		pand		mm5,mm3
		pandn		mm3,mm6
		por			mm3,mm5
		movq		mm6,mm3

		//C3
		inc			edi
			//L1
		movq		mm0,[eax+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm3,mm6
		movq		mm4,mm5
		pcmpgtw		mm3,mm4
		pand		mm5,mm3
		pandn		mm3,mm6
		por			mm3,mm5
		movq		mm6,mm3
			//L2
		movq		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm4,mm5
		pcmpgtw		mm4,mm6
		pand		mm5,mm4
		pandn		mm4,mm6
		por			mm4,mm5
		movq		mm6,mm4
			//L3
		movq		mm0,[ecx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm3,mm6
		movq		mm4,mm5
		pcmpgtw		mm3,mm4
		pand		mm5,mm3
		pandn		mm3,mm6
		por			mm3,mm5
		movq		mm6,mm3

		//Move back to C2 and store result
		dec			edi
		packuswb	mm6,mm7
		movd		[edx+edi],mm6

		add			edi,3
		cmp			edi,[row_size]
		jl			loophere

		mov			edi,0
		add			edx,row_size
		mov			[edx+1],di
		mov			eax,ebx
		mov			ebx,ecx
		sub			edx,row_size
		add			ecx,[src_pitch]
		add			edx,[dst_pitch]

		dec			esi
		jnz			loophere
	}
/*
	srcp += height*src_pitch;
	dstp += height*dst_pitch;
	srcpp -= src_pitch;
*/
	//Bottom left corner
	min = srcpp[0];
	temp = srcpp[1];
	if (temp < min) min = temp;
	temp = srcp[0];
	if (temp < min) min = temp;
	temp = srcp[1];
	if (temp < min) min = temp;
	dstp[0] = min;

	//Bottom line
	__asm {
		mov			eax,[srcpp]
		mov			edx,[dstp]
		mov			ebx,[srcp]
		mov			edi,0
		pxor		mm7,mm7
loopbot:
		pxor		mm5,mm5
		pxor		mm6,mm6

		//C1
			//L1
		movd		mm0,[eax+edi]
		movq		mm6,mm0
		punpckhbw	mm6,mm7

			//L2
		movd		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm3,mm6
		movq		mm4,mm5
		pcmpgtw		mm3,mm4
		pand		mm5,mm3
		pandn		mm3,mm6
		por			mm3,mm5
		movq		mm6,mm3

		//C2
		inc			edi
			//L1
		movq		mm0,[eax+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm3,mm6
		movq		mm4,mm5
		pcmpgtw		mm3,mm4
		pand		mm5,mm3
		pandn		mm3,mm6
		por			mm3,mm5
		movq		mm6,mm3
			//L2
		movq		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm3,mm6
		movq		mm4,mm5
		pcmpgtw		mm3,mm4
		pand		mm5,mm3
		pandn		mm3,mm6
		por			mm3,mm5
		movq		mm6,mm3

		//C3
		inc			edi
			//L1
		movq		mm0,[eax+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm3,mm6
		movq		mm4,mm5
		pcmpgtw		mm3,mm4
		pand		mm5,mm3
		pandn		mm3,mm6
		por			mm3,mm5
		movq		mm6,mm3
			//L2
		movq		mm0,[ebx+edi]
		movq		mm5,mm0
		punpckhbw	mm5,mm7
		movq		mm3,mm6
		movq		mm4,mm5
		pcmpgtw		mm3,mm4
		pand		mm5,mm3
		pandn		mm3,mm6
		por			mm3,mm5
		movq		mm6,mm3

		//Move back to C2 and store result
		dec			edi
		packuswb	mm6,mm7
		movd		[edx+edi],mm6

		add			edi,3
		cmp			edi,[row_size]
		jl			loopbot

		emms
	}

	//Bottom right corner
	min = srcpp[row_size-2];
	temp = srcpp[row_size-1];
	if (temp < min) min = temp;
	temp = srcp[row_size-2];
	if (temp < min) min = temp;
	temp = srcp[row_size-1];
	if (temp < min) min = temp;
	dstp[row_size-1] = min;

}

void Inflate_C(const unsigned char *srcpinit, unsigned char *dstpinit,
			const int src_pitch, const int dst_pitch,
			int row_size, const int height)
{
	const unsigned char *srcpn, *srcpp;
	int x, y;
	unsigned char temp;
	unsigned char *dstp = dstpinit;
	const unsigned char *srcp = srcpinit;

	srcpp = srcp;
	srcp += src_pitch;
	srcpn = srcp + src_pitch;

	temp = (srcpp[1] + srcp[0] + srcp[1]) / 3;
	if (temp > srcpp[0]) dstp[0] = temp;
	else dstp[0] = srcpp[0];

	for ( x = 1; x < row_size-1; x++)
	{
		temp = (srcpp[x-1] + srcpp[x+1] + srcp[x-1] + srcp[x+1] + srcp[x]) / 5;
		if (temp > srcpp[x]) dstp[x] = temp;
		else dstp[x] = srcpp[x];
	}

	temp = (srcpp[row_size-2] + srcp[row_size-2] + srcp[row_size-1]) / 3;
	if (temp > srcpp[row_size-1]) dstp[row_size-1] = temp;
	else dstp[row_size-1] = srcpp[row_size-1];

	dstp += dst_pitch;

	for (y = 1; y < height-1; y++) 
	{
		temp = (srcpp[0] + srcpp[1] + srcp[0] + srcpn[0] + srcpn[1])/5;
		if (temp>srcp[0]) dstp[0] = temp;
		else dstp[0] = srcp[0];

		for (x = 1; x < row_size-1; x++)
		{
			temp = (srcpp[x-1] + srcpp[x] + srcpp[x+1] +
					srcp[x-1] +				srcp[x+1] +
					srcpn[x-1] + srcpn[x] + srcpn[x+1])>>3;
			if (temp>srcp[x]) dstp[x] = temp;
			else dstp[x] = srcp[x];
		}

		temp = (srcpp[row_size-2] + srcpp[row_size-1] + srcp[row_size-2] +
				srcpn[row_size-2] + srcpn[row_size-1])/5;
		if (temp>srcp[row_size-1]) dstp[row_size-1] = temp;
		else dstp[row_size-1] = srcp[row_size-1];

		dstp += dst_pitch;
		srcpp = srcp;
		srcp = srcpn;
		srcpn += src_pitch;
	}

	temp = (srcpp[1] + srcp[0] + srcp[1]) / 3;
	if (temp > srcpp[0]) dstp[0] = temp;
	else dstp[0] = srcpp[0];

	for ( x = 1; x < row_size-1; x++)
	{
		temp = (srcpp[x-1] + srcpp[x+1] + srcp[x-1] + srcp[x+1] + srcp[x]) / 5;
		if (temp > srcpp[x]) dstp[x] = temp;
		else dstp[x] = srcpp[x];
	}

	temp = (srcpp[row_size-2] + srcp[row_size-2] + srcp[row_size-1]) / 3;
	if (temp > srcpp[row_size-1]) dstp[row_size-1] = temp;
	else dstp[row_size-1] = srcpp[row_size-1];

}

void Inflate_MMX(const unsigned char *srcpinit, unsigned char *dstpinit,
			const int src_pitch, const int dst_pitch,
			int row_size, const int height)
{
	unsigned char *dstp = dstpinit;
	const unsigned char *srcp = srcpinit;
	row_size -= 2;
	__asm
	{
		mov			eax,[srcp]
		mov			edx,[dstp]
		mov			ebx,eax
		add			edx,[dst_pitch]
		add			ebx,[src_pitch]
		mov			esi,[height]
		mov			ecx,ebx		
		mov			edi,0
		add			ecx,[src_pitch]
		sub			esi,2
		pxor		mm7,mm7
		mov			[edx],di			//set to 0 1st pixel by default
loophere:

		movd		mm6,[eax+edi]
		movd		mm5,[ebx+edi]
		punpcklbw	mm6,mm7
		punpcklbw	mm5,mm7
		movd		mm4,[ecx+edi]
		paddusw		mm6,mm5
		punpcklbw	mm4,mm7
		inc			edi
		paddusw		mm6,mm4

		movd		mm5,[ebx+edi]
		movd		mm0,[ecx+edi]
		punpcklbw	mm5,mm7
		punpcklbw	mm0,mm7
		paddusw		mm6,mm5
		movd		mm3,[ecx+edi]
		inc			edi
		paddusw		mm6,mm0
		punpcklbw	mm3,mm7

		movd		mm5,[ebx+edi]
		movd		mm4,[ecx+edi]
		paddusw		mm6,mm3
		punpcklbw	mm5,mm7
		punpcklbw	mm4,mm7
		paddusw		mm6,mm5
		movd		mm3,[ecx+edi]
		paddusw		mm6,mm4
		punpcklbw	mm3,mm7
		movq		mm1,mm0
		paddusw		mm6,mm3

		dec			edi
		psrlw		mm6,3
		movq		mm5,mm6
		pcmpgtw		mm6,mm1
		pand		mm5,mm6
		pandn		mm6,mm0
		por			mm6,mm5

		packuswb	mm6,mm7
		movd		[edx+edi],mm6

		add			edi,3
		cmp			edi,[row_size]
		jl			loophere

		mov			edi,0
		mov			eax,ebx
		mov			ebx,ecx
		add			edx,[dst_pitch]
		add			ecx,[src_pitch]

		dec			esi
		jnz			loophere

		emms
	}
}

void Deflate_C(const unsigned char *srcpinit, unsigned char *dstpinit,
			const int src_pitch, const int dst_pitch,
			int row_size, const int height)
{
	const unsigned char *srcpn, *srcpp;
	int x, y;
	unsigned char temp;
	unsigned char *dstp = dstpinit;
	const unsigned char *srcp = srcpinit;

	srcpp = srcp;
	srcp += src_pitch;
	srcpn = srcp + src_pitch;

	temp = (srcpp[1] + srcp[0] + srcp[1]) / 3;
	if (temp < srcpp[0]) dstp[0] = temp;
	else dstp[0] = srcpp[0];

	for ( x = 1; x < row_size-1; x++)
	{
		temp = (srcpp[x-1] + srcpp[x+1] + srcp[x-1] + srcp[x+1] + srcp[x]) / 5;
		if (temp < srcpp[x]) dstp[x] = temp;
		else dstp[x] = srcpp[x];
	}

	temp = (srcpp[row_size-2] + srcp[row_size-2] + srcp[row_size-1]) / 3;
	if (temp < srcpp[row_size-1]) dstp[row_size-1] = temp;
	else dstp[row_size-1] = srcpp[row_size-1];

	dstp += dst_pitch;

	for (y = 1; y < height-1; y++) 
	{
		temp = (srcpp[0] + srcpp[1] + srcp[0] + srcpn[0] + srcpn[1])/5;
		if (temp<srcp[0]) dstp[0] = temp;
		else dstp[0] = srcp[0];

		for (x = 1; x < row_size-1; x++)
		{
			temp = (srcpp[x-1] + srcpp[x] + srcpp[x+1] +
					srcp[x-1] +				srcp[x+1] +
					srcpn[x-1] + srcpn[x] + srcpn[x+1])>>3;
			if (temp<srcp[x]) dstp[x] = temp;
			else dstp[x] = srcp[x];
		}

		temp = (srcpp[row_size-2] + srcpp[row_size-1] + srcp[row_size-2] +
				srcpn[row_size-2] + srcpn[row_size-1])/5;
		if (temp<srcp[row_size-1]) dstp[row_size-1] = temp;
		else dstp[row_size-1] = srcp[row_size-1];

		dstp += dst_pitch;
		srcpp = srcp;
		srcp = srcpn;
		srcpn += src_pitch;
	}

	temp = (srcpp[1] + srcp[0] + srcp[1]) / 3;
	if (temp < srcpp[0]) dstp[0] = temp;
	else dstp[0] = srcpp[0];

	for ( x = 1; x < row_size-1; x++)
	{
		temp = (srcpp[x-1] + srcpp[x+1] + srcp[x-1] + srcp[x+1] + srcp[x]) / 5;
		if (temp < srcpp[x]) dstp[x] = temp;
		else dstp[x] = srcpp[x];
	}

	temp = (srcpp[row_size-2] + srcp[row_size-2] + srcp[row_size-1]) / 3;
	if (temp < srcpp[row_size-1]) dstp[row_size-1] = temp;
	else dstp[row_size-1] = srcpp[row_size-1];

}

void Deflate_MMX(const unsigned char *srcpinit, unsigned char *dstpinit,
			const int src_pitch, const int dst_pitch,
			int row_size, const int height)
{
	unsigned char *dstp = dstpinit;
	const unsigned char *srcp = srcpinit;
	row_size -= 2;
	__asm
	{
		mov			eax,[srcp]
		mov			edx,[dstp]
		mov			ebx,eax
		add			edx,[dst_pitch]
		add			ebx,[src_pitch]
		mov			esi,[height]
		mov			ecx,ebx		
		mov			edi,0
		add			ecx,[src_pitch]
		sub			esi,2
		pxor		mm7,mm7
		mov			[edx],di			//set to 0 1st pixel by default
loophere:

		movd		mm6,[eax+edi]
		movd		mm5,[ebx+edi]
		punpcklbw	mm6,mm7
		punpcklbw	mm5,mm7
		movd		mm4,[ecx+edi]
		paddusw		mm6,mm5
		punpcklbw	mm4,mm7
		inc			edi
		paddusw		mm6,mm4

		movd		mm5,[ebx+edi]
		movd		mm0,[ecx+edi]
		punpcklbw	mm5,mm7
		punpcklbw	mm0,mm7
		paddusw		mm6,mm5
		movd		mm3,[ecx+edi]
		inc			edi
		paddusw		mm6,mm0
		punpcklbw	mm3,mm7

		movd		mm5,[ebx+edi]
		movd		mm4,[ecx+edi]
		paddusw		mm6,mm3
		punpcklbw	mm5,mm7
		punpcklbw	mm4,mm7
		paddusw		mm6,mm5
		movd		mm3,[ecx+edi]
		paddusw		mm6,mm4
		punpcklbw	mm3,mm7
		movq		mm1,mm0
		paddusw		mm6,mm3

		dec			edi
		psrlw		mm6,3
		pcmpgtw		mm1,mm6
		pand		mm6,mm1
		pandn		mm1,mm0
		por			mm1,mm6

		packuswb	mm1,mm7
		movd		[edx+edi],mm1

		add			edi,3
		cmp			edi,[row_size]
		jl			loophere

		mov			edi,0
		mov			eax,ebx
		mov			ebx,ecx
		add			edx,[dst_pitch]
		add			ecx,[src_pitch]

		dec			esi
		jnz			loophere

		emms
	}
}