// Logic - very basic mask logic manipulations
// Logic (C) 2003 Kurosu (kurosu@inforezo.org)

// See legal notice in Interface.cpp and Readme.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include <memory.h>

void And_C4(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	unsigned char *src1pn = src1p + src1_pitch;
	const unsigned char *src2pn = src2p + src2_pitch;
	for (int y = 0; y < height; y += 2)
	{
		for (int x = 0; x < row_size; x += 4)
		{
			*(unsigned int *)(src1p+x) &= *(unsigned int *)(src2p+x);
			*(unsigned int *)(src1pn+x) &= *(unsigned int *)(src2pn+x);
		}
		src2p += src2_pitch << 1;
		src1p += src1_pitch << 1;
		src2pn += src2_pitch << 1;
		src1pn += src1_pitch << 1;
	}
}

void And_C1(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	for (int y = 0; y < height; y ++)
	{
		for (int x = 0; x < row_size; x ++)
			*(src1p+x) &= *(src2p+x);
		src2p += src2_pitch;
		src1p += src1_pitch;
	}
}


void And_MMX(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	__asm {
		mov			eax,[src1p]
		mov			ecx,[src2p]
		mov			ebx,eax
		mov			edx,ecx
		mov			esi,[height]
		add			ebx,[src1_pitch]
		mov			edi,0
		add			edx,[src2_pitch]
		shl			[src1_pitch],1
		shl			[src2_pitch],1
		shr			esi, 1
loopx:
		movq		mm0,[eax+edi]
		movq		mm1,[ebx+edi]
		pand		mm0,[ecx+edi]
		pand		mm1,[edx+edi]
		movq		[eax+edi],mm0
		movq		[ebx+edi],mm1

		add			edi,8
		cmp			edi,[row_size]
		jl			loopx

		mov			edi,0
		add			eax,src1_pitch
		add			ecx,src2_pitch
		add			ebx,src1_pitch
		add			edx,src2_pitch
		dec			esi

		jnz			loopx

		emms
	}
}

void Or_C4(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	unsigned char *src1pn = src1p + src1_pitch;
	const unsigned char *src2pn = src2p + src2_pitch;
	for (int y = 0; y < height; y += 2)
	{
		for (int x = 0; x < row_size; x += 4)
		{
			*(unsigned int *)(src1p+x) |= *(unsigned int *)(src2p+x);
			*(unsigned int *)(src1pn+x) |= *(unsigned int *)(src2pn+x);
		}
		src2p += src2_pitch << 1;
		src1p += src1_pitch << 1;
		src2pn += src2_pitch << 1;
		src1pn += src1_pitch << 1;
	}
}

void Or_C1(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	for (int y = 0; y < height; y ++)
	{
		for (int x = 0; x < row_size; x ++)
			*(src1p+x) |= *(src2p+x);
		src2p += src2_pitch;
		src1p += src1_pitch;
	}
}

void Or_MMX(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	__asm {
		mov			eax,[src1p]
		mov			ecx,[src2p]
		mov			ebx,eax
		mov			edx,ecx
		mov			esi,[height]
		add			ebx,[src1_pitch]
		mov			edi,0
		add			edx,[src2_pitch]
		shl			[src1_pitch],1
		shl			[src2_pitch],1
		shr			esi, 1
loopx:
		movq		mm0,[eax+edi]
		movq		mm1,[ebx+edi]
		por			mm0,[ecx+edi]
		por			mm1,[edx+edi]
		movq		[eax+edi],mm0
		movq		[ebx+edi],mm1

		add			edi,8
		cmp			edi,[row_size]
		jl			loopx

		mov			edi,0
		add			eax,src1_pitch
		add			ecx,src2_pitch
		add			ebx,src1_pitch
		add			edx,src2_pitch
		dec			esi

		jnz			loopx

		emms
	}
}

void Xor_C4(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	unsigned char *src1pn = src1p + src1_pitch;
	const unsigned char *src2pn = src2p + src2_pitch;
	for (int y=0; y<height; y+=2)
	{
		for (int x=0; x<row_size; x+=4)
		{
			*(unsigned int *)(src1p+x) ^= *(unsigned int *)(src2p+x);
			*(unsigned int *)(src1pn+x) ^= *(unsigned int *)(src2pn+x);
		}
		src2p += src2_pitch<<1;
		src1p += src1_pitch<<1;
		src2pn += src2_pitch<<1;
		src1pn += src1_pitch<<1;
	}
}

void Xor_C1(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	for (int y = 0; y < height; y ++)
	{
		for (int x = 0; x < row_size; x ++)
			*(src1p+x) ^= *(src2p+x);
		src2p += src2_pitch;
		src1p += src1_pitch;
	}
}

void Xor_MMX(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	__asm {
		mov			eax,[src1p]
		mov			ecx,[src2p]
		mov			ebx,eax
		mov			edx,ecx
		mov			esi,[height]
		add			ebx,[src1_pitch]
		mov			edi,0
		add			edx,[src2_pitch]
		shl			[src1_pitch],1
		shl			[src2_pitch],1
		shr			esi, 1
loopx:
		movq		mm0,[eax+edi]
		movq		mm1,[ebx+edi]
		pxor		mm0,[ecx+edi]
		pxor		mm1,[edx+edi]
		movq		[eax+edi],mm0
		movq		[ebx+edi],mm1

		add			edi,8
		cmp			edi,[row_size]
		jl			loopx

		mov			edi,0
		add			eax,src1_pitch
		add			ecx,src2_pitch
		add			ebx,src1_pitch
		add			edx,src2_pitch
		dec			esi

		jnz			loopx

		emms
	}
}

void Andn_C4(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	unsigned char *src1pn = src1p + src1_pitch;
	const unsigned char *src2pn = src2p + src2_pitch;
	for (int y=0; y<height; y+=2)
	{
		for (int x=0; x<row_size; x+=4)
		{
			*(unsigned int *)(src1p+x) &= !(*(unsigned int *)(src2p+x));
			*(unsigned int *)(src1pn+x) &= !(*(unsigned int *)(src2pn+x));
		}
		src2p += src2_pitch<<1;
		src1p += src1_pitch<<1;
		src2pn += src2_pitch<<1;
		src1pn += src1_pitch<<1;
	}
}

void Andn_C1(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	for (int y = 0; y < height; y ++)
	{
		for (int x = 0; x < row_size; x ++)
			*(src1p+x) &= !(*(src2p+x));
		src2p += src2_pitch;
		src1p += src1_pitch;
	}
}

void Andn_MMX(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	__asm {
		mov			eax,[src1p]
		mov			ecx,[src2p]
		mov			ebx,eax
		mov			edx,ecx
		mov			esi,[height]
		add			ebx,[src1_pitch]
		mov			edi,0
		add			edx,[src2_pitch]
		shl			[src1_pitch],1
		shl			[src2_pitch],1
		shr			esi, 1
loopx:
		movq		mm0,[ecx+edi]
		movq		mm1,[edx+edi]
		pandn		mm0,[eax+edi]
		pandn		mm1,[ebx+edi]
		movq		[eax+edi],mm0
		movq		[ebx+edi],mm1

		add			edi,8
		cmp			edi,[row_size]
		jl			loopx

		mov			edi,0
		add			eax,src1_pitch
		add			ecx,src2_pitch
		add			ebx,src1_pitch
		add			edx,src2_pitch
		dec			esi

		jnz			loopx

		emms
	}
}

void Min_C(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	for (int y = 0; y < height; y ++)
	{
		for (int x = 0; x < row_size; x ++)
			*(src1p+x) = (*(src2p + x) < *(src1p + x)) ? *(src2p + x) : *(src1p + x);
		src2p += src2_pitch;
		src1p += src1_pitch;
	}
}

void Max_C(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	for (int y = 0; y < height; y ++)
	{
		for (int x = 0; x < row_size; x ++)
			*(src1p+x) = (*(src2p + x) > *(src1p + x)) ? *(src2p + x) : *(src1p + x);
		src2p += src2_pitch;
		src1p += src1_pitch;
	}
}


void Min_iSSE(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	__asm {
		mov			eax,[src1p]
		mov			ecx,[src2p]
		mov			ebx,eax
		mov			edx,ecx
		mov			esi,[height]
		add			ebx,[src1_pitch]
		mov			edi,0
		add			edx,[src2_pitch]
		shl			[src1_pitch],1
		shl			[src2_pitch],1
		shr			esi, 1
loopx:
		movq		mm0,[ecx+edi]
		movq		mm1,[edx+edi]
		pminub		mm0,[eax+edi]
		pminub		mm1,[ebx+edi]
		movq		[eax+edi],mm0
		movq		[ebx+edi],mm1

		add			edi,8
		cmp			edi,[row_size]
		jl			loopx

		mov			edi,0
		add			eax,src1_pitch
		add			ecx,src2_pitch
		add			ebx,src1_pitch
		add			edx,src2_pitch
		dec			esi

		jnz			loopx

		emms
	}
}

void Min_MMX(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	__asm {
		mov			eax,[src1p]
		mov			ecx,[src2p]
		mov			ebx,eax
		mov			edx,ecx
		mov			esi,[height]
		add			ebx,[src1_pitch]
		mov			edi,0
		add			edx,[src2_pitch]
		shl			[src1_pitch],1
		shl			[src2_pitch],1
		shr			esi, 1
loopx:
		movq		mm0,[ecx+edi]
		movq		mm1,[edx+edi]
		movq		mm2,[ecx+edi]
		movq		mm3,[edx+edi]

		pcmpgtb		mm0, [eax+edi]
		pcmpgtb		mm1, [ebx+edi]
		movq		mm4, mm0
		movq		mm5, mm1

		pand		mm0, [eax+edi]
		pand		mm1, [ebx+edi]
		pandn		mm2, mm4
		pandn		mm3, mm5

		paddb		mm0, mm2
		paddb		mm1, mm3

		movq		[eax+edi],mm0
		movq		[ebx+edi],mm1

		add			edi,8
		cmp			edi,[row_size]
		jl			loopx

		mov			edi,0
		add			eax,src1_pitch
		add			ecx,src2_pitch
		add			ebx,src1_pitch
		add			edx,src2_pitch
		dec			esi

		jnz			loopx

		emms
	}
}

void Max_iSSE(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	__asm {
		mov			eax,[src1p]
		mov			ecx,[src2p]
		mov			ebx,eax
		mov			edx,ecx
		mov			esi,[height]
		add			ebx,[src1_pitch]
		mov			edi,0
		add			edx,[src2_pitch]
		shl			[src1_pitch],1
		shl			[src2_pitch],1
		shr			esi, 1
loopx:
		movq		mm0,[ecx+edi]
		movq		mm1,[edx+edi]
		pmaxub		mm0,[eax+edi]
		pmaxub		mm1,[ebx+edi]
		movq		[eax+edi],mm0
		movq		[ebx+edi],mm1

		add			edi,8
		cmp			edi,[row_size]
		jl			loopx

		mov			edi,0
		add			eax,src1_pitch
		add			ecx,src2_pitch
		add			ebx,src1_pitch
		add			edx,src2_pitch
		dec			esi

		jnz			loopx

		emms
	}
}

void Max_MMX(unsigned char *src1p, const unsigned char *src2p, const int src1_pitch,
		   const int src2_pitch, const int row_size, const int height)
{
	__asm {
		mov			eax,[src1p]
		mov			ecx,[src2p]
		mov			ebx,eax
		mov			edx,ecx
		mov			esi,[height]
		add			ebx,[src1_pitch]
		mov			edi,0
		add			edx,[src2_pitch]
		shl			[src1_pitch],1
		shl			[src2_pitch],1
		shr			esi, 1
loopx:
		movq		mm0,[ecx+edi]
		movq		mm1,[edx+edi]
		movq		mm2,[ecx+edi]
		movq		mm3,[edx+edi]

		pcmpgtb		mm0, [eax+edi]
		pcmpgtb		mm1, [ebx+edi]
		movq		mm4, mm0
		movq		mm5, mm1

		pandn		mm0, [eax+edi]
		pandn		mm1, [ebx+edi]
		pand		mm2, mm4
		pand		mm3, mm5

		paddb		mm0, mm2
		paddb		mm1, mm3

		movq		[eax+edi],mm0
		movq		[ebx+edi],mm1

		add			edi,8
		cmp			edi,[row_size]
		jl			loopx

		mov			edi,0
		add			eax,src1_pitch
		add			ecx,src2_pitch
		add			ebx,src1_pitch
		add			edx,src2_pitch
		dec			esi

		jnz			loopx

		emms
	}
}



