// Logic - very basic mask logic manipulations
// Logic (C) 2003 Kurosu (kurosu@inforezo.org)

// See legal notice in Interface.cpp and Readme.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include "GenericTools.h"
#include "Logic.h"
#include "LGC_func.h"

Logic::Logic(PClip cl1, PClip cl2, const char *string, int y, int u, int v,
				 IScriptEnvironment* env) : 
GenericVideoFilter(cl1), clip1(cl1), clip2(cl2), Y(y), V(v), U(u)
{
	const VideoInfo& vi1 = cl1->GetVideoInfo();
	const VideoInfo& vi2 = cl2->GetVideoInfo();
	if (vi1.width != vi2.width || vi1.height != vi2.height)
		env->ThrowError("Logic: Image dimensions don't match");
	if (!vi1.IsYV12() || !vi2.IsYV12())
		env->ThrowError("Logic: Needs YV12 data");
	
	if (!Verify(Y)) env->ThrowError("Logic: Unvalid operation mode on Y");
	if (!Verify(U)) env->ThrowError("Logic: Unvalid operation mode on U");
	if (!Verify(V)) env->ThrowError("Logic: Unvalid operation mode on V");

	if (!lstrcmpi(string,"And"))
	{
		if (env->GetCPUFlags() & CPUF_MMX)
		{
			if (!( vi1.height % 4 ) && !( vi1.width % 16 )) 
			{ ProcessY = And_MMX; ProcessUV = And_MMX; }
			else if (!( vi1.height % 4 ) && !( vi1.width % 8 ))
			{ ProcessY = And_MMX; ProcessUV = And_C4; }
			else if (!( vi1.height % 2 ) && !( vi1.width % 8 ))
			{ ProcessY = And_MMX; ProcessUV = And_C1; }
			else if (!( vi1.width % 4))
			{ ProcessY = And_C4; ProcessUV = And_C1; }
			else { ProcessY = And_C1; ProcessUV = And_C1; }
		}
		else if (!( vi1.height % 4 ) && !( vi1.width % 8 ))
		{ ProcessY = And_C4; ProcessUV = And_C4; }
		else if (!( vi1.height % 2 ) && !( vi1.width % 4 ))
		{ ProcessY = And_C4; ProcessUV = And_C1; }
		else { ProcessY = And_C1; ProcessUV = And_C1; }
	}
	else if (!lstrcmpi(string,"Or"))
	{
		if (env->GetCPUFlags() & CPUF_MMX)
		{
			if (!( vi1.height % 4 ) && !( vi1.width % 16 )) 
			{ ProcessY = Or_MMX; ProcessUV = Or_MMX; }
			else if (!( vi1.height % 4 ) && !( vi1.width % 8 ))
			{ ProcessY = Or_MMX; ProcessUV = Or_C4; }
			else if (!( vi1.height % 2 ) && !( vi1.width % 8 ))
			{ ProcessY = Or_MMX; ProcessUV = Or_C1; }
			else if (!( vi1.width % 4))
			{ ProcessY = Or_C4; ProcessUV = Or_C1; }
			else { ProcessY = Or_C1; ProcessUV = Or_C1; }
		}
		else if (!( vi1.height % 4 ) && !( vi1.width % 8 ))
		{ ProcessY = Or_C4; ProcessUV = Or_C4; }
		else if (!( vi1.height % 2 ) && !( vi1.width % 4 ))
		{ ProcessY = Or_C4; ProcessUV = Or_C1; }
		else { ProcessY = Or_C1; ProcessUV = Or_C1; }
	}
	else if (!lstrcmpi(string,"Xor"))
	{
		if (env->GetCPUFlags() & CPUF_MMX)
		{
			if (!( vi1.height % 4 ) && !( vi1.width % 16 )) 
			{ ProcessY = Xor_MMX; ProcessUV = Xor_MMX; }
			else if (!( vi1.height % 4 ) && !( vi1.width % 8 ))
			{ ProcessY = Xor_MMX; ProcessUV = Xor_C4; }
			else if (!( vi1.height % 2 ) && !( vi1.width % 8 ))
			{ ProcessY = Xor_MMX; ProcessUV = Xor_C1; }
			else if (!( vi1.width % 4))
			{ ProcessY = Xor_C4; ProcessUV = Xor_C1; }
			else { ProcessY = Xor_C1; ProcessUV = Xor_C1; }
		}
		else if (!( vi1.height % 4 ) && !( vi1.width % 8 ))
		{ ProcessY = Xor_C4; ProcessUV = Xor_C4; }
		else if (!( vi1.height % 2 ) && !( vi1.width % 4 ))
		{ ProcessY = Xor_C4; ProcessUV = Xor_C1; }
		else { ProcessY = Xor_C1; ProcessUV = Xor_C1; }
	}
	else if (!lstrcmpi(string,"Andn"))
	{
		if (env->GetCPUFlags() & CPUF_MMX)
		{
			if (!( vi1.height % 4 ) && !( vi1.width % 16 )) 
			{ ProcessY = Andn_MMX; ProcessUV = Andn_MMX; }
			else if (!( vi1.height % 4 ) && !( vi1.width % 8 ))
			{ ProcessY = Andn_MMX; ProcessUV = Andn_C4; }
			else if (!( vi1.height % 2 ) && !( vi1.width % 8 ))
			{ ProcessY = Andn_MMX; ProcessUV = Andn_C1; }
			else if (!( vi1.width % 4))
			{ ProcessY = Andn_C4; ProcessUV = Andn_C1; }
			else { ProcessY = Andn_C1; ProcessUV = Andn_C1; }
		}
		else if (!( vi1.height % 4 ) && !( vi1.width % 8 ))
		{ ProcessY = Andn_C4; ProcessUV = Andn_C4; }
		else if (!( vi1.height % 2 ) && !( vi1.width % 4 ))
		{ ProcessY = Andn_C4; ProcessUV = Andn_C1; }
		else { ProcessY = Andn_C1; ProcessUV = Andn_C1; }
	}
	else if (!lstrcmpi(string,"Min"))
	{
		if (env->GetCPUFlags() & CPUF_INTEGER_SSE)
		{
			if (!( vi1.height % 4 ) && !( vi1.width % 16 )) 
			{ ProcessY = Min_iSSE; ProcessUV = Min_iSSE; }
			else if (!( vi1.height % 2 ) && !( vi1.width % 8 ))
			{ ProcessY = Min_iSSE; ProcessUV = Min_C; }
			else { ProcessY = Min_C; ProcessUV = Min_C; }
		}
		else if (env->GetCPUFlags() & CPUF_MMX)
		{
			if (!( vi1.height % 4 ) && !( vi1.width % 16 )) 
			{ ProcessY = Min_MMX; ProcessUV = Min_MMX; }
			else if (!( vi1.height % 2 ) && !( vi1.width % 8 ))
			{ ProcessY = Min_MMX; ProcessUV = Min_C; }
			else { ProcessY = Min_C; ProcessUV = Min_C; }
		}
		else { ProcessY = Min_C; ProcessUV = Min_C; }
	}
	else if (!lstrcmpi(string,"Max"))
	{
		if (env->GetCPUFlags() & CPUF_INTEGER_SSE)
		{
			if (!( vi1.height % 4 ) && !( vi1.width % 16 )) 
			{ ProcessY = Max_iSSE; ProcessUV = Max_iSSE; }
			else if (!( vi1.height % 2 ) && !( vi1.width % 8 ))
			{ ProcessY = Max_iSSE; ProcessUV = Max_C; }
			else { ProcessY = Max_C; ProcessUV = Max_C; }
		}
		else if (env->GetCPUFlags() & CPUF_MMX)
		{
			if (!( vi1.height % 4 ) && !( vi1.width % 16 )) 
			{ ProcessY = Max_MMX; ProcessUV = Max_MMX; }
			else if (!( vi1.height % 2 ) && !( vi1.width % 8 ))
			{ ProcessY = Max_MMX; ProcessUV = Max_C; }
			else { ProcessY = Max_C; ProcessUV = Max_C; }
		}
		else { ProcessY = Max_C; ProcessUV = Max_C; }
	}
	else env->ThrowError("Logic: unvalid operator");
}

PVideoFrame __stdcall Logic::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame			src1 = clip1->GetFrame(n, env);
	env->MakeWritable(&src1);
	unsigned char		*src1p_y = src1->GetWritePtr(PLANAR_Y);
	unsigned char		*src1p_u = src1->GetWritePtr(PLANAR_U);
	unsigned char		*src1p_v = src1->GetWritePtr(PLANAR_V);
	const int			src1_pitch_y = src1->GetPitch(PLANAR_Y);
	const int			src1_pitch_uv = src1->GetPitch(PLANAR_U);
	
	PVideoFrame			src2 = clip2->GetFrame(n, env);
	const unsigned char	*src2p_y = src2->GetReadPtr(PLANAR_Y);
	const unsigned char	*src2p_u = src2->GetReadPtr(PLANAR_U);
	const unsigned char	*src2p_v = src2->GetReadPtr(PLANAR_V);
	const int			src2_pitch_y = src2->GetPitch(PLANAR_Y);
	const int			src2_pitch_uv = src2->GetPitch(PLANAR_U);
	
	const int			row_size = src1->GetRowSize(PLANAR_Y);
	const int			height = src1->GetHeight(PLANAR_Y);

	switch (Y)
	{
	case OPERATION:
		ProcessY( src1p_y, src2p_y, src1_pitch_y, src2_pitch_y, row_size, height);
		break;
	case COPY_SOURCE:
		env->BitBlt(src1p_y, src1_pitch_y, src2p_y, src2_pitch_y, row_size, height);
		break;
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(src1p_y, -Y, row_size, height, 0, 0, src1_pitch_y);
		break;
	}
	
	switch (U)
	{
	case OPERATION:
		ProcessUV( src1p_u, src2p_u, src1_pitch_uv, src2_pitch_uv, row_size>>1, height>>1);
		break;
	case COPY_SOURCE:
		env->BitBlt(src1p_u, src1_pitch_uv, src2p_u, src2_pitch_uv, row_size>>1, height>>1);
		break;
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(src1p_u, -U, row_size / 2, height / 2, 0, 0, src1_pitch_uv);
		break;
	}
	
	switch (V)
	{
	case OPERATION:
		ProcessUV( src1p_v, src2p_v, src1_pitch_uv, src2_pitch_uv, row_size>>1, height>>1);
		break;
	case COPY_SOURCE:
		env->BitBlt(src1p_v, src1_pitch_uv, src2p_v, src2_pitch_uv, row_size>>1, height>>1);
		break;
	case NO_OPERATION:
		break;
	default:
		MemZoneSet(src1p_v, -V, row_size / 2, height / 2, 0, 0, src1_pitch_uv);
		break;
	}
	
	return src1;
}
