// YV12Subtract - expansion of AviSynth's Subtract
// YV12Subtract (C) 2003 Kurosu (kurosu@inforezo.org)
// Subtract (C) AviSynth's team

// See legal notice in Interface.cpp and Readme.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include <math.h>
#include "YV12Substract.h"
//Yeah I know, 2nd version doesn't work if sub result is <127 or >127
void Substract_MMX(unsigned char *dstp, const unsigned char *srcp,
				const int dst_pitch, const int src_pitch,
				const int row_size, const int height, unsigned char tol, unsigned char *LUT)
{
	//static const __int64 Addi = 0x0080008000800080i64;
	static const __int64 Addi2 = 0x8080808080808080i64;
	//static const __int64 Keep = 0x00FF00FF00FF00FFi64;

	__asm {
		mov			eax,[srcp]
		mov			ecx,[dstp]
		mov			esi,[height]
		pxor		mm6,mm6
		pxor		mm5,mm5
		//Doesn't work because of the final conversion
		//signed word->unsigned byte
#if 0
		//movq		mm4,[Keep]
		movq		mm7,[Addi]
		mov			edi,0
loopx:
		movd		mm2,[ecx+edi]
		movd		mm0,[eax+edi]
		punpcklbw	mm2,mm5
		punpcklbw	mm0,mm6

		paddsw		mm2,mm7
		psubsw		mm2,mm0
		//pand		mm2,mm4
		packsswb	mm2,mm6

		movq		[ecx+edi],mm2

		add			edi,4
#else
		movq		mm7,[Addi2]
		mov			edi,0
loopx:
		movq		mm2,[ecx+edi]
		movq		mm0,[eax+edi]

		psubb		mm2,mm0
		paddb		mm2,mm7
		movq		[ecx+edi],mm2

		add			edi,8
#endif
		cmp			edi,[row_size]
		jl			loopx

		add			eax,[src_pitch]
		add			ecx,[dst_pitch]
		mov			edi,0
		dec			esi

		jnz			loopx

		emms
	}
}

void Substract_C2(unsigned char *dstp, const unsigned char *srcp,
					const int dst_pitch, const int src_pitch,
					const int row_size, const int height, unsigned char tol, unsigned char *LUT)
{
	unsigned char *dstpn = dstp + dst_pitch;
	const unsigned char *srcpn = srcp + src_pitch;

	for (int y=0; y<height; y+=2)
	{
		for (int x=0; x<row_size; x+=2)
		{
			dstp[x] = (dstp[x] - srcp[x]) + 128;
			dstp[x+1] = (dstp[x+1] - srcp[x+1]) + 128;
			dstpn[x] = (dstpn[x] - srcpn[x]) + 128;
			dstpn[x+1] = (dstpn[x+1] - srcpn[x+1]) + 128;
		}
		dstp += dst_pitch<<1;
		srcp += src_pitch<<1;
		dstpn += dst_pitch<<1;
		srcpn += src_pitch<<1;
	}
}

void Substract_C1(unsigned char *dstp, const unsigned char *srcp,
					const int dst_pitch, const int src_pitch,
					const int row_size, const int height, unsigned char tol, unsigned char *LUT)
{
	for (int y=0; y<height; y++)
	{
		for (int x=0; x<row_size; x++)
			dstp[x] = (dstp[x] - srcp[x]) + 128;
		dstp += dst_pitch;
		srcp += src_pitch;
	}
}

void SubstractW_C2(unsigned char *dstp, const unsigned char *srcp,
					const int dst_pitch, const int src_pitch,
					const int row_size, const int height, unsigned char tol, unsigned char *LUT)
{
	unsigned char *dstpn = dstp + dst_pitch;
	const unsigned char *srcpn = srcp + src_pitch;

	for (int y=0; y<height; y+=2)
	{
		for (int x=0; x<row_size; x+=2)
		{
			dstp[x] = LUT[(255 + dstp[x] - srcp[x])];
			dstp[x+1] = LUT[(255 + dstp[x+1] - srcp[x+1])];
			dstpn[x] = LUT[(255 + dstpn[x] - srcpn[x])];
			dstpn[x+1] = LUT[(255 + dstpn[x+1] - srcpn[x+1])];
		}
		dstp += dst_pitch<<1;
		srcp += src_pitch<<1;
		dstpn += dst_pitch<<1;
		srcpn += src_pitch<<1;
	}
}

void SubstractW_MMX(unsigned char *dstp, const unsigned char *srcp,
					const int dst_pitch, const int src_pitch,
					const int row_size, const int height, unsigned char tol, unsigned char *LUT)
{
	unsigned char *dstpn = dstp + dst_pitch;
	const unsigned char *srcpn = srcp + src_pitch;

	for (int y=0; y<height; y+=2)
	{
		for (int x=0; x<row_size; x+=2)
		{
			dstp[x] = LUT[(255 + dstp[x] - srcp[x])];
			dstp[x+1] = LUT[(255 + dstp[x+1] - srcp[x+1])];
			dstpn[x] = LUT[(255 + dstpn[x] - srcpn[x])];
			dstpn[x+1] = LUT[(255 + dstpn[x+1] - srcpn[x+1])];
		}
		dstp += dst_pitch<<1;
		srcp += src_pitch<<1;
		dstpn += dst_pitch<<1;
		srcpn += src_pitch<<1;
	}
}

void SubstractW_C1(unsigned char *dstp, const unsigned char *srcp,
					const int dst_pitch, const int src_pitch,
					const int row_size, const int height, unsigned char tol, unsigned char *LUT)
{
	for (int y=0; y<height; y++)
	{
		for (int x=0; x<row_size; x++)
			dstp[x] = LUT[(255 + dstp[x] - srcp[x])];
		dstp += dst_pitch;
		srcp += src_pitch;
	}
}


void Diff_C(unsigned char *dstp, const unsigned char *srcp,
					const int dst_pitch, const int src_pitch,
					const int row_size, const int height, unsigned char tol, unsigned char *LUT)
{
	short int temp;

	for (int y=0; y<height; y++)
	{
		for (int x=0; x<row_size; x++)
		{
			temp = abs(dstp[x] - srcp[x]) - tol;
			if (temp<0) dstp[x] = 0;
			else dstp[x] = (unsigned char)temp;
		}
		dstp += dst_pitch;
		srcp += src_pitch;
	}
}

void Diff_MMX(unsigned char *dstp, const unsigned char *srcp,
				const int dst_pitch, const int src_pitch,
				const int row_size, const int height, unsigned char T, unsigned char *LUT)
{
	static const __int64 Tolerance = ((__int64)T<<56) + ((__int64)T<<48) + ((__int64)T<<40) +
									((__int64)T<<32) + ((__int64)T<<24) + ((__int64)T<<16) +
									((__int64)T<<8) + ((__int64)T);
	__asm {
		mov			eax,[srcp]
		mov			ecx,[dstp]
		mov			ebx,[srcp]
		mov			edx,[dstp]
		mov			esi,[height]
		add			ebx,[src_pitch]
		add			edx,[dst_pitch]
		shr			esi,1
		shl			src_pitch,1
		shl			dst_pitch,1
		mov			edi,0
		movq		mm7,[Tolerance]
loopx:
		movq		mm0,[eax+edi]
		movq		mm2,[ecx+edi]
		movq		mm4,mm0
		movq		mm1,[ebx+edi]
		psubusb		mm0,mm2
		movq		mm3,[edx+edi]
		movq		mm5,mm1
		psubusb		mm2,mm4
		psubusb		mm1,mm3
		por			mm0,mm2
		psubusb		mm3,mm5
		psubusb		mm0,mm7
		por			mm1,mm3
		psubusb		mm1,mm7

		movq		[ecx+edi],mm0
		movq		[edx+edi],mm1

		add			edi,8
		cmp			edi,[row_size]
		jl			loopx

		mov			edi,0
		add			eax,src_pitch
		add			ebx,src_pitch
		add			ecx,dst_pitch
		add			edx,dst_pitch
		dec			esi

		jnz			loopx

		emms
	}
}