/*
**                    TIVTC v0.9.9 for AviSynth 2.5.x
**
**   TIVTC includes a field matching filter (TFM) and a decimation
**   filter (TDecimate) which can be used together to achieve an
**   IVTC or for other uses. TIVTC currently supports YV12 and 
**   YUY2 colorspaces.
**   
**   Copyright (C) 2004-2005 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "TDecimate.h"

PVideoFrame __stdcall TDecimate::GetFrame(int n, IScriptEnvironment *env)
{
	if (n<0) n=0;
	else if (n>nfrmsN) n=nfrmsN;
	int np = vi.IsYV12() ? 3 : 1;
	if (mode < 2 ) return GetFrameMode01(n, env, np);     // most similar/longest string
	else if (mode == 2) return GetFrameMode2(n, env, np); // arbitrary framerate
	else if (mode == 3) return GetFrameMode3(n, env, np); // single pass mkv-vfr
	else if (mode == 4) return GetFrameMode4(n, env, np); // metrics output
	else if (mode == 5) return GetFrameMode5(n, env, np); // second pass of two pass hybrid
	else env->ThrowError("TDecimate:  unknown error (no such mode)!");
	return NULL;
}

PVideoFrame TDecimate::GetFrameMode01(int n, IScriptEnvironment* env, int np)
{
	int EvalGroup = ((int)(n / (cycle-cycleR))) * cycle; 
	if (curr.frame != EvalGroup)
	{
		prev = curr;
		if (prev.frame != EvalGroup - cycle)
		{
			prev.setFrame(EvalGroup - cycle);
			getOvrCycle(prev, false);
			calcMetricCycle(prev, env, np, true, true);
			if (hybrid > 0)
			{
				checkVideoMatches(prev, prev);
				checkVideoMetrics(prev, vidThresh);
			}
			if (*output) addMetricCycle(prev);
		}
		curr = next;
		if (curr.frame != EvalGroup)
		{
			curr.setFrame(EvalGroup);
			getOvrCycle(curr, false);
			calcMetricCycle(curr, env, np, true, true);
			if (hybrid > 0)
			{
				checkVideoMatches(prev, curr);
				checkVideoMetrics(curr, vidThresh);
			}
			if (*output) addMetricCycle(curr);
		}
		next.setFrame(EvalGroup + cycle);
		getOvrCycle(next, false);
		calcMetricCycle(next, env, np, true, true);
		if (hybrid > 0)
		{
			checkVideoMatches(curr, next);
			checkVideoMetrics(next, vidThresh);
		}
		if (*output) addMetricCycle(next);
		if (hybrid > 0 && curr.type > 1)
		{
			int scenetest = curr.sceneDetect(prev, next, sceneThreshU);
			bool isVid = ((curr.type == 2 || curr.type == 4) && !curr.isfilmd2v && // matches
						  (prev.type == 5 || (prev.type == 2 && (vidDetect == 0 || vidDetect == 2)) || prev.type == 4 ||
						   next.type == 5 || (next.type == 2 && (vidDetect == 0 || vidDetect == 2)) || next.type == 4 || 
						   conCycle == 1 || scenetest != -20));
			bool isVid2 = ((curr.type == 3 || curr.type == 4) && !curr.isfilmd2v && // metrics
						  (prev.type == 5 || (prev.type == 3 && (vidDetect == 1 || vidDetect == 2)) || prev.type == 4 ||
						   next.type == 5 || (next.type == 3 && (vidDetect == 1 || vidDetect == 2)) || next.type == 4 || 
						   conCycle == 1 || scenetest != -20));
			if (curr.type == 5 || (vidDetect == 0 && isVid) || (vidDetect == 1 && isVid2) ||
			    (vidDetect == 2 && (isVid2 || isVid)) || (vidDetect == 3 && (isVid2 && isVid)))
			{
				int temp = curr.sceneDetect(prev, next, sceneThreshU);
				if (temp != -20)
				{
					for (int p=curr.cycleS; p<curr.cycleE; ++p) curr.decimate[p] = curr.decimate2[p] = 0;
					curr.decimate[temp] = curr.decimate2[temp] = 1;
					curr.blend = 2;
					curr.decSet = true;
				}
				else curr.blend = 1;
			}
			else { goto novidjump; }
		}
		else
		{
novidjump:
			if (mode == 0)
			{
				mostSimilarDecDecision(prev, curr, next);
			}
			else
			{
				prev.setDups(dupThresh);
				curr.setDups(dupThresh);
				next.setDups(dupThresh);
				findDupStrings(prev, curr, next);
			}
			if (curr.blend == 3)
			{
				int tscene = curr.sceneDetect(prev, next, sceneThreshU);
				if (tscene != -20 && curr.decimate[tscene] == 1)
				{
					curr.decimate[tscene] = curr.decimate2[tscene] = 0;
					curr.blend = 0;
				}
			}
			if (curr.blend != 3) curr.blend = 0;
		}
		if (debug) debugOutput1(n, curr.blend == 1 ? false : true, curr.blend);
	}
	if (curr.blend == 3)
	{
		int ret = n%(cycle-cycleR), y, f1 = 0, f2 = 0, jk;
		double a1 = 0.0, a2 = 0.0;
		int tscene = curr.sceneDetect(prev, next, sceneThreshU);
		if (tscene != -20)
		{
			for (jk=-1, y=curr.cycleS; y<curr.cycleE; ++y)
			{
				if (curr.decimate[y] == 0) ++jk;
				if (y == tscene && jk < ret) ++jk;
				if (ret == jk)
				{
					ret = y;
					break;
				}
			}
		}
		else
		{
			int d1 = -20, d2 = -20;
			for (y=curr.cycleS; y<curr.cycleE; ++y)
			{
				if (curr.decimate[y] == 1 && d1 == -20) d1 = y;
				else if (curr.decimate[y] == 1 && d2 == -20) { d2 = y; break; }
			}
			if (curr.diffMetricsU[d1] > curr.diffMetricsU[d2]) d1 = d2;
			for (jk=0, y=curr.cycleS; y<curr.cycleE; ++y)
			{
				if (ret == jk && y != d1)
				{
					if (curr.decimate[y] == 1)
					{
						f1 = curr.frameSO+y-1;
						f2 = curr.frameSO+y+1;
						a1 = a2 = 0.5f;
					}
					else ret = y;
					break;
				}
				if (y != d1) ++jk;
			}
		}
		if (f1 != 0)
		{
			PVideoFrame dst = env->NewVideoFrame(vi);
			blendFrames(child->GetFrame(f1,env),child->GetFrame(f2,env),dst,a1,a2,np,env);
			if (display) displayOutput(env, dst, n, 0, true, a1, a2, f1, f2, np);
			if (debug) debugOutput2(n, 0, true, f1, f2, a1, a2);
			return dst;
		}
		if (debug) debugOutput2(n, curr.frame+ret, true, f1, f2, a1, a2);
		if (display) 
		{
			PVideoFrame dst = env->NewVideoFrame(vi);
			copyFrame(dst, child->GetFrame(curr.frame+ret, env), env, np);
			displayOutput(env, dst, n, curr.frame+ret, true, a1, a2, f1, f2, np);
			return dst;
		}
		return child->GetFrame(curr.frame+ret, env);
	}
	else if (curr.blend != 1)
	{
		int ret = curr.getNonDec(n%(cycle-cycleR));
		if (ret == -1)
		{
			curr.debugOutput();
			curr.debugMetrics(curr.length);
			env->ThrowError("TDecimate:  ret less then 0 (tritical is a dumbass)!");
		}
		if (debug) debugOutput2(n, curr.frame+ret, curr.blend==2?false:true, 0, 0, 0.0, 0.0);
		if (display)
		{
			PVideoFrame dst = env->NewVideoFrame(vi);
			copyFrame(dst, child->GetFrame(curr.frame+ret, env), env, np);
			displayOutput(env, dst, n, curr.frame+ret, curr.blend==2?false:true, 0.0, 0.0, 0, 0, np);
			return dst;
		}
		return child->GetFrame(curr.frame+ret, env);
	}
	else
	{
		PVideoFrame dst = env->NewVideoFrame(vi);
		double a1, a2;
		int f1, f2;
		calcBlendRatios(a1, a2, f1, f2, n, curr.frame, curr.cycleE-curr.cycleS);
		if (a1 >= 1.0) copyFrame(dst, child->GetFrame(f1,env), env, np);
		else if (a2 >= 1.0) copyFrame(dst, child->GetFrame(f2,env), env, np);
		else blendFrames(child->GetFrame(f1,env),child->GetFrame(f2,env),dst,a1,a2,np,env);
		if (debug) debugOutput2(n, 0, false, f1, f2, a1, a2);
		if (display) displayOutput(env, dst, n, 0, false, a1, a2, f1, f2, np);
		return dst;
	}
}

PVideoFrame TDecimate::GetFrameMode3(int n, IScriptEnvironment *env, int np)
{
	static int vidC = 0;
	static int filmC = 0;
	static int longestT = 0;
	static int longestV = 0;
	static int countVT = 0;
	static double timestamp = 0.0f;
	if (n == 0) 
	{
		vidC = filmC = longestT = longestV = countVT = 0;
		timestamp = 0.0f;
	}
	if (linearCount != n) env->ThrowError("TDecimate:  non-linear access detected in mode 3!");
	++linearCount;
	if (n == 0 || n-lastGroup == retFrames)
	{
		lastGroup = n;
		lastCycle += cycle;
		prev = curr;
		if (prev.frame != lastCycle - cycle)
		{
			prev.setFrame(lastCycle - cycle);
			getOvrCycle(prev, false);
			calcMetricCycle(prev, env, np, true, true);
			checkVideoMatches(prev, prev);
			checkVideoMetrics(prev, vidThresh);
			if (*output) addMetricCycle(prev);
		}
		curr = next;
		if (curr.frame != lastCycle)
		{
			curr.setFrame(lastCycle);
			getOvrCycle(curr, false);
			calcMetricCycle(curr, env, np, true, true);
			checkVideoMatches(prev, curr);
			checkVideoMetrics(curr, vidThresh);
			if (*output) addMetricCycle(curr);
		}
		next.setFrame(lastCycle + cycle);
		getOvrCycle(next, false);
		calcMetricCycle(next, env, np, true, true);
		checkVideoMatches(curr, next);
		checkVideoMetrics(next, vidThresh);
		if (*output) addMetricCycle(next);
		int scenetest = curr.sceneDetect(prev, next, sceneThreshU);
		bool isVid = ((curr.type == 2 || curr.type == 4) && !curr.isfilmd2v && // matches
					  (prev.type == 5 || (prev.type == 2 && (vidDetect == 0 || vidDetect == 2)) || prev.type == 4 ||
					   next.type == 5 || (next.type == 2 && (vidDetect == 0 || vidDetect == 2)) || next.type == 4 || 
					   conCycle == 1 || scenetest != -20));
		bool isVid2 = ((curr.type == 3 || curr.type == 4) && !curr.isfilmd2v && // metrics
					  (prev.type == 5 || (prev.type == 3 && (vidDetect == 1 || vidDetect == 2)) || prev.type == 4 ||
					   next.type == 5 || (next.type == 3 && (vidDetect == 1 || vidDetect == 2)) || next.type == 4 || 
					   conCycle == 1 || scenetest != -20));
		if (curr.type == 5 || (vidDetect == 0 && isVid) || (vidDetect == 1 && isVid2) ||
		    (vidDetect == 2 && (isVid2 || isVid)) || (vidDetect == 3 && (isVid2 && isVid)))
		{
			retFrames = cycle;
			vidC += (curr.frame+cycle <= nfrms ? cycle : nfrms-curr.frame+1);
			longestT += (curr.frame+cycle <= nfrms ? cycle : nfrms-curr.frame+1);
			if (!tcfv1)
			{
				int stop = (lastCycle+cycle <= nfrms ? cycle : nfrms-lastCycle+1);
				for (int u=0; u<stop; ++u) 
				{
					fprintf(mkvOutF, "%3.1f\n", timestamp);
					timestamp += 1000.0f/fps;
				}
			}
		}
		else
		{
			if (vfrDec != 1)
			{
				mostSimilarDecDecision(prev, curr, next);
			}
			else
			{
				prev.setDups(dupThresh);
				curr.setDups(dupThresh);
				next.setDups(dupThresh);
				findDupStrings(prev, curr, next);
			}
			filmC += (curr.frame+cycle <= nfrms ? cycle : nfrms-curr.frame+1);
			if (retFrames == cycle)
			{
				if (longestT > longestV) longestV = longestT;
				++countVT;
				longestT = 0;
			}
			if (curr.blend != 3) 
			{
				if (!tcfv1)
				{
					int stop = (lastCycle+cycle <= nfrms ? cycle-cycleR : nfrms-lastCycle+1-cycleR);
					for (int u=0; u<stop; ++u) 
					{
						fprintf(mkvOutF, "%3.1f\n", timestamp);
						timestamp += 1000.0f/mkvfps;
					}
				}
				retFrames = cycle-cycleR;
			}
			else 
			{
				if (lastType > 0 && tcfv1)
					fprintf(mkvOutF, "%d,%d,%4.3f\n", lastGroup-(lastType*(cycle-cycleR)),lastGroup-1,mkvfps);
				if (!tcfv1)
				{
					int stop = (lastCycle+cycle <= nfrms ? cycle-cycleR-1 : nfrms-lastCycle+1-cycleR-1);
					for (int u=0; u<stop; ++u) 
					{
						fprintf(mkvOutF, "%3.1f\n", timestamp);
						timestamp += 1000.0f/mkvfps2;
					}
				}
				else fprintf(mkvOutF, "%d,%d,%4.3f\n", lastGroup, lastGroup+cycle-cycleR-2, mkvfps2);
				retFrames = cycle-cycleR-1;
			}
		}
		if (retFrames == cycle && lastType > 0 && tcfv1)
			fprintf(mkvOutF, "%d,%d,%4.3f\n", lastGroup-(lastType*(cycle-cycleR)),lastGroup-1,mkvfps);
		if (retFrames == cycle-cycleR) ++lastType;
		else lastType = 0;
		if (debug) debugOutput1(n, retFrames == cycle ? false : true, curr.blend);
	}
	if (retFrames == cycle)
	{
		if (lastCycle+(n-lastGroup) > nfrms) 
		{ 
			retFrames = -1; 
			lastFrame = n-1;
			fprintf(mkvOutF,"# TDecimate Mode 3:  Last Frame = %d\n", lastFrame);
		}
		else
		{
			if (debug) debugOutput2(n, lastCycle+(n-lastGroup), false, 0, 0, 0.0, 0.0);
			if (display) 
			{
				PVideoFrame dst = env->NewVideoFrame(vi);
				copyFrame(dst, child->GetFrame(lastCycle+(n-lastGroup), env), env, np);
				displayOutput(env, dst, n, lastCycle+(n-lastGroup), false, 0.0, 0.0, 0, 0, np);
				return dst;
			}
			return child->GetFrame(lastCycle+(n-lastGroup), env);
		}
	}
	else if (retFrames == cycle-cycleR || (curr.blend == 3 && retFrames == cycle-cycleR-1))
	{
		int ret = curr.getNonDec(n-lastGroup);
		if ((ret >= 0 && curr.frame+ret > nfrms) || ret < 0) 
		{ 
			retFrames = -1; 
			lastFrame = n-1;
			if (lastType > 0)
			{
				if (tcfv1) fprintf(mkvOutF, "%d,%d,%4.3f\n", lastGroup-((lastType-1)*(cycle-cycleR)),lastFrame,mkvfps);
				lastType = 0;
			}
			fprintf(mkvOutF,"# TDecimate Mode 3:  Last Frame = %d\n", lastFrame);
		}
		else
		{
			if (debug) debugOutput2(n, curr.frame+ret, true, 0, 0, 0.0, 0.0);
			if (display) 
			{
				PVideoFrame dst = env->NewVideoFrame(vi);
				copyFrame(dst, child->GetFrame(curr.frame+ret, env), env, np);
				displayOutput(env, dst, n, curr.frame+ret, true, 0.0, 0.0, curr.blend == 3 ? 5 : 0, 0, np);
				return dst;
			}
			return child->GetFrame(curr.frame+ret, env);
		}
	}
	if (retFrames == -1 && mkvOutF != NULL)
	{
		double filmCf = ((double)(filmC)/(double)(nfrms+1))*100.0f;
		double videoCf = ((double)(vidC)/(double)(nfrms+1))*100.0f;
		fprintf(mkvOutF, "# vfr stats:  %05.2f%c film  %05.2f%c video\n", filmCf, '%', videoCf, '%');
		fprintf(mkvOutF, "# vfr stats:  %d - film  %d - video  %d - total\n", filmC, vidC, nfrms+1);
		fprintf(mkvOutF, "# vfr stats:  longest vid section - %d frames\n", longestV);
		fprintf(mkvOutF, "# vfr stats:  # of detected vid sections - %d", countVT);
		fclose(mkvOutF);
		mkvOutF = NULL;
	}
	PVideoFrame dst = env->NewVideoFrame(vi);
	setBlack(dst, np);
	if (retFrames <= -305)
	{
		if (retFrames <= -306 && se)
			env->ThrowError("TDecimate:  mode 3 finished (early termination)!");
		sprintf(buf, "Mode 3:  Last Actual Frame = %d", lastFrame);
		Draw(dst, 2, 1, buf, np);
	}
	--retFrames;
	return dst;
}

PVideoFrame TDecimate::GetFrameMode4(int n, IScriptEnvironment *env, int np)
{
	PVideoFrame prv = child->GetFrame(n>0?n-1:0,env);
	PVideoFrame src = child->GetFrame(n, env);
	int blockN = -20, xblocks;
	unsigned long metricU = ULONG_MAX, metricF = ULONG_MAX;
	getOvrFrame(n, metricU, metricF);
	if (metricU == ULONG_MAX || metricF == ULONG_MAX || display) 
		metricU = calcMetric(prv, src, np, blockN, xblocks, metricF, env, true);
	double metricN = (metricU*100.0f)/MAX_DIFF;
	if (debug)
	{
		sprintf(buf,"TDecimate:  frame %d  metric = %3.2f  metricF = %u (%3.2f)", n, metricN, metricF,
			(float)metricF*100.0f/(float)sceneDivU);
		OutputDebugString(buf);
	}
	if (*output && metricsOutArray != NULL) 
	{
		metricsOutArray[n<<1] = metricU;
		metricsOutArray[(n<<1)+1] = metricF;
	}
	if (display)
	{
		PVideoFrame dst = env->NewVideoFrame(vi);
		copyFrame(dst, src, env, np);
		if (blockN != -20) drawBox(dst, blockN, xblocks, np);
		sprintf(buf, "TDecimate %s by tritical", VERSION);
		Draw(dst, 0, 0, buf, np);
		sprintf(buf, "Mode: 4 (metrics output)");
		Draw(dst, 0, 1, buf, np);
		sprintf(buf, "chroma = %s", chroma ? "true" : "false");
		Draw(dst, 0, 2, buf, np);
		sprintf(buf,"Frame %d:  %3.2f  %3.2f", n, metricN, (float)metricF*100.0f/(float)sceneDivU);
		Draw(dst, 0, 3, buf, np);
		return dst;
	}
	return src;
}

PVideoFrame TDecimate::GetFrameMode5(int n, IScriptEnvironment *env, int np)
{
	int frame = aLUT[n];
	if (debug)
	{
		sprintf(buf,"TDecimate:  inframe = %d  useframe = %d  (mode = %d)", n, frame, mode);
		OutputDebugString(buf);
	}
	if (display)
	{
		PVideoFrame dst = env->NewVideoFrame(vi);
		copyFrame(dst, child->GetFrame(frame, env), env, np);
		sprintf(buf, "TDecimate %s by tritical", VERSION);
		Draw(dst, 0, 0, buf, np);
		sprintf(buf, "Mode: %d (vfr)  Hybrid = %d", mode, hybrid);
		Draw(dst, 0, 1, buf, np);
		sprintf(buf,"inframe = %d  useframe = %d", n, frame);
		Draw(dst, 0, 2, buf, np);
		return dst;
	}
	return child->GetFrame(frame, env);
}

unsigned long TDecimate::calcMetric(PVideoFrame &prev, PVideoFrame &curr, int np, 
		int &blockNI, int &xblocksI, unsigned long &metricF, IScriptEnvironment *env,
		bool scene)
{
	const unsigned char *prvp, *curp;
	int prv_pitch, cur_pitch, width, height, y, x, difft;
	int xblocks = ((vi.width+xhalfS)>>xshiftS) + 1;
	int xblocks4 = xblocks<<2;
	int yblocks = ((vi.height+yhalfS)>>yshiftS) + 1;
	int arraysize = (xblocks*yblocks)<<2;
	int temp1, temp2, box1, box2, stop, inc;
	int yhalf, xhalf, yshift, xshift, b, plane;
	unsigned long highestDiff = 0;
	memset(diff,0,arraysize*sizeof(unsigned long));
	stop = chroma ? np : 1;
	inc = np == 3 ? 1 : chroma ? 1 : 2;
	for (b=0; b<stop; ++b)
	{
		if (b == 0) plane = PLANAR_Y;
		else if (b == 1) plane = PLANAR_U;
		else plane = PLANAR_V;
		prvp = prev->GetReadPtr(plane);
		prv_pitch = prev->GetPitch(plane);
		width = prev->GetRowSize(plane);
		height = prev->GetHeight(plane);
		curp = curr->GetReadPtr(plane);
		cur_pitch = curr->GetPitch(plane);
		if (b == 0)
		{
			yshift = yshiftS;
			yhalf = yhalfS;
			if (np == 3)
			{
				xshift = xshiftS;
				xhalf = xhalfS;
			}
			else
			{
				xshift = xshiftS+1;
				xhalf = xhalfS<<1;
			}
		}
		else
		{
			yshift = yshiftS-1;
			yhalf = yhalfS>>1;
			xshift = xshiftS-1;
			xhalf = xhalfS>>1;
		}
		for (y=0; y<height; ++y)
		{
			temp1 = (y>>yshift)*xblocks4;
			temp2 = ((y+yhalf)>>yshift)*xblocks4;
			for (x=0; x<width; x+=inc)
			{
				difft = abs(prvp[x]-curp[x]);
				if (difft > nt)
				{
					box1 = (x>>xshift)<<2;
					box2 = ((x+xhalf)>>xshift)<<2;
					diff[temp1+box1+0] += difft;
					diff[temp1+box2+1] += difft;
					diff[temp2+box1+2] += difft;
					diff[temp2+box2+3] += difft;
				}
			}
			prvp += prv_pitch;
			curp += cur_pitch;
		}
		if (b == 0)
		{
			metricF = 0;
			if (scene)
			{
				if (np == 3 || !chroma)
				{
					for (x=0; x<arraysize; x+=4) metricF += diff[x];
				}
				else
				{
					metricF = calcLumaDiffYUY2(prev->GetReadPtr(),curr->GetReadPtr(),
						prev->GetRowSize(),prev->GetHeight(),prev->GetPitch(),curr->GetPitch(),env);
				}
			}
		}
	}
	blockNI = -20;
	xblocksI = xblocks4;
	for (x=0; x<arraysize; ++x)
	{
		if (diff[x] > highestDiff)
		{
			highestDiff = diff[x];
			blockNI = x;
		}
	}
	if (blockNI == -20) blockNI = 0;
	return highestDiff;
}

void TDecimate::calcMetricCycle(Cycle &current, IScriptEnvironment *env, int np,
								bool scene, bool hnt)
{
	if (current.mSet || current.cycleS == current.cycleE) return;
	int xblocks = ((vi.width+xhalfS)>>xshiftS) + 1;
	int xblocks4 = xblocks<<2;
	int yblocks = ((vi.height+yhalfS)>>yshiftS) + 1;
	int arraysize = (xblocks*yblocks)<<2;
	int temp1, temp2, box1, box2;
	int yhalf, xhalf, yshift, xshift, b, plane;
	int prv_pitch, cur_pitch, width, height, difft;
	int i, y, x, w, stop, inc;
	unsigned long highestDiff;
	PVideoFrame prev, next;
	const unsigned char *prvp, *curp;
	for (w=current.frameSO, i=current.cycleS; i<current.cycleE; ++i, ++w)
	{
		if ((current.match[i] != -20 || !hnt) && current.diffMetricsU[i] != ULONG_MAX &&
			(current.diffMetricsUF[i] != ULONG_MAX || !scene)) continue;
		prev = child->GetFrame(w>0?w-1:0,env);
		next = child->GetFrame(w, env);
		if (current.match[i] == -20 && hnt) current.match[i] = getHint(next, current.filmd2v[i]);
		if (current.diffMetricsU[i] != ULONG_MAX && 
			(current.diffMetricsUF[i] != ULONG_MAX || !scene)) continue;
		memset(diff,0,arraysize*sizeof(unsigned long));
		stop = chroma ? np : 1;
		inc = np == 3 ? 1 : chroma ? 1 : 2;
		for (b=0; b<stop; ++b)
		{
			if (b == 0) plane = PLANAR_Y;
			else if (b == 1) plane = PLANAR_U;
			else plane = PLANAR_V;
			prvp = prev->GetReadPtr(plane);
			prv_pitch = prev->GetPitch(plane);
			width = prev->GetRowSize(plane);
			height = prev->GetHeight(plane);
			curp = next->GetReadPtr(plane);
			cur_pitch = next->GetPitch(plane);
			if (b == 0)
			{
				yshift = yshiftS;
				yhalf = yhalfS;
				if (np == 3)
				{
					xshift = xshiftS;
					xhalf = xhalfS;
				}
				else
				{
					xshift = xshiftS+1;
					xhalf = xhalfS<<1;
				}
			}
			else
			{
				yshift = yshiftS-1;
				yhalf = yhalfS>>1;
				xshift = xshiftS-1;
				xhalf = xhalfS>>1;
			}
			for (y=0; y<height; ++y)
			{
				temp1 = (y>>yshift)*xblocks4;
				temp2 = ((y+yhalf)>>yshift)*xblocks4;
				for (x=0; x<width; x+=inc)
				{
					difft = abs(prvp[x]-curp[x]);
					if (difft > nt)
					{
						box1 = (x>>xshift)<<2;
						box2 = ((x+xhalf)>>xshift)<<2;
						diff[temp1+box1+0] += difft;
						diff[temp1+box2+1] += difft;
						diff[temp2+box1+2] += difft;
						diff[temp2+box2+3] += difft;
					}
				}
				prvp += prv_pitch;
				curp += cur_pitch;
			}
			if (b == 0)
			{
				current.diffMetricsUF[i] = 0;
				if (scene)
				{
					if (np == 3 || !chroma)
					{
						for (x=0; x<arraysize; x+=4) current.diffMetricsUF[i] += diff[x];
					}
					else 
					{
						current.diffMetricsUF[i] = calcLumaDiffYUY2(prev->GetReadPtr(),next->GetReadPtr(),
							prev->GetRowSize(),prev->GetHeight(),prev->GetPitch(),next->GetPitch(), env);
					}
				}
			}
		}
		highestDiff = 0;
		for (x=0; x<arraysize; ++x)
		{
			if (diff[x] > highestDiff) highestDiff = diff[x];
		}
		current.diffMetricsU[i] = highestDiff;
		current.diffMetricsN[i] = (highestDiff * 100.0f) / MAX_DIFF;
	}
	current.mSet = true;
	current.setIsFilmD2V();
}

unsigned long TDecimate::calcLumaDiffYUY2(const unsigned char *prvp, const unsigned char *nxtp,
	int width, int height, int prv_pitch, int nxt_pitch, IScriptEnvironment *env)
{
	unsigned long diff = 0;
	long cpu = env->GetCPUFlags();
	if ((cpu&CPUF_INTEGER_SSE) && ((width&15) == 0) && nt == 0) 
		calcLumaDiffYUY2_ISSE_16(prvp, nxtp, width, height, prv_pitch, nxt_pitch, diff);
	else if ((cpu&CPUF_INTEGER_SSE) && ((width&7) == 0) && nt == 0)
		calcLumaDiffYUY2_ISSE_8(prvp, nxtp, width, height, prv_pitch, nxt_pitch, diff);
	else
	{
		int x, y, temp;
		for (y=0; y<height; ++y)
		{
			for (x=0; x<width; x+=2)
			{
				temp = abs(prvp[x]-nxtp[x]);
				if (temp > nt) diff += temp;
			}
			prvp += prv_pitch;
			nxtp += nxt_pitch;
		}
	}
	return diff;
}

void TDecimate::copyFrame(PVideoFrame &dst, PVideoFrame &src, IScriptEnvironment *env, int np)
{
	int b, plane;
	for (b=0; b<np; ++b)
	{
		if (b == 0) plane = PLANAR_Y;
		else if (b == 1) plane = PLANAR_V;
		else plane = PLANAR_U;
		env->BitBlt(dst->GetWritePtr(plane), dst->GetPitch(plane), src->GetReadPtr(plane),
			src->GetPitch(plane), src->GetRowSize(plane), src->GetHeight(plane));
	}
}

int TDecimate::getHint(PVideoFrame &src, int &d2vfilm)
{
	const unsigned char *p = src->GetReadPtr(PLANAR_Y);
	unsigned int i, magic_number = 0, hint = 0;
	int match = -20, field = 0;
	d2vfilm = 0;
	for (i=0; i<32; ++i)
	{
		magic_number |= ((*p++ & 1) << i);
	}
	if (magic_number != MAGIC_NUMBER) return match;
	for (i=0; i<32; ++i)
	{
		hint |= ((*p++ & 1) << i);
	}
	if (hint&0xFFFFFFC0) return match;
	if (hint&TOP_FIELD) field = 1;
	if (hint&D2VFILM) d2vfilm = 1;
	hint &= 0x00000007;
	if (hint == ISC) match = ISC;
	else if (hint == ISP) match = ISP;
	else if (hint == ISN) match = ISN;
	else if (hint == ISB) match = ISB;
	else if (hint == ISU) match = ISU;
	else if (hint == ISDB) match = ISDB;
	else if (hint == ISDT) match = ISDT;
	if (match != -20 && field != 0)
	{
		if (match == 0) match = 3;
		else if (match == 2) match = 4;
		else if (match == 3) match = 0;
		else if (match == 4) match = 2;
	}
	return match;
}

bool TDecimate::checkForObviousDecFrame(Cycle &p, Cycle &c, Cycle &n)
{
	int i, v, dups=0, mc, mp, saved = -20, saved2 = -20;
	unsigned long lowest_metric = ULONG_MAX;
	for (i=c.cycleS; i<c.cycleE; ++i)
	{
		mp = i == c.cycleS ? (p.cycleE > 0 ? p.match[p.cycleE-1] : -20) : mc;
		mc = c.match[i];
		if (checkMatchDup(mp, mc)) ++dups;
		if (dups > 1) return false;
		if (dups == 1 && saved == -20) saved = i;
		if (c.diffMetricsU[i] < lowest_metric) 
		{
			lowest_metric = c.diffMetricsU[i];
			saved2 = i;
		}
	}
	if (dups != 1) return false;
	if (saved != saved2 || saved2 == -20) return false;
	lowest_metric = ULONG_MAX;
	int cp = -20, cn = -20;
	for (dups=0, v=-1, i=p.cycleS; i<p.cycleE; ++i)
	{
		mp = i == p.cycleS ? -20 : mc;
		mc = p.match[i];
		if (checkMatchDup(mp, mc)) { ++dups; cp = i; }
		if (p.diffMetricsU[i] < lowest_metric) { v = i; lowest_metric = p.diffMetricsU[i]; }
	}
	if (dups != 1 || cp != v) cp = -20;
	if (cp == -20)
	{
		lowest_metric = ULONG_MAX;
		for (dups=0, v=-1, i=n.cycleS; i<n.cycleE; ++i)
		{
			mp = i == n.cycleS ? c.match[c.cycleE-1] : mc;
			mc = n.match[i];
			if (checkMatchDup(mp, mc)) { ++dups; cn = i; }
			if (n.diffMetricsU[i] < lowest_metric) { v = i; lowest_metric = n.diffMetricsU[i]; }
		}
		if (dups != 1 || cn != v) return false;
	}
	if (saved != cp && saved != cn) return false;
	if (debug)
	{
		sprintf(buf,"TDecimate:  obvious dec frame found  %d - %d!\n", saved, c.frameSO);
		OutputDebugString(buf);
	}
	c.decimate[saved] = c.decimate2[saved] = 1;
	c.decSet = true;
	return true;
}

int TDecimate::checkForD2VDecFrame(Cycle &p, Cycle &c)
{
	int i, v = 0, mp, savedV = -20, savedL = -20, savedM = -20;
	unsigned int lowest = ULONG_MAX;
	for (i=c.cycleS; i<c.cycleE; ++i)
	{
		if (c.filmd2v[i] == 1) 
		{
			++v;
			if (v > 1) return -20;
			savedV = i;
			mp = i == c.cycleS ? (p.cycleE > 0 ? p.match[p.cycleE-1] : -20) : c.match[i-1];
			if (checkMatchDup(mp, c.match[i])) savedM = i;
		}
		if (c.diffMetricsU[i] < lowest)
		{
			lowest = c.diffMetricsU[i];
			savedL = i;
		}
	}
	if (v != 1 || (savedV != savedM && savedV != savedL)) return -20;
	if (debug)
	{
		sprintf(buf,"TDecimate:  d2v dec frame found  %d - %d!\n", savedV, c.frameSO);
		OutputDebugString(buf);
	}
	return savedV;
}

bool TDecimate::checkForTwoDropLongestString(Cycle &p, Cycle &c, Cycle &n)
{
	int dupsP = 0, savedp = -20, dupsC = 0, dupsN = 0, savedn = -20;
	int c1 = -20, c2 = -20, i, v, mp, mc;
	unsigned int lowest = ULONG_MAX;
	for (v=-1, i=p.cycleS; i<p.cycleE; ++i)
	{
		mp = i == p.cycleS ? -20 : mc;
		mc = p.match[i];
		if (checkMatchDup(mp, mc)) { ++dupsP; savedp = i; }
		if (p.diffMetricsU[i] < lowest) { lowest = p.diffMetricsU[i]; v = i; }
	}
	if (dupsP == 0 && v == p.cycleS) { dupsP = 1; savedp = v; }
	if (dupsP != 1 || savedp != v) return false;
	lowest = ULONG_MAX;
	for (v=-1, i=n.cycleS; i<n.cycleE; ++i)
	{
		mp = i == n.cycleS ? c.match[c.cycleE-1] : mc;
		mc = n.match[i];
		if (checkMatchDup(mp, mc)) { ++dupsN; savedn = i; }
		if (n.diffMetricsU[i] < lowest) { lowest = n.diffMetricsU[i]; v = i; }
	}
	if (dupsN != 1 || savedn != v || savedn == savedp) return false;
	for (v=0,c1=-1,c2=-1,i=c.cycleS; i<c.cycleE; ++i)
	{
		mp = i == c.cycleS ? (p.cycleE > 0 ? p.match[p.cycleE-1] : -20) : mc;
		mc = c.match[i];
		if (checkMatchDup(mp, mc))
		{
			++v;
			if (c1 == -1) c1 = i;
			else if (c2 == -1) c2 = i;
		}
	}
	if (v != 2 || c1 == -1 || c2 == -1) return false;
	if (c1 != savedp || c2 != savedn) return false;
	if ((c1 == c.cycleS && (p.dupArray[p.cycleE>0?p.cycleE-1:0] == 1 || c.dupArray[c1+1] == 1)) ||
		(c1 != c.cycleS && (c.dupArray[c1-1] == 1 || c.dupArray[c1+1] == 1))) return false;
	if ((c2 == c.cycleE-1 && (n.dupArray[n.cycleS] == 1 || c.dupArray[c2-1] == 1)) ||
		(c2 != c.cycleE-1 && (c.dupArray[c2-1] == 1 || c.dupArray[c2+1] == 1))) return false;
	c.blend = 3;
	c.decimate[c1] = c.decimate2[c1] = 1;
	c.decimate[c2] = c.decimate2[c2] = 1;
	c.decSet = true;
	if (debug)
	{
		sprintf(buf,"TDecimate:  drop two frames longest string  %d:%d - %d!\n", c1, c2, c.frameSO);
		OutputDebugString(buf);
	}
	return true;
}

void TDecimate::mostSimilarDecDecision(Cycle &p, Cycle &c, Cycle &n)
{
	if (!c.lowSet) c.setLowest(false);
	if (cycleR != 1)
	{
		c.setDecimateLow(cycleR);
		return;
	}
	if (!p.dupsSet) p.setDupsMatches(p);
	if (!c.dupsSet) c.setDupsMatches(p);
	if (!n.dupsSet) n.setDupsMatches(c);
	int d2vdecf = checkForD2VDecFrame(p, c);
	if (c.dupCount <= 0 && d2vdecf == -20)
	{
		c.setDecimateLow(cycleR);
		return;
	}
	int i, ovrdups;
	for (ovrdups=0, i=c.cycleS; i<c.cycleE; ++i)
	{
		if (c.decimate[i] != 1) c.decimate[i] = c.decimate2[i] = 0;
		else ++ovrdups;
	}
	if (ovrdups != 0)
	{
		c.setDecimateLow(cycleR);
		return;
	}
	for (i=0; i<c.cycleS; ++i) c.decimate[i] = c.decimate2[i] = -20;
	for (i=c.cycleE; i<c.length; ++i) c.decimate[i] = c.decimate2[i] = -20;
	if (d2vdecf != -20)
	{
		c.decimate[d2vdecf] = c.decimate2[d2vdecf] = 1;
		c.decSet = true;
		return;
	}
	if (c.dupCount == 1)
	{
		unsigned long lowest = ULONG_MAX, lowest2 = ULONG_MAX;
		int savedc, savedp, savedn, v;
		for (v=-1, i=c.cycleS; i<c.cycleE; ++i)
		{
			if (c.dupArray[i] == 1) savedc = i;
			if (c.diffMetricsU[i] < lowest) { v = i; lowest = c.diffMetricsU[i]; }
		}
		if (savedc == v || (double)(c.diffMetricsU[savedc]*0.9) <= (double)lowest ||
			fabs(((double)lowest*100.0/(double)MAX_DIFF)-c.diffMetricsN[savedc]) < 0.20)
		{
			c.decimate[savedc] = c.decimate2[savedc] = 1;
			c.decSet = true;
			return;
		}
		if (p.dupCount == 1 && n.dupCount == 1)
		{
			for (lowest2 = ULONG_MAX, i=p.cycleS; i<p.cycleE; ++i)
			{
				if (p.dupArray[i] == 1) savedp = i;
				if (p.diffMetricsU[i] < lowest2) { v = i; lowest2 = p.diffMetricsU[i]; }
			}
			if (savedp != v)
			{
				c.setDecimateLow(cycleR);
				return;
			}
			for (lowest2 = ULONG_MAX, i=n.cycleS; i<n.cycleE; ++i)
			{
				if (n.dupArray[i] == 1) savedn = i;
				if (n.diffMetricsU[i] < lowest2) { v = i; lowest2 = n.diffMetricsU[i]; }
			}
			if (savedn != v || savedp != savedn || savedp != savedc) 
			{
				c.setDecimateLow(cycleR);
				return;
			}
			c.decimate[savedc] = c.decimate2[savedc] = 1;
			c.decSet = true;
			return;
		}
		c.setDecimateLow(cycleR);
		return;
	}
	else
	{
		unsigned long lowestp = ULONG_MAX, lowestn = ULONG_MAX;
		int savedp, savedn, savedc1, savedc2, v;
		if (c.dupCount == 2 && p.dupCount == 1 && n.dupCount == 1)
		{
			for (v=-1, i=p.cycleS; i<p.cycleE; ++i)
			{
				if (p.dupArray[i] == 1) savedp = i;
				if (p.diffMetricsU[i] < lowestp) { v = i; lowestp = p.diffMetricsU[i]; }
			}
			if (savedp != v || v == -1) goto tryother;
			for (v=-1, i=n.cycleS; i<n.cycleE; ++i)
			{
				if (n.dupArray[i] == 1) savedn = i;
				if (n.diffMetricsU[i] < lowestn) { v = i; lowestn = n.diffMetricsU[i]; }
			}
			if (savedn != v || v == -1 || savedn == savedp) goto tryother;
			for (savedc1=-1,savedc2=-1,i=c.cycleS; i<c.cycleE; ++i)
			{
				if (c.dupArray[i] == 1)
				{
					if (savedc1 == -1) savedc1 = i;
					else if (savedc2 == -1) savedc2 = i;
				}
			}
			if (savedc1 != savedp || savedc2 != savedn) goto tryother;
			if (savedc1 != c.lowest[0] && savedc1 != c.lowest[1]) goto tryother;
			if (savedc2 != c.lowest[0] && savedc2 != c.lowest[1]) goto tryother;
			if (abs(savedc1-savedc2) <= 1) goto tryother;
			c.blend = 3;
			c.decimate[savedc1] = c.decimate2[savedc1] = 1;
			c.decimate[savedc2] = c.decimate2[savedc2] = 1;
			c.decSet = true;
			if (debug)
			{
				sprintf(buf,"TDecimate:  drop two frames most similar  %d:%d - %d!\n", savedc1, savedc2, c.frameSO);
				OutputDebugString(buf);
			}
			return;
		}
tryother:
		unsigned long metricP, metricN, metricC, metricPt, metricNt;
		int savedc = -1;
		for (v=0, i=c.cycleS; i<c.cycleE; ++i)
		{
			if (c.dupArray[i] == 1)
			{
				if (((i == c.cycleS && p.dupArray[p.cycleE>0?p.cycleE-1:0] == 0) || (i != c.cycleS && c.dupArray[i-1] == 0)) &&
					((i == c.cycleE-1 && n.dupArray[n.cycleS] == 0) || (i != c.cycleE-1 && c.dupArray[i+1] == 0)))
				{
					++v;
					metricPt = i == c.cycleS ? p.diffMetricsU[p.cycleE>0?p.cycleE-1:0] : c.diffMetricsU[i-1];
					metricNt = i == c.cycleE-1 ? n.diffMetricsU[n.cycleS] : c.diffMetricsU[i+1];
					if (savedc == -1 || metricPt+metricNt > metricP+metricN)
					{
						savedc = i;
						metricP = metricPt;
						metricN = metricNt;
						metricC = c.diffMetricsU[i];
					}
				}
			}
		}
		bool check = false;
		for (i=0; i<v; ++i)
		{
			if (savedc == c.lowest[i]) { check = true; break; }
		}
		if (!check || savedc == -1)
		{
			c.setDecimateLow(cycleR);
			return;
		}
		c.decimate[savedc] = c.decimate2[savedc] = 1;
		c.decSet = true;
		return;
	}
	c.setDecimateLow(cycleR);
}

void TDecimate::findDupStrings(Cycle &p, Cycle &c, Cycle &n)
{
	if (!p.dupsSet) p.setDups(dupThresh);
	if (!c.dupsSet) c.setDups(dupThresh);
	if (!n.dupsSet) n.setDups(dupThresh);
	int dupStrings[(DEFAULT_CYCLE_SIZE+1)>>1][3];
	unsigned long lowest;
	int temp, i, g, b, f, forward, back, v, w = 0, j;
	int temp1, temp2, temp3, y, dups, ovrdups = 0, d2vdecf = -20;
	if (cycleR == 1) d2vdecf = checkForD2VDecFrame(p, c);
	for (temp=0, i=c.cycleS; i<c.cycleE; ++i) temp += c.dupArray[i];
	if (temp == 0 && d2vdecf == -20) 
	{
		if (!c.lowSet) c.setLowest(false);
		c.setDecimateLow(cycleR);
		return;
	}
	for (ovrdups=0, i=c.cycleS; i<c.cycleE; ++i)
	{
		if (c.decimate[i] != 1) c.decimate[i] = c.decimate2[i] = 0;
		else ++ovrdups;
	}
	for (i=0; i<c.cycleS; ++i) c.decimate[i] = c.decimate2[i] = -20;
	for (i=c.cycleE; i<c.length; ++i) c.decimate[i] = c.decimate2[i] = -20;
	if (ovrdups >= cycleR) { c.decSet = true; return; }
	if (cycleR == 1 && checkForObviousDecFrame(p, c, n)) return;
	if (cycleR == 1 && checkForTwoDropLongestString(p, c, n)) return;
	if (cycleR == 1 && cycle > 2 && ovrdups == 0 && c.dupCount > 1 && 
		(p.dupCount == 1 || p.dupCount == 0 || n.dupCount == 1 || n.dupCount == 0)) 
	{
		int p1 = -20, c1 = -20, c2 = -20, n1 = -20, dupcp = 0, usecp = 0;
		for (dupcp=0,i=c.cycleS; i<c.cycleE && dupcp==0; ++i)
		{
			if (c.dupArray[i] == 1 && 
				((i>c.cycleS && c.dupArray[i-1] == 0) || (i==c.cycleS && p.dupArray[p.cycleE>0?p.cycleE-1:0] == 0)) && 
				((i<c.cycleE-1 && c.dupArray[i+1] == 0) || ((i==c.cycleE-1 && n.dupArray[n.cycleS] == 0)))) 
			{ c1 = i; break; }
			if (c.dupArray[i] == 1) ++dupcp;
		}
		for (dupcp=0,i=c.cycleE-1; i>=c.cycleS && dupcp==0; --i)
		{
			if (c.dupArray[i] == 1 && 
				((i>c.cycleS && c.dupArray[i-1] == 0) || (i==c.cycleS && p.dupArray[p.cycleE>0?p.cycleE-1:0] == 0)) && 
				((i<c.cycleE-1 && c.dupArray[i+1] == 0) || ((i==c.cycleE-1 && n.dupArray[n.cycleS] == 0)))) 
			{ c2 = i; break; }
			if (c.dupArray[i] == 1) ++dupcp;
		}
		if ((p.dupCount == 1 || p.dupCount == 0) && c1 != -20)
		{
			if (p.dupCount == 0) p1 = c1;
			else
			{
				for (i=p.cycleS; i<p.cycleE; ++i)
				{
					if (p.dupArray[i] == 1) { p1 = i; break; }
				}
			}
			if (p1 == c1) usecp += 1;
		}
		if ((n.dupCount == 1 || n.dupCount == 0) && c2 != -20)
		{
			if (n.dupCount == 0) n1 = c2;
			else
			{
				for (i=n.cycleS; i<n.cycleE; ++i)
				{
					if (n.dupArray[i] == 1) { n1 = i; break; }
				}
			}
			if (n1 == c2) usecp += 5;
		}
		if (usecp == 1 || usecp == 5)
		{
			if (usecp == 5) c1 = c2;
			c.decimate[c1] = c.decimate2[c1] = 1;
			if (debug)
			{
				sprintf(buf,"TDecimate:  usecp case %d - %d!\n", usecp, c.frameSO);
				OutputDebugString(buf);
			}
			c.decSet = true;
			return;
		}
		else if (usecp == 6)
		{
			c.blend = 3;
			c.decimate[c1] = c.decimate2[c1] = 1;
			c.decimate[c2] = c.decimate2[c2] = 1;
			if (debug)
			{
				sprintf(buf,"TDecimate:  usecp case %d - %d!\n", usecp, c.frameSO);
				OutputDebugString(buf);
			}
			c.decSet = true;
			return;
		}
	}
	if (d2vdecf != -20)
	{
		c.decimate[d2vdecf] = c.decimate2[d2vdecf] = 1;
		c.decSet = true;
		return;
	}
	for (i=0; i<((DEFAULT_CYCLE_SIZE+1)>>1); ++i) 
		dupStrings[i][0] = dupStrings[i][1] = dupStrings[i][2] = -20;
	for (w=0, i=c.cycleS; i<c.cycleE; ++i)
	{
		if (c.dupArray[i] == 0) continue;
		f = b = i;
		forward = back = 0;
		while (c.dupArray[f] == 1 && f < c.cycleE) 
		{
			++forward;
			++f;
		}
		if (f == c.cycleE)
		{
			g = 0;
			while (n.dupArray[g] == 1 && g < n.cycleE)
			{
				++g;
				++forward;
			}
		}
		while (c.dupArray[b] == 1 && b >= 0) 
		{
			++back;
			--b;
		}
		if (b < 0)
		{
			g = cycle-1;
			while ((p.dupArray[g] == 1 && p.decimate2[g] != 1) && g >= 0) 
			{
				++back;
				--g;
			}
		}
		i = f;
		++b;
		dupStrings[w][0] = back + forward - 1;
		dupStrings[w][1] = b;
		dupStrings[w][2] = f;
		++w;
	}
	if (ovrArray != NULL)
	{
		for (i=c.cycleS; i<c.cycleE; ++i)
		{
			if (c.decimate[i] == 1)
			{
				for (v=0; v<w; ++v)
				{
					if (i >= dupStrings[v][1] && i < dupStrings[v][2]) 
					{
						if (dupStrings[v][2]-dupStrings[v][1]-1 <= 0) dupStrings[v][0] = -20;
						else if (dupStrings[v][0] != 2555) --dupStrings[v][0];
						else dupStrings[v][0] = -20;
					}
				}
			}
		}
	}
	for (i=1; i<w; ++i) 
	{
		j = i;
		temp1 = dupStrings[i][0];
		temp2 = dupStrings[i][1];
		temp3 = dupStrings[i][2];
		while (j>0 && (dupStrings[j-1][0]<temp1 || (dupStrings[j-1][0] == temp1 && 
			dupStrings[j-1][1]>temp2))) 
		{
			dupStrings[j][0] = dupStrings[j-1][0];
			dupStrings[j][1] = dupStrings[j-1][1];
			dupStrings[j][2] = dupStrings[j-1][2];
			--j;
		}
		dupStrings[j][0] = temp1;
		dupStrings[j][1] = temp2;
		dupStrings[j][2] = temp3;
	}
	for (v=0; v<c.dupCount && v<cycleR-ovrdups; ++v)
	{
		if (dupStrings[0][0] < 1) break;
		lowest = ULONG_MAX;
		f = dupStrings[0][1];
		for (dups=0, i=dupStrings[0][1]; i<dupStrings[0][2]; ++i)
		{
			if (c.diffMetricsU[i] < lowest && c.decimate[i] == 0)
			{
				lowest = c.diffMetricsU[i];
				f = i;
			}
			if (c.decimate[i] == 1) ++dups;
		}
		c.decimate[f] = 1;
		y = dupStrings[0][1];
		while (c.decimate2[y] == 1) ++y;
		c.decimate2[y] = 1;
		if (dupStrings[0][2]-dupStrings[0][1]-dups-1 <= 0) dupStrings[0][0] = -20;
		else if (dupStrings[0][0] != 2555) --dupStrings[0][0];
		else dupStrings[0][0] = -20;
		j = 0;
		temp1 = dupStrings[0][0];
		temp2 = dupStrings[0][1];
		temp3 = dupStrings[0][2];
		while (j<w-1 && (dupStrings[j+1][0]>temp1 || (dupStrings[j+1][0] == temp1 && 
			dupStrings[j+1][1]<temp2))) 
		{
			dupStrings[j][0] = dupStrings[j+1][0];
			dupStrings[j][1] = dupStrings[j+1][1];
			dupStrings[j][2] = dupStrings[j+1][2];
			++j;
		}
		dupStrings[j][0] = temp1;
		dupStrings[j][1] = temp2;
		dupStrings[j][2] = temp3;
	}
	c.decSet = true;
	if (v<cycleR-ovrdups)
	{
		c.setLowest(true);
		c.setDecimateLowP(cycleR-ovrdups-v);
	}
}

void TDecimate::checkVideoMatches(Cycle &p, Cycle &c)
{
	if (!p.mSet || !c.mSet || (c.type != 3 && c.type > 0)) return;
	int dups = 0, mp, mc, i;
	for (i=c.cycleS; i<c.cycleE && dups<=0; ++i)
	{
		if (i == c.cycleS) 
		{
			if (p.frame != c.frame) mp = p.cycleE > 0 ? p.match[p.cycleE-1] : -20;
			else mp = -20;
		}
		else mp = c.match[i-1];
		mc = c.match[i];
		if (mp == 0 && mc == 3) ++dups;
		else if (mp == 1 && (mc == 0 || mc == 3)) ++dups;
		else if (mp == 2 && (mc == 1 || mc == 3 || mc == 4 || mc == 6)) ++dups;
		else if (mp == 3 && mc == 0) ++dups;
		else if (mp == 4 && (mc == 0 || mc == 1 || mc == 2 || mc == 5)) ++dups;
		else if (mp == 5 && mc == 3) ++dups;
		else if (mp == 6 && mc == 0) ++dups;
		else if (mc == -20) ++dups;
	}
	if (dups == 0)
	{
		if (c.type == -1) c.type = 2;
		else if (c.type == 0) c.type = 2;
		else if (c.type == 3) c.type = 4;
	}
	else if (c.type == -1) c.type = 0;
}

bool TDecimate::checkMatchDup(int mp, int mc)
{
	if (mp == 0 && mc == 3) return true;
	else if (mp == 1 && (mc == 0 || mc == 3)) return true;
	else if (mp == 2 && (mc == 1 || mc == 3 || mc == 4 || mc == 6)) return true;
	else if (mp == 3 && mc == 0) return true;
	else if (mp == 4 && (mc == 0 || mc == 1 || mc == 2 || mc == 5)) return true;
	else if (mp == 5 && mc == 3) return true;
	else if (mp == 6 && mc == 0) return true;
	else if (mc == -20) return true;
	return false;
}

void TDecimate::checkVideoMetrics(Cycle &c, double thresh)
{
	if (!c.mSet || (c.type > 0 && c.type != 2)) return;
	int dups = 0, f = c.cycleS, i;
	if (c.frame == 0) ++f;
	for (i=f; i<c.cycleE && dups<=0; ++i)
	{
		if (c.diffMetricsN[i] <= thresh) ++dups;
	}
	if (dups == 0)
	{
		if (c.type == -1) c.type = 3;
		else if (c.type == 0) c.type = 3;
		else if (c.type == 2) c.type = 4;
	}
	else if (c.type == -1) c.type = 0;
}

void TDecimate::getOvrCycle(Cycle &current, bool mode2)
{
	if (mode2) current.dupCount = 0;
	if (ovrArray == NULL && metricsArray == NULL && metricsOutArray == NULL) return;
	int b = current.cycleS, v = 0, i, p = 0, d = 0, value;
	for (i=current.frameSO; i<current.frameEO; ++i, ++b)
	{
		if (ovrArray != NULL)
		{
			value = ovrArray[i];
			if (value&DROP_FRAME && (d < cycleR || mode2))
			{
				if (mode2)
				{
					current.dupArray[b] = 1;
					++current.dupCount;
				}
				else
				{
					current.decimate[b] = current.decimate2[b] = 1;
					++d;
					current.type = 1;
				}
			}
			if (value&VIDEO) ++v;
			if (value&FILM) ++p;
			if (value&ISD2VFILM) current.filmd2v[b] = 1;
			if ((value&ISMATCH) != 0x70)
			{
				value = (value&ISMATCH)>>4;
				if (value == ISC) current.match[b] = ISC;
				else if (value == ISP) current.match[b] = ISP;
				else if (value == ISN) current.match[b] = ISN;
				else if (value == ISB) current.match[b] = ISB;
				else if (value == ISU) current.match[b] = ISU;
				else if (value == ISDB) current.match[b] = ISDB;
				else if (value == ISDT) current.match[b] = ISDT;
			}
		}
		bool foundM = false;
		if (metricsArray != NULL)
		{
			if (metricsArray[i<<1] != ULONG_MAX) 
			{
				current.diffMetricsU[b] = metricsArray[i<<1];
				current.diffMetricsN[b] = (metricsArray[i<<1] * 100.0f) / MAX_DIFF;
				foundM = true;
			}
			if (metricsArray[(i<<1)+1] != ULONG_MAX) current.diffMetricsUF[b] = metricsArray[(i<<1)+1];
		}
		if (metricsOutArray != NULL && !foundM)
		{
			if (metricsOutArray[i<<1] != ULONG_MAX) 
			{
				current.diffMetricsU[b] = metricsOutArray[i<<1];
				current.diffMetricsN[b] = (metricsOutArray[i<<1] * 100.0f) / MAX_DIFF;
			}
			if (metricsOutArray[(i<<1)+1] != ULONG_MAX) current.diffMetricsUF[b] = metricsOutArray[(i<<1)+1];
		}
	}
	if (v > 0 && v == current.cycleE-current.cycleS && current.type != 1) current.type = 5;
	if (p > 0 && p == current.cycleE-current.cycleS && current.type != 5) current.type = 1;
	current.setIsFilmD2V();
}

void TDecimate::getOvrFrame(int n, unsigned long &metricU, unsigned long &metricF)
{
	metricU = metricF = ULONG_MAX;
	if (metricsArray != NULL)
	{
		if (metricsArray[n<<1] != ULONG_MAX) metricU = metricsArray[n<<1];
		if (metricsArray[(n<<1)+1] != ULONG_MAX) metricF = metricsArray[(n<<1)+1];
	}
	if (metricU != ULONG_MAX && metricF != ULONG_MAX) return;
	if (metricsOutArray != NULL)
	{
		if (metricU == ULONG_MAX && metricsOutArray[n<<1] != ULONG_MAX) metricU = metricsOutArray[n<<1];
		if (metricF == ULONG_MAX && metricsOutArray[(n<<1)+1] != ULONG_MAX) metricF = metricsOutArray[(n<<1)+1];
	}
}

void TDecimate::calcBlendRatios(double &amount1, double &amount2, int &frame1, int &frame2, int n,
								int bframe, int cycleI)
{
	double stepsize = ((double)cycleI)/((double)(cycleI-cycleR));
	double offset = ((cycleI-1)-(stepsize*(cycleI-cycleR-1)))*0.5f;
	double pos = bframe + (n%(cycle-cycleR))*stepsize + offset;
	double posf = pos - (int)(pos);
	frame1 = (int)(pos);
	frame2 = (int)(pos+1.0f);
	amount1 = 1.0f - posf;
	amount2 = posf;
}

void TDecimate::blendFrames(PVideoFrame &src1, PVideoFrame &src2, PVideoFrame &dst,
			double amount1, double amount2, int np, IScriptEnvironment *env)
{
	const unsigned char *srcp1, *srcp2;
	unsigned char *dstp;
	int b, plane, x, y, width, height;
	int s1_pitch, dst_pitch, s2_pitch, widthM;
	bool useMMX = ((env->GetCPUFlags()&CPUF_MMX) != 0 ? true : false);
	bool useSSE2 = ((env->GetCPUFlags()&CPUF_SSE2) != 0 ? true : false);
	for (b=0; b<np; ++b)
	{
		if (b == 0) plane = PLANAR_Y;
		else if (b == 1) plane = PLANAR_V;
		else plane = PLANAR_U;
		srcp1 = src1->GetReadPtr(plane);
		s1_pitch = src1->GetPitch(plane);
		width = src1->GetRowSize(plane);
		height = src1->GetHeight(plane);
		srcp2 = src2->GetReadPtr(plane);
		s2_pitch = src2->GetPitch(plane);
		dstp = dst->GetWritePtr(plane);
		dst_pitch = dst->GetPitch(plane);
		if (useSSE2 && (s1_pitch%16) == 0 && (s2_pitch%16) == 0)
		{
			widthM = width + ((width%16) == 0 ? 0 : 16-(width%16));
			blend_SSE2_16(dstp, srcp1, srcp2, widthM, height, dst_pitch, s1_pitch, s2_pitch, amount1, amount2);
		}
		else if (useMMX)
		{
			widthM = width + ((width%8) == 0 ? 0 : 8-(width%8));
			blend_MMX_8(dstp, srcp1, srcp2, widthM, height, dst_pitch, s1_pitch, s2_pitch, amount1, amount2);
		}
		else
		{
			for (y=0; y<height; ++y)
			{
				for (x=0; x<width; ++x)
				{
					dstp[x] = (int)(amount1*srcp1[x]+amount2*srcp2[x]+0.5f);
				}
				srcp1 += s1_pitch;
				srcp2 += s2_pitch;
				dstp += dst_pitch;
			}
		}
	}
}

void TDecimate::drawBox(PVideoFrame &dst, int blockN, int xblocks, int np)
{
	if (np == 3) drawBoxYV12(dst, blockN, xblocks);
	else drawBoxYUY2(dst, blockN, xblocks);
}

void TDecimate::Draw(PVideoFrame &dst, int x1, int y1, const char *s, int np)
{
	if (np == 3) DrawYV12(dst, x1, y1, s);
	else DrawYUY2(dst, x1, y1, s);
}

void TDecimate::setBlack(PVideoFrame &dst, int np)
{
	unsigned char *dstp;
	int b, plane, x, y, pitch, width, height;
	for (b=0; b<np; ++b)
	{
		if (b == 0) plane = PLANAR_Y;
		else if (b == 1) plane = PLANAR_V;
		else plane = PLANAR_U;
		dstp = dst->GetWritePtr(plane);
		pitch = dst->GetPitch(plane);
		height = dst->GetHeight(plane);
		if (np == 3)
		{
			if (b == 0) memset(dstp, 0, pitch*height);
			else memset(dstp, 128, pitch*height);
		}
		else 
		{
			width = dst->GetRowSize(plane)>>1;
			for (y=0; y<height; ++y)
			{
				for (x=0; x<width; ++x) ((unsigned short *)dstp)[x] = 0x8000;
				dstp += pitch;
			}
		}
	}
}

AVSValue __cdecl Create_TDecimate(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	int cc = args[9].IsInt() ? args[9].AsInt() : 3;
	if (cc == 3) cc = 1;
	else cc = 2;
	bool chroma = args[26].IsBool() ? args[26].AsBool() : true;
	return new TDecimate(args[0].AsClip(),args[1].AsInt(0),args[2].AsInt(1),args[3].AsInt(5),
		args[4].AsFloat(23.976),args[5].AsFloat(chroma ? 1.1 : 1.4),args[6].AsFloat(chroma ? 1.1 : 1.4),
		args[7].AsFloat(15),args[8].AsInt(0),args[9].AsInt(3),args[10].AsInt(cc),
		args[11].AsInt(cc),args[12].AsString(""),args[13].AsString(""),args[14].AsString(""),
		args[15].AsString(""),args[16].AsString(""),args[17].AsInt(0),args[18].AsInt(32),args[19].AsInt(32),
		args[20].AsBool(false),args[21].AsBool(false),args[22].AsInt(1),args[23].AsBool(false),
		args[24].AsBool(true),args[25].AsBool(false),args[26].AsBool(true),args[27].AsBool(false),
		args[28].AsInt(-200),args[29].AsBool(false),env);
}

TDecimate::TDecimate(PClip _child, int _mode, int _cycleR, int _cycle, double _rate, 
	double _dupThresh, double _vidThresh, double _sceneThresh, int _hybrid,
	int _vidDetect, int _conCycle, int _conCycleTP, const char* _ovr, 
	const char* _output, const char* _input, const char* _tfmIn, const char* _mkvOut,
	int _nt, int _blockx, int _blocky, bool _debug, bool _display, int _vfrDec, 
	bool _batch, bool _tcfv1, bool _se, bool _chroma, bool _exPP, int _maxndl, bool _m2PA, 
	IScriptEnvironment* env) : GenericVideoFilter(_child), mode(_mode), cycleR(_cycleR), cycle(_cycle), 
	rate(_rate), dupThresh(_dupThresh), vidThresh(_vidThresh), sceneThresh(_sceneThresh),
	hybrid(_hybrid), vidDetect(_vidDetect), conCycle(_conCycle), conCycleTP(_conCycleTP),
	ovr(_ovr), output(_output), input(_input), tfmIn(_tfmIn), mkvOut(_mkvOut), nt(_nt), 
	blockx(_blockx), blocky(_blocky), debug(_debug), display(_display), vfrDec(_vfrDec),
	batch(_batch), tcfv1(_tcfv1), se(_se), chroma(_chroma), exPP(_exPP), maxndl(_maxndl), m2PA(_m2PA)
{
	diff = metricsArray = metricsOutArray = mode2_metrics = NULL;
	aLUT = mode2_decA = mode2_order = NULL;
	ovrArray = NULL;
	mkvOutF = NULL;
	FILE *f = NULL;
	char linein[81], *linep, *linet;
	fps = (double)(vi.fps_numerator) / (double)(vi.fps_denominator);
	if (!vi.IsYV12() && !vi.IsYUY2())
		env->ThrowError("TDecimate:  YV12 and YUY2 colorspaces only!");
	if (mode < 0 || mode > 5)
		env->ThrowError("TDecimate:  mode must be set to 0, 1, 2, 3, 4, or 5!");
	if (mode == 3 && !(*mkvOut))
		env->ThrowError("TDecimate:  an mkvOut file must be specified in mode 3!");
	if (mode == 3 && hybrid != 2)
		env->ThrowError("TDecimate:  hybrid must be set to 2 when in mode 3!");
	if (cycleR >= cycle && (mode == 0 || mode == 1))
		env->ThrowError("TDecimate:  cycleR must be less then cycle!");
	if ((cycle < 2 || cycle > DEFAULT_CYCLE_SIZE) && (mode == 0 || mode == 1))
		env->ThrowError("TDecimate:  cycle must be at least 2 and less then 101!");
	if (rate >= fps && mode == 2)
		env->ThrowError("TDecimate:  mode 2 - new rate must be less then current rate!");
	if (hybrid < 0 || hybrid > 3)
		env->ThrowError("TDecimate:  hybrid must be set to 0, 1, or 2!");
	if (hybrid > 0 && cycleR > 1)
		env->ThrowError("TDecimate:  hybrid processing is currently limited to cycleR=1 cases only!");
	if (mode < 2 && hybrid > 1)
		env->ThrowError("TDecimate:  only hybrid = 0 or 1 is supported in modes 0 and 1!");
	if (vidDetect < 0 || vidDetect > 3)
		env->ThrowError("TDecimate:  vidDetect must be set to 0, 1, 2, or 3!");
	if (conCycle > 2)
		env->ThrowError("TDecimate:  conCycle cannot be greater then 2!");
	if (mode == 5 && (!(*tfmIn) || !(*input)))
		env->ThrowError("TDecimate:  an input file and a tfmIn file must be specified in mode 5!");
	if (mode == 5 && hybrid != 2)
		env->ThrowError("TDecimate:  mode 5 can only be used with hybrid = 2!");
	if (mode == 5 && !(*mkvOut))
		env->ThrowError("TDecimate:  an mkvOut file must be specified in mode 5!");
	if (mode == 4 && (*ovr || *tfmIn))
		env->ThrowError("TDecimate:  cannot use an ovr or tfmIn file when in mode 4!");
	if (vfrDec != 0 && vfrDec != 1)
		env->ThrowError("TDecimate:  vfrDec must be set to 0 or 1!");
	if (hybrid == 0 && *tfmIn)
		env->ThrowError("TDecimate:  a tfmIn file will not be used if hybrid = 0!");
	if (hybrid == 1 && (sceneThresh < 0.0 || sceneThresh > 100.0))
		env->ThrowError("TDecimate:  sceneThresh must be in the range 0 to 100!");
	if (*output && mode == 5)
		env->ThrowError("TDecimate:  output not supported in mode 5 (you should already have the metrics)!");
	if (blockx != 4 && blockx != 8 && blockx != 16 && blockx != 32 && blockx != 64 && 
		blockx != 128 && blockx != 256 && blockx != 512 && blockx != 1024 && blockx != 2048)
		env->ThrowError("TDecimate:  illegal blockx size!");
	if (blocky != 4 && blocky != 8 && blocky != 16 && blocky != 32 && blocky != 64 && 
		blocky != 128 && blocky != 256 && blocky != 512 && blocky != 1024 && blocky != 2048)
		env->ThrowError("TDecimate:  illegal blocky size!");
	if (mode == 2 && maxndl != -200 && (maxndl < 1 || maxndl > 99))
		env->ThrowError("TDecimate:  maxndl must be set to a value between 1 and 99 inclusive!");
	if (debug)
	{
		sprintf(buf, "TDecimate:  %s by tritical\n", VERSION);
		OutputDebugString(buf);
	}
	if (mode == 4 || mode == 5) child->SetCacheHints(CACHE_RANGE, 1);
	else if (mode != 2)
	{
		int cacheRange = cycle*2+1;
		if (cacheRange < 1) cacheRange = 1;
		if (cacheRange > 201) child->SetCacheHints(CACHE_ALL, 999);
		else child->SetCacheHints(CACHE_RANGE, cacheRange);
	}
	linearCount = -342;
	mode2_num = mode2_den = mode2_numCycles = -20; 
	memset(mode2_cfs, 0, 10*sizeof(int));
	nfrms = nfrmsN = vi.num_frames - 1;
	prev.length = curr.length = next.length = cycle;
	prev.maxFrame = curr.maxFrame = next.maxFrame = nfrms;
	xshiftS = blockx == 4 ? 2 : blockx == 8 ? 3 : blockx == 16 ? 4 : blockx == 32 ? 5 :
		blockx == 64 ? 6 : blockx == 128 ? 7 : blockx == 256 ? 8 : blockx == 512 ? 9 : 
		blockx == 1024 ? 10 : 11;
	yshiftS = blocky == 4 ? 2 : blocky == 8 ? 3 : blocky == 16 ? 4 : blocky == 32 ? 5 :
		blocky == 64 ? 6 : blocky == 128 ? 7 : blocky == 256 ? 8 : blocky == 512 ? 9 : 
		blocky == 1024 ? 10 : 11;
	yhalfS = blocky >> 1;
	xhalfS = blockx >> 1;
	useTFMPP = false;
	try 
	{
		AVSValue getTFMPP = env->GetVar("TFMPPValue");
		if (getTFMPP.IsInt() && getTFMPP.AsInt() > 1) useTFMPP = true;
	}
	catch (IScriptEnvironment::NotFound) {  }
	if (exPP) useTFMPP = true;
	if (vi.IsYV12()) 
	{
		if (chroma)
		{
			MAX_DIFF = (219*blockx*blocky) + (224*xhalfS*yhalfS*2);
			//sceneThreshU = (unsigned long)((sceneThresh*(224*((vi.width*vi.height+1)/2.0f) + 219*vi.height*vi.width))/100.0f);
			//sceneDivU = (unsigned long)(224*((vi.width*vi.height+1)/2.0f) + 219*vi.height*vi.width);
		}
		else
		{
			MAX_DIFF = (219*blockx*blocky);
			//sceneThreshU = (unsigned long)((sceneThresh*219*vi.height*vi.width)/100.0f);
			//sceneDivU = (unsigned long)(219*vi.width*vi.height);
		}
		sceneThreshU = (unsigned long)((sceneThresh*219*vi.height*vi.width)/100.0f);
		sceneDivU = (unsigned long)(219*vi.width*vi.height);
	}
	else
	{
		if (chroma)
		{
			MAX_DIFF = (219*blockx*blocky) + (int)(224*xhalfS*yhalfS*4*0.625);
			//sceneThreshU = (unsigned long)((sceneThresh*(224*vi.width*vi.height + 219*vi.width*vi.height))/100.0f);
			//sceneDivU = (unsigned long)(224*vi.width*vi.height + 219*vi.width*vi.height);
		}
		else
		{
			MAX_DIFF = (219*blockx*blocky);
			//sceneThreshU = (unsigned long)((sceneThresh*219*vi.width*vi.height)/100.0f);
			//sceneDivU = (unsigned long)(219*vi.width*vi.height);
		}
		sceneThreshU = (unsigned long)((sceneThresh*219*vi.width*vi.height)/100.0f);
		sceneDivU = (unsigned long)(219*vi.width*vi.height);
	}
	if (mode < 5)
	{
		diff = (unsigned long *)_aligned_malloc((((vi.width+xhalfS)>>xshiftS)+1)*(((vi.height+yhalfS)>>yshiftS)+1)*4*sizeof(unsigned long), 16);
		if (diff == NULL) env->ThrowError("TDecimate:  malloc failure (diff)!");
	}
	if (*output)
	{
		if ((f = fopen(output, "w")) != NULL) 
		{
			_fullpath(outputFull, output, MAX_PATH);
			calcCRC(child, 15, outputCrc, env);
			fclose(f);
			f = NULL;
			// unlink(outputFull);
			metricsOutArray = (unsigned long *)malloc(vi.num_frames*2*sizeof(unsigned long));
			if (metricsOutArray == NULL)
				env->ThrowError("TDecimate:  malloc failure (metricsOutArray)!");
			for (int h=0; h<vi.num_frames*2; ++h) metricsOutArray[h] = ULONG_MAX;
		}
		else env->ThrowError("TDecimate:  output error (cannot create output file)!");
	}
	if (*input)
	{
		metricsArray = (unsigned long *)malloc(vi.num_frames*2*sizeof(unsigned long));
		if (metricsArray == NULL) env->ThrowError("TDecimate:  malloc failure (metricsArray)!");
		for (int h=0; h<vi.num_frames*2; ++h) 
		{
			if (!batch) metricsArray[h] = ULONG_MAX;
			else metricsArray[h] = 0;
		}
		if ((f = fopen(input, "r")) != NULL)
		{
			unsigned long metricU, metricF;
			int w;
			while (fgets(linein, 80, f) != NULL)
			{
				if (linein[0] == 0 || linein[0] == '\n' || linein[0] == '\r' || linein[0] == '#' || linein[0] == ';') 
					continue;
				linep = linein;
				while (*linep != ' ' && *linep != 0 && *linep != 'c') *linep++;
				if (*linep == 'c')
				{
					if (strnicmp(linein, "crc32 = ", 8) == 0)
					{
						linet = linein;
						while(*linet != ' ') *linet++;
						*linet++;
						while(*linet != ' ') *linet++;
						*linet++;
						unsigned int z, tempCrc;
						sscanf(linet, "%x", &z);
						calcCRC(child, 15, tempCrc, env);
						if (tempCrc != z)
						{
							fclose(f);
							f = NULL;
							env->ThrowError("TDecimate:  crc32 in input file does not match that of the current clip!");
						}
						linep = linein;
						while (*linep != ',' && linep != 0) *linep++;
						if (*linep == 0) continue;
						*linep++; *linep++;
						int j;
						if (strnicmp(linep, "blockx = ", 9) == 0)
						{
							while(*linep != '=') *linep++;
							*linep++; *linep++;
							sscanf(linep, "%d", &j);
							if (j != blockx)
							{
								fclose(f);
								f = NULL;
								env->ThrowError("TDecimate:  current blockx value does not match" \
										        " that which was used to create the given input file!");
							}
						}
						linep = linein;
						while (*linep != ',' && *linep != 0) *linep++;
						if (*linep == 0) continue;
						*linep++;
						while (*linep != ',' && *linep != 0) *linep++;
						if (*linep == 0) continue;
						*linep++; *linep++;
						if (strnicmp(linep, "blocky = ", 9) == 0)
						{
							while(*linep != '=') *linep++;
							*linep++; *linep++;
							sscanf(linep, "%d", &j);
							if (j != blocky)
							{
								fclose(f);
								f = NULL;
								env->ThrowError("TDecimate:  current blocky value does not match" \
										        " that which was used to create the given input file!");
							}
						}
						linep = linein;
						while (*linep != ',' && *linep != 0) *linep++;
						if (*linep == 0) continue;
						*linep++;
						while (*linep != ',' && *linep != 0) *linep++;
						if (*linep == 0) continue;
						*linep++;
						while (*linep != ',' && *linep != 0) *linep++;
						if (*linep == 0) continue;
						*linep++; *linep++;
						if (strnicmp(linep, "chroma = ", 9) == 0)
						{
							while(*linep != '=') *linep++;
							*linep++; *linep++;
							sscanf(linep, "%c", &j);
							if (((j == 'T' || j == 't') && !chroma) || ((j == 'F' || j == 'f') && chroma))
							{
								fclose(f);
								f = NULL;
								env->ThrowError("TDecimate:  current chroma setting does not match" \
										        " that which was used to create the given input file!");
							}
						}
					}
				}
				else if (*linep == ' ' && *(linep+1) != 0 && *(linep+1) != ' ')
				{
					sscanf(linein, "%d %u %u", &w, &metricU, &metricF);
					if (w < 0 || w > nfrms) 
					{
						free(metricsArray);
						metricsArray = NULL;
						fclose(f);
						f = NULL;
						env->ThrowError("TDecimate:  input error (out of range frame #)!");
					}
					metricsArray[w*2] = metricU;
					metricsArray[w*2+1] = metricF;
				}
			}
			fclose(f);
			f = NULL;
			if (mode == 5 && !batch)
			{
				for (int h=0; h<vi.num_frames*2; h+=2) 
				{
					if (metricsArray[h] == ULONG_MAX)
					{
						free(metricsArray);
						metricsArray = NULL;
						env->ThrowError("TDecimate:  input error (mode 5, all frames must have entries)!");
					}
				}
			}
		}
		else 
		{
			free(metricsArray);
			metricsArray = NULL;
			env->ThrowError("TDecimate:  input error (cannot open input file)!");
		}
	}
	if (*ovr)
	{
		if ((f = fopen(ovr, "r")) != NULL)
		{		
			if (ovrArray == NULL)
			{
				ovrArray = (unsigned char *)malloc(vi.num_frames*sizeof(unsigned char));
				if (ovrArray == NULL)
				{
					fclose(f);
					f = NULL;
					env->ThrowError("TDecimate:  malloc failure (ovrArray, ovr)!");
				}
				if (!batch) memset(ovrArray, 112, vi.num_frames);
				else memset(ovrArray, 0, vi.num_frames);
			}
			int q, w, z, count = 0;
			while(fgets(linein, 80, f) != 0)
			{
				if (linein[0] == 0 || linein[0] == '\n' || linein[0] == '\r' || linein[0] == ';' || linein[0] == '#') 
					continue;
				linep = linein;
				while (*linep != 0 && *linep != ' ' && *linep != ',') *linep++;
				if (*linep == ' ')
				{
					linet = linein;
					while(*linet != 0)
					{
						if (*linet != ' ' && *linet != 10) break;
						*linet++;
					}
					if (*linet == 0) continue;
					*linep++;
					if (*linep == '-' || *linep == '+')
					{
						sscanf(linein, "%d", &z);
						if (z<0 || z>nfrms) 
						{
							fclose(f);
							f = NULL;
							env->ThrowError("TDecimate:  ovr file error (out of range frame #)!");
						}
						linep = linein;
						while (*linep != ' ' && *linep != 0) *linep++;
						if (*linep != 0)
						{
							*linep++;
							q = *linep;
							if (q == 45) q = DROP_FRAME;
							else if (q == 43) q = KEEP_FRAME;
							else 
							{
								fclose(f);
								f = NULL;
								env->ThrowError("TDecimate:  ovr file error (invalid specifier)!");
							}
							ovrArray[z] &= 0xFC;
							ovrArray[z] |= q;
						}
					}
					else if (*linep == 'f' || *linep == 'v')
					{
						sscanf(linein, "%d", &z);
						if (z<0 || z>nfrms) 
						{
							fclose(f);
							f = NULL;
							env->ThrowError("TDecimate:  ovr file error (out of range frame #)!");
						}
						linep = linein;
						while (*linep != ' ' && *linep != 0) *linep++;
						if (*linep != 0)
						{
							*linep++;
							q = *linep;
							if (q == 102) q = FILM;
							else if (q == 118) q = VIDEO;
							else 
							{
								fclose(f);
								f = NULL;
								env->ThrowError("TDecimate:  ovr file error (invalid symbol)!");
							}
							ovrArray[z] &= 0xF3;
							ovrArray[z] |= q;
						}
					}
				}
				else if (*linep == ',')
				{
					while (*linep != ' ' && *linep != 0) *linep++;
					if (*linep == 0) continue;
					*linep++;
					if (*linep == 'f' || *linep == 'v')
					{
						sscanf(linein, "%d,%d", &z, &w);
						if (w == 0) w = nfrms;
						if (z<0 || z>nfrms || w<0 || w>nfrms || w<z) 
						{
							fclose(f);
							f = NULL;
							env->ThrowError("TDecimate:  input file error (out of range frame #)!");
						}
						linep = linein;
						while (*linep != ' ' && *linep != 0) *linep++;
						if (*linep != 0)
						{
							*linep++;
							q = *linep;
							if (q == 102) q = FILM;
							else if (q == 118) q = VIDEO;
							else 
							{
								fclose(f);
								f = NULL;
								env->ThrowError("TDecimate:  input file error (invalid specifier)!");
							}
							while (z<=w)
							{
								ovrArray[z] &= 0xF3;
								ovrArray[z] |= q;
								++z;
							}
						}
					}
					else if (*linep == '-' || *linep == '+')
					{
						sscanf(linein, "%d,%d", &z, &w);
						if (w == 0) w = nfrms;
						if (z<0 || z>nfrms || w<0 || w>nfrms || w<z) 
						{
							fclose(f);
							f = NULL;
							env->ThrowError("TDecimate:  input file error (out of range frame #)!");
						}
						linep = linein;
						while (*linep != ' ' && *linep != 0) *linep++;
						*linep++;
						if (*(linep+1) == '-' || *(linep+1) == '+')
						{
							count = 0;
							while ((*linep == '-' || *linep == '+') && (z+count<=w))
							{
								q = *linep;
								if (q == 45) q = DROP_FRAME;
								else if (q == 43) q = KEEP_FRAME;
								else 
								{
									fclose(f);
									f = NULL;
									env->ThrowError("TDecimate:  input file error (invalid specifier)!");
								}
								ovrArray[z+count] &= 0xFC;
								ovrArray[z+count] |= q;
								++count;
								*linep++;
							}
							while (z+count<=w)
							{
								ovrArray[z+count] &= 0xFC;
								ovrArray[z+count] |= (ovrArray[z]&0x03);
								++z;
							}
						}
						else
						{
							q = *linep;
							if (q == 45) q = DROP_FRAME;
							else if (q == 43) q = KEEP_FRAME;
							else 
							{
								fclose(f);
								f = NULL;
								env->ThrowError("TDecimatee:  input file error (invalid specifier)!");
							}
							while (z<=w)
							{
								ovrArray[z] &= 0xFC;
								ovrArray[z] |= q;
								++z;
							}
						}
					}
				}
			}
			fclose(f);
			f = NULL;
		}
		else env->ThrowError("TDecimate:  ovr error (could not open ovr file)!");
	}
	if (*tfmIn)
	{
		bool d2vmarked;
		if ((f = fopen(tfmIn, "r")) != NULL)
		{
			int fieldt, firstLine, z, q, r;
			if (ovrArray == NULL)
			{
				ovrArray = (unsigned char *)malloc(vi.num_frames*sizeof(unsigned char));
				if (ovrArray == NULL)
				{
					fclose(f);
					f = NULL;
					env->ThrowError("TDecimate:  malloc failure (ovrArray, tfmIn)!");
				}
				if (!batch) memset(ovrArray,112,vi.num_frames);
				else memset(ovrArray,0,vi.num_frames);
			}
			fieldt = firstLine = 0;
			while (fgets(linein, 80, f) != NULL)
			{
				if (linein[0] == 0 || linein[0] == '\n' || linein[0] == '\r' ||  linein[0] == ';' || linein[0] == '#') 
					continue;
				++firstLine;
				linep = linein;
				while (*linep != 'f' && *linep != 'F' && *linep != 0 && *linep != ' ' && *linep != 'c') *linep++;
				if (*linep == 'f' || *linep == 'F')
				{
					if (firstLine == 1)
					{
						if (strnicmp(linein, "field = top", 11) == 0) fieldt = 1;
						else if (strnicmp(linein, "field = bottom", 14) == 0) fieldt = 0;
					}
				}
				else if (*linep == ' ')
				{
					linet = linein;
					while(*linet != 0)
					{
						if (*linet != ' ' && *linet != 10) break;
						*linet++;
					}
					if (*linet == 0) { --firstLine; continue; }
					sscanf(linein, "%d", &z);
					linep = linein;
					while (*linep != 'p' && *linep != 'c' && *linep != 'n' && *linep != 'u' &&
							*linep != 'b' && *linep != 'l' && *linep != 'h' && *linep != 0) *linep++;
					if (*linep != 0)
					{
						if (z<0 || z>nfrms)
						{
							fclose(f);
							f = NULL;
							env->ThrowError("TDecimate:  tfmIn file error (out of range frame #)!");
						}
						linep = linein;
						while (*linep != ' ' && *linep != 0) *linep++;
						if (*linep != 0)
						{
							*linep++;
							q = *linep;
							if (q == 112) q = 0;
							else if (q == 99) q = 1;
							else if (q == 110) q = 2;
							else if (q == 98) q = 3;
							else if (q == 117) q = 4;
							else if (q == 108) q = 5;
							else if (q == 104) q = 6;
							else 
							{
								fclose(f);
								f = NULL;
								env->ThrowError("TDecimate:  tfmIn file error (invalid match specifier)!");
							}
							if (fieldt != 0)
							{
								if (q == 0) q = 3;
								else if (q == 2) q = 4;
								else if (q == 3) q = 0;
								else if (q == 4) q = 2;
							}
							d2vmarked = false;
							*linep++;
							while (*linep == ' ' && *linep != 0) *linep++;
							if (*linep != 0)
							{
								r = *linep;
								if (r == 45 && useTFMPP) q = q;
								else if (r == 43 && q < 5 && useTFMPP)
								{
									if (fieldt == 0) q = 5;
									else q = 6;
								}
								else if (r == '1') d2vmarked = true;
								else if (r != 43 && r != 45)
								{
									fclose(f);
									f = NULL;
									env->ThrowError("TDecimate:  tfmIn file error (invalid specifier)!");
								}
							}
							if (!d2vmarked && *linep != 0)
							{
								*linep++;
								while (*linep == ' ' && *linep != 0) *linep++;
								if (*linep != 0) 
								{
									r = *linep;
									if (r == '1') d2vmarked = true;
								}
							}
							if (d2vmarked) ovrArray[z] |= ISD2VFILM;
							ovrArray[z] |= 0x70;
							ovrArray[z] &= ((q<<4)|0x8F);
						}
					}
				}
			}
			fclose(f);
			f = NULL;
			if (mode == 5 && !batch)
			{
				for (int h=0; h<vi.num_frames; ++h)
				{
					if ((ovrArray[h]&ISMATCH) == 0x70)
					{
						free(ovrArray);
						ovrArray = NULL;
						env->ThrowError("TDecimate:  tfmIn error (mode 5, all frames must have an entry)!");
					}
				}
			}
		}
		else env->ThrowError("TDecimate:  tfmIn file error (could not open file)!");
	}
	if (mode < 2)
	{
		vi.num_frames = (vi.num_frames * (cycle-cycleR)) / cycle;
		nfrmsN = vi.num_frames - 1;
		vi.SetFPS(vi.fps_numerator*(cycle-cycleR), vi.fps_denominator*cycle);
	}
	else if (mode == 2)
	{
		if (metricsOutArray == NULL)
		{
			metricsOutArray = (unsigned long *)malloc(vi.num_frames*2*sizeof(unsigned long));
			if (metricsOutArray == NULL) env->ThrowError("TDecimate:  malloc failure (metricsOutArray)!");
			for (int h=0; h<vi.num_frames*2; ++h) metricsOutArray[h] = ULONG_MAX;
		}
		mode2_decA = (int *)malloc(vi.num_frames*sizeof(int));
		if (mode2_decA == NULL) env->ThrowError("TDecimate:  malloc failure (mode2_decA)!");
		for (int j=0; j<vi.num_frames; ++j) mode2_decA[j] = -20;
		buildDecStrategy(env);
		if (mode2_numCycles > 0) 
		{
			if (curr.length < 0)
				env->ThrowError("TDecimate:  unknown error with mode 2!");
			if (curr.length <= 100) child->SetCacheHints(CACHE_RANGE, (curr.length*2)+2);
			else child->SetCacheHints(CACHE_ALL, 999);
			mode2_order = (int*)malloc(max(curr.length+10,100)*sizeof(int));
			mode2_metrics = (unsigned long*)malloc(max(curr.length+10,100)*sizeof(unsigned long));
		}
		else child->SetCacheHints(CACHE_RANGE, 1);
		int d = 1;
		double n = rate;
		while (n < 16777216 && d < 16777216)
		{
			n *= 2;
			d *= 2;
		}
		vi.SetFPS(int(n+0.5f), d);
		vi.num_frames = (int)(vi.num_frames * (rate/fps));
		nfrmsN = vi.num_frames - 1;
	}
	else if (mode == 3)
	{
		mkvfps = (fps*(cycle-cycleR))/cycle;
		mkvfps2 = (fps*(cycle-cycleR-1))/cycle;
		lastGroup = -1;
		lastCycle = -cycle;
		retFrames = -200;
		lastType = linearCount = 0;
		if ((mkvOutF = fopen(mkvOut, "w")) != NULL)
		{
			if (tcfv1)
			{
				fprintf(mkvOutF, "# timecode format v1\n");
				fprintf(mkvOutF, "Assume %4.3f\n", fps);
			}
			else fprintf(mkvOutF, "# timecode format v2\n");
			fprintf(mkvOutF, "# TDecimate %s by tritical\n", VERSION);
			fprintf(mkvOutF, "# Mode 3 - Auto-generated mkv timecodes file\n");
		}
		else env->ThrowError("TDecimate:  mode 3 error (cannot create mkvOut file)!");
	}
	else if (mode == 5)
	{
		mkvfps = (fps*(cycle-cycleR))/cycle;
		mkvfps2 = (fps*(cycle-cycleR-1))/cycle;
		int *input = NULL;
		input = (int*)malloc(vi.num_frames*sizeof(int));
		if (input == NULL) env->ThrowError("TDecimate:  malloc failure (mode 5, input)!");
		memset(input, 0, vi.num_frames*sizeof(int));
		Cycle prevM, currM, nextM;
		prevM.length = currM.length = nextM.length = cycle;
		prevM.maxFrame = currM.maxFrame = nextM.maxFrame = nfrms;
		bool vid, prevVid;
		int i, h, w, firstkv, countprev, filmC, videoC, longestT, longestV, countVT;
		int count, b, np = vi.IsYV12() ? 3 : 1, passThrough = 0;
twopassrun:
		++passThrough;
		count = 0;
		for (b=0; b<=nfrms; b+=cycle)
		{
			if (b == 0)
			{
				currM.setFrame(0);
				getOvrCycle(currM, false);
				calcMetricCycle(currM, env, np, true, true);
				checkVideoMatches(currM, currM);
				checkVideoMetrics(currM, vidThresh);
			}
			else
			{
				prevM = currM;
				currM = nextM;
			}
			nextM.setFrame(b+cycle);
			getOvrCycle(nextM, false);
			calcMetricCycle(nextM, env, np, true, true);
			checkVideoMatches(currM, nextM);
			checkVideoMetrics(nextM, vidThresh);
			if (passThrough == 1)
			{
				if (currM.type == 5 || (!currM.isfilmd2v && ((currM.type == 2 && (vidDetect == 0 || vidDetect == 2)) ||
					(currM.type == 3 && (vidDetect == 1 || vidDetect == 2)) || (currM.type == 4 && vidDetect == 3))))
				{
					if (currM.type == 5) input[b] = 8;
					if (currM.sceneDetect(prevM, nextM, sceneThreshU) != -20) input[b] = 8;
				}
				else
				{
					if (vfrDec != 1) 
					{
						mostSimilarDecDecision(prevM, currM, nextM);
					}
					else
					{
                        prevM.setDups(dupThresh);
						currM.setDups(dupThresh);
						nextM.setDups(dupThresh);
						findDupStrings(prevM, currM, nextM);
					}
					for (w=0, i=b; i<b+cycle && i<=nfrms; ++i, ++w)
					{
						if (currM.decimate[w] == 1) input[i] = 2;
					}
				}
			}
			else
			{
				for (vid=true, i=b; i<=nfrms && i<b+cycle; ++i)
				{
					if (input[i] == 2) vid = false;
				}
				if (!vid)
				{
					if (vfrDec != 1) 
					{
						mostSimilarDecDecision(prevM, currM, nextM);
					}
					else
					{
                        prevM.setDups(dupThresh);
						currM.setDups(dupThresh);
						nextM.setDups(dupThresh);
						findDupStrings(prevM, currM, nextM);
					}
					for (w=0, i=b; i<b+cycle && i<=nfrms; ++i, ++w) 
					{   
						if (currM.decimate[w] == 1)
						{
							input[i] = 2;
							++count;
						}
						else input[i] = 0;
					}
				}
				else
				{
					for (i=b; i<b+cycle && i<=nfrms; ++i) input[i] = 0;
				}
			}
		}
		if (passThrough == 2) { goto finishTP; }
		for (w=0, h=0; h<=nfrms; h+=cycle)
		{
			for (vid=true, i=h; i<h+cycle && i<=nfrms; ++i)
			{
				if (input[i] == 2) vid = false;
			}
			if (vid) ++w;
			else
			{
				if (w > 0 && w < conCycleTP)
				{
					for (i=max(0,h-w*cycle); i<h && i <= nfrms; i+=cycle) 
					{
						if (input[i] != 8) input[i] = 2;
					}
				}
				w = 0;
			}
		}
		if (w > 0 && w < conCycleTP)
		{
			for (i=h-w*cycle; i<h && i <= nfrms; i+=cycle) 
			{
				if (input[i] != 8) input[i] = 2;
			}
		}
		goto twopassrun;
finishTP:
		if (metricsArray != NULL)
		{
			free(metricsArray);
			metricsArray = NULL;
		}
		if (ovrArray != NULL)
		{
			free(ovrArray);
			ovrArray = NULL;
		}
		vi.num_frames = vi.num_frames - count;
		if ((f = fopen(mkvOut, "w")) != NULL)
		{
			double timestamp = 0.0f;
			double sample1 = 1000.0f/fps;
			double sample2 = 1000.0f/mkvfps;
			double sample3 = 1000.0f/mkvfps2;
			int ddup;
			if (tcfv1)
			{
				fprintf(f, "# timecode format v1\n");
				fprintf(f, "Assume %4.3f\n", fps);
			}
			else fprintf(f, "# timecode format v2\n");
			fprintf(f, "# TDecimate %s by tritical\n", VERSION);
			fprintf(f, "# Mode 5 - Auto-generated mkv timecodes file\n");
			firstkv = countprev = 0;
			vid = prevVid = true;
			filmC = videoC = longestT = longestV = countVT = 0;
			for (count=0, b=0; b<=nfrms; b+=cycle)
			{
				prevVid = vid;
				countprev = count;
				vid = true;
				for (i=b,ddup=0; i<b+cycle && i<=nfrms; ++i)
				{
					if (input[i] == 2) 
					{
						++ddup;
						if (ddup < 2) filmC += (b+cycle <= nfrms ? cycle : nfrms-b+1);
						vid = false;
					}
					else ++count;
				}
				if (vid)
				{
					if (!tcfv1)
					{
						int stop = (b+cycle <= nfrms ? cycle : nfrms-b+1);
						for (int x=0; x<stop; ++x)
						{
							fprintf(f, "%3.1f\n", timestamp);
							timestamp += sample1;
						}
					}
					videoC += (b+cycle <= nfrms ? cycle : nfrms-b+1);
					longestT += (b+cycle <= nfrms ? cycle : nfrms-b+1);
				}
				else if (!tcfv1)
				{
					if (ddup == 1)
					{
						int stop = (b+cycle <= nfrms ? cycle-cycleR : nfrms-b+1-cycleR);
						for (int x=0; x<stop; ++x)
						{
							fprintf(f, "%3.1f\n", timestamp);
							timestamp += sample2;
						}
					}
					else if (ddup == 2)
					{
						int stop = (b+cycle <= nfrms ? cycle-cycleR-1 : nfrms-b+1-cycleR-1);
						for (int x=0; x<stop; ++x)
						{
							fprintf(f, "%3.1f\n", timestamp);
							timestamp += sample3;
						}
					}
					else env->ThrowError("TDecimate:  unknown mode 5 error (tc file creation)!");
				}
				else if (ddup == 2)
				{
					if (!prevVid) fprintf(f, "%d,%d,%4.3f\n", firstkv, countprev-1, mkvfps);
					fprintf(f, "%d,%d,%4.3f\n", countprev, countprev+cycle-cycleR-2, mkvfps2);
					firstkv = countprev+cycle-cycleR-1;
				}
				if (prevVid != vid && countprev != 0 && ddup != 2 && countprev > firstkv)
				{
					if (!prevVid && tcfv1) fprintf(f, "%d,%d,%4.3f\n", firstkv, countprev-1, mkvfps);
					firstkv = countprev;
				}
				else if (prevVid != vid && ddup != 2) firstkv = countprev;
				if (prevVid != vid && prevVid && countprev != 0)
				{
					if (longestT > longestV) longestV = longestT;
					++countVT;
					longestT = 0;
				}
			}
			if (!vid && tcfv1) fprintf(f, "%d,%d,%4.3f\n", firstkv, count-1, mkvfps);
			double filmCf = ((double)(filmC)/(double)(nfrms+1))*100.0f;
			double videoCf = ((double)(videoC)/(double)(nfrms+1))*100.0f;
			fprintf(f, "# vfr stats:  %05.2f%c film  %05.2f%c video\n", filmCf, '%', videoCf, '%');
			fprintf(f, "# vfr stats:  %d - film  %d - video  %d - total\n", filmC, videoC, nfrms+1);
			fprintf(f, "# vfr stats:  longest vid section - %d frames\n", longestV);
			fprintf(f, "# vfr stats:  # of detected vid sections - %d", countVT);
			fclose(f);
			f = NULL;
		}
		else
		{
			free(input);
			input = NULL;
			env->ThrowError("TDecimate:  mkvOut file output error (cannot create file)!");
		}
		if (aLUT != NULL) 
		{
			free(aLUT);
			aLUT = NULL;
		}
		aLUT = (int *)malloc((vi.num_frames+1)*sizeof(int));
		if (aLUT == NULL) 
		{
			free(input);
			input = NULL;
			env->ThrowError("TDecimate:  malloc failure (aLUT, mode 5)!");
		}
		memset(aLUT,0,(vi.num_frames+1)*sizeof(int));
		i = w = 0;
		while(i <= nfrms && w <= vi.num_frames-1)
		{
			if (input[i] != 2)
			{
				aLUT[w] = i;
				++w;
			}
			++i;
		}
		free(input);
		input = NULL;
		nfrmsN = vi.num_frames - 1;
	}
	if (f != NULL) fclose(f);
}

TDecimate::~TDecimate()
{
	if (diff != NULL) _aligned_free(diff);
	if (aLUT != NULL) free(aLUT);
	if (ovrArray != NULL) free(ovrArray);
	if (metricsArray != NULL) free(metricsArray);
	if (metricsOutArray != NULL)
	{
		if (*output)
		{
			FILE *f = NULL;
			if ((f = fopen(outputFull, "w")) != NULL)
			{
				unsigned long metricU, metricF;
				fprintf(f, "#TDecimate %s by tritical\n", VERSION);
				fprintf(f, "crc32 = %x, blockx = %d, blocky = %d, chroma = %c\n", outputCrc, blockx, blocky,
					chroma ? 'T' : 'F');
				for (int h=0; h<(nfrms+1)*2; h+=2)
				{
					metricU = metricF = ULONG_MAX;
					if (metricsOutArray[h] != ULONG_MAX) metricU = metricsOutArray[h];
					if (metricsOutArray[h+1] != ULONG_MAX) metricF = metricsOutArray[h+1];
					if (metricU != ULONG_MAX || metricF != ULONG_MAX)
						fprintf(f, "%d %u %u\n", h>>1, metricU, metricF);
				}
				fclose(f);
				f = NULL;
			}
			if (f != NULL) fclose(f);
		}
		free(metricsOutArray);
	}
	if (mkvOutF != NULL) fclose(mkvOutF);
	if (mode2_decA != NULL) free(mode2_decA);
	if (mode2_metrics != NULL) free(mode2_metrics);
	if (mode2_order != NULL) free(mode2_order);
}