/*
**                    TIVTC v0.9.9 for AviSynth 2.5.x
**
**   TIVTC includes a field matching filter (TFM) and a decimation
**   filter (TDecimate) which can be used together to achieve an
**   IVTC or for other uses. TIVTC currently supports YV12 and 
**   YUY2 colorspaces.
**   
**   Copyright (C) 2004-2005 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <windows.h>
#include <stdio.h>
#include <malloc.h>
#include "internal.h"
#include "font.h"
#include "Cycle.h"
#include "calcCRC.h"
#include "profUtil.h"
#include "math.h"
#define ISP 0x00000000 // p
#define ISC 0x00000001 // c
#define ISN 0x00000002 // n
#define ISB 0x00000003 // b
#define ISU 0x00000004 // u
#define ISDB 0x00000005 // l = (deinterlaced c bottom field)
#define ISDT 0x00000006 // h = (deinterlaced c top field)
#define MTC(n) n == 0 ? 'p' : n == 1 ? 'c' : n == 2 ? 'n' : n == 3 ? 'b' : n == 4 ? 'u' : \
			   n == 5 ? 'l' : n == 6 ? 'h' : 'x'
#define TOP_FIELD 0x00000008
#define D2VFILM 0x00000020
#define MAGIC_NUMBER (0xdeadfeed)
#define DROP_FRAME 0x00000001 // ovr array - bit 1
#define KEEP_FRAME 0x00000002 // ovr array - 2
#define FILM 0x00000004	// ovr array - bit 3
#define VIDEO 0x00000008 // ovr array - bit 4
#define ISMATCH 0x00000070 // ovr array - bits 5-7
#define ISD2VFILM 0x00000080 // ovr array - bit 8
#define VERSION "v0.9.9"

class TDecimate : public GenericVideoFilter 
{
private:
	int nfrms, nfrmsN, nt, blockx, blocky, linearCount, maxndl;
	int yshiftS, xshiftS, xhalfS, yhalfS, mode, conCycleTP;
	int cycleR, cycle, hybrid, vidDetect, conCycle, vfrDec;
	int lastFrame, lastCycle, lastGroup, lastType, retFrames;
	unsigned long MAX_DIFF, sceneThreshU, sceneDivU;
	double rate, fps, mkvfps, mkvfps2, dupThresh, vidThresh, sceneThresh;
	bool debug, display, useTFMPP, batch, tcfv1, se, chroma, exPP, m2PA;
	unsigned long *diff, *metricsArray, *metricsOutArray, *mode2_metrics;
	int *aLUT, *mode2_decA, *mode2_order;
	unsigned int outputCrc;
	unsigned char *ovrArray;
	const char *ovr, *input, *output, *mkvOut, *tfmIn;
	int mode2_num, mode2_den, mode2_numCycles, mode2_cfs[10];
	Cycle prev, curr, next;
	FILE *mkvOutF;
	char buf[2048], outputFull[270];
	void TDecimate::setBlack(PVideoFrame &dst, int np);
	void TDecimate::checkVideoMetrics(Cycle &c, double thresh);
	void TDecimate::checkVideoMatches(Cycle &p, Cycle &c);
	bool TDecimate::checkMatchDup(int mp, int mc);
	unsigned long TDecimate::calcMetric(PVideoFrame &prev, PVideoFrame &curr, int np, 
		int &blockNI, int &xblocksI, unsigned long &metricF, IScriptEnvironment *env, bool scene);
	void TDecimate::calcMetricCycle(Cycle &current, IScriptEnvironment *env, int np, bool scene, bool hnt);
	void TDecimate::copyFrame(PVideoFrame &dst, PVideoFrame &src, IScriptEnvironment *env, int np);
	void TDecimate::findDupStrings(Cycle &p, Cycle &c, Cycle &n);
	int TDecimate::getHint(PVideoFrame &src, int &d2vfilm);
	void TDecimate::blendFrames(PVideoFrame &src1, PVideoFrame &src2, PVideoFrame &dst,
			double amount1, double amount2, int np, IScriptEnvironment *env);
	void TDecimate::calcBlendRatios(double &amount1, double &amount2, int &frame1, int &frame2, int n,
				int bframe, int cycleI);
	void TDecimate::drawBoxYUY2(PVideoFrame &dst, int blockN, int xblocks);
	void TDecimate::drawBoxYV12(PVideoFrame &dst, int blockN, int xblocks);
	void TDecimate::drawBox(PVideoFrame &dst, int blockN, int xblocks, int np);
	void TDecimate::DrawYUY2(PVideoFrame &dst, int x1, int y1, const char *s);
	void TDecimate::DrawYV12(PVideoFrame &dst, int x1, int y1, const char *s);
	void TDecimate::Draw(PVideoFrame &dst, int x1, int y1, const char *s, int np);
	PVideoFrame TDecimate::GetFrameMode01(int n, IScriptEnvironment *env, int np);
	PVideoFrame TDecimate::GetFrameMode2(int n, IScriptEnvironment *env, int np);
	PVideoFrame TDecimate::GetFrameMode3(int n, IScriptEnvironment *env, int np);
	PVideoFrame TDecimate::GetFrameMode4(int n, IScriptEnvironment *env, int np);
	PVideoFrame TDecimate::GetFrameMode5(int n, IScriptEnvironment *env, int np);
	void TDecimate::getOvrFrame(int n, unsigned long &metricU, unsigned long &metricF);
	void TDecimate::getOvrCycle(Cycle &current, bool mode2);
	void TDecimate::displayOutput(IScriptEnvironment* env, PVideoFrame &dst, int n, 
			int ret, bool film, double amount1, double amount2, int f1, int f2, int np);
	void TDecimate::formatMetrics(Cycle &current);
	void TDecimate::formatDups(Cycle &current);
	void TDecimate::formatDecs(Cycle &current);
	void TDecimate::formatMatches(Cycle &current);
	void TDecimate::formatMatches(Cycle &current, Cycle &previous);
	void TDecimate::debugOutput1(int n, bool film, int blend);
	void TDecimate::debugOutput2(int n, int ret, bool film, int f1, int f2, double amount1,
		double amount2);
	void TDecimate::addMetricCycle(Cycle &j);
	void TDecimate::blend_MMX_8(unsigned char* dstp, const unsigned char* srcp,  
			const unsigned char* nxtp, int width, int height, int dst_pitch, 
			int src_pitch, int nxt_pitch, double w1, double w2);
	void TDecimate::blend_SSE2_16(unsigned char* dstp, const unsigned char* srcp,  
			const unsigned char* nxtp, int width, int height, int dst_pitch, 
			int src_pitch, int nxt_pitch, double w1, double w2);
	void TDecimate::calcLumaDiffYUY2_ISSE_16(const unsigned char *prvp, const unsigned char *nxtp, 
		int width, int height, int prv_pitch, int nxt_pitch, unsigned long &diff);
	void TDecimate::calcLumaDiffYUY2_ISSE_8(const unsigned char *prvp, const unsigned char *nxtp, 
		int width, int height, int prv_pitch, int nxt_pitch, unsigned long &diff);
	bool TDecimate::checkForObviousDecFrame(Cycle &p, Cycle &c, Cycle &n);
	void TDecimate::mostSimilarDecDecision(Cycle &p, Cycle &c, Cycle &n);
	unsigned long TDecimate::calcLumaDiffYUY2(const unsigned char *prvp, const unsigned char *nxtp,
		int width, int height, int prv_pitch, int nxt_pitch, IScriptEnvironment *env);
	int TDecimate::checkForD2VDecFrame(Cycle &p, Cycle &c);
	bool TDecimate::checkForTwoDropLongestString(Cycle &p, Cycle &c, Cycle &n);
	int TDecimate::getNonDecMode2(int n, int start, int stop);
	void TDecimate::buildDecStrategy(IScriptEnvironment *env);
	void TDecimate::mode2MarkDecFrames(int cycleF);
	void TDecimate::removeMinN(int m, int n, int start, int stop);
	void TDecimate::removeMinN(int m, int n, unsigned long *metricsT, int *orderT, int &ovrC);
	int TDecimate::findDivisor(double decRatio, int min_den);
	int TDecimate::findNumerator(double decRatio, int divisor);
	void TDecimate::findCorrectionFactors(double decRatio, int num, int den, int rc[10], IScriptEnvironment *env);
	void TDecimate::sortMetrics(unsigned long *metrics, int *order, int length);
	void TDecimate::SedgeSort(unsigned long *metrics, int *order, int length);
	void TDecimate::pQuickerSort(unsigned long *metrics, int *order, int lower, int upper);

public:
	PVideoFrame __stdcall TDecimate::GetFrame(int n, IScriptEnvironment *env);
	TDecimate::TDecimate(PClip _child, int _mode, int _cycleR, int _cycle, double _rate, 
		double _dupThresh, double _vidThresh, double _sceneThresh, int _hybrid,
		int _vidDetect, int _conCycle, int _conCycleTP, const char* _ovr, 
		const char* _output, const char* _input, const char* _tfmIn, const char* _mkvOut,
		int _nt, int _blockx, int _blocky, bool _debug, bool _display, int _vfrDec, 
		bool _batch, bool _tcfv1, bool _se, bool _chroma, bool _exPP, int _maxndl, 
		bool _m2PA, IScriptEnvironment* env);
	TDecimate::~TDecimate();
};