/*
**                    TIVTC v0.9.9 for AviSynth 2.5.x
**
**   TIVTC includes a field matching filter (TFM) and a decimation
**   filter (TDecimate) which can be used together to achieve an
**   IVTC or for other uses. TIVTC currently supports YV12 and 
**   YUY2 colorspaces.
**   
**   Copyright (C) 2004-2005 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "TDecimate.h"

__declspec(align(16)) const __int64 lumamask = 0x00FF00FF00FF00FF;

// Leak's mmx blend routine
void TDecimate::blend_MMX_8(unsigned char* dstp, const unsigned char* srcp,  
			const unsigned char* nxtp, int width, int height, int dst_pitch, 
			int src_pitch, int nxt_pitch, double w1, double w2)
{
	int iw1t=(int)(w1*65536.0); iw1t+=(iw1t << 16);
	__int64 iw1=(__int64)iw1t; iw1+=(iw1 << 32);
	int iw2t=(int)(w2*65536.0); iw2t+=(iw2t << 16);
	__int64 iw2=(__int64)iw2t; iw2+=(iw2 << 32);
	__asm
	{
 		mov ebx,height
		mov eax,width
 		mov esi,srcp
 		mov edi,dstp
 		mov edx,nxtp
		add esi,eax
		add edi,eax
		add edx,eax
		neg eax
 		movq mm6,iw1
 		movq mm7,iw2
		mov ecx,eax
 		align 16
 xloop:
		movq mm0,[esi+ecx]
		movq mm1,[edx+ecx]
		movq mm2,mm0
		movq mm3,mm1
		punpcklbw mm0,mm0
		punpcklbw mm1,mm1
		pmulhuw mm0,mm6
		punpckhbw mm2,mm2
		pmulhuw mm1,mm7
		punpckhbw mm3,mm3
		paddusw mm0,mm1
		pmulhuw mm2,mm6
		pmulhuw mm3,mm7
 		psrlw mm0,8
		paddusw mm2,mm3
		psrlw mm2,8
		add ecx,8
		packuswb mm0,mm2
		movq [edi+ecx-8],mm0
		jnz xloop
 		add edi,dst_pitch
 		add esi,src_pitch
 		add edx,nxt_pitch
 		dec ebx
		mov ecx,eax
 		jnz xloop
 		emms
	}
}

// Leak's sse2 blend routine
void TDecimate::blend_SSE2_16(unsigned char* dstp, const unsigned char* srcp,  
			const unsigned char* nxtp, int width, int height, int dst_pitch, 
			int src_pitch, int nxt_pitch, double w1, double w2)
{
	int iw1t=(int)(w1*65536.0); iw1t+=(iw1t << 16);
	__int64 iw1t2=(__int64)iw1t; iw1t2+=(iw1t2 << 32);
	int iw2t=(int)(w2*65536.0); iw2t+=(iw2t << 16);
	__int64 iw2t2=(__int64)iw2t; iw2t2+=(iw2t2 << 32);
	__int64 iw1[]={ iw1t2, iw1t2 };
	__int64 iw2[]={ iw2t2, iw2t2 };
	__asm
	{
 		mov ebx,height
		mov eax,width
 		mov esi,srcp
 		mov edi,dstp
 		mov edx,nxtp
		add esi,eax
		add edi,eax
		add edx,eax
		neg eax
 		movdqu xmm6,iw1
 		movdqu xmm7,iw2
		mov ecx,eax
 		align 16
 xloop:
		movdqu xmm0,[esi+ecx]
		movdqu xmm1,[edx+ecx]
		movdqa xmm2,xmm0
		movdqa xmm3,xmm1
		punpcklbw xmm0,xmm0
		punpcklbw xmm1,xmm1
		pmulhuw xmm0,xmm6
		punpckhbw xmm2,xmm2
		pmulhuw xmm1,xmm7
		punpckhbw xmm3,xmm3
		paddusw xmm0,xmm1
		pmulhuw xmm2,xmm6
		pmulhuw xmm3,xmm7
 		psrlw xmm0,8
		paddusw xmm2,xmm3
		psrlw xmm2,8
		add ecx,16
		packuswb xmm0,xmm2
		movdqu [edi+ecx-16],xmm0
		jnz xloop
 		add edi,dst_pitch
 		add esi,src_pitch
 		add edx,nxt_pitch
 		dec ebx
		mov ecx,eax
 		jnz xloop
 		emms
	}
}

void TDecimate::calcLumaDiffYUY2_ISSE_16(const unsigned char *prvp, const unsigned char *nxtp, 
	int width, int height, int prv_pitch, int nxt_pitch, unsigned long &diff)
{
	__asm
	{
		mov edi,prvp
 		mov esi,nxtp
 		mov ecx,width
		mov edx,prv_pitch
		mov ebx,nxt_pitch
		movq mm4,lumamask
		movq mm5,lumamask
		pxor mm6,mm6
		pxor mm7,mm7
yloop:
		xor eax,eax
 		align 16
xloop:
		movq mm0,[edi+eax]
		movq mm2,[edi+eax+8]
		movq mm1,[esi+eax]
		movq mm3,[esi+eax+8]
		pand mm0,mm4
		pand mm1,mm5
		pand mm2,mm4
		pand mm3,mm5
		psadbw mm0,mm1
		psadbw mm2,mm3
		paddd mm6,mm0
		paddd mm7,mm2
		add eax,16
		cmp eax,ecx
		jl xloop
 		add edi,edx
 		add esi,ebx
 		dec height
 		jnz yloop
		paddd mm6,mm7
		mov eax,diff
		movd [eax],mm6
 		emms
	}
}

void TDecimate::calcLumaDiffYUY2_ISSE_8(const unsigned char *prvp, const unsigned char *nxtp, 
	int width, int height, int prv_pitch, int nxt_pitch, unsigned long &diff)
{
	__asm
	{
		mov edi,prvp
 		mov esi,nxtp
 		mov ecx,width
		mov edx,prv_pitch
		mov ebx,nxt_pitch
		movq mm2,lumamask
		movq mm3,lumamask
		pxor mm4,mm4
yloop:
		xor eax,eax
 		align 16
xloop:
		movq mm0,[edi+eax]
		movq mm1,[esi+eax]
		pand mm0,mm2
		pand mm1,mm3
		psadbw mm0,mm1
		paddd mm4,mm0
		add eax,8
		cmp eax,ecx
		jl xloop
 		add edi,edx
 		add esi,ebx
 		dec height
 		jnz yloop
		mov eax,diff
		movd [eax],mm4
 		emms
	}
}