/*
**                    TIVTC v0.9.9 for AviSynth 2.5.x
**
**   TIVTC includes a field matching filter (TFM) and a decimation
**   filter (TDecimate) which can be used together to achieve an
**   IVTC or for other uses. TIVTC currently supports YV12 and 
**   YUY2 colorspaces.
**   
**   Copyright (C) 2004-2005 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "TDecimate.h"

PVideoFrame TDecimate::GetFrameMode2(int n, IScriptEnvironment *env, int np)
{
	int ret, cycleF, x;
	if (mode2_numCycles >= 0)
	{
		for (x=0; x<mode2_numCycles; ++x)
		{
			if (aLUT[x*5+1] <= n && aLUT[x*5+3] > n) 
			{ 
				cycleF = x; 
				break; 
			}
		}
		if (cycleF > 0 && prev.frame != aLUT[(cycleF-1)*5])
		{
			if (curr.frame == aLUT[(cycleF-1)*5]) prev = curr;
			else
			{
				prev.setFrame(aLUT[(cycleF-1)*5]);
				getOvrCycle(prev, true);
				calcMetricCycle(prev, env, np, true, false);
				addMetricCycle(prev);
			}
		}
		else if (cycleF <= 0) prev.setFrame(-prev.length);
		if (curr.frame != aLUT[cycleF*5])
		{
			if (next.frame == aLUT[cycleF*5]) curr = next;
			else
			{
				curr.setFrame(aLUT[cycleF*5]);
				getOvrCycle(curr, true);
				calcMetricCycle(curr, env, np, true, false);
				addMetricCycle(curr);
			}
		}
		if (cycleF < mode2_numCycles-1 && next.frame != aLUT[(cycleF+1)*5])
		{
			next.setFrame(aLUT[(cycleF+1)*5]);
			getOvrCycle(next, true);
			calcMetricCycle(next, env, np, true, false);
			addMetricCycle(next);
		}
		else if (cycleF >= mode2_numCycles-1) next.setFrame(-next.length);
		mode2MarkDecFrames(cycleF);
		ret = getNonDecMode2(n-aLUT[cycleF*5+1], aLUT[cycleF*5], aLUT[cycleF*5+2]);
		if (ret < 0) env->ThrowError("TDecimate:  mode 2 error, ret less then 0!");
	}
	else ret = aLUT[n];
	if (debug)
	{
		sprintf(buf,"TDecimate:  inframe = %d  useframe = %d  rate = %3.3f\n", n, ret, rate);
		OutputDebugString(buf);
	}
	if (display)
	{
		PVideoFrame dst = env->NewVideoFrame(vi);
		copyFrame(dst, child->GetFrame(ret, env), env, np);
		sprintf(buf, "TDecimate %s by tritical", VERSION);
		Draw(dst, 0, 0, buf, np);
		sprintf(buf, "Mode: 2  Rate = %3.3f", rate);
		Draw(dst, 0, 1, buf, np);
		sprintf(buf,"inframe = %d  useframe = %d", n, ret);
		Draw(dst, 0, 2, buf, np);
		return dst;
	}
	return child->GetFrame(ret, env);
}

int TDecimate::getNonDecMode2(int n, int start, int stop)
{
	int i, count, ret;
	for (count=-1,ret=-1,i=start; i<stop; ++i)
	{
		if (mode2_decA[i] == 0) ++count;
		if (count == n) { ret = i; break; }
	}
	return ret;
}

void TDecimate::mode2MarkDecFrames(int cycleF)
{
	for (int i=curr.frame; i<curr.frameEO; ++i)
	{
		if (mode2_decA[i] != -20) return;
		mode2_decA[i] = 0;
	}
	removeMinN(mode2_num, mode2_den, curr.frame, curr.frameEO);
	for (int x=0; x<10; ++x)
	{
		if (mode2_cfs[x] <= 0) break;
		if (aLUT[cycleF*5+4]&(1<<x)) 
		{
			if (mode2_cfs[x] <= curr.length) removeMinN(1, mode2_cfs[x], curr.frame, curr.frameEO);
			else removeMinN(1, curr.length, curr.frame, curr.frameEO);
		}
	}
}

void TDecimate::removeMinN(int m, int n, int start, int stop)
{
	int i, x, v, dec, t, b, stop2;
	double pM, nM, cM;
	for (x=start; x<stop; x+=n)
	{
		dec = t = 0;
		stop2 = n;
		if (x+n-1 > nfrms) 
		{
			if (m <= 1) continue;
			else
			{
				m = (int)(double(m)*(double(nfrms-x+1)/double(n)));
				if (m < 1) continue;
				stop2 = nfrms-x+1;
			}
		}
		if (curr.dupCount > 0)
		{
			for (b=x-start, i=0; i<stop2; ++i, ++b)
			{
				if (curr.dupArray[b] == 1 && dec<m) 
				{
					mode2_decA[x+i] = 1;
					--curr.dupCount;
					curr.dupArray[b] = 0;
					++dec;
				}
			}
			if (dec >= m) continue;
		}
		for (i=0; i<stop2; ++i)
		{
			if (mode2_decA[x+i] == 0)
			{
				v = 1;
				cM = (metricsOutArray[(x+i)<<1]*100.0f) / MAX_DIFF;
				pM = nM = -20.0;
				while (pM < 0 || nM < 0)
				{
					if (pM < 0)
					{
						if (x+i-v >= 0)
						{
							if (mode2_decA[x+i-v] == 0 || mode2_decA[x+i-v] == -20) 
								pM = (metricsOutArray[(x+i-v)<<1]*100.0f)/MAX_DIFF;
						}
						else pM = 1.0f;
					}
					if (nM < 0)
					{
						if (x+i+v <= nfrms)
						{
							if (mode2_decA[x+i+v] == 0 || mode2_decA[x+i+v] == -20) 
								nM = (metricsOutArray[(x+i+v)<<1]*100.0f)/MAX_DIFF;
						}
						else nM = 1.0f;
					}
					++v;
				}
				if (pM >= 3.0 && nM >= 3.0 && cM < 3.0 && pM*0.5f > cM && nM*0.5f > cM)
				{
					mode2_order[t] = i;
					mode2_metrics[t] = (int)(min(pM-cM,nM-cM)*10000.0f+0.5f);
					++t;
				}
			}
		}
		if (t > 0)
		{
			if (t < 150) sortMetrics(mode2_metrics, mode2_order, t);
			else SedgeSort(mode2_metrics, mode2_order, t);
			for (i=0; i<t && dec<m; ++i)
			{
				if (mode2_decA[x+mode2_order[t-1-i]] != 1)
				{
					mode2_decA[x+mode2_order[t-1-i]] = 1;
					++dec;
				}
			}
		}
		if (dec >= m) continue;
		for (i=0; i<stop2; ++i) 
		{
			mode2_order[i] = i;
			mode2_metrics[i] = metricsOutArray[(x+i)<<1];
		}
		if (stop2 < 150) sortMetrics(mode2_metrics, mode2_order, n);
		else SedgeSort(mode2_metrics, mode2_order, n);
		for (i=0; i<stop2 && dec<m; ++i)
		{
			if (mode2_decA[x+mode2_order[i]] != 1)
			{
				mode2_decA[x+mode2_order[i]] = 1;
				++dec;
			}
		}
	}
}

void TDecimate::removeMinN(int m, int n, unsigned long *metricsT, int *orderT, int &ovrC)
{
	int i, x, v, dec, t, stop2;
	double pM, nM, cM;
	int start = 0, stop = vi.num_frames;
	for (x=start; x<stop; x+=n)
	{
		dec = t = 0;
		stop2 = n;
		if (x+n-1 > nfrms) 
		{
			if (m <= 1) continue;
			else 
			{
				m = (int)(double(m)*(double(nfrms-x+1)/double(n)));
				if (m < 1) continue;
				stop2 = nfrms-x+1;
			}
		}
		if (ovrC > 0 && ovrArray != NULL)
		{
			for (i=0; i<stop2; ++i)
			{
				if ((ovrArray[x+i]&DROP_FRAME) && dec<m) 
				{
					mode2_decA[x+i] = 1;
					--ovrC;
					ovrArray[x+i] &= ~DROP_FRAME;
					++dec;
				}
			}
			if (dec >= m) continue;
		}
		for (i=0; i<stop2; ++i)
		{
			if (mode2_decA[x+i] == 0)
			{
				v = 1;
				cM = (metricsArray[(x+i)<<1]*100.0f) / MAX_DIFF;
				pM = nM = -20.0;
				while (pM < 0 || nM < 0)
				{
					if (pM < 0)
					{
						if (x+i-v >= 0)
						{
							if (mode2_decA[x+i-v] != 1) 
								pM = (metricsArray[(x+i-v)<<1]*100.0f)/MAX_DIFF;
						}
						else pM = 1.0f;
					}
					if (nM < 0)
					{
						if (x+i+v <= nfrms)
						{
							if (mode2_decA[x+i+v] != 1) 
								nM = (metricsArray[(x+i+v)<<1]*100.0f)/MAX_DIFF;
						}
						else nM = 1.0f;
					}
					++v;
				}
				if (pM >= 3.0 && nM >= 3.0 && cM < 3.0 && pM*0.5f > cM && nM*0.5f > cM)
				{
					orderT[t] = i;
					metricsT[t] = (int)(min(pM-cM,nM-cM)*10000.0f+0.5f);
					++t;
				}
			}
		}
		if (t > 0)
		{
			if (t < 150) sortMetrics(metricsT, orderT, t);
			else SedgeSort(metricsT, orderT, t);
			for (i=0; i<t && dec<m; ++i)
			{
				if (mode2_decA[x+orderT[t-1-i]] != 1)
				{
					mode2_decA[x+orderT[t-1-i]] = 1;
					++dec;
				}
			}
		}
		if (dec >= m) continue;
		for (i=0; i<stop2; ++i) 
		{
			orderT[i] = i;
			metricsT[i] = metricsArray[(x+i)<<1];
		}
		if (stop2 < 150) sortMetrics(metricsT, orderT, stop2);
		else SedgeSort(metricsT, orderT, stop2);
		for (i=0; i<stop2 && dec<m; ++i)
		{
			if (mode2_decA[x+orderT[i]] != 1)
			{
				mode2_decA[x+orderT[i]] = 1;
				++dec;
			}
		}
	}
}

void TDecimate::sortMetrics(unsigned long *metrics, int *order, int length)
{
	int i, j, temp2;
	unsigned long temp1;
	for (i=1; i<length; ++i) 
	{
		j = i;
		temp1 = metrics[j];
		temp2 = order[j];
		while (j>0 && metrics[j-1]>temp1) 
		{
			metrics[j] = metrics[j-1];
			order[j] = order[j-1];
			--j;
		}
		metrics[j] = temp1;
		order[j] = temp2;
	}
}

void TDecimate::SedgeSort(unsigned long *metrics, int *order, int length)
{
    pQuickerSort(metrics, order, 0, length-1);
    sortMetrics(metrics, order, length);
}

void TDecimate::pQuickerSort(unsigned long *metrics, int *order, int lower, int upper)
{
    if (upper-lower > 15) 
	{
		int	i, j, temp2;
		unsigned long pivot, temp1;
		temp1 = metrics[lower];
		metrics[lower] = metrics[(upper+lower)>>1];
		metrics[(upper+lower)>>1] = temp1;
		temp2 = order[lower];
		order[lower] = order[(upper+lower)>>1];
		order[(upper+lower)>>1] = temp2;
		i = lower;  
		j = upper+1;  
		pivot = metrics[lower];
		while (1) 
		{
			do i++; while (metrics[i]<pivot);
			do j--; while (metrics[j]>pivot);
			if (j<i) break;
			temp1 = metrics[i];
			metrics[i] = metrics[j];
			metrics[j] = temp1;
			temp2 = order[i];
			order[i] = order[j];
			order[j] = temp2;
		}
		temp1 = metrics[lower];
		metrics[lower] = metrics[j];
		metrics[j] = temp1;
		temp2 = order[lower];
		order[lower] = order[j];
		order[j] = temp2;
		pQuickerSort(metrics, order, lower, j-1);
		pQuickerSort(metrics, order, i, upper);
    }
}

int TDecimate::findDivisor(double decRatio, int min_den)
{
	int ret = -20;
	double num = 1.0f, lowest = 5.0f, temp;
	double offset = 0.00000001;
	for (int x=min_den; x<=100; ++x)
	{
		temp = num/((double)x);
		if (temp > decRatio+offset) continue;
		if (fabs(temp-decRatio) < lowest)
		{
			lowest = fabs(temp-decRatio);
			ret = x;
		}
	}
	return ret;
}

int TDecimate::findNumerator(double decRatio, int divisor)
{
	int ret = -20;
	double den = (double)divisor, lowest = 5.0f, temp;
	double offset = 0.00000001;
	for (int x=1; x<divisor; ++x)
	{
		temp = ((double)x)/den;
		if (temp > decRatio+offset) continue;
		if (fabs(temp-decRatio) < lowest)
		{
			lowest = fabs(temp-decRatio);
			ret = x;
		}
	}
	return ret;
}

void TDecimate::findCorrectionFactors(double decRatio, int num, int den, int rc[10], IScriptEnvironment *env)
{
	double approx = ((double)num)/((double)den), error, length;
	memset(rc, 0, 10*sizeof(int));
	for (int x=0; x<10; ++x)
	{
		error = decRatio-approx;
		if (error <= 0.0f) break;
		length = 1.0f/error;
		if (length > vi.num_frames) break;
		rc[x] = ((int)((length/((double)(x == 0 ? den : rc[x-1]))) + 0.9999999f))*(x == 0 ? den : rc[x-1]);
		if (rc[x] > vi.num_frames) rc[x] = vi.num_frames;
		approx += 1.0f/((double)rc[x]);
	}
	if ((1.0f/fabs(decRatio-approx)) < vi.num_frames)
		env->ThrowError("TDecimate:  mode 2 error, unable to achieve a completely synced result!");
}

void TDecimate::buildDecStrategy(IScriptEnvironment *env)
{
	double frRatio = fps/rate;
	double rfRatio = rate/fps;
	double max = 99.0f/100.f, min = 1.0f/100.0f;
	if (rfRatio >= max || rfRatio <= min)
		env->ThrowError("TDecimate:  mode 2 error, unable to achieve desired decimation ratio!");
	double decRatio = 1.0f-rfRatio;
	if (frRatio < 3.0f)
	{
		mode2_num = 1;
		mode2_den = findDivisor(decRatio, maxndl > 0 ? maxndl < 99 ? maxndl+1 : 2 : 2);
	}
	else
	{
		mode2_den = (int)frRatio;
		mode2_num = findNumerator(decRatio, mode2_den);
		if (maxndl > 0 && maxndl < 99 && mode2_num-mode2_den < maxndl) mode2_den = mode2_num+maxndl;
	}
	if (mode2_den <= 0 || mode2_num <= 0 || mode2_num > 100 || mode2_den > 100 || mode2_num >= mode2_den)
		env->ThrowError("TDecimate:  mode 2 invalid num and den results!");
	int clength = mode2_den, cdrop = 0, add, rc[10], arc[10];
	int i, x, temp, dropCount = 0, rstart = -20;
	findCorrectionFactors(decRatio, mode2_num, mode2_den, rc, env);
	for (x=0; x<10; ++x)
	{
		if (rc[x] > 0 && (rc[x] <= 100 || m2PA) && rc[x] > clength) clength = rc[x];
	}
	if (clength == mode2_den && rc[0] > 0 && (mode2_den-mode2_num <= 1 || mode2_den <= 25))
	{
		while (clength <= 50) clength *= 2;
	}
	cdrop += ((int)(clength/mode2_den))*mode2_num;
	for (x=0; x<10; ++x)
	{
		if (rc[x] > 0 && rc[x] > clength) { rstart = x; break; }
		if (rc[x] > 0 && rc[x] <= clength) cdrop += (int)(clength/rc[x]);
	}
	if (rstart == -20) rstart = 11;
	mode2_numCycles = (int)((((double)vi.num_frames)/((double)clength))+1.0f);
	bool allMetrics = true;
	if (metricsArray != NULL)
	{
		for (int h=0; h<vi.num_frames*2; h+=2) 
		{
			if (metricsArray[h] == ULONG_MAX) { allMetrics = false; break; }
		}
	}
	else allMetrics = false;
	if (aLUT != NULL) free(aLUT);
	if (allMetrics)
	{
		aLUT = (int *)malloc(((int)(vi.num_frames*rate/fps))*sizeof(int));
		if (aLUT == NULL) env->ThrowError("TDecimate:  mode 2 error, malloc failure (aLUT)!");
		memset(aLUT, 0, ((int)(vi.num_frames*rate/fps))*sizeof(int));
		int *orderT = (int *)malloc(vi.num_frames*sizeof(int));
		if (orderT == NULL) env->ThrowError("TDecimate:  mode 2 error, malloc failure (orderT)!");
		memset(orderT, 0, vi.num_frames*sizeof(int));
		unsigned long *metricsT = (unsigned long *)malloc(vi.num_frames*sizeof(unsigned long));
		if (metricsT == NULL) env->ThrowError("TDecimate:  mode 2 error, malloc failure (metricsT)!");
		memset(metricsT, 0, vi.num_frames*sizeof(unsigned long));
		memset(mode2_decA, 0, vi.num_frames*sizeof(int));
		int ovrC = 0;
		if (ovrArray != NULL)
		{
			for (i=0; i<vi.num_frames; ++i)
			{
				if (ovrArray[i]&DROP_FRAME) ++ovrC;
			}
		}
		removeMinN(mode2_num, mode2_den, metricsT, orderT, ovrC);
		for (x=0; x<10; ++x)
		{
			if (rc[x] > 0) removeMinN(1, rc[x], metricsT, orderT, ovrC);
		}
		int v = 0, tc = (int)(vi.num_frames*rate/fps);
		for (i=0; i<vi.num_frames && v<tc; ++i)
		{
			if (mode2_decA[i] != 1)
			{
				aLUT[v] = i;
				++v;
			}
		}
		free(orderT);
		free(metricsT);
		free(mode2_decA);
		mode2_decA = NULL;
		if (v != (int)(vi.num_frames * (rate/fps)))
			env->ThrowError("TDecimate:  mode 2 error, number of frames after decimation doesn't match!");
		mode2_numCycles = -20;
	}
	else
	{
		aLUT = (int *)malloc(mode2_numCycles*5*sizeof(int));
		memset(aLUT, 0, mode2_numCycles*5*sizeof(int));
		temp = 0;
		for (x=0; x<10; ++x) 
		{
			if (rc[x] > 0 && rc[x] <= clength) temp |= (1<<x);
		}
		for (x=0; x<mode2_numCycles; ++x) aLUT[x*5+4] = temp;
		memset(arc, 0, 10*sizeof(int));
		for (x=0; x<mode2_numCycles; ++x)
		{
			temp = x*5;
			add = x*clength+clength-vi.num_frames >= 0 ? vi.num_frames-x*clength : clength;
			aLUT[temp+0] = x*clength;
			aLUT[temp+1] = x*clength-dropCount;
			aLUT[temp+2] = x*clength+add;
			dropCount += cdrop;
			for (i=rstart; i<10; ++i)
			{
				if (rc[i] > 0 && arc[i]+add > rc[i])
				{
					++dropCount;
					arc[i] = 0;
					aLUT[temp+4] |= (1<<i);
				}
				else if (rc[i] > 0) arc[i] += add;
			}
			aLUT[temp+3] = add == clength ? x*clength+add-dropCount : (int)(vi.num_frames*rate/fps);
		}
		if (clength > DEFAULT_CYCLE_SIZE)
		{
			prev.setSize(clength);
			curr.setSize(clength);
			next.setSize(clength);
		}
		prev.length = curr.length = next.length = clength;
	}
	memcpy(mode2_cfs,rc,10*sizeof(int));
	if (debug)
	{
		sprintf(buf,"mode2_num = %d  mode2_den = %d  numCycles = %d  clength = %d\n", mode2_num, mode2_den, mode2_numCycles, clength);
		OutputDebugString(buf);
		for (x=0; x<10; ++x)
		{
			if (mode2_cfs[x] <= 0) break;
			sprintf(buf,"mode2_cfs %d = %d\n", x, mode2_cfs[x]);
			OutputDebugString(buf);
		}
	}
}