/*
**                    TIVTC v0.9.9 for AviSynth 2.5.x
**
**   TIVTC includes a field matching filter (TFM) and a decimation
**   filter (TDecimate) which can be used together to achieve an
**   IVTC or for other uses. TIVTC currently supports YV12 and 
**   YUY2 colorspaces.
**   
**   Copyright (C) 2004-2005 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "TFM.h"

PVideoFrame __stdcall TFM::GetFrame(int n, IScriptEnvironment* env)
{
	if (n < 0) n = 0;
	else if (n > nfrms) n = nfrms;
	PVideoFrame prv = child->GetFrame(n > 0 ? n-1 : 0, env);
	PVideoFrame src = child->GetFrame(n, env);
	PVideoFrame nxt = child->GetFrame(n < nfrms ? n+1 : nfrms, env);
	PVideoFrame dst = env->NewVideoFrame(vi);
	PVideoFrame tmp = env->NewVideoFrame(vi);
	int mmatch1, nmatch1, nmatch2, mmatch2, fmatch, tmatch;
	int blockN = -20, xblocks = -20;
	int tblockN = -20, txblocks = -20;
	int combed = -1, tcombed = -1;
	int MICount = -1, tMICount = -1;
	bool d2vfilm = false;
	int np = vi.IsYV12() ? 3 : 1;
	order = orderS; 
	mode = modeS; 
	field = fieldS;
	PP = PPS;
	getSettingOvr(n);
	if (order == -1) order = child->GetParity(n) ? 1 : 0;
	if (field == -1) field = order;
	int frstT = field^order ? 2 : 0;
	int scndT = field^order ? 0 : 2;
	if (debug)
	{
		sprintf(buf,"TFM:  ----------------------------------------\n");
		OutputDebugString(buf);
	}
	if (getMatchOvr(n, fmatch, combed))
	{
		createWeaveFrame(dst, prv, src, nxt, env, fmatch, np);
		if (PP > 0 && combed == -1) 
		{
			if (checkCombed(dst, n, env, np, fmatch, blockN, xblocks, MICount)) combed = 2;
			else combed = 0;
		}
		d2vfilm = d2vduplicate(fmatch, combed, n);
		fileOut(fmatch, combed, d2vfilm, n);
		if (display) writeDisplay(dst, np, n, fmatch, combed, true, blockN, xblocks, MICount);
		if (debug)
		{
			char buft[20];
			if (MICount < 0) sprintf(buft,"N/A");
			else sprintf(buft,"%d",MICount);
			sprintf(buf,"TFM:  frame %d  - final match = %c  MIC = %s  (OVR)\n", n, MTC(fmatch), buft);
			OutputDebugString(buf);
			sprintf(buf,"TFM:  mode = %d  field = %d  order = %d  d2vfilm = %c\n", mode, field, order,
				d2vfilm ? 'T' : 'F');
			OutputDebugString(buf);
			if (combed != -1)
			{
				if (combed == 1) sprintf(buf,"TFM:  CLEAN FRAME  (forced!)\n");
				else if (combed == 5) sprintf(buf,"TFM:  COMBED FRAME  (forced!)\n");
				else if (combed == 0) sprintf(buf,"TFM:  CLEAN FRAME\n");
				else sprintf(buf,"TFM:  COMBED FRAME\n");
				OutputDebugString(buf);
			}
		}
		putHint(dst, fmatch, combed, d2vfilm);
		lastMatch.frame = n;
		lastMatch.match = fmatch;
		lastMatch.field = field;
		lastMatch.combed = combed;
		return dst;
	}
	if (!slow) fmatch = compareFields(prv, src, nxt, 1, frstT, nmatch1, nmatch2, mmatch1, mmatch2, np, n);
	else fmatch = compareFieldsSlow(prv, src, nxt, 1, frstT, nmatch1, nmatch2, mmatch1, mmatch2, np, n, env);
	createWeaveFrame(dst, prv, src, nxt, env, fmatch, np);
	if (mode > 2 || (mode > 0 && checkCombed(dst, n, env, np, fmatch, blockN, xblocks, MICount)))
	{
		if (mode < 3) tcombed = 2;
		if (!slow) tmatch = compareFields(prv, src, nxt, fmatch, scndT, nmatch1, nmatch2, mmatch1, mmatch2, np, n);
		else tmatch = compareFieldsSlow(prv, src, nxt, fmatch, scndT, nmatch1, nmatch2, mmatch1, mmatch2, np, n, env);
		if (tmatch == scndT) 
		{
			if (mode > 2)
			{
				fmatch = tmatch;
				createWeaveFrame(dst, prv, src, nxt, env, tmatch, np);
			}
			else
			{
				createWeaveFrame(tmp, prv, src, nxt, env, tmatch, np);			
				if (!checkCombed(tmp, n, env, np, tmatch, tblockN, txblocks, tMICount))
				{
					fmatch = tmatch;
					MICount = tMICount;
					blockN = tblockN;
					xblocks = txblocks;
					tcombed = 0;
					copyFrame(dst, tmp, env, np);
				}
			}
		}
		if ((mode == 2 && tcombed == 2) || (mode == 4 && checkCombed(dst, n, env, np, fmatch, blockN, xblocks, MICount)))
		{
			tcombed = 2;
			if (!slow) tmatch = compareFields(prv, src, nxt, 3, 4, nmatch1, nmatch2, mmatch1, mmatch2, np, n);
			else tmatch = compareFieldsSlow(prv, src, nxt, 3, 4, nmatch1, nmatch2, mmatch1, mmatch2, np, n, env);
			createWeaveFrame(tmp, prv, src, nxt, env, tmatch, np);
			if (!checkCombed(tmp, n, env, np, tmatch, tblockN, txblocks, tMICount))
			{
				fmatch = tmatch;
				MICount = tMICount;
				blockN = tblockN;
				xblocks = txblocks;
				tcombed = 0;
				copyFrame(dst, tmp, env, np);
			}
		}
		if (mode == 4 && tcombed == -1) tcombed = 0;
	}
	if ((mode == 1 || mode == 2) && tcombed == -1) tcombed = 0;
	if (combed == -1 && PP > 0) combed = tcombed;
	if (PP > 0 && combed == -1)
	{
		if (checkCombed(dst, n, env, np, fmatch, blockN, xblocks, MICount)) combed = 2;
		else combed = 0;
	}
	d2vfilm = d2vduplicate(fmatch, combed, n);
	fileOut(fmatch, combed, d2vfilm, n);
	if (display) writeDisplay(dst, np, n, fmatch, combed, false, blockN, xblocks, MICount);
	if (debug)
	{
		char buft[20];
		if (MICount < 0) sprintf(buft,"N/A");
		else sprintf(buft,"%d",MICount);
		sprintf(buf,"TFM:  frame %d  - final match = %c  MIC = %s\n", n, MTC(fmatch), buft);
		OutputDebugString(buf);
		sprintf(buf,"TFM:  mode = %d  field = %d  order = %d  d2vfilm = %c\n", mode, field, order,
			d2vfilm ? 'T' : 'F');
		OutputDebugString(buf);
		if (combed != -1)
		{
			if (combed == 1) sprintf(buf,"TFM:  CLEAN FRAME  (forced!)\n");
			else if (combed == 5) sprintf(buf,"TFM:  COMBED FRAME  (forced!)\n");
			else if (combed == 0) sprintf(buf,"TFM:  CLEAN FRAME\n");
			else sprintf(buf,"TFM:  COMBED FRAME\n");
			OutputDebugString(buf);
		}
	}
	putHint(dst, fmatch, combed, d2vfilm);
	lastMatch.frame = n;
	lastMatch.match = fmatch;
	lastMatch.field = field;
	lastMatch.combed = combed;
	return dst;
}

void TFM::writeDisplay(PVideoFrame &dst, int np, int n, int fmatch, int combed, bool over,
					   int blockN, int xblocks, int MICount)
{
	if (combed > 1 && PP > 1) return;
	if (combed > 1 && PP == 1 && blockN != -20)
	{
		if (np == 3) drawBoxYV12(dst, blockN, xblocks);
		else drawBoxYUY2(dst, blockN, xblocks);
	}
	sprintf(buf, "TFM %s by tritical ", VERSION);
	if (np == 3) DrawYV12(dst, 0, 0, buf);
	else DrawYUY2(dst, 0, 0, buf);
	sprintf(buf, "order = %d  field = %d  mode = %d ", order, field, mode);
	if (np == 3) DrawYV12(dst, 0, 1, buf);
	else DrawYUY2(dst, 0, 1, buf);
	if (!over) sprintf(buf, "frame: %d  match = %c ", n, MTC(fmatch));
	else sprintf(buf, "frame: %d  match = %c (OVR) ", n, MTC(fmatch));
	if (np == 3) DrawYV12(dst, 0, 2, buf);
	else DrawYUY2(dst, 0, 2, buf);
	if (combed != -1)
	{
		if (combed == 1) sprintf(buf, "PP = %d  CLEAN FRAME (forced!) ", PP);
		else if (combed == 5) sprintf(buf, "PP = %d  COMBED FRAME  (forced!) ", PP);
		else if (combed == 0) sprintf(buf, "PP = %d  CLEAN FRAME ", PP);
		else sprintf(buf, "PP = %d  COMBED FRAME ", PP);
		if (MICount >= 0) 
		{
			char buft[20];
			sprintf(buft, " MIC = %d ", MICount);
			strcat(buf,buft);
		}
		if (np == 3) DrawYV12(dst, 0, 3, buf);
		else DrawYUY2(dst, 0, 3, buf);
	}
}

void TFM::getSettingOvr(int n)
{
	if (setArray == NULL || setArraySize <= 0) return;
	for (int x=0; x<setArraySize; x+=4)
	{
		if (n >= setArray[x+1] && n <= setArray[x+2])
		{
			if (setArray[x] == 111) order = setArray[x+3];
			else if (setArray[x] == 102) field = setArray[x+3];
			else if (setArray[x] == 109) mode = setArray[x+3];
			else if (setArray[x] == 80) PP = setArray[x+3];
		}
	}
}

bool TFM::getMatchOvr(int n, int &match, int &combed)
{
	if (ovrArray == NULL || ovrArray[n] == 255) return false;
	int value = ovrArray[n], temp;
	temp = value&0x00000020;
	if (temp == 0 && PP > 0)
	{
		if (value&0x00000010) combed = 5;
		else combed = 1;
	}
	temp = value&0x00000007;
	if (temp >= 0 && temp <= 6)
	{
		match = temp;
		if (field != fieldO)
		{
			if (match == 0) match = 3;
			else if (match == 2) match = 4;
			else if (match == 3) match = 0;
			else if (match == 4) match = 2;
		}
		if (match == 5) { combed = 5; match = 1; field = 0; }
		else if (match == 6) { combed = 5; match = 1; field = 1; }
		return true;
	}
	return false;
}

bool TFM::d2vduplicate(int match, int combed, int n)
{
	if (d2vfilmarray == NULL || d2vfilmarray[n] == 0) return false;
	if (n-1 != lastMatch.frame)
		lastMatch.field = lastMatch.frame = lastMatch.combed = lastMatch.match = -20;
	if (d2vfilmarray[n] == 0x3) // indicates possible top field duplicate
	{
		if (lastMatch.field == 1)
		{
			if ((lastMatch.combed > 1 || lastMatch.match != 3) && field == 1 && 
				(match != 4 || combed > 1)) return true;
			else if ((lastMatch.combed > 1 || lastMatch.match != 3) && field == 0 && 
				combed < 2 && match != 2) return true;
		}
		else if (lastMatch.field == 0)
		{
			if (lastMatch.combed < 2 && lastMatch.match != 0 && field == 1 && 
				(match != 4 || combed > 1)) return true;
			else if (lastMatch.combed < 2 && lastMatch.match != 0 && field == 0 && 
				combed < 2 && match != 2) return true;
		}
		else
		{
			if (field == 1 && (match != 4 || combed > 1)) return true;
			else if (field == 0 && combed < 2 && match != 2) return true;
		}
	}
	else if (d2vfilmarray[n] == 0x1) // indicates possible bottom field duplicate
	{
		if (lastMatch.field == 1)
		{
			if (lastMatch.combed < 2 && lastMatch.match != 0 && field == 0 && 
				(match != 4 || combed > 1)) return true;
			else if (lastMatch.combed < 2 && lastMatch.match != 0 && field == 1 && 
				combed < 2 && match != 2) return true;
		}
		else if (lastMatch.field == 0)
		{
			if ((lastMatch.combed > 1 || lastMatch.match != 3) && field == 0 && 
				(match != 4 || combed > 1)) return true;
			else if ((lastMatch.combed > 1 || lastMatch.match != 3) && field == 1 && 
				combed < 2 && match != 2) return true;
		}
		else
		{
			if (field == 0 && (match != 4 || combed > 1)) return true;
			else if (field == 1 && combed < 2 && match != 2) return true;
		}
	}
	return false;
}

void TFM::fileOut(int match, int combed, bool d2vfilm, int n)
{
	if (outArray == NULL) return;
	if (*output || *outputC)
	{
		if (field != fieldO)
		{
			if (match == 0) match = 3;
			else if (match == 2) match = 4;
			else if (match == 3) match = 0;
			else if (match == 4) match = 2;
		}
		if (match == 1 && combed > 1 && field == 0) match = 5;
		else if (match == 1 && combed > 1  && field == 1) match = 6;
		unsigned char hint = 0;
		hint |= match;
		if (combed > 1) hint |= FILE_COMBED;
		else if (combed >= 0) hint |= FILE_NOTCOMBED;
		if (d2vfilm) hint |= FILE_D2V;
		hint |= FILE_ENTRY;
		outArray[n] = hint;
	}
}

bool TFM::checkCombed(PVideoFrame &src, int n, IScriptEnvironment *env, int np, int match,
					  int &blockN, int &xblocksi, int &MICount)
{
	if (np == 1) return checkCombedYUY2(src, n, env, match, blockN, xblocksi, MICount);
	else if (np == 3) return checkCombedYV12(src, n, env, match, blockN, xblocksi, MICount);
	else env->ThrowError("TFM:  an unknown error occured (unknown colorspace)!");
	return false;
}

void TFM::copyFrame(PVideoFrame &dst, PVideoFrame &src, IScriptEnvironment *env, int np)
{
	int b, plane;
	for (b=0; b<np; ++b)
	{
		if (b == 0) plane = PLANAR_Y;
		else if (b == 1) plane = PLANAR_V;
		else plane = PLANAR_U;
		env->BitBlt(dst->GetWritePtr(plane), dst->GetPitch(plane), src->GetReadPtr(plane),
			src->GetPitch(plane), src->GetRowSize(plane), src->GetHeight(plane));
	}
}

int TFM::compareFields(PVideoFrame &prv, PVideoFrame &src, PVideoFrame &nxt, int match1, 
		int match2, int &norm1, int &norm2, int &mtn1, int &mtn2, int np, int n) 
{
	int b, plane, ret;
	const unsigned char *prvp, *srcp, *nxtp;
	const unsigned char *curpf, *curf, *curnf;
	const unsigned char *prvpf, *prvnf, *nxtpf, *nxtnf;
	int prv_pitch, src_pitch, Width, Height, nxt_pitch;
	int prvf_pitch, nxtf_pitch, curf_pitch, stopx, y0a, y1a;
	int x, y, temp1, temp2, diff1, diff2, startx;
	int inc = np == 3 ? 1 : mChroma ? 1 : 2;
	int stop = np == 3 ? mChroma ? 3 : 1 : 1;
	unsigned long accumPc = 0, accumNc = 0, accumPm = 0, accumNm = 0;
	norm1 = norm2 = mtn1 = mtn2 = 0;
	for (b=0; b<stop; ++b)
	{
		if (b == 0) plane = PLANAR_Y;
		else if (b == 1) plane = PLANAR_V;
		else plane = PLANAR_U;
		prvp = prv->GetReadPtr(plane);
		prv_pitch = prv->GetPitch(plane);
		srcp = src->GetReadPtr(plane);
		src_pitch = src->GetPitch(plane);
		Width = src->GetRowSize(plane);
		Height = src->GetHeight(plane);
		nxtp = nxt->GetReadPtr(plane);
		nxt_pitch = nxt->GetPitch(plane);
		startx = np > 1 ? (b == 0 ? 8 : 4) : 16;
		stopx = Width - startx;
		curf_pitch = src_pitch<<1;
		if (match1 < 3) curf = srcp + ((3-field)*src_pitch);
		if (b == 0) { y0a = y0; y1a = y1; }
		else { y0a = y0>>1; y1a = y1>>1; }
		if (match1 == 0)
		{
			prvf_pitch = prv_pitch<<1;
			prvpf = prvp + ((field == 1 ? 1 : 2)*prv_pitch);
		}
		else if (match1 == 1)
		{
			prvf_pitch = src_pitch<<1;
			prvpf = srcp + ((field == 1 ? 1 : 2)*src_pitch);
		}
		else if (match1 == 2)
		{
			prvf_pitch = nxt_pitch<<1;
			prvpf = nxtp + ((field == 1 ? 1 : 2)*nxt_pitch);
		}
		else if (match1 == 3)
		{
			curf = srcp + ((2+field)*src_pitch);
			prvf_pitch = prv_pitch<<1;
			prvpf = prvp + ((field == 1 ? 2 : 1)*prv_pitch);
		}
		else if (match1 == 4)
		{
			curf = srcp + ((2+field)*src_pitch);
			prvf_pitch = nxt_pitch<<1;
			prvpf = nxtp + ((field == 1 ? 2 : 1)*nxt_pitch);
		}
		if (match2 == 0)
		{
			nxtf_pitch = prv_pitch<<1;
			nxtpf = prvp + ((field == 1 ? 1 : 2)*prv_pitch);
		}
		else if (match2 == 1)
		{
			nxtf_pitch = src_pitch<<1;
			nxtpf = srcp + ((field == 1 ? 1 : 2)*src_pitch);
		}
		else if (match2 == 2)
		{
			nxtf_pitch = nxt_pitch<<1;
			nxtpf = nxtp + ((field == 1 ? 1 : 2)*nxt_pitch);
		}
		else if (match2 == 3)
		{
			nxtf_pitch = prv_pitch<<1;
			nxtpf = prvp + ((field == 1 ? 2 : 1)*prv_pitch);
		}
		else if (match2 == 4)
		{
			nxtf_pitch = nxt_pitch<<1;
			nxtpf = nxtp + ((field == 1 ? 2 : 1)*nxt_pitch);
		}
		prvnf = prvpf + prvf_pitch;
		curpf = curf - curf_pitch;
		curnf = curf + curf_pitch;
		nxtnf = nxtpf + nxtf_pitch;
		for (y=2; y<Height-2; y+=2)
		{
			if (y0a == y1a || y < y0a || y > y1a)
			{
				for (x=startx; x<stopx; x+=inc)
				{
					diff1 = abs(prvpf[x]-nxtpf[x]);
					diff2 = abs(prvnf[x]-nxtnf[x]);
					if (diff1 > 3 || diff2 > 3)
					{ 
						temp1 = curpf[x]+(curf[x]<<2)+curnf[x];
						temp2 = abs(3*(prvpf[x]+prvnf[x])-temp1);
						if (temp2 > 23) accumPc += temp2;
						if (temp2 > 42 && (diff1 > 19 || diff2 > 19)) accumPm += temp2;
						temp2 = abs(3*(nxtpf[x]+nxtnf[x])-temp1);
						if (temp2 > 23) accumNc += temp2;
						if (temp2 > 42 && (diff1 > 19 || diff2 > 19)) accumNm += temp2;
					}
				}
			}
			prvpf += prvf_pitch;
			prvnf += prvf_pitch;
			curpf += curf_pitch;
			curf += curf_pitch;
			curnf += curf_pitch;
			nxtpf += nxtf_pitch;
			nxtnf += nxtf_pitch;
		}
	}
	norm1 = (int)((accumPc / 6.0f) + 0.5f);
	norm2 = (int)((accumNc / 6.0f) + 0.5f);
	mtn1 = (int)((accumPm / 6.0f) + 0.5f);
	mtn2 = (int)((accumNm / 6.0f) + 0.5f);
	if ((mtn1 >= 500 || mtn2 >= 500) && ((mtn1<<1) < mtn2 || (mtn2<<1) < mtn1))
	{
		if (mtn1 > mtn2) ret = match2;
		else ret = match1;
	}
	else
	{
		if (norm1 > norm2) ret = match2;
		else ret = match1;
	}
	if (debug)
	{
		sprintf(buf,"TFM:  frame %d  - comparing %c to %c\n", n, MTC(match1), MTC(match2));
		OutputDebugString(buf);
		sprintf(buf,"TFM:  frame %d  - nmatches:  %d vs %d   mmatches:  %d vs %d\n", n, norm1, norm2, mtn1, mtn2);
		OutputDebugString(buf);
	}
	return ret;
}

int TFM::compareFieldsSlow(PVideoFrame &prv, PVideoFrame &src, PVideoFrame &nxt, int match1, 
		int match2, int &norm1, int &norm2, int &mtn1, int &mtn2, int np, int n, IScriptEnvironment *env)
{
	PVideoFrame map = env->NewVideoFrame(vi);
	int b, plane, ret;
	const unsigned char *prvp, *srcp, *nxtp;
	const unsigned char *curpf, *curf, *curnf;
	const unsigned char *prvpf, *prvnf, *nxtpf, *nxtnf;
	unsigned char *mapp, *mapn;
	int prv_pitch, src_pitch, Width, Height, nxt_pitch;
	int prvf_pitch, nxtf_pitch, curf_pitch, stopx, map_pitch;
	int x, y, temp1, temp2, startx, y0a, y1a;
	int inc = np == 3 ? 1 : mChroma ? 1 : 2;
	int stop = np == 3 ? mChroma ? 3 : 1 : 1;
	unsigned long accumPc = 0, accumNc = 0, accumPm = 0, accumNm = 0;
	norm1 = norm2 = mtn1 = mtn2 = 0;
	for (b=0; b<stop; ++b)
	{
		if (b == 0) plane = PLANAR_Y;
		else if (b == 1) plane = PLANAR_V;
		else plane = PLANAR_U;
		mapp = map->GetWritePtr(plane);
		map_pitch = map->GetPitch(plane);
		prvp = prv->GetReadPtr(plane);
		prv_pitch = prv->GetPitch(plane);
		srcp = src->GetReadPtr(plane);
		src_pitch = src->GetPitch(plane);
		Width = src->GetRowSize(plane);
		Height = src->GetHeight(plane);
		nxtp = nxt->GetReadPtr(plane);
		nxt_pitch = nxt->GetPitch(plane);
		memset(mapp,0,Height*map_pitch);
		startx = np > 1 ? (b == 0 ? 8 : 4) : 16;
		stopx = Width - startx;
		curf_pitch = src_pitch<<1;
		if (b == 0) { y0a = y0; y1a = y1; }
		else { y0a = y0>>1; y1a = y1>>1; }
		if (match1 < 3) 
		{
			curf = srcp + ((3-field)*src_pitch);
			mapp = mapp + ((field == 1 ? 1 : 2)*map_pitch);
		}
		if (match1 == 0)
		{
			prvf_pitch = prv_pitch<<1;
			prvpf = prvp + ((field == 1 ? 1 : 2)*prv_pitch);
		}
		else if (match1 == 1)
		{
			prvf_pitch = src_pitch<<1;
			prvpf = srcp + ((field == 1 ? 1 : 2)*src_pitch);
		}
		else if (match1 == 2)
		{
			prvf_pitch = nxt_pitch<<1;
			prvpf = nxtp + ((field == 1 ? 1 : 2)*nxt_pitch);
		}
		else if (match1 == 3)
		{
			curf = srcp + ((2+field)*src_pitch);
			prvf_pitch = prv_pitch<<1;
			prvpf = prvp + ((field == 1 ? 2 : 1)*prv_pitch);
			mapp = mapp + ((field == 1 ? 2 : 1)*map_pitch);
		}
		else if (match1 == 4)
		{
			curf = srcp + ((2+field)*src_pitch);
			prvf_pitch = nxt_pitch<<1;
			prvpf = nxtp + ((field == 1 ? 2 : 1)*nxt_pitch);
			mapp = mapp + ((field == 1 ? 2 : 1)*map_pitch);
		}
		if (match2 == 0)
		{
			nxtf_pitch = prv_pitch<<1;
			nxtpf = prvp + ((field == 1 ? 1 : 2)*prv_pitch);
		}
		else if (match2 == 1)
		{
			nxtf_pitch = src_pitch<<1;
			nxtpf = srcp + ((field == 1 ? 1 : 2)*src_pitch);
		}
		else if (match2 == 2)
		{
			nxtf_pitch = nxt_pitch<<1;
			nxtpf = nxtp + ((field == 1 ? 1 : 2)*nxt_pitch);
		}
		else if (match2 == 3)
		{
			nxtf_pitch = prv_pitch<<1;
			nxtpf = prvp + ((field == 1 ? 2 : 1)*prv_pitch);
		}
		else if (match2 == 4)
		{
			nxtf_pitch = nxt_pitch<<1;
			nxtpf = nxtp + ((field == 1 ? 2 : 1)*nxt_pitch);
		}
		prvnf = prvpf + prvf_pitch;
		curpf = curf - curf_pitch;
		curnf = curf + curf_pitch;
		nxtnf = nxtpf + nxtf_pitch;
		map_pitch <<= 1;
		mapn = mapp + map_pitch;
		if (np == 3)
		{
			if ((match1 >= 3 && field == 1) || (match1 < 3 && field != 1))
				buildDiffMapPlaneYV12(prvpf,nxtpf,mapp,prvf_pitch,nxtf_pitch,map_pitch,Height,Width);
			else
				buildDiffMapPlaneYV12(prvnf,nxtnf,mapn,prvf_pitch,nxtf_pitch,map_pitch,Height,Width);
		}
		else
		{
			if ((match1 >= 3 && field == 1) || (match1 < 3 && field != 1))
				buildDiffMapPlaneYUY2(prvpf,nxtpf,mapp,prvf_pitch,nxtf_pitch,map_pitch,Height,Width);
			else
				buildDiffMapPlaneYUY2(prvnf,nxtnf,mapn,prvf_pitch,nxtf_pitch,map_pitch,Height,Width);
		}
		for (y=2; y<Height-2; y+=2)
		{
			if (y0a == y1a || y < y0a || y > y1a)
			{
				for (x=startx; x<stopx; x+=inc)
				{
					if (mapp[x] > 0 || mapn[x] > 0)
					{
						temp1 = curpf[x]+(curf[x]<<2)+curnf[x];
						temp2 = abs(3*(prvpf[x]+prvnf[x])-temp1);
						if (temp2 > 23 && ((mapp[x]&1) || (mapn[x]&1))) accumPc += temp2;
						if (temp2 > 42 && ((mapp[x]&2) || (mapn[x]&2))) accumPm += temp2;
						temp2 = abs(3*(nxtpf[x]+nxtnf[x])-temp1);
						if (temp2 > 23 && ((mapp[x]&1) || (mapn[x]&1))) accumNc += temp2;
						if (temp2 > 42 && ((mapp[x]&2) || (mapn[x]&2))) accumNm += temp2;
					}
				}
			}
			prvpf += prvf_pitch;
			prvnf += prvf_pitch;
			curpf += curf_pitch;
			curf += curf_pitch;
			curnf += curf_pitch;
			nxtpf += nxtf_pitch;
			nxtnf += nxtf_pitch;
			mapp += map_pitch;
			mapn += map_pitch;
		}
	}
	norm1 = (int)((accumPc / 6.0f) + 0.5f);
	norm2 = (int)((accumNc / 6.0f) + 0.5f);
	mtn1 = (int)((accumPm / 6.0f) + 0.5f);
	mtn2 = (int)((accumNm / 6.0f) + 0.5f);
	if ((mtn1 >= 500 || mtn2 >= 500) && ((mtn1<<1) < mtn2 || (mtn2<<1) < mtn1))
	{
		if (mtn1 > mtn2) ret = match2;
		else ret = match1;
	}
	else
	{
		if (norm1 > norm2) ret = match2;
		else ret = match1;
	}
	if (debug)
	{
		sprintf(buf,"TFM:  frame %d  - comparing %c to %c  (SLOW)\n", n, MTC(match1), MTC(match2));
		OutputDebugString(buf);
		sprintf(buf,"TFM:  frame %d  - nmatches:  %d vs %d   mmatches:  %d vs %d\n", n, norm1, norm2, mtn1, mtn2);
		OutputDebugString(buf);
	}
	return ret;
}

void TFM::createWeaveFrame(PVideoFrame &dst, PVideoFrame &prv, PVideoFrame &src, 
		PVideoFrame &nxt, IScriptEnvironment *env, int match, int np)
{
	int b, plane;
	for (b=0; b<np; ++b)
	{
		if (b == 0) plane = PLANAR_Y;
		else if (b == 1) plane = PLANAR_V;
		else plane = PLANAR_U;
		if (match == 0)
		{
			env->BitBlt(dst->GetWritePtr(plane)+(1-field)*dst->GetPitch(plane),dst->GetPitch(plane)<<1,
				src->GetReadPtr(plane)+(1-field)*src->GetPitch(plane),src->GetPitch(plane)<<1, 
				src->GetRowSize(plane),src->GetHeight(plane)>>1);
			env->BitBlt(dst->GetWritePtr(plane)+field*dst->GetPitch(plane),dst->GetPitch(plane)<<1, 
				prv->GetReadPtr(plane)+field*prv->GetPitch(plane),prv->GetPitch(plane)<<1, 
				prv->GetRowSize(plane),prv->GetHeight(plane)>>1);
		}
		else if (match == 1)
		{
			env->BitBlt(dst->GetWritePtr(plane),dst->GetPitch(plane),src->GetReadPtr(plane), 
				src->GetPitch(plane),src->GetRowSize(plane),src->GetHeight(plane));
		}
		else if (match == 2)
		{
			env->BitBlt(dst->GetWritePtr(plane)+(1-field)*dst->GetPitch(plane),dst->GetPitch(plane)<<1,
				src->GetReadPtr(plane)+(1-field)*src->GetPitch(plane),src->GetPitch(plane)<<1, 
				src->GetRowSize(plane),src->GetHeight(plane)>>1);
			env->BitBlt(dst->GetWritePtr(plane)+field*dst->GetPitch(plane),dst->GetPitch(plane)<<1, 
				nxt->GetReadPtr(plane)+field*nxt->GetPitch(plane),nxt->GetPitch(plane)<<1,
				nxt->GetRowSize(plane),nxt->GetHeight(plane)>>1);
		}
		else if (match == 3)
		{
			env->BitBlt(dst->GetWritePtr(plane)+field*dst->GetPitch(plane),dst->GetPitch(plane)<<1,
				src->GetReadPtr(plane)+field*src->GetPitch(plane),src->GetPitch(plane)<<1, 
				src->GetRowSize(plane),src->GetHeight(plane)>>1);
			env->BitBlt(dst->GetWritePtr(plane)+(1-field)*dst->GetPitch(plane),dst->GetPitch(plane)<<1, 
				prv->GetReadPtr(plane)+(1-field)*prv->GetPitch(plane),prv->GetPitch(plane)<<1, 
				prv->GetRowSize(plane),prv->GetHeight(plane)>>1);
		}
		else if (match == 4)
		{
			env->BitBlt(dst->GetWritePtr(plane)+field*dst->GetPitch(plane),dst->GetPitch(plane)<<1,
				src->GetReadPtr(plane)+field*src->GetPitch(plane),src->GetPitch(plane)<<1, 
				src->GetRowSize(plane),src->GetHeight(plane)>>1);
			env->BitBlt(dst->GetWritePtr(plane)+(1-field)*dst->GetPitch(plane),dst->GetPitch(plane)<<1, 
				nxt->GetReadPtr(plane)+(1-field)*nxt->GetPitch(plane),nxt->GetPitch(plane)<<1, 
				nxt->GetRowSize(plane),nxt->GetHeight(plane)>>1);
		}
		else env->ThrowError("TFM:  an unknown error occurred (no such match!)");
	}
}

void TFM::putHint(PVideoFrame &dst, int match, int combed, bool d2vfilm)
{
	unsigned char *p = dst->GetWritePtr(PLANAR_Y);
	unsigned int i, hint = 0;
	if (match == 0) hint |= ISP;
	else if (match == 1 && combed < 2) hint |= ISC;
	else if (match == 2) hint |= ISN;
	else if (match == 3) hint |= ISB;
	else if (match == 4) hint |= ISU;
	else if (match == 1 && combed > 1 && field == 0) hint |= ISDB;
	else if (match == 1 && combed > 1 && field == 1) hint |= ISDT;
	if (field == 1) hint |= TOP_FIELD;
	if (combed > 1) hint |= COMBED;
	if (d2vfilm) hint |= D2VFILM;
	for (i=0; i<32; ++i)
	{
		*p &= ~1;
		*p++ |= ((MAGIC_NUMBER & (1 << i)) >> i);
	}
	for (i=0; i<32; ++i)
	{
		*p &= ~1;
		*p++ |= ((hint & (1 << i)) >> i);
	}
}

AVSValue __cdecl Create_TFM(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	AVSValue v = new TFM(args[0].AsClip(),args[1].AsInt(-1),args[2].AsInt(-1),args[3].AsInt(1),
		args[4].AsInt(6),args[5].AsString(""),args[6].AsString(""),args[7].AsString(""),args[8].AsString(""),
		args[9].AsBool(false),args[10].AsBool(false),args[11].AsBool(true),args[12].AsBool(true),
		args[13].AsInt(15),args[14].AsInt(10),args[15].AsInt(85),args[16].AsBool(false),args[17].AsInt(16),
		args[18].AsInt(16),args[19].AsInt(0),args[20].AsInt(0),args[23].AsString(""),args[24].AsInt(0),env);
	if (!args[4].IsInt() || args[4].AsInt() >= 2)
	{
		if (!args[4].IsInt() || args[4].AsInt() > 4)
		{
			try { v = env->Invoke("InternalCache", v).AsClip(); } 
			catch (IScriptEnvironment::NotFound) {  }
		}
		v = new TFMPP(v.AsClip(),args[4].AsInt(6),args[21].AsInt(5),args[5].AsString(""),
				args[10].AsBool(false),(args[22].IsClip() ? args[22].AsClip() : NULL),env);
	}
	return v;
}

TFM::TFM(PClip _child, int _order, int _field, int _mode, int _PP, const char* _ovr, const char* _input, 
	const char* _output, const char * _outputC, bool _debug, bool _display, bool _slow, 
	bool _mChroma, int _cNum, int _cthresh, int _MI, bool _chroma, int _blockx, int _blocky, 
	int _y0, int _y1, const char* _d2v, int _ovrDefault, IScriptEnvironment* env) : GenericVideoFilter(_child), 
	order(_order), field(_field), mode(_mode), PP(_PP), ovr(_ovr), input(_input), output(_output), 
	outputC(_outputC), debug(_debug), display(_display), slow(_slow), mChroma(_mChroma), cNum(_cNum), 
	cthresh(_cthresh), MI(_MI), chroma(_chroma), blockx(_blockx), blocky(_blocky), y0(_y0), y1(_y1),
	d2v(_d2v), ovrDefault(_ovrDefault)
{
	cArray = setArray = NULL;
	ovrArray = outArray = NULL;
	d2vfilmarray = NULL;
	int z, w, q, b, i, count, last, fieldt, firstLine, qt;
	int countOvrS, countOvrM;
	char linein[81];
	char *linep, *linet;
	FILE *f = NULL;
	if (!vi.IsYV12() && !vi.IsYUY2())
		env->ThrowError("TFM:  YV12 and YUY2 data only!");
	if (vi.height&1 || vi.width&1)
		env->ThrowError("TFM:  height and width must be divisible by 2!");
	if (vi.height < 6 || vi.width < 64)
		env->ThrowError("TFM:  frame dimensions too small!");
	if (mode < 0 || mode > 4)
		env->ThrowError("TFM:  mode must be set to 0, 1, 2, 3, or 4!");
	if (field < -1 || field > 1)
		env->ThrowError("TFM:  field must be set to -1, 0, or 1!");
	if (PP < 0 || PP > 7)
		env->ThrowError("TFM:  PP must be at least 0 and less than 8!");
	if (order < -1 || order > 1)
		env->ThrowError("TFM:  order must be set to -1, 0, or 1!");
	if (blockx != 4 && blockx != 8 && blockx != 16 && blockx != 32 && blockx != 64 && 
		blockx != 128 && blockx != 256 && blockx != 512 && blockx != 1024 && blockx != 2048)
		env->ThrowError("TFM:  illegal blockx size!");
	if (blocky != 4 && blocky != 8 && blocky != 16 && blocky != 32 && blocky != 64 && 
		blocky != 128 && blocky != 256 && blocky != 512 && blocky != 1024 && blocky != 2048)
		env->ThrowError("TFM:  illegal blocky size!");
	if (y0 != y1 && (y0 < 0 || y1 < 0 || y0 > y1 || y1 > vi.height || y0 > vi.height))
		env->ThrowError("TFM:  bad y0 and y1 exclusion band values!");
	if (ovrDefault < 0 || ovrDefault > 2)
		env->ThrowError("TFM:  ovrDefault must be set to 0, 1, or 2!");
	if (debug)
	{
		sprintf(buf, "TFM:  %s by tritical\n", VERSION);
		OutputDebugString(buf);
	}
	child->SetCacheHints(CACHE_RANGE, 1);
	lastMatch.frame = lastMatch.field = lastMatch.combed = lastMatch.match = -20;
	nfrms = vi.num_frames-1;
	f = NULL;
	modeS = mode;
	PPS = PP;
	vidCount = setArraySize = 0;
	xhalf = blockx >> 1;
	yhalf = blocky >> 1;
	xshift = blockx == 4 ? 2 : blockx == 8 ? 3 : blockx == 16 ? 4 : blockx == 32 ? 5 :
		blockx == 64 ? 6 : blockx == 128 ? 7 : blockx == 256 ? 8 : blockx == 512 ? 9 : 
		blockx == 1024 ? 10 : 11;
	yshift = blocky == 4 ? 2 : blocky == 8 ? 3 : blocky == 16 ? 4 : blocky == 32 ? 5 :
		blocky == 64 ? 6 : blocky == 128 ? 7 : blocky == 256 ? 8 : blocky == 512 ? 9 : 
		blocky == 1024 ? 10 : 11;
	if (mode == 1 || mode == 2 || mode == 4 || PP > 0)
		cArray = (int *)_aligned_malloc((((vi.width+xhalf)>>xshift)+1)*(((vi.height+yhalf)>>yshift)+1)*4*sizeof(int), 16);
	if (vi.IsYUY2())
	{
		xhalf *= 2;
		++xshift;
	}
	if (*d2v) parseD2V(env);
	orderS = order;
	fieldS = fieldO = field;
	if (fieldO == -1) 
	{
		if (order == -1) fieldO = child->GetParity(0) ? 1 : 0;
		else fieldO = order;
	}
	if (*input)
	{
		bool d2vmarked;
		if ((f = fopen(input, "r")) != NULL)
		{
			ovrArray = (unsigned char *)malloc(vi.num_frames*sizeof(unsigned char));
			if (ovrArray == NULL) 
			{
				fclose(f);
				f = NULL;
				env->ThrowError("TFM:  malloc failure (ovrArray)!");
			}
			memset(ovrArray,255,vi.num_frames);
			if (d2vfilmarray == NULL)
			{
				d2vfilmarray = (unsigned char *)malloc((vi.num_frames+1)*sizeof(unsigned char));
				if (d2vfilmarray == NULL) env->ThrowError("TFM:  malloc failure (d2vfilmarray)!");
				memset(d2vfilmarray,0,(vi.num_frames+1)*sizeof(unsigned char));
			}
			fieldt = fieldO;
			firstLine = 0;
			if (debug)
			{
				sprintf(buf,"TFM:  successfully opened input file.  Field defaulting to - %s.\n", 
					fieldt == 0 ? "bottom" : "top");
				OutputDebugString(buf);
			}
			while (fgets(linein, 80, f) != NULL)
			{
				if (linein[0] == 0 || linein[0] == '\n' || linein[0] == '\r' ||  linein[0] == ';' || linein[0] == '#') 
					continue;
				++firstLine;
				linep = linein;
				while (*linep != 'f' && *linep != 'F' && *linep != 0 && *linep != ' ' && *linep != 'c') *linep++;
				if (*linep == 'f' || *linep == 'F')
				{
					if (firstLine == 1)
					{
						bool changed = false;
						if (strnicmp(linein, "field = top", 11) == 0) { fieldt = 1; changed = true; }
						else if (strnicmp(linein, "field = bottom", 14) == 0) { fieldt = 0; changed = true; }
						if (debug && changed)
						{
							sprintf(buf,"TFM:  detected field for input file - %s.\n", 
								fieldt == 0 ? "bottom" : "top");
							OutputDebugString(buf);
						}
					}
				}
				else if (*linep == 'c')
				{
					if (strnicmp(linein, "crc32 = ", 8) == 0)
					{
						linet = linein;
						while(*linet != ' ') *linet++;
						*linet++;
						while(*linet != ' ') *linet++;
						*linet++;
						unsigned int m, tempCrc;
						sscanf(linet, "%x", &m);
						calcCRC(child, 15, tempCrc, env);
						if (tempCrc != m)
						{
							fclose(f);
							f = NULL;
							env->ThrowError("TFM:  crc32 in input file does not match that of the current clip!");
						}
					}
				}
				else if (*linep == ' ')
				{
					linet = linein;
					while(*linet != 0)
					{
						if (*linet != ' ' && *linet != 10) break;
						*linet++;
					}
					if (*linet == 0) { --firstLine; continue; }
					sscanf(linein, "%d", &z);
					linep = linein;
					while (*linep != 'p' && *linep != 'c' && *linep != 'n' && *linep != 'u' &&
							*linep != 'b' && *linep != 'l' && *linep != 'h' && *linep != 0) *linep++;
					if (*linep != 0)
					{
						if (z<0 || z>nfrms)
						{
							fclose(f);
							f = NULL;
							env->ThrowError("TFM:  input file error (out of range or non-ascending frame #)!");
						}
						linep = linein;
						while (*linep != ' ' && *linep != 0) *linep++;
						if (*linep != 0)
						{
							qt = -1;
							d2vmarked = false;
							*linep++;
							q = *linep;
							if (q == 112) q = 0;
							else if (q == 99) q = 1;
							else if (q == 110) q = 2;
							else if (q == 98) q = 3;
							else if (q == 117) q = 4;
							else if (q == 108) q = 5;
							else if (q == 104) q = 6;
							else 
							{
								fclose(f);
								f = NULL;
								env->ThrowError("TFM:  input file error (invalid match specifier)!");
							}
							*linep++;
							*linep++;
							if (*linep != 0)
							{
								qt = *linep;
								if (qt == 45) qt = 0;
								else if (qt == 43) qt = COMBED;
								else if (qt == '1') { d2vmarked = true; qt = -1; }
								else
								{
									fclose(f);
									f = NULL;
									env->ThrowError("TFM:  input file error (invalid specifier)!");
								}
							}
							if (fieldt != fieldO)
							{
								if (q == 0) q = 3;
								else if (q == 2) q = 4;
								else if (q == 3) q = 0;
								else if (q == 4) q = 2;
							}
							if (!d2vmarked && qt != -1)
							{
								*linep++;
								*linep++;
								if (*linep == '1') d2vmarked = true;
							}
							if (d2vmarked) d2vfilmarray[z] = fieldt == 1 ? 0x3 : 0x1;
							ovrArray[z] |= 0x07;
							ovrArray[z] &= (q|0xF8);
							if (qt != -1)
							{
								ovrArray[z] &= 0xDF;
								ovrArray[z] |= 0x10;
								ovrArray[z] &= (qt|0xEF);
							}
						}
					}
				}
			}
			fclose(f);
			f = NULL;
		}
		else env->ThrowError("TFM:  input file error (could not open file)!");
	}
	if (*ovr)
	{
		if ((f = fopen(ovr, "r")) != NULL)
		{
			countOvrS = countOvrM = 0;
			while(fgets(linein, 80, f) != 0)
			{
				if (linein[0] == 0 || linein[0] == '\n' || linein[0] == '\r' || linein[0] == ';' || linein[0] == '#') 
					continue;
				linep = linein;
				while (*linep != 'c' && *linep != 'p' && *linep != 'n' && *linep != 'b' && 
					*linep != 'u' && *linep != 'l' && *linep != 'h' && *linep != '+' && *linep != '-' && *linep != 0) *linep++;
				if (*linep == 0) ++countOvrS;
				else ++countOvrM;
			}
			fclose(f);
			f = NULL;
			if (ovrDefault != 0 && ovrArray != NULL)
			{
				if (ovrDefault == 1) q = 0;
				else if (ovrDefault == 2) q = COMBED;
				for (int h=0; h<vi.num_frames; ++h)
				{
					ovrArray[h] &= 0xDF;
					ovrArray[h] |= 0x10;
					ovrArray[h] &= (q|0xEF);
				}
			}
			if (countOvrS == 0 && countOvrM == 0) { goto emptyovr; }
			if (countOvrS > 0)
			{
				++countOvrS;
				countOvrS *= 4;
				setArray = (int *)malloc(countOvrS*sizeof(int));
				if (setArray == NULL) env->ThrowError("TFM:  malloc failure (setArray)!");
				memset(setArray,255,countOvrS*sizeof(int));
				setArraySize = countOvrS;
			}
			if (countOvrM > 0 && ovrArray == NULL)
			{
				ovrArray = (unsigned char *)malloc(vi.num_frames*sizeof(unsigned char));
				if (ovrArray == NULL) env->ThrowError("TFM:  malloc failure (ovrArray)!");
				memset(ovrArray,255,vi.num_frames);
				if (ovrDefault != 0)
				{
					if (ovrDefault == 1) q = 0;
					else if (ovrDefault == 2) q = COMBED;
					for (int h=0; h<vi.num_frames; ++h)
					{
						ovrArray[h] &= 0xDF;
						ovrArray[h] |= 0x10;
						ovrArray[h] &= (q|0xEF);
					}
				}
			}
			last = -1;
			fieldt = fieldO;
			firstLine = 0;
			i = 0;
			if ((f = fopen(ovr, "r")) != NULL)
			{
				if (debug)
				{
					sprintf(buf,"TFM:  successfully opened ovr file.  Field defaulting to - %s.\n", 
						fieldt == 0 ? "bottom" : "top");
					OutputDebugString(buf);
				}
				while (fgets(linein, 80, f) != NULL)
				{
					if (linein[0] == 0 || linein[0] == '\n' || linein[0] == '\r' ||  linein[0] == ';' || linein[0] == '#') 
						continue;
					++firstLine;
					linep = linein;
					while (*linep != 'f' && *linep != 'F' && *linep != 0 && *linep != ' ' && *linep != ',') *linep++;
					if (*linep == 'f' || *linep == 'F')
					{
						if (firstLine == 1)
						{
							bool changed = false;
							if (strnicmp(linein, "field = top", 11) == 0) { fieldt = 1; changed = true; }
							else if (strnicmp(linein, "field = bottom", 14) == 0) { fieldt = 0; changed = true; }
							if (debug && changed)
							{
								sprintf(buf,"TFM:  detected field for ovr file - %s.\n", 
									fieldt == 0 ? "bottom" : "top");
								OutputDebugString(buf);
							}
						}
					}
					else if (*linep == ' ')
					{
						linet = linein;
						while(*linet != 0)
						{
							if (*linet != ' ' && *linet != 10) break;
							*linet++;
						}
						if (*linet == 0) { --firstLine; continue; }
						*linep++;
						if (*linep == 'p' || *linep == 'c' || *linep == 'n' || *linep == 'b' || *linep == 'u' || *linep == 'l' || *linep == 'h')
						{
							sscanf(linein, "%d", &z);
							if (z<0 || z>nfrms || z <= last) 
							{
								fclose(f);
								f = NULL;
								env->ThrowError("TFM:  ovr file error (out of range or non-ascending frame #)!");
							}
							linep = linein;
							while (*linep != ' ' && *linep != 0) *linep++;
							if (*linep != 0)
							{
								*linep++;
								q = *linep;
								if (q == 112) q = 0;
								else if (q == 99) q = 1;
								else if (q == 110) q = 2;
								else if (q == 98) q = 3;
								else if (q == 117) q = 4;
								else if (q == 108) q = 5;
								else if (q == 104) q = 6;
								else 
								{
									fclose(f);
									f = NULL;
									env->ThrowError("TFM:  ovr file error (invalid match specifier)!");
								}
								if (fieldt != fieldO)
								{
									if (q == 0) q = 3;
									else if (q == 2) q = 4;
									else if (q == 3) q = 0;
									else if (q == 4) q = 2;
								}
								ovrArray[z] |= 0x07;
								ovrArray[z] &= (q|0xF8);
								last = z;
							}
						}
						else if (*linep == '-' || *linep == '+')
						{
							sscanf(linein, "%d", &z);
							if (z<0 || z>nfrms) 
							{
								fclose(f);
								f = NULL;
								env->ThrowError("TFM:  ovr file error (out of range or non-ascending frame #)!");
							}
							linep = linein;
							while (*linep != ' ' && *linep != 0) *linep++;
							if (*linep != 0)
							{
								*linep++;
								q = *linep;
								if (q == 45) q = 0;
								else if (q == 43) q = COMBED;
								else 
								{
									fclose(f);
									f = NULL;
									env->ThrowError("TFM:  ovr file error (invalid symbol)!");
								}
								ovrArray[z] &= 0xDF;
								ovrArray[z] |= 0x10;
								ovrArray[z] &= (q|0xEF);
							}
						}
						else
						{
							sscanf(linein, "%d", &z);
							if (z<0 || z>nfrms) 
							{
								fclose(f);
								f = NULL;
								env->ThrowError("TFM:  ovr input error (out of range frame #)!");
							}
							linep = linein;
							while (*linep != ' ' && *linep != 0) *linep++;
							if (*linep != 0)
							{
								*linep++;
								if (*linep == 'f' || *linep == 'm' || *linep == 'o' || *linep == 'P')
								{
									q = *linep;
									*linep++; 
									*linep++;
									if (*linep == 0) continue;
									sscanf(linep, "%d", &b);
									if (q == 102 && b != 0 && b != 1 && b != -1) 
									{
										fclose(f);
										f = NULL;
										env->ThrowError("TFM:  ovr input error (bad field value)!");
									}
									else if (q == 111 && b != 0 && b != 1 && b != -1) 
									{
										fclose(f);
										f = NULL;
										env->ThrowError("TFM:  ovr input error (bad order value)!");
									}
									else if (q == 109 && (b < 0 || b > 4))
									{
										fclose(f);
										f = NULL;
										env->ThrowError("TFM:  ovr input error (bad mode value)!");
									}
									else if (q == 80 && (b < 0 || b > 7))
									{
										fclose(f);
										f = NULL;
										env->ThrowError("TFM:  ovr input error (bad PP value)!");
									}
									setArray[i] = q; ++i;
									setArray[i] = z; ++i;
									setArray[i] = z; ++i;
									setArray[i] = b; ++i;
								}
							}
						}
					}
					else if (*linep == ',')
					{
						while (*linep != ' ' && *linep != 0) *linep++;
						if (*linep == 0) continue;
						*linep++;
						if (*linep == 'p' || *linep == 'c' || *linep == 'n' || *linep == 'u' || *linep == 'b' || *linep == 'l' || *linep == 'h')
						{
							sscanf(linein, "%d,%d", &z, &w);
							if (w == 0) w = nfrms;
							if (z<0 || z>nfrms || w<0 || w>nfrms || w<z || z <= last) 
							{
								fclose(f);
								f = NULL;
								env->ThrowError("TFM:  input file error (out of range or non-ascending frame #)!");
							}
							linep = linein;
							while (*linep != ' ' && *linep != 0) *linep++;
							if (*linep != 0)
							{
								*linep++;
								if (*(linep+1) == 'p' || *(linep+1) == 'c' || *(linep+1) == 'n' || *(linep+1) == 'b' || *(linep+1) == 'u' || *(linep+1) == 'l' || *(linep+1) == 'h')
								{
									count = 0;
									while ((*linep == 'p' || *linep == 'c' || *linep == 'n' || *linep == 'b' || *linep == 'u' || *linep == 'l' || *linep == 'h') && (z+count<=w))
									{
										q = *linep;
										if (q == 112) q = 0;
										else if (q == 99) q = 1;
										else if (q == 110) q = 2;
										else if (q == 98) q = 3;
										else if (q == 117) q = 4;
										else if (q == 108) q = 5;
										else if (q == 104) q = 6;
										else 
										{
											fclose(f);
											f = NULL;
											env->ThrowError("TFM:  input file error (invalid match specifier)!");
										}
										if (fieldt != fieldO)
										{
											if (q == 0) q = 3;
											else if (q == 2) q = 4;
											else if (q == 3) q = 0;
											else if (q == 4) q = 2;
										}
										ovrArray[z+count] |= 0x07;
										ovrArray[z+count] &= (q|0xF8);
										++count;
										*linep++;
									}
									while (z+count<=w)
									{
										ovrArray[z+count] |= 0x07;
										ovrArray[z+count] &= (ovrArray[z]|0xF8);
										++z;
									}
									last = w;
								}
								else
								{
									q = *linep;
									if (q == 112) q = 0;
									else if (q == 99) q = 1;
									else if (q == 110) q = 2;
									else if (q == 98) q = 3;
									else if (q == 117) q = 4;
									else if (q == 108) q = 5;
									else if (q == 104) q = 6;
									else 
									{
										fclose(f);
										f = NULL;
										env->ThrowError("TFM:  input file error (invalid match specifier)!");
									}
									if (fieldt != fieldO)
									{
										if (q == 0) q = 3;
										else if (q == 2) q = 4;
										else if (q == 3) q = 0;
										else if (q == 4) q = 2;
									}
									while (z<=w)
									{
										ovrArray[z] |= 0x07;
										ovrArray[z] &= (q|0xF8);
										++z;
									}
									last = w;
								}
							}
						}
						else if (*linep == '-' || *linep == '+')
						{
							sscanf(linein, "%d,%d", &z, &w);
							if (w == 0) w = nfrms;
							if (z<0 || z>nfrms || w<0 || w>nfrms || w<z) 
							{
								fclose(f);
								f = NULL;
								env->ThrowError("TFM:  input file error (out of range or non-ascending frame #)!");
							}
							linep = linein;
							while (*linep != ' ' && *linep != 0) *linep++;
							if (*linep != 0)
							{
								*linep++;
								if (*(linep+1) == '-' || *(linep+1) == '+')
								{
									count = 0;
									while ((*linep == '-' || *linep == '+') && (z+count<=w))
									{
										q = *linep;
										if (q == 45) q = 0;
										else if (q == 43) q = COMBED;
										else 
										{
											fclose(f);
											f = NULL;
											env->ThrowError("TFM:  input file error (invalid symbol)!");
										}
										ovrArray[z+count] &= 0xDF;
										ovrArray[z+count] |= 0x10;
										ovrArray[z+count] &= (q|0xEF);
										++count;
										*linep++;
									}
									while (z+count<=w)
									{
										ovrArray[z+count] &= 0xDF;
										ovrArray[z+count] |= 0x10;
										ovrArray[z+count] &= (ovrArray[z]|0xEF);
										++z;
									}
								}
								else
								{
									q = *linep;
									if (q == 45) q = 0;
									else if (q == 43) q = COMBED;
									else 
									{
										fclose(f);
										f = NULL;
										env->ThrowError("TFM:  input file error (invalid symbol)!");
									}
									while (z<=w)
									{
										ovrArray[z] &= 0xDF;
										ovrArray[z] |= 0x10;
										ovrArray[z] &= (q|0xEF);
										++z;
									}
								}
							}
						}
						else
						{
							sscanf(linein, "%d,%d", &z, &w);
							if (w == 0) w = nfrms;
							if (z<0 || z>nfrms || w<0 || w>nfrms || w < z) 
							{
								fclose(f);
								f = NULL;
								env->ThrowError("TFM: ovr input error (invalid frame range)!");
							}
							linep = linein;
							while (*linep != ' ' && *linep != 0) *linep++;
							if (*linep != 0)
							{
								*linep++;
								if (*linep == 'f' || *linep == 'm' || *linep == 'o' || *linep == 'P')
								{
									q = *linep;
									*linep++; 
									*linep++;
									if (*linep == 0) continue;
									sscanf(linep, "%d", &b);
									if (q == 102 && b != 0 && b != 1 && b != -1) 
									{
										fclose(f);
										f = NULL;
										env->ThrowError("TFM:  ovr input error (bad field value)!");
									}
									else if (q == 111 && b != 0 && b != 1 && b != -1) 
									{
										fclose(f);
										f = NULL;
										env->ThrowError("TFM:  ovr input error (bad order value)!");
									}
									else if (q == 109 && (b < 0 || b > 4))
									{
										fclose(f);
										f = NULL;
										env->ThrowError("TFM:  ovr input error (bad mode value)!");
									}
									else if (q == 80 && (b < 0 || b > 7))
									{
										fclose(f);
										f = NULL;
										env->ThrowError("TFM:  ovr input error (bad PP value)!");
									}
									setArray[i] = q; ++i;
									setArray[i] = z; ++i;
									setArray[i] = w; ++i;
									setArray[i] = b; ++i;
								}
							}
						}
					}
				}
				fclose(f);
				f = NULL;
			}
			else env->ThrowError("TFM:  ovr file error (could not open file)!");
		}
		else env->ThrowError("TFM:  ovr input error (could not open ovr file)!");
	}
emptyovr:
	if (*output)
	{
		if ((f = fopen(output, "w")) != NULL)
		{
			_fullpath(outputFull, output, MAX_PATH);
			calcCRC(child, 15, outputCrc, env);
			fclose(f);
			f = NULL;
			//unlink(outputFull);
			outArray = (unsigned char *)malloc(vi.num_frames*sizeof(unsigned char));
			if (outArray == NULL)
				env->ThrowError("TFM:  malloc failure (outArray, output)!");
			memset(outArray, 0, vi.num_frames);
		}
		else env->ThrowError("TFM:  output file error (cannot create file)!");
	}
	if (*outputC)
	{
		if ((f = fopen(outputC, "w")) != NULL) 
		{
			_fullpath(outputCFull, outputC, MAX_PATH);
			fclose(f);
			f = NULL;
			//unlink(outputCFull);
			if (outArray == NULL)
			{
				outArray = (unsigned char *)malloc(vi.num_frames*sizeof(unsigned char));
				if (outArray == NULL)
					env->ThrowError("TFM:  malloc failure (outArray, outputC)!");
				memset(outArray, 0, vi.num_frames);
			}
		}
		else env->ThrowError("TFM:  outputC file error (cannot create file)!");
	}
	AVSValue tfmPassValue(PP);
	const char *varname = "TFMPPValue";
	env->SetVar(varname,tfmPassValue);
	if (f != NULL) fclose(f);
}

TFM::~TFM()
{
	if (cArray != NULL) _aligned_free(cArray);
	if (setArray != NULL) free(setArray);
	if (ovrArray != NULL) free(ovrArray);
	if (d2vfilmarray != NULL) free(d2vfilmarray);
	if (outArray != NULL)
	{
		FILE *f = NULL;
		if (*output)
		{
			if ((f = fopen(outputFull, "w")) != NULL)
			{
				char tempBuf[40];
				int match;
				fprintf(f, "#TFM %s by tritical\n", VERSION);
				fprintf(f, "field = %s\n", fieldO == 1 ? "top" : "bottom");
				fprintf(f, "crc32 = %x\n", outputCrc);
				for (int h=0; h<=nfrms; ++h)
				{
					if (outArray[h]&FILE_ENTRY)
					{
						match = (outArray[h]&0x07);
						sprintf(tempBuf, "%d %c", h, MTC(match));
						if (outArray[h]&0x20)
						{
							if (outArray[h]&0x10) strcat(tempBuf, " +");
							else strcat(tempBuf, " -");
						}
						if (outArray[h]&FILE_D2V) strcat(tempBuf, " 1");
						strcat(tempBuf, "\n");
						fprintf(f, tempBuf);
					}
				}
				fclose(f);
				f = NULL;
			}
		}
		if (*outputC)
		{
			if ((f = fopen(outputCFull, "w")) != NULL)
			{
				int count = 0, match;
				fprintf(f, "#TFM %s by tritical\n", VERSION);
				for (int h=0; h<=nfrms; ++h)
				{
					if (outArray[h]&FILE_ENTRY) match = (outArray[h]&0x07);
					else match = 0;
					if (match == 1 || match == 5 || match == 6) ++count;
					else
					{
						if (count > cNum) fprintf(f, "%d,%d\n", h-count, h-1);
						count = 0;
					}
				}
				if (count > cNum) fprintf(f, "%d,%d\n", nfrms-count+1, nfrms);
				fclose(f);
				f = NULL;
			}
		}
		if (f != NULL) fclose(f);
		free(outArray);
	}
}