/*
**                    TIVTC v0.9.9 for AviSynth 2.5.x
**
**   TIVTC includes a field matching filter (TFM) and a decimation
**   filter (TDecimate) which can be used together to achieve an
**   IVTC or for other uses. TIVTC currently supports YV12 and 
**   YUY2 colorspaces.
**
**   Copyright (C) 2004-2005 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <windows.h>
#include <stdio.h>
#include <malloc.h>
#include "Font.h"
#include "calcCRC.h"
#include "internal.h"
#include "profUtil.h"
#define TFM_INCLUDED
#ifndef TFMPP_INCLUDED
#include "TFMPP.h"
#endif
#define ISP 0x00000000 // p
#define ISC 0x00000001 // c
#define ISN 0x00000002 // n
#define ISB 0x00000003 // b
#define ISU 0x00000004 // u
#define ISDB 0x00000005 // l
#define ISDT 0x00000006 // h
#define MTC(n) n == 0 ? 'p' : n == 1 ? 'c' : n == 2 ? 'n' : n == 3 ? 'b' : n == 4 ? 'u' : \
			   n == 5 ? 'l' : n == 6 ? 'h' : 'x'
#define TOP_FIELD 0x00000008
#define COMBED 0x00000010
#define D2VFILM 0x00000020
#define MAGIC_NUMBER (0xdeadfeed)
#define FILE_COMBED 0x00000030
#define FILE_NOTCOMBED 0x00000020
#define FILE_ENTRY 0x00000080
#define FILE_D2V 0x00000008
#ifdef VERSION
#undef VERSION
#endif
#define VERSION "v0.9.7.3"

struct MTRACK {
	int frame, match;
	int field, combed;
};

class TFM : public GenericVideoFilter
{
private:
	int order, field, mode, cthresh, MI, y0, y1, PP, PPS;
	const char *ovr, *input, *output, *outputC, *d2v;
	bool debug, slow, chroma, mChroma, display;
	int cNum, nfrms, orderS, fieldS, modeS, blockx, blocky;
	int xhalf, yhalf, xshift, yshift, ovrDefault;
	int vidCount, setArraySize, fieldO;
	unsigned int outputCrc;
	int *cArray, *setArray;
	unsigned char *ovrArray, *outArray, *d2vfilmarray;
	MTRACK lastMatch;
	char buf[270], outputFull[270], outputCFull[270];
	void TFM::buildDiffMapPlaneYV12(const unsigned char *prvp, const unsigned char *nxtp, 
		unsigned char *dstp, int prv_pitch, int nxt_pitch, int dst_pitch, int Height, int Width);
	void TFM::buildDiffMapPlaneYUY2(const unsigned char *prvp, const unsigned char *nxtp, 
		unsigned char *dstp, int prv_pitch, int nxt_pitch, int dst_pitch, int Height, int Width);
	void TFM::fileOut(int match, int combed, bool d2vfilm, int n);
	void TFM::copyFrame(PVideoFrame &dst, PVideoFrame &src, IScriptEnvironment *env, int np);
	int TFM::compareFields(PVideoFrame &prv, PVideoFrame &src, PVideoFrame &nxt, 
		int match1, int match2, int &norm1, int &norm2, int &mtn1, int &mtn2, int np, int n);
	int TFM::compareFieldsSlow(PVideoFrame &prv, PVideoFrame &src, PVideoFrame &nxt, int match1, 
		int match2, int &norm1, int &norm2, int &mtn1, int &mtn2, int np, int n, IScriptEnvironment *env);
	void TFM::createWeaveFrame(PVideoFrame &dst, PVideoFrame &prv, PVideoFrame &src, 
		PVideoFrame &nxt, IScriptEnvironment *env, int match, int np);
	bool TFM::checkCombed(PVideoFrame &src, int n, IScriptEnvironment *env, int np, int match,
		int &blockN, int &xblocksi, int &MICount);
	bool TFM::getMatchOvr(int n, int &match, int &combed);
	void TFM::getSettingOvr(int n);
	bool TFM::checkCombedYV12(PVideoFrame &src, int n, IScriptEnvironment *env, int match,
		int &blockN, int &xblocksi, int &MICount);
	bool TFM::checkCombedYUY2(PVideoFrame &src, int n, IScriptEnvironment *env, int match,
		int &blockN, int &xblocksi, int &MICount);
	void TFM::writeDisplay(PVideoFrame &dst, int np, int n, int fmatch, int combed, bool over,
		int blockN, int xblocks, int MICount);
	void TFM::putHint(PVideoFrame &dst, int match, int combed, bool d2vfilm);
	void TFM::drawBoxYUY2(PVideoFrame &dst, int blockN, int xblocks);
	void TFM::drawBoxYV12(PVideoFrame &dst, int blockN, int xblocks);
	void TFM::parseD2V(IScriptEnvironment *env);
	int TFM::D2V_find_and_correct(int *array, bool &found, int &tff);
	void TFM::D2V_find_fix(int a1, int a2, int sync, int &f1, int &f2, int &change);
	bool TFM::D2V_check_illegal(int a1, int a2);
	int TFM::D2V_check_final(int *array);
	int TFM::D2V_initialize_array(int *&array, int &d2vtype);
	int TFM::D2V_write_array(int *array, char wfile[]);
	int TFM::D2V_get_output_filename(char wfile[]);
	int TFM::D2V_fill_d2vfilmarray(int *array);
	bool TFM::d2vduplicate(int match, int combed, int n);

public:
	PVideoFrame __stdcall TFM::GetFrame(int n, IScriptEnvironment* env);
	AVSValue TFM::ConditionalIsCombedTIVTC(int n, IScriptEnvironment* env);
	static void TFM::DrawYV12(PVideoFrame &dst, int x1, int y1, const char *s);
	static void TFM::DrawYUY2(PVideoFrame &dst, int x1, int y1, const char *s);
	TFM::TFM(PClip _child, int _order, int _field, int _mode, int _PP, const char* _ovr, const char* _input, 
		const char* _output, const char * _outputC, bool _debug, bool _display, bool _slow, 
		bool _mChroma, int _cNum, int _cthresh, int _MI, bool _chroma, int _blockx, int _blocky, 
		int _y0, int _y1, const char* _d2v, int _ovrDefault, IScriptEnvironment* env);
	TFM::~TFM();
};