/*
**                    TIVTC v0.9.9 for AviSynth 2.5.x
**
**   TIVTC includes a field matching filter (TFM) and a decimation
**   filter (TDecimate) which can be used together to achieve an
**   IVTC or for other uses. TIVTC currently supports YV12 and 
**   YUY2 colorspaces.
**   
**   Copyright (C) 2004-2005 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "TFMPP.h"

PVideoFrame __stdcall TFMPP::GetFrame(int n, IScriptEnvironment *env)
{
	if (n < 0) n = 0;
	else if (n > nfrms) n = nfrms;
	bool combed;
	int fieldSrc, field;
	unsigned int hint;
	PVideoFrame src = child->GetFrame(n, env);
	getHint(src, fieldSrc, combed, hint);
	if (!combed) return src;
	int np = vi.IsYV12() ? 3 : 1;
	getSetOvr(n);
	PVideoFrame dst = env->NewVideoFrame(vi);
	PVideoFrame mask = env->NewVideoFrame(vi);
	if (PP > 4)
	{
		int use = 0;
		unsigned int hintt;
		PVideoFrame prv = child->GetFrame(n>0?n-1:0, env);
		getHint(prv, field, combed, hintt);
		if (!combed && field != -1 && n != 0) ++use;
		PVideoFrame nxt = child->GetFrame(n<nfrms?n+1:nfrms, env);
		getHint(nxt, field, combed, hintt);
		if (!combed && field != -1 && n != nfrms) use += 2;
		if (use > 0)
		{
			buildMotionMask(prv, src, nxt, mask, use, np);
			if (uC2) maskClip2(src, clip2->GetFrame(n, env), mask, dst, np);
			else
			{
				if (PP == 5) BlendDeint(src, mask, dst, false, np);
				else
				{
					copyFrame(dst, src, env, np);
					if (PP == 6) CubicDeint(src, mask, dst, false, fieldSrc, np);
					else elaDeint(dst, mask, src, false, fieldSrc, np);
				}
			}
		}
		else
		{
			if (uC2) dst = clip2->GetFrame(n, env);
			else
			{
				if (PP == 5) BlendDeint(src, mask, dst, true, np);
				else
				{
					copyFrame(dst, src, env, np);
					if (PP == 6) CubicDeint(src, mask, dst, true, fieldSrc, np);
					else elaDeint(dst, mask, src, true, fieldSrc, np);
				}
			}
		}
	}
	else
	{
		if (uC2) dst = clip2->GetFrame(n, env);
		else
		{
			if (PP == 2) BlendDeint(src, mask, dst, true, np);
			else 
			{
				copyFrame(dst, src, env, np);
				if (PP == 3) CubicDeint(src, mask, dst, true, fieldSrc, np);
				else elaDeint(dst, mask, src, true, fieldSrc, np);
			}
		}
	}
	if (display) writeDisplay(dst, np, n, fieldSrc);
	putHint(dst, fieldSrc, hint);
	return dst;
}

void TFMPP::buildMotionMask(PVideoFrame &prv, PVideoFrame &src, PVideoFrame &nxt, 
		PVideoFrame &mask, int use, int np)
{
	const unsigned char *srcpp, *srcp, *srcpn;
	const unsigned char *prvpp, *prvp, *prvpn;
	const unsigned char *nxtpp, *nxtp, *nxtpn;
	unsigned char *maskw;
	int b, plane, x, y, width, height;
	int src_pitch, msk_pitch, nxt_pitch, prv_pitch;
	for (b=0; b<np; ++b)
	{
		if (b == 0) plane = PLANAR_Y;
		else if (b == 1) plane = PLANAR_U;
		else plane = PLANAR_V;
		prvpp = prv->GetReadPtr(plane);
		prv_pitch = prv->GetPitch(plane);
		prvp = prvpp + prv_pitch;
		prvpn = prvp + prv_pitch;
		srcpp = src->GetReadPtr(plane);
		src_pitch = src->GetPitch(plane);
		width = src->GetRowSize(plane);
		height = src->GetHeight(plane);
		srcp = srcpp + src_pitch;
		srcpn = srcp + src_pitch;
		nxtpp = nxt->GetReadPtr(plane);
		nxt_pitch = nxt->GetPitch(plane);
		nxtp = nxtpp + nxt_pitch;
		nxtpn = nxtp + nxt_pitch;
		maskw = mask->GetWritePtr(plane);
		msk_pitch = mask->GetPitch(plane);
		memset(maskw, 0xFF, msk_pitch*height);
		maskw += msk_pitch;
		for (y=1; y<height-1; ++y)
		{
			if (use == 1)
			{
				for (x=0; x<width; ++x) 
				{
					if (!(abs(prvpp[x] - srcpp[x]) > mthresh || abs(prvp[x] - srcp[x]) > mthresh || 
						abs(prvpn[x] - srcpn[x]) > mthresh)) maskw[x] = 0;
				}
			}
			else if (use == 2)
			{
				for (x=0; x<width; ++x) 
				{
					if (!(abs(nxtpp[x] - srcpp[x]) > mthresh || abs(nxtp[x] - srcp[x]) > mthresh || 
						abs(nxtpn[x] - srcpn[x]) > mthresh)) maskw[x] = 0;
				}
			}
			else
			{
				for (x=0; x<width; ++x) 
				{
					if (!(((abs(prvp[x] - srcp[x]) > mthresh) && (abs(nxtpp[x] - srcpp[x]) > mthresh || 
						abs(nxtp[x] - srcp[x]) > mthresh || abs(nxtpn[x] - srcpn[x]) > mthresh)) || 
						((abs(nxtp[x] - srcp[x]) > mthresh) && (abs(prvpp[x] - srcpp[x]) > mthresh || 
						abs(prvp[x] - srcp[x]) > mthresh || abs(prvpn[x] - srcpn[x]) > mthresh)) || 
						(abs(prvpp[x] - srcpp[x]) > mthresh && abs(prvpn[x] - srcpn[x]) > mthresh && 
						(abs(nxtpp[x] - srcpp[x]) > mthresh || abs(nxtpn[x] - srcpn[x]) > mthresh)) ||
						((abs(prvpp[x] - srcpp[x]) > mthresh || abs(prvpn[x] - srcpn[x]) > mthresh) && 
						abs(nxtpp[x] - srcpp[x]) > mthresh && abs(nxtpn[x] - srcpn[x]) > mthresh)))
						maskw[x] = 0;
				}
			}
			prvpp += prv_pitch;
			prvp += prv_pitch;
			prvpn += prv_pitch;
			srcpp += src_pitch;
			srcp += src_pitch;
			srcpn += src_pitch;
			nxtpp += nxt_pitch;
			nxtp += nxt_pitch;
			nxtpn += nxt_pitch;
			maskw += msk_pitch;
		}
	}
	if (np == 3)
	{
		denoiseYV12(mask);
		linkYV12(mask);
	}
	else
	{
		denoiseYUY2(mask);
		linkYUY2(mask);
	}
}

void TFMPP::denoiseYUY2(PVideoFrame &mask) 
{
	unsigned char *maskw = mask->GetWritePtr();
	int mask_pitch = mask->GetPitch();
	int Height = mask->GetHeight();
	int Width = mask->GetRowSize();
	unsigned char *maskwp = maskw - mask_pitch;
	unsigned char *maskwn = maskw + mask_pitch;
	int x, y, sum, v, xhi;
	for (y=1; y<Height-1; ++y)
	{
		maskwp = maskw;
		maskw = maskwn;
		maskwn += mask_pitch;
		for (x=4; x<Width-4; ++x)
		{
			if (maskw[x] == 0xFF)
			{
				for (sum=0,v=x-2,xhi=x+3; v<xhi && sum<2; v+=2)
				{
					if (maskwp[v] == 0xFF) ++sum;
					if (maskw[v] == 0xFF) ++sum;
					if (maskwn[v] == 0xFF) ++sum;
				}
				if (sum < 2) maskw[x] = 0;
			}
			++x;
			if (maskw[x] == 0xFF)
			{
				for (sum=0,v=x-4,xhi=x+5; v<xhi && sum<2; v+=4)
				{
					if (maskwp[v] == 0xFF) ++sum;
					if (maskw[v] == 0xFF) ++sum;
					if (maskwn[v] == 0xFF) ++sum;
				}
				if (sum < 2) maskw[x] = 0;
			}
		}
	}
}

void TFMPP::linkYUY2(PVideoFrame &mask) 
{
	unsigned char *maskw = mask->GetWritePtr();
	int mask_pitch = mask->GetPitch();
	int Height = mask->GetHeight();
	int Width = mask->GetRowSize(), x, y;
	for (y=1; y<Height-1; ++y)
	{
		maskw += mask_pitch;
		for (x=0; x<Width; x+=4)
		{
			if (maskw[x] == 0xFF && maskw[x+2] == 0xFF)
			{
				maskw[x+1] = maskw[x+3] = 0xFF;
			}
		}
	}
}

void TFMPP::denoiseYV12(PVideoFrame &mask) 
{
	unsigned char *maskp, *maskpp, *maskpn;
	int msk_pitch, Height, Width, sum, xhi, b, v, x, y, plane;
	for (b=0; b<3; ++b)
	{
		if (b == 0) plane = PLANAR_Y;
		else if (b == 1) plane = PLANAR_V;
		else plane = PLANAR_U;
		maskpp = mask->GetWritePtr(plane);
		msk_pitch = mask->GetPitch(plane);
		maskp = maskpp + msk_pitch;
		maskpn = maskp + msk_pitch;
		Height = mask->GetHeight(plane);
		Width = mask->GetRowSize(plane);
		for (y=1; y<Height-1; ++y)
		{
			for (x=1; x<Width-1; ++x)
			{
				if (maskp[x] == 0xFF)
				{
					for (sum=0,v=x-1,xhi=x+2; v<xhi && sum<2; ++v)
					{
						if (maskpp[v] == 0xFF) ++sum;
						if (maskp[v] == 0xFF) ++sum;
						if (maskpn[v] == 0xFF) ++sum;
					}
					if (sum < 2) maskp[x] = 0;
				}
			}
			maskpp += msk_pitch;
			maskp += msk_pitch;
			maskpn += msk_pitch;
		}
	}
}

void TFMPP::linkYV12(PVideoFrame &mask) 
{
	unsigned char *maskpY = mask->GetWritePtr(PLANAR_Y);
	unsigned char *maskpV = mask->GetWritePtr(PLANAR_V);
	unsigned char *maskpU = mask->GetWritePtr(PLANAR_U);
	int mask_pitchY = mask->GetPitch(PLANAR_Y);
	int mask_pitchUV = mask->GetPitch(PLANAR_V);
	int HeightUV = mask->GetHeight(PLANAR_V);
	int WidthUV = mask->GetRowSize(PLANAR_V), x, y;
	unsigned char *maskppY = maskpY - mask_pitchY;
	unsigned char *maskpnY = maskpY + mask_pitchY;
	unsigned char *maskpnnY = maskpnY + mask_pitchY;
	mask_pitchY <<= 1;
	for (y=1; y<HeightUV-1; ++y)
	{
		maskppY = maskpnY;
		maskpY = maskpnnY;
		maskpnY += mask_pitchY;
		maskpnnY += mask_pitchY;
		maskpV += mask_pitchUV;
		maskpU += mask_pitchUV;
		for (x=0; x<WidthUV; ++x)
		{
			if ((((unsigned short*)maskpY)[x] == (unsigned short) 0xFFFF) && 
				(((unsigned short*)maskpnY)[x] == (unsigned short) 0xFFFF) && 
				(((y&1) && (((unsigned short*)maskppY)[x] == (unsigned short) 0xFFFF)) || 
				(!(y&1) && (((unsigned short*)maskpnnY)[x] == (unsigned short) 0xFFFF))))
			{
				maskpV[x] = maskpU[x] = 0xFF;
			}
		}
	}
}

void TFMPP::BlendDeint(PVideoFrame &src, PVideoFrame &mask, PVideoFrame &dst, bool nomask, int np)
{
	const unsigned char *srcpp, *srcp, *srcpn, *maskp;
	unsigned char *dstp;
	int b, plane, x, y, width, height, src_pitch, dst_pitch, msk_pitch;
	for (b=0; b<np; ++b)
	{
		if (b == 0) plane = PLANAR_Y;
		else if (b == 1) plane = PLANAR_U;
		else plane = PLANAR_V;
		srcp = src->GetReadPtr(plane);
		src_pitch = src->GetPitch(plane);
		width = src->GetRowSize(plane);
		height = src->GetHeight(plane);
		srcpp = srcp - src_pitch;
		srcpn = srcp + src_pitch;
		dstp = dst->GetWritePtr(plane);
		dst_pitch = dst->GetPitch(plane);
		maskp = mask->GetReadPtr(plane);
		msk_pitch = mask->GetPitch(plane);
		for (x=0; x<width; ++x) dstp[x] = (srcp[x]+srcpn[x]+1)>>1;
		srcpp += src_pitch;
		srcp += src_pitch;
		srcpn += src_pitch;
		dstp += dst_pitch;
		maskp += msk_pitch;
		for (y=1; y<height-1; ++y)
		{
			for (x=0; x<width; ++x) 
			{
				if (nomask || maskp[x] == 0xFF) 
					dstp[x] = (srcpp[x]+srcp[x]+srcp[x]+srcpn[x]+2)>>2;
				else dstp[x] = srcp[x];
			}
			srcpp += src_pitch;
			srcp += src_pitch;
			srcpn += src_pitch;
			dstp += dst_pitch;
			maskp += msk_pitch;
		}
		for (x=0; x<width; ++x) dstp[x] = (srcpp[x]+srcp[x]+1)>>1;
	}
}

void TFMPP::CubicDeint(PVideoFrame &src, PVideoFrame &mask, PVideoFrame &dst, bool nomask, 
					   int field, int np)
{
	const unsigned char *srcppp, *srcpp, *srcp, *srcpn, *maskp;
	unsigned char *dstp;
	int b, plane, x, y, width, height, src_pitch, dst_pitch, msk_pitch;
	for (b=0; b<np; ++b)
	{
		if (b == 0) plane = PLANAR_Y;
		else if (b == 1) plane = PLANAR_U;
		else plane = PLANAR_V;
		srcp = src->GetReadPtr(plane);
		src_pitch = src->GetPitch(plane);
		width = src->GetRowSize(plane);
		height = src->GetHeight(plane);
		dstp = dst->GetWritePtr(plane);
		dst_pitch = dst->GetPitch(plane);
		maskp = mask->GetReadPtr(plane);
		msk_pitch = mask->GetPitch(plane);
		srcp += src_pitch*(3-field);
		dstp += dst_pitch*(2-field);
		maskp += msk_pitch*(2-field);
		src_pitch <<= 1;
		dst_pitch <<= 1;
		msk_pitch <<= 1;
		srcpp = srcp - src_pitch;
		srcppp = srcpp - src_pitch;
		srcpn = srcp + src_pitch;
		for (x=0; x<width; ++x)
		{
			if (nomask || maskp[x] == 0xFF) dstp[x] = (srcp[x]+srcpp[x]+1)>>1;
		}
		srcppp += src_pitch;
		srcpp += src_pitch;
		srcp += src_pitch;
		srcpn += src_pitch;
		maskp += msk_pitch;
		dstp += dst_pitch;
		for (y=4-field; y<height-3; y+=2)
		{
			for (x=0; x<width; ++x)
			{
				if (nomask || maskp[x] == 0xFF) 
					dstp[x] = cubicInt(srcppp[x],srcpp[x],srcp[x],srcpn[x]);
			}
			srcppp += src_pitch;
			srcpp += src_pitch;
			srcp += src_pitch;
			srcpn += src_pitch;
			maskp += msk_pitch;
			dstp += dst_pitch;
		}
		for (x=0; x<width; ++x)
		{
			if (nomask || maskp[x] == 0xFF) dstp[x] = (srcp[x]+srcpp[x]+1)>>1;
		}
	}
}

void TFMPP::putHint(PVideoFrame &dst, int field, unsigned int hint)
{
	unsigned char *p = dst->GetWritePtr(PLANAR_Y);
	unsigned int i;
	hint &= D2VFILM;
	if (field == 1) 
	{
		hint |= TOP_FIELD;
		hint |= ISDT;
	}
	else hint |= ISDB;
	for (i=0; i<32; ++i)
	{
		*p &= ~1;
		*p++ |= ((MAGIC_NUMBER & (1 << i)) >> i);
	}
	for (i=0; i<32; ++i)
	{
		*p &= ~1;
		*p++ |= ((hint & (1 << i)) >> i);
	}
}

void TFMPP::getHint(PVideoFrame &src, int &field, bool &combed, unsigned int &hint)
{
	field = -1; combed = false; hint = 0;
	const unsigned char *srcp = src->GetReadPtr(PLANAR_Y);
	unsigned int i, magic_number = 0;
	for (i=0; i<32; ++i)
	{
		magic_number |= ((*srcp++ & 1) << i);
	}
	if (magic_number != MAGIC_NUMBER) return;
	for (i=0; i<32; ++i)
	{
		hint |= ((*srcp++ & 1) << i);
	}
	if (hint&0xFFFFFFC0) return;
	if (hint&TOP_FIELD) field = 1;
	else field = 0;
	if (hint&COMBED) combed = true;
	int value = hint&0x07;
	if (value == 5) { combed = true; field = 0; }
	else if (value == 6) { combed = true; field = 1; }
}

void TFMPP::getSetOvr(int n)
{
	if (setArray == NULL || setArraySize <= 0) return;
	mthresh = mthreshS;
	PP = PPS;
	for (int x=0; x<setArraySize; x+=4)
	{
		if (n >= setArray[x+1] && n <= setArray[x+2])
		{
			if (setArray[x] == 80) PP = setArray[x+3];
			else if (setArray[x] == 77) mthresh = setArray[x+3];
		}
	}
}

void TFMPP::copyFrame(PVideoFrame &dst, PVideoFrame &src, IScriptEnvironment *env, int np)
{
	int b, plane;
	for (b=0; b<np; ++b)
	{
		if (b == 0) plane = PLANAR_Y;
		else if (b == 1) plane = PLANAR_V;
		else plane = PLANAR_U;
		env->BitBlt(dst->GetWritePtr(plane), dst->GetPitch(plane), src->GetReadPtr(plane),
			src->GetPitch(plane), src->GetRowSize(plane), src->GetHeight(plane));
	}
}

unsigned char TFMPP::cubicInt(unsigned char p1, unsigned char p2, unsigned char p3, unsigned char p4)
{
	int temp = (int)((19*(p2+p3)-3*(p1+p4)+16)>>5);
	if (temp>255) temp = 255; 
	else if (temp<0) temp = 0;
	return (unsigned char)temp;
}

void TFMPP::writeDisplay(PVideoFrame &dst, int np, int n, int field)
{
	sprintf(buf, "TFMPP %s by tritical ", VERSION);
	if (np == 3) TFM::DrawYV12(dst, 0, 0, buf);
	else TFM::DrawYUY2(dst, 0, 0, buf);
	sprintf(buf, "field = %d  PP = %d  mthresh = %d ", field, PP, mthresh);
	if (np == 3) TFM::DrawYV12(dst, 0, 1, buf);
	else TFM::DrawYUY2(dst, 0, 1, buf);
	sprintf(buf, "frame: %d  (COMBED - DEINTERLACED)! ", n);
	if (np == 3) TFM::DrawYV12(dst, 0, 2, buf);
	else TFM::DrawYUY2(dst, 0, 2, buf);
}

AVSValue __cdecl Create_TFMPP(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	return new TFMPP(args[0].AsClip(),args[1].AsInt(6),args[2].AsInt(5),
		args[3].AsString(""),args[4].AsBool(false),(args[5].IsClip() ? args[5].AsClip() : NULL),env);
}

void TFMPP::elaDeint(PVideoFrame &dst, PVideoFrame &mask, PVideoFrame &src, bool nomask, int field, int np)
{
	if (np == 3) elaDeintYV12(dst, mask, src, nomask, field);
	else elaDeintYUY2(dst, mask, src, nomask, field);
}

void TFMPP::elaDeintYV12(PVideoFrame &dst, PVideoFrame &mask, PVideoFrame &src, bool nomask, int field)
{
	const unsigned char *srcpY = src->GetReadPtr(PLANAR_Y);
	const unsigned char *srcpV = src->GetReadPtr(PLANAR_V);
	const unsigned char *srcpU = src->GetReadPtr(PLANAR_U);
	int src_pitchY = src->GetPitch(PLANAR_Y);
	int src_pitchUV = src->GetPitch(PLANAR_V);
	int WidthY = src->GetRowSize(PLANAR_Y);
	int WidthUV = src->GetRowSize(PLANAR_V);
	int HeightY = src->GetHeight(PLANAR_Y);
	int HeightUV = src->GetHeight(PLANAR_V);
	unsigned char *dstpY = dst->GetWritePtr(PLANAR_Y);
	unsigned char *dstpV = dst->GetWritePtr(PLANAR_V);
	unsigned char *dstpU = dst->GetWritePtr(PLANAR_U);
	int dst_pitchY = dst->GetPitch(PLANAR_Y);
	int dst_pitchUV = dst->GetPitch(PLANAR_V);
	const unsigned char *maskpY = mask->GetReadPtr(PLANAR_Y);
	const unsigned char *maskpV = mask->GetReadPtr(PLANAR_V);
	const unsigned char *maskpU = mask->GetReadPtr(PLANAR_U);
	int mask_pitchY = mask->GetPitch(PLANAR_Y);
	int mask_pitchUV = mask->GetPitch(PLANAR_V);
	srcpY += src_pitchY*(3-field);
	srcpV += src_pitchUV*(3-field);
	srcpU += src_pitchUV*(3-field);
	dstpY += dst_pitchY*(2-field);
	dstpV += dst_pitchUV*(2-field);
	dstpU += dst_pitchUV*(2-field);
	maskpY += mask_pitchY*(2-field);
	maskpV += mask_pitchUV*(2-field);
	maskpU += mask_pitchUV*(2-field);
	src_pitchY <<= 1;
	src_pitchUV <<= 1;
	dst_pitchY <<= 1;
	dst_pitchUV <<= 1;
	mask_pitchY <<= 1;
	mask_pitchUV <<= 1;
	const unsigned char *srcppY = srcpY - src_pitchY;
	const unsigned char *srcpppY = srcppY - src_pitchY;
	const unsigned char *srcpnY = srcpY + src_pitchY;
	const unsigned char *srcppV = srcpV - src_pitchUV;
	const unsigned char *srcpppV = srcppV - src_pitchUV;
	const unsigned char *srcpnV = srcpV + src_pitchUV;
	const unsigned char *srcppU = srcpU - src_pitchUV;
	const unsigned char *srcpppU = srcppU - src_pitchUV;
	const unsigned char *srcpnU = srcpU + src_pitchUV;
	int stopx = WidthY;
	int startxuv = 0, x, y;
	int stopxuv = WidthUV;
	int Iy1, Iy2, Iye, Ix1, Ix2, edgeS1, edgeS2, sum, sumsq, temp, temp1, temp2, minN, maxN;
	double dir1, dir2, dir, dirF;
	for (y=2-field; y<HeightY-1; y+=2)
	{
		for (x=0; x<stopx; ++x)
		{
			if (nomask || maskpY[x] == 0xFF)
			{
				if (y>2 && y<HeightY-3 && x>3 && x<WidthY-4)
				{
					Iy1 = -srcpY[x-1]-srcpY[x]-srcpY[x]-srcpY[x+1]+srcpppY[x-1]+srcpppY[x]+srcpppY[x]+srcpppY[x+1];
					Iy2 = -srcpnY[x-1]-srcpnY[x]-srcpnY[x]-srcpnY[x+1]+srcppY[x-1]+srcppY[x]+srcppY[x]+srcppY[x+1];
					Ix1 = srcpppY[x+1]+srcppY[x+1]+srcppY[x+1]+srcpY[x+1]-srcpppY[x-1]-srcppY[x-1]-srcppY[x-1]-srcpY[x-1];
					Ix2 = srcppY[x+1]+srcpY[x+1]+srcpY[x+1]+srcpnY[x+1]-srcppY[x-1]-srcpY[x-1]-srcpY[x-1]-srcpnY[x-1];
					edgeS1 = Ix1*Ix1 + Iy1*Iy1;
					edgeS2 = Ix2*Ix2 + Iy2*Iy2;
					if (edgeS1 < 1600 && edgeS2 < 1600)
					{
						dstpY[x] = (srcppY[x]+srcpY[x]+1)>>1;
						continue;
					}
					if (abs(srcppY[x]-srcpY[x]) < 10 && (edgeS1 < 1600 || edgeS2 < 1600))
					{
						dstpY[x] = (srcppY[x]+srcpY[x]+1)>>1;
						continue;
					}
					sum = srcppY[x-1] + srcppY[x] + srcppY[x+1] + srcpY[x-1] + srcpY[x] + srcpY[x+1];
					sumsq = srcppY[x-1]*srcppY[x-1] + srcppY[x]*srcppY[x] + srcppY[x+1]*srcppY[x+1] +
							srcpY[x-1]*srcpY[x-1] + srcpY[x]*srcpY[x] + srcpY[x+1]*srcpY[x+1];
					if ((6*sumsq - sum*sum) < 432)
					{
						dstpY[x] = (srcppY[x]+srcpY[x]+1)>>1;
						continue;
					}
					if (Ix1 == 0) dir1 = 3.1415926;
					else 
					{
						dir1 = atan(Iy1/(Ix1*2.0f)) + 1.5707963;
						if (Iy1 >= 0) { if (Ix1 < 0) dir1 += 3.1415927; }
						else { if (Ix1 >= 0) dir1 += 3.1415927; }
						if (dir1 >= 3.1415927) dir1 -= 3.1415927;
					}
					if (Ix2 == 0) dir2 = 3.1415926;
					else 
					{
						dir2 = atan(Iy2/(Ix2*2.0f)) + 1.5707963;
						if (Iy2 >= 0) { if (Ix2 < 0) dir2 += 3.1415927; }
						else { if (Ix2 >= 0) dir2 += 3.1415927; }
						if (dir2 >= 3.1415927) dir2 -= 3.1415927;
					}
					if (fabs(dir1-dir2) < 0.5f) 
					{
						if (edgeS1 >= 3600 && edgeS2 >= 3600) dir = (dir1 + dir2) * 0.5f;
						else dir = edgeS1 >= edgeS2 ? dir1 : dir2;
					}
					else 
					{
						if (edgeS1 >= 5000 && edgeS2 >= 5000)
						{
							Iye = -srcpY[x-1]-srcpY[x]-srcpY[x]-srcpY[x+1]+srcppY[x-1]+srcppY[x]+srcppY[x]+srcppY[x+1];
							if ((Iy1*Iye > 0) && (Iy2*Iye < 0)) dir = dir1;
							else if ((Iy1*Iye < 0) && (Iy2*Iye > 0)) dir = dir2;
							else 
							{
								if (abs(Iye-Iy1) <= abs(Iye-Iy2)) dir = dir1;
								else dir = dir2;
							}
						}
						else dir = edgeS1 >= edgeS2 ? dir1 : dir2;
					}
					dirF = 0.5f/tan(dir);
					if (dirF >= 0.0f)
					{
						if (dirF >= 0.5f)
						{
							if (dirF >= 1.0f)
							{
								if (dirF >= 1.5f)
								{
									if (dirF >= 2.0f)
									{
										if (dirF <= 2.50f)
										{
											temp1 = srcppY[x+4];
											temp2 = srcpY[x-4];
											temp = (srcppY[x+4]+srcpY[x-4]+1)>>1;
										}
										else 
										{
											temp1 = temp2 = srcpY[x];
											temp = cubicInt(srcpppY[x],srcppY[x],srcpY[x],srcpnY[x]);
										}
									}
									else 
									{
										temp1 = (int)((dirF-1.5f)*(srcppY[x+4]) + (2.0f-dirF)*(srcppY[x+3]) + 0.5f);
										temp2 = (int)((dirF-1.5f)*(srcpY[x-4]) + (2.0f-dirF)*(srcpY[x-3]) + 0.5f);
										temp = (int)((dirF-1.5f)*(srcppY[x+4]+srcpY[x-4]) + (2.0f-dirF)*(srcppY[x+3]+srcpY[x-3]) + 0.5f);
									}
								}
								else 
								{
									temp1 = (int)((dirF-1.0f)*(srcppY[x+3]) + (1.5f-dirF)*(srcppY[x+2]) + 0.5f);
									temp2 = (int)((dirF-1.0f)*(srcpY[x-3]) + (1.5f-dirF)*(srcpY[x-2]) + 0.5f);
									temp = (int)((dirF-1.0f)*(srcppY[x+3]+srcpY[x-3]) + (1.5f-dirF)*(srcppY[x+2]+srcpY[x-2]) + 0.5f);
								}
							}
							else 
							{
								temp1 = (int)((dirF-0.5f)*(srcppY[x+2]) + (1.0f-dirF)*(srcppY[x+1]) + 0.5f);
								temp2 = (int)((dirF-0.5f)*(srcpY[x-2]) + (1.0f-dirF)*(srcpY[x-1]) + 0.5f);
								temp = (int)((dirF-0.5f)*(srcppY[x+2]+srcpY[x-2]) + (1.0f-dirF)*(srcppY[x+1]+srcpY[x-1]) + 0.5f);
							}
						}
						else 
						{
							temp1 = (int)(dirF*(srcppY[x+1]) + (0.5f-dirF)*(srcppY[x]) + 0.5f);
							temp2 = (int)(dirF*(srcpY[x-1]) + (0.5f-dirF)*(srcpY[x]) + 0.5f);
							temp = (int)(dirF*(srcppY[x+1]+srcpY[x-1]) + (0.5f-dirF)*(srcppY[x]+srcpY[x]) + 0.5f);
						}
					}
					else
					{
						if (dirF <= -0.5f)
						{
							if (dirF <= -1.0f)
							{
								if (dirF <= -1.5f)
								{
									if (dirF <= -2.0f)
									{
										if (dirF >= -2.50f) 
										{
											temp1 = srcppY[x-4];
											temp2 = srcpY[x+4];
											temp = (srcppY[x-4]+srcpY[x+4]+1)>>1;
										}
										else 
										{
											temp1 = temp2 = srcpY[x];
											temp = cubicInt(srcpppY[x],srcppY[x],srcpY[x],srcpnY[x]);
										}
									}
									else
									{
										temp1 = (int)((-dirF-1.5f)*(srcppY[x-4]) + (2.0f+dirF)*(srcppY[x-3]) + 0.5f);
										temp2 = (int)((-dirF-1.5f)*(srcpY[x+4]) + (2.0f+dirF)*(srcpY[x+3]) + 0.5f);
										temp = (int)((-dirF-1.5f)*(srcppY[x-4]+srcpY[x+4]) + (2.0f+dirF)*(srcppY[x-3]+srcpY[x+3]) + 0.5f);
									}
								}
								else 
								{
									temp1 = (int)((-dirF-1.0f)*(srcppY[x-3]) + (1.5f+dirF)*(srcppY[x-2]) + 0.5f);
									temp2 = (int)((-dirF-1.0f)*(srcpY[x+3]) + (1.5f+dirF)*(srcpY[x+2]) + 0.5f);
									temp = (int)((-dirF-1.0f)*(srcppY[x-3]+srcpY[x+3]) + (1.5f+dirF)*(srcppY[x-2]+srcpY[x+2]) + 0.5f);
								}
							}
							else 
							{
								temp1 = (int)((-dirF-0.5f)*(srcppY[x-2]) + (1.0f+dirF)*(srcppY[x-1]) + 0.5f);
								temp2 = (int)((-dirF-0.5f)*(srcpY[x+2]) + (1.0f+dirF)*(srcpY[x+1]) + 0.5f);
								temp = (int)((-dirF-0.5f)*(srcppY[x-2]+srcpY[x+2]) + (1.0f+dirF)*(srcppY[x-1]+srcpY[x+1]) + 0.5f);
							}
						}
						else 
						{
							temp1 = (int)((-dirF)*(srcppY[x-1]) + (0.5f+dirF)*(srcppY[x]) + 0.5f);
							temp2 = (int)((-dirF)*(srcpY[x+1]) + (0.5f+dirF)*(srcpY[x]) + 0.5f);
							temp = (int)((-dirF)*(srcppY[x-1]+srcpY[x+1]) + (0.5f+dirF)*(srcppY[x]+srcpY[x]) + 0.5f);
						}
					}
					minN = min(srcppY[x],srcpY[x]) - 25;
					maxN = max(srcppY[x],srcpY[x]) + 25;
					if (abs(temp1-temp2) > 20 || abs(srcppY[x]+srcpY[x]-temp-temp) > 60 || temp < minN || temp > maxN)
					{
						temp = cubicInt(srcpppY[x],srcppY[x],srcpY[x],srcpnY[x]);
					}
					if (temp > 255) temp = 255;
					else if (temp < 0) temp = 0;
					dstpY[x] = temp;
				}
				else
				{
					if (y<3 || y>HeightY-4) dstpY[x] = ((srcpY[x] + srcppY[x] + 1) >> 1);
					else dstpY[x] = cubicInt(srcpppY[x],srcppY[x],srcpY[x],srcpnY[x]);
				}
			}
		}
		srcpppY = srcppY;
		srcppY = srcpY;
		srcpY = srcpnY;
		srcpnY += src_pitchY;
		maskpY += mask_pitchY;
		dstpY += dst_pitchY;
	}
	for (y=2-field; y<HeightUV-1; y+=2)
	{
		for (x=startxuv; x<stopxuv; ++x)
		{
			if (nomask || maskpV[x] == 0xFF)
			{
				if (y<3 || y>HeightUV-4) dstpV[x] = ((srcpV[x] + srcppV[x] + 1) >> 1);
				else dstpV[x] = cubicInt(srcpppV[x],srcppV[x],srcpV[x],srcpnV[x]);
			}
			if (nomask || maskpU[x] == 0xFF)
			{
				if (y<3 || y>HeightUV-4) dstpU[x] = ((srcpU[x] + srcppU[x] + 1) >> 1);
				else dstpU[x] = cubicInt(srcpppU[x],srcppU[x],srcpU[x],srcpnU[x]);
			}
		}
		srcpppV = srcppV;
		srcppV = srcpV;
		srcpV = srcpnV;
		srcpnV += src_pitchUV;
		srcpppU = srcppU;
		srcppU = srcpU;
		srcpU = srcpnU;
		srcpnU += src_pitchUV;
		maskpV += mask_pitchUV;
		maskpU += mask_pitchUV;
		dstpV += dst_pitchUV;
		dstpU += dst_pitchUV;
	}
}

void TFMPP::elaDeintYUY2(PVideoFrame &dst, PVideoFrame &mask, PVideoFrame &src, bool nomask, int field)
{
	const unsigned char *srcp = src->GetReadPtr();
	int src_pitch = src->GetPitch();
	int Width = src->GetRowSize();
	int Height = src->GetHeight();
	unsigned char *dstp = dst->GetWritePtr();
	int dst_pitch = dst->GetPitch();
	const unsigned char *maskp = mask->GetReadPtr();
	int mask_pitch = mask->GetPitch();
	srcp += src_pitch*(3-field);
	dstp += dst_pitch*(2-field);
	maskp += mask_pitch*(2-field);
	src_pitch <<= 1;
	dst_pitch <<= 1;
	mask_pitch <<= 1;
	const unsigned char *srcpp = srcp - src_pitch;
	const unsigned char *srcppp = srcpp - src_pitch;
	const unsigned char *srcpn = srcp + src_pitch;
	int stopx = Width;
	int Iy1, Iy2, Iye, Ix1, Ix2, edgeS1, edgeS2, sum, sumsq, temp, temp1, temp2, minN, maxN, x, y;
	double dir1, dir2, dir, dirF;
	for (y=2-field; y<Height-1; y+=2)
	{
		for (x=0; x<stopx; ++x)
		{
			if (nomask || maskp[x] == 0xFF)
			{
				if (y>2 && y<Height-3 && x>7 && x<Width-9)
				{
					Iy1 = -srcp[x-2]-srcp[x]-srcp[x]-srcp[x+2]+srcppp[x-2]+srcppp[x]+srcppp[x]+srcppp[x+2];
					Iy2 = -srcpn[x-2]-srcpn[x]-srcpn[x]-srcpn[x+2]+srcpp[x-2]+srcpp[x]+srcpp[x]+srcpp[x+2];
					Ix1 = srcppp[x+2]+srcpp[x+2]+srcpp[x+2]+srcp[x+2]-srcppp[x-2]-srcpp[x-2]-srcpp[x-2]-srcp[x-2];
					Ix2 = srcpp[x+2]+srcp[x+2]+srcp[x+2]+srcpn[x+2]-srcpp[x-2]-srcp[x-2]-srcp[x-2]-srcpn[x-2];
					edgeS1 = Ix1*Ix1 + Iy1*Iy1;
					edgeS2 = Ix2*Ix2 + Iy2*Iy2;
					if (edgeS1 < 1600 && edgeS2 < 1600) 
					{
						dstp[x] = (srcpp[x]+srcp[x]+1)>>1;
						goto chromajump;
					}
					if (abs(srcpp[x]-srcp[x]) < 10 && (edgeS1 < 1600 || edgeS2 < 1600))
					{
						dstp[x] = (srcpp[x]+srcp[x]+1)>>1;
						goto chromajump;
					}
					sum = srcpp[x-2] + srcpp[x] + srcpp[x+2] + srcp[x-2] + srcp[x] + srcp[x+2];
					sumsq = srcpp[x-2]*srcpp[x-2] + srcpp[x]*srcpp[x] + srcpp[x+2]*srcpp[x+2] +
							srcp[x-2]*srcp[x-2] + srcp[x]*srcp[x] + srcp[x+2]*srcp[x+2];
					if ((6*sumsq - sum*sum) < 432)
					{
						dstp[x] = (srcpp[x]+srcp[x]+1)>>1;
						goto chromajump;
					}
					if (Ix1 == 0) dir1 = 3.1415926;
					else 
					{
						dir1 = atan(Iy1/(Ix1*2.0f)) + 1.5707963;
						if (Iy1 >= 0) { if (Ix1 < 0) dir1 += 3.1415927; }
						else { if (Ix1 >= 0) dir1 += 3.1415927; }
						if (dir1 >= 3.1415927) dir1 -= 3.1415927;
					}
					if (Ix2 == 0) dir2 = 3.1415926;
					else 
					{
						dir2 = atan(Iy2/(Ix2*2.0f)) + 1.5707963;
						if (Iy2 >= 0) { if (Ix2 < 0) dir2 += 3.1415927; }
						else { if (Ix2 >= 0) dir2 += 3.1415927; }
						if (dir2 >= 3.1415927) dir2 -= 3.1415927;
					}
					if (fabs(dir1-dir2) < 0.5f) 
					{
						if (edgeS1 >= 3600 && edgeS2 >= 3600) dir = (dir1 + dir2) * 0.5f;
						else dir = edgeS1 >= edgeS2 ? dir1 : dir2;
					}
					else 
					{
						if (edgeS1 >= 5000 && edgeS2 >= 5000)
						{
							Iye = -srcp[x-2]-srcp[x]-srcp[x]-srcp[x+2]+srcpp[x-2]+srcpp[x]+srcpp[x]+srcpp[x+2];
							if ((Iy1*Iye > 0) && (Iy2*Iye < 0)) dir = dir1;
							else if ((Iy1*Iye < 0) && (Iy2*Iye > 0)) dir = dir2;
							else 
							{
								if (abs(Iye-Iy1) <= abs(Iye-Iy2)) dir = dir1;
								else dir = dir2;
							}
						}
						else dir = edgeS1 >= edgeS2 ? dir1 : dir2;
					}
					dirF = 0.5f/tan(dir);
					if (dirF >= 0.0f)
					{
						if (dirF >= 0.5f)
						{
							if (dirF >= 1.0f)
							{
								if (dirF >= 1.5f)
								{
									if (dirF >= 2.0f)
									{
										if (dirF <= 2.50f) 
										{
											temp1 = srcpp[x+8];
											temp2 = srcp[x-8];
											temp = (srcpp[x+8]+srcp[x-8]+1)>>1;
										}
										else 
										{
											temp1 = temp2 = srcp[x];
											temp = cubicInt(srcppp[x],srcpp[x],srcp[x],srcpn[x]);
										}
									}
									else 
									{
										temp1 = (int)((dirF-1.5f)*(srcpp[x+8]) + (2.0f-dirF)*(srcpp[x+6]) + 0.5f);
										temp2 = (int)((dirF-1.5f)*(srcp[x-8]) + (2.0f-dirF)*(srcp[x-6]) + 0.5f); 
										temp = (int)((dirF-1.5f)*(srcpp[x+8]+srcp[x-8]) + (2.0f-dirF)*(srcpp[x+6]+srcp[x-6]) + 0.5f);
									}
								}
								else 
								{
									temp1 = (int)((dirF-1.0f)*(srcpp[x+6]) + (1.5f-dirF)*(srcpp[x+4]) + 0.5f);
									temp2 = (int)((dirF-1.0f)*(srcp[x-6]) + (1.5f-dirF)*(srcp[x-4]) + 0.5f);
									temp = (int)((dirF-1.0f)*(srcpp[x+6]+srcp[x-6]) + (1.5f-dirF)*(srcpp[x+4]+srcp[x-4]) + 0.5f);
								}
							}
							else 
							{
								temp1 = (int)((dirF-0.5f)*(srcpp[x+4]) + (1.0f-dirF)*(srcpp[x+2]) + 0.5f);
								temp2 = (int)((dirF-0.5f)*(srcp[x-4]) + (1.0f-dirF)*(srcp[x-2]) + 0.5f);
								temp = (int)((dirF-0.5f)*(srcpp[x+4]+srcp[x-4]) + (1.0f-dirF)*(srcpp[x+2]+srcp[x-2]) + 0.5f);
							}
						}
						else 
						{
							temp1 = (int)(dirF*(srcpp[x+2]) + (0.5f-dirF)*(srcpp[x]) + 0.5f);
							temp2 = (int)(dirF*(srcp[x-2]) + (0.5f-dirF)*(srcp[x]) + 0.5f);
							temp = (int)(dirF*(srcpp[x+2]+srcp[x-2]) + (0.5f-dirF)*(srcpp[x]+srcp[x]) + 0.5f);
						}
					}
					else
					{
						if (dirF <= -0.5f)
						{
							if (dirF <= -1.0f)
							{
								if (dirF <= -1.5f)
								{
									if (dirF <= -2.0f)
									{
										if (dirF >= -2.50f) 
										{
											temp1 = srcpp[x-8];
											temp2 = srcp[x+8];
											temp = (srcpp[x-8]+srcp[x+8]+1)>>1;
										}
										else 
										{
											temp1 = temp2 = srcp[x];
											temp = cubicInt(srcppp[x],srcpp[x],srcp[x],srcpn[x]);
										}
									}
									else 
									{
										temp1 = (int)((-dirF-1.5f)*(srcpp[x-8]) + (2.0f+dirF)*(srcpp[x-6]) + 0.5f);
										temp2 = (int)((-dirF-1.5f)*(srcp[x+8]) + (2.0f+dirF)*(srcp[x+6]) + 0.5f);
										temp = (int)((-dirF-1.5f)*(srcpp[x-8]+srcp[x+8]) + (2.0f+dirF)*(srcpp[x-6]+srcp[x+6]) + 0.5f);
									}
								}
								else 
								{
									temp1 = (int)((-dirF-1.0f)*(srcpp[x-6]) + (1.5f+dirF)*(srcpp[x-4]) + 0.5f);
									temp2 = (int)((-dirF-1.0f)*(srcp[x+6]) + (1.5f+dirF)*(srcp[x+4]) + 0.5f);
									temp = (int)((-dirF-1.0f)*(srcpp[x-6]+srcp[x+6]) + (1.5f+dirF)*(srcpp[x-4]+srcp[x+4]) + 0.5f);
								}
							}
							else 
							{
								temp1 = (int)((-dirF-0.5f)*(srcpp[x-4]) + (1.0f+dirF)*(srcpp[x-2]) + 0.5f);
								temp2 = (int)((-dirF-0.5f)*(srcp[x+4]) + (1.0f+dirF)*(srcp[x+2]) + 0.5f);
								temp = (int)((-dirF-0.5f)*(srcpp[x-4]+srcp[x+4]) + (1.0f+dirF)*(srcpp[x-2]+srcp[x+2]) + 0.5f);
							}
						}
						else 
						{
							temp1 = (int)((-dirF)*(srcpp[x-2]) + (0.5f+dirF)*(srcpp[x]) + 0.5f);
							temp2 = (int)((-dirF)*(srcp[x+2]) + (0.5f+dirF)*(srcp[x]) + 0.5f);
							temp = (int)((-dirF)*(srcpp[x-2]+srcp[x+2]) + (0.5f+dirF)*(srcpp[x]+srcp[x]) + 0.5f);
						}
					}
					minN = min(srcpp[x],srcp[x]) - 25;
					maxN = max(srcpp[x],srcp[x]) + 25;
					if (abs(temp1-temp2) > 20 || abs(srcpp[x]+srcp[x]-temp-temp) > 60 || temp < minN || temp > maxN)
					{
						temp = cubicInt(srcppp[x],srcpp[x],srcp[x],srcpn[x]);
					}
					if (temp > 255) temp = 255;
					else if (temp < 0) temp = 0;
					dstp[x] = temp;
				}
				else
				{
					if (y<3 || y>Height-4) dstp[x] = ((srcp[x] + srcpp[x] + 1) >> 1);
					else dstp[x] = cubicInt(srcppp[x],srcpp[x],srcp[x],srcpn[x]);
				}
			}
chromajump:
			++x;
			if (nomask || maskp[x] == 0xFF)
			{
				if (y<3 || y>Height-4) dstp[x] = ((srcp[x] + srcpp[x] + 1) >> 1);
				else dstp[x] = cubicInt(srcppp[x],srcpp[x],srcp[x],srcpn[x]);
			}
		}
		srcppp = srcpp;
		srcpp = srcp;
		srcp = srcpn;
		srcpn += src_pitch;
		maskp += mask_pitch;
		dstp += dst_pitch;
	}
}

void TFMPP::maskClip2(PVideoFrame &src, PVideoFrame &deint, PVideoFrame &mask, PVideoFrame &dst, int np)
{
	const unsigned char *srcp, *maskp, *dntp;
	unsigned char *dstp;
	int b, plane, x, y, width, height;
	int src_pitch, msk_pitch, dst_pitch, dnt_pitch;
	for (b=0; b<np; ++b)
	{
		if (b==0) plane = PLANAR_Y;
		else if (b == 1) plane = PLANAR_V;
		else plane = PLANAR_U;
		srcp = src->GetReadPtr(plane);
		width = src->GetRowSize(plane);
		height = src->GetHeight(plane);
		src_pitch = src->GetPitch(plane);
		maskp = mask->GetReadPtr(plane);
		msk_pitch = mask->GetPitch(plane);
		dntp = deint->GetReadPtr(plane);
		dnt_pitch = deint->GetPitch(plane);
		dstp = dst->GetWritePtr(plane);
		dst_pitch = dst->GetPitch(plane);
		for (y=0; y<height; ++y)
		{
			for (x=0; x<width; ++x)
			{
				if (maskp[x] == 0xFF) dstp[x] = dntp[x];
				else dstp[x] = srcp[x];
			}
			maskp += msk_pitch;
			srcp += src_pitch;
			dntp += dnt_pitch;
			dstp += dst_pitch;
		}
	}
}

TFMPP::TFMPP(PClip _child, int _PP, int _mthresh, const char* _ovr, bool _display, PClip _clip2, 
	IScriptEnvironment* env) : GenericVideoFilter(_child), PP(_PP), mthresh(_mthresh), 
	ovr(_ovr), display(_display), clip2(_clip2)
{
	setArray = NULL;
	int w, i, z, b, q, countOvrS;
	char linein[80], *linep, *linet;
	FILE *f = NULL;
	if (!vi.IsYV12() && !vi.IsYUY2())
		env->ThrowError("TFMPP:  YV12 and YUY2 data only!");
	if (vi.height&1 || vi.width&1)
		env->ThrowError("TFMPP:  height and width must be divisible by 2!");
	if (PP < 2 || PP > 7)
		env->ThrowError("TFMPP:  PP must be set to 2, 3, 4, 5, 6, or 7!");
	if (clip2)
	{
		uC2 = true;
		const VideoInfo &vi2 = clip2->GetVideoInfo();
		if (!vi2.IsYV12() && !vi2.IsYUY2())
			env->ThrowError("TFMPP:  clip2 must be in YUY2 or YV12 colorspace!");
		if ((vi.IsYV12() && vi2.IsYUY2()) || (vi.IsYUY2() && vi2.IsYV12()))
			env->ThrowError("TFMPP:  clip2 colorspace must be the same as input clip!");
		if (vi2.height != vi.height || vi2.width != vi.width)
			env->ThrowError("TFMPP:  clip2 frame dimensions do not match input clip!");
		if (vi2.num_frames != vi.num_frames)
			env->ThrowError("TFMPP:  clip2 does not have the same number of frames as input clip!");
	}
	else uC2 = false;
	child->SetCacheHints(CACHE_RANGE, 1);
	nfrms = vi.num_frames-1;
	PPS = PP;
	mthreshS = mthresh;
	setArraySize = 0;
	i = 0;
	if (*ovr)
	{
		if ((f = fopen(ovr, "r")) != NULL)
		{
			countOvrS = 0;
			while(fgets(linein, 80, f) != 0)
			{
				if (linein[0] == 0 || linein[0] == '\n' || linein[0] == '\r' || linein[0] == ';' || linein[0] == '#') 
					continue;
				linep = linein;
				while (*linep != 'M' && *linep != 'P' && *linep != 0) *linep++;
				if (*linep != 0) ++countOvrS;
			}
			fclose(f);
			f = NULL;
			if (countOvrS == 0) { goto emptyovrFM; }
			++countOvrS;
			countOvrS *= 4;
			setArray = (int *)malloc(countOvrS*sizeof(int));
			if (setArray == NULL) env->ThrowError("TFMPP:  malloc failure (setArray)!");
			memset(setArray,255,countOvrS*sizeof(int));
			setArraySize = countOvrS;
			if ((f = fopen(ovr, "r")) != NULL)
			{
				while (fgets(linein, 80, f) != NULL)
				{
					if (linein[0] == 0 || linein[0] == '\n' || linein[0] == '\r' ||  linein[0] == ';' || linein[0] == '#') 
						continue;
					linep = linein;
					while (*linep != 0 && *linep != ' ' && *linep != ',') *linep++;
					if (*linep == ' ')
					{
						linet = linein;
						while(*linet != 0)
						{
							if (*linet != ' ' && *linet != 10) break;
							*linet++;
						}
						if (*linet == 0) { continue; }
						*linep++;
						if (*linep == 'M' || *linep == 'P')
						{
							sscanf(linein, "%d", &z);
							if (z<0 || z>nfrms) 
							{
								fclose(f);
								f = NULL;
								env->ThrowError("TFMPP:  ovr input error (out of range frame #)!");
							}
							linep = linein;
							while (*linep != ' ' && *linep != 0) *linep++;
							if (*linep != 0)
							{
								*linep++;
								if (*linep == 'P' || *linep == 'M')
								{
									q = *linep;
									*linep++; 
									*linep++;
									if (*linep == 0) continue;
									sscanf(linep, "%d", &b);
									if (q == 80 && (b < 0 || b > 7))
									{
										fclose(f);
										f = NULL;
										env->ThrowError("TFMPP:  ovr input error (bad PP value)!");
									}
									else if (q != 80 && q != 77) continue;
									setArray[i] = q; ++i;
									setArray[i] = z; ++i;
									setArray[i] = z; ++i;
									setArray[i] = b; ++i;
								}
							}
						}
					}
					else if (*linep == ',')
					{
						while (*linep != ' ' && *linep != 0) *linep++;
						if (*linep == 0) continue;
						*linep++;
						if (*linep == 'P' || *linep == 'M')
						{
							sscanf(linein, "%d,%d", &z, &w);
							if (w == 0) w = nfrms;
							if (z<0 || z>nfrms || w<0 || w>nfrms || w < z) 
							{
								fclose(f);
								f = NULL;
								env->ThrowError("TFMPP: ovr input error (invalid frame range)!");
							}
							linep = linein;
							while (*linep != ' ' && *linep != 0) *linep++;
							if (*linep != 0)
							{
								*linep++;
								if (*linep == 'M' || *linep == 'P')
								{
									q = *linep;
									*linep++; 
									*linep++;
									if (*linep == 0) continue;
									sscanf(linep, "%d", &b);
									if (q == 80 && (b < 0 || b > 7))
									{
										fclose(f);
										f = NULL;
										env->ThrowError("TFMPP:  ovr input error (bad PP value)!");
									}
									else if (q != 77 && q != 80) continue;
									setArray[i] = q; ++i;
									setArray[i] = z; ++i;
									setArray[i] = w; ++i;
									setArray[i] = b; ++i;
								}
							}
						}
					}
				}
				fclose(f);
				f = NULL;
			}
			else env->ThrowError("TFMPP:  ovr file error (could not open file)!");
		}
		else env->ThrowError("TFMPP:  ovr input error (could not open ovr file)!");
	}
emptyovrFM:
	if (f != NULL) fclose(f);
}

TFMPP::~TFMPP()
{
	if (setArray != NULL) free(setArray);
}