/*
**                    TIVTC v0.9.9 for AviSynth 2.5.x
**
**   TIVTC includes a field matching filter (TFM) and a decimation
**   filter (TDecimate) which can be used together to achieve an
**   IVTC or for other uses. TIVTC currently supports YV12 and 
**   YUY2 colorspaces.
**   
**   Copyright (C) 2004-2005 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "TFM.h"

bool TFM::checkCombedYUY2(PVideoFrame &src, int n, IScriptEnvironment *env, int match,
						  int &blockN, int &xblocksi, int &MICount) 
{
	int cthresh6 = cthresh*6;
	const unsigned char *srcp = src->GetReadPtr();
	int src_pitch = src->GetPitch();
	int Width = src->GetRowSize();
	int Height = src->GetHeight();
	const unsigned char *srcpp = srcp - src_pitch;
	const unsigned char *srcppp = srcpp - src_pitch;
	const unsigned char *srcpn = srcp + src_pitch;
	const unsigned char *srcpnn = srcpn + src_pitch;
	PVideoFrame cmask = env->NewVideoFrame(vi);
	unsigned char *cmkp = cmask->GetWritePtr();
	unsigned char *cmk_saved = cmkp;
	unsigned char *cmkpp, *cmkpn;
	int cmk_pitch = cmask->GetPitch();
	int inc = chroma ? 1 : 2;
	int xblocks = ((Width+xhalf)>>xshift) + 1;
	int xblocks4 = xblocks<<2;
	int yblocks = ((Height+yhalf)>>yshift) + 1;
	int arraysize = (xblocks*yblocks)<<2;
	int x, y, sFirst, sSecond, temp1, temp2, box1, box2;
	memset(cmkp,0,Height*cmk_pitch);
	for (x=0; x<Width; x+=inc)
	{
		sFirst = srcp[x] - srcpn[x];
		if (sFirst > cthresh || sFirst < -cthresh)
		{
			sFirst = abs(srcpnn[x]+(srcp[x]<<2)+srcpnn[x]-(3*(srcpn[x]+srcpn[x])));
			if (sFirst > cthresh6) cmkp[x] = 0xFF;
		}
	}
	srcppp += src_pitch;
	srcpp += src_pitch;
	srcp += src_pitch;
	srcpn += src_pitch;
	srcpnn += src_pitch;
	cmkp += cmk_pitch;
	for (x=0; x<Width; x+=inc)
	{
		sFirst = srcp[x] - srcpp[x];
		sSecond = srcp[x] - srcpn[x];
		if ((sFirst > cthresh && sSecond > cthresh) || (sFirst < -cthresh && sSecond < -cthresh))
		{
			sFirst = abs(srcpnn[x]+(srcp[x]<<2)+srcpnn[x]-(3*(srcpp[x]+srcpn[x])));
			if (sFirst > cthresh6) cmkp[x] = 0xFF;
		}
	}
	srcppp += src_pitch;
	srcpp += src_pitch;
	srcp += src_pitch;
	srcpn += src_pitch;
	srcpnn += src_pitch;
	cmkp += cmk_pitch;
	for (y=2; y<Height-2; ++y)
	{
		for (x=0; x<Width; x+=inc)
		{
			sFirst = srcp[x] - srcpp[x];
			sSecond = srcp[x] - srcpn[x];
			if ((sFirst > cthresh && sSecond > cthresh) || (sFirst < -cthresh && sSecond < -cthresh))
			{
				sFirst = abs(srcppp[x]+(srcp[x]<<2)+srcpnn[x]-(3*(srcpp[x]+srcpn[x])));
				if (sFirst > cthresh6) cmkp[x] = 0xFF;
			}	
		}
		srcppp += src_pitch;
		srcpp += src_pitch;
		srcp += src_pitch;
		srcpn += src_pitch;
		srcpnn += src_pitch;
		cmkp += cmk_pitch;
	}
	for (x=0; x<Width; x+=inc)
	{
		sFirst = srcp[x] - srcpp[x];
		sSecond = srcp[x] - srcpn[x];
		if ((sFirst > cthresh && sSecond > cthresh) || (sFirst < -cthresh && sSecond < -cthresh))
		{
			sFirst = abs(srcppp[x]+(srcp[x]<<2)+srcppp[x]-(3*(srcpp[x]+srcpn[x])));
			if (sFirst > cthresh6) cmkp[x] = 0xFF;
		}
	}
	srcppp += src_pitch;
	srcpp += src_pitch;
	srcp += src_pitch;
	srcpn += src_pitch;
	srcpnn += src_pitch;
	cmkp += cmk_pitch;
	for (x=0; x<Width; x+=inc)
	{
		sFirst = srcp[x] - srcpp[x];
		if (sFirst > cthresh || sFirst < -cthresh)
		{
			sFirst = abs(srcppp[x]+(srcp[x]<<2)+srcppp[x]-(3*(srcpp[x]+srcpp[x])));
			if (sFirst > cthresh6) cmkp[x] = 0xFF;
		}
	}
	if (chroma) 
	{
		cmkp = cmk_saved + cmk_pitch;
		cmkpp = cmkp - cmk_pitch;
		cmkpn = cmkp + cmk_pitch;
		for (y=1; y<Height-1; ++y)
		{
			for (x=4; x<Width-4; x+=4)
			{
				if ((cmkp[x+1] == 0xFF && (cmkpp[x-3] == 0xFF || cmkpp[x+1] == 0xFF || cmkpp[x+5] == 0xFF || 
					cmkp[x-3] == 0xFF || cmkp[x+5] == 0xFF || cmkpn[x-3] == 0xFF || cmkpn[x+1] == 0xFF || 
					cmkpn[x+5] == 0xFF)) || (cmkp[x+3] == 0xFF && (cmkpp[x-1] == 0xFF || cmkpp[x+3] == 0xFF || 
					cmkpp[x+7] == 0xFF || cmkp[x-1] == 0xFF || cmkp[x+7] == 0xFF || cmkpn[x-1] == 0xFF || 
					cmkpn[x+3] == 0xFF || cmkpn[x+7] == 0xFF))) cmkp[x] = cmkp[x+2] = 0xFF;
			}
			cmkpp += cmk_pitch;
			cmkp += cmk_pitch;
			cmkpn += cmk_pitch;
		}
	}
	cmkp = cmk_saved + cmk_pitch;
	cmkpp = cmkp - cmk_pitch;
	cmkpn = cmkp + cmk_pitch;
	memset(cArray,0,arraysize*sizeof(int));
	for (y=1; y<Height-1; ++y)
	{
		temp1 = (y>>yshift)*xblocks4;
		temp2 = ((y+yhalf)>>yshift)*xblocks4;
		for (x=0; x<Width; x+=2)
		{
			if (cmkpp[x] == 0xFF && cmkp[x] == 0xFF && cmkpn[x] == 0xFF)
			{
				box1 = (x>>xshift)<<2;
				box2 = ((x+xhalf)>>xshift)<<2;
				++cArray[temp1+box1+0];
				++cArray[temp1+box2+1];
				++cArray[temp2+box1+2];
				++cArray[temp2+box2+3];
			}
		}
		cmkpp += cmk_pitch;
		cmkp += cmk_pitch;
		cmkpn += cmk_pitch;
	}
	MICount = -1;
	for (x=0; x<arraysize; ++x)
	{
		if (cArray[x] > MICount) 
		{
			MICount = cArray[x];
			blockN = x;
			xblocksi = xblocks4;
		}
	}
	if (MICount > MI)
	{
		if (debug)
		{
			sprintf(buf,"TFM:  match %c  - Detected As Combed! (%d > %d)\n", MTC(match), MICount, MI);
			OutputDebugString(buf);
		}
		return true;
	}
	if (debug)
	{
		sprintf(buf,"TFM:  match %c  - Detected As NOT Combed! (%d <= %d)\n", MTC(match), MICount, MI);
		OutputDebugString(buf);
	}
	blockN = xblocksi = -20;
	return false;
}

void TFM::buildDiffMapPlaneYUY2(const unsigned char *prvp, const unsigned char *nxtp, 
		unsigned char *dstp, int prv_pitch, int nxt_pitch, int dst_pitch, int Height, int Width)
{
	const unsigned char *prvpp = prvp - prv_pitch;
	const unsigned char *prvpn = prvp + prv_pitch;
	const unsigned char *nxtpp = nxtp - nxt_pitch;
	const unsigned char *nxtpn = nxtp + nxt_pitch;
	int x, y, u, diff, count;
	for (y=2; y<Height-2; y+=2)
	{
		for (x=4; x<Width-4; ++x)
		{
			diff = abs(prvp[x]-nxtp[x]);
			if (diff > 3)
			{
				for (count=0, u=x-2; u<x+3 && count < 2; u+=2)
				{
					if (abs(prvpp[u]-nxtpp[u]) > 3) ++count;
					if (abs(prvp[u]-nxtp[u]) > 3) ++count;
					if (abs(prvpn[u]-nxtpn[u]) > 3) ++count;
				}
				if (count > 1) ++dstp[x];
				if (diff > 19)
				{
					for (count=0, u=x-2; u<x+3 && count < 3; u+=2)
					{
						if (abs(prvpp[u]-nxtpp[u]) > 19) ++count;
						if (abs(prvp[u]-nxtp[u]) > 19) ++count;
						if (abs(prvpn[u]-nxtpn[u]) > 19) ++count;
					}
					if (count > 2) dstp[x] += 2;
				}
			}
			++x;
			diff = abs(prvp[x]-nxtp[x]);
			if (diff > 3)
			{
				for (count=0, u=x-4; u<x+5 && count < 2; u+=4)
				{
					if (abs(prvpp[u]-nxtpp[u]) > 3) ++count;
					if (abs(prvp[u]-nxtp[u]) > 3) ++count;
					if (abs(prvpn[u]-nxtpn[u]) > 3) ++count;
				}
				if (count > 1) ++dstp[x];
				if (diff > 19)
				{
					for (count=0, u=x-4; u<x+5 && count < 3; u+=4)
					{
						if (abs(prvpp[u]-nxtpp[u]) > 19) ++count;
						if (abs(prvp[u]-nxtp[u]) > 19) ++count;
						if (abs(prvpn[u]-nxtpn[u]) > 19) ++count;
					}
					if (count > 2) dstp[x] += 2;
				}
			}
		}
		prvpp += prv_pitch;
		prvp += prv_pitch;
		prvpn += prv_pitch;
		nxtpp += nxt_pitch;
		nxtp += nxt_pitch;
		nxtpn += nxt_pitch;
		dstp += dst_pitch;
	}
}

void TFM::DrawYUY2(PVideoFrame &dst, int x1, int y1, const char *s) 
{
	int x, y = y1 * 20, num, pitch = dst->GetPitch();
	unsigned char *dp;
	unsigned int width = dst->GetRowSize();
	int height = dst->GetHeight();
	if (y+20 >= height) return;
	for (int xx = 0; *s; ++s, ++xx) 
	{
		x = (x1 + xx) * 10; 
		if ((x+10)*2 >= (int)(width)) return;
		num = *s - ' ';
		for (int tx = 0; tx < 10; tx++) 
		{
			for (int ty = 0; ty < 20; ty++) 
			{
				dp = &dst->GetWritePtr()[(x + tx) * 2 + (y + ty) * pitch];
				if (font[num][ty] & (1 << (15 - tx))) 
				{
					if (tx & 1) 
					{
						dp[0] = 235;
						dp[-1] = 128;
						dp[1] = 128;
					} 
					else 
					{
						dp[0] = 235;
						dp[1] = 128;
						dp[3] = 128;
					}
				} 
				else 
				{
					if (tx & 1) 
					{
						dp[0] = (unsigned char) (dp[0] >> 1);
						dp[-1] = (unsigned char) ((dp[-1] + 128) >> 1);
						dp[1] = (unsigned char) ((dp[1] + 128) >> 1);
					} 
					else 
					{
						dp[0] = (unsigned char) (dp[0] >> 1);
						dp[1] = (unsigned char) ((dp[1] + 128) >> 1);
						dp[3] = (unsigned char) ((dp[3] + 128) >> 1);
					}
				}
			}
		}
	}
}

void TFM::drawBoxYUY2(PVideoFrame &dst, int blockN, int xblocks)
{
	unsigned char *dstp = dst->GetWritePtr();
	int pitch = dst->GetPitch();
	int width = dst->GetRowSize();
	int height = dst->GetHeight();
	int cordy, cordx, x, y, temp, xlim, ylim;
	cordy = blockN / xblocks;
	cordx = blockN - (cordy*xblocks);
	temp = cordx%4;
	cordx = (cordx>>2);
	cordy *= blocky;
	cordx *= (blockx<<1);
	if (temp == 1) cordx -= blockx;
	else if (temp == 2) cordy -= (blocky>>1);
	else if (temp == 3) { cordx -= blockx; cordy -= (blocky>>1); }
	xlim = cordx + 2*blockx;
	if (xlim > width) xlim = width;
	ylim = cordy + blocky;
	if (ylim > height) ylim = height;
	for (y=max(cordy,0), temp=cordx+2*(blockx-1); y<ylim; ++y)
	{
		(dstp+y*pitch)[cordx] = 255;
		if (temp < width) (dstp+y*pitch)[temp] = 255;
	}
	for (x=max(cordx,0), temp=cordy+blocky-1; x<xlim; x+=4)
	{
		(dstp+cordy*pitch)[x] = 255;
		(dstp+cordy*pitch)[x+1] = 128;
		(dstp+cordy*pitch)[x+2] = 255;
		(dstp+cordy*pitch)[x+3] = 128;
		if (temp < height)
		{
			(dstp+temp*pitch)[x] = 255;
			(dstp+temp*pitch)[x+1] = 128;
			(dstp+temp*pitch)[x+2] = 255;
			(dstp+temp*pitch)[x+3] = 128;
		}
	}
}