/*
**                    TIVTC v0.9.9 for AviSynth 2.5.x
**
**   TIVTC includes a field matching filter (TFM) and a decimation
**   filter (TDecimate) which can be used together to achieve an
**   IVTC or for other uses. TIVTC currently supports YV12 and 
**   YUY2 colorspaces.
**   
**   Copyright (C) 2004-2005 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "TFM.h"

bool TFM::checkCombedYV12(PVideoFrame &src, int n, IScriptEnvironment *env, int match,
						  int &blockN, int &xblocksi, int &MICount) 
{
	int cthresh6 = cthresh*6;
	PVideoFrame cmask = env->NewVideoFrame(vi);
	const unsigned char *srcp, *srcpp, *srcppp, *srcpn, *srcpnn;
	unsigned char *cmkp, *cmkpp, *cmkpn, *cmkpV, *cmkpU;
	unsigned char *cmkpnn, *cmkppV, *cmkppU, *cmkpnV, *cmkpnU;
	int cmk_pitch, src_pitch, Width, Height, plane, box1, box2;
	int xblocks, xblocks4, yblocks, arraysize, cmk_pitchUV;
	int x, y, b, sFirst, sSecond, temp1, temp2;
	for (b=chroma ? 3 : 1; b>0; --b)
	{
		if (b == 3) plane = PLANAR_U;
		else if (b == 2) plane = PLANAR_V;
		else plane = PLANAR_Y;
		srcp = src->GetReadPtr(plane);
		src_pitch = src->GetPitch(plane);
		Width = src->GetRowSize(plane);
		Height = src->GetHeight(plane);
		srcpp = srcp - src_pitch;
		srcppp = srcpp - src_pitch;
		srcpn = srcp + src_pitch;
		srcpnn = srcpn + src_pitch;
		cmkp = cmask->GetWritePtr(plane);
		cmk_pitch = cmask->GetPitch(plane);
		memset(cmkp,0,Height*cmk_pitch);
		for (x=0; x<Width; ++x)
		{
			sFirst = srcp[x] - srcpn[x];
			if (sFirst > cthresh || sFirst < -cthresh)
			{
				sFirst = abs(srcpnn[x]+(srcp[x]<<2)+srcpnn[x]-(3*(srcpn[x]+srcpn[x])));
				if (sFirst > cthresh6) cmkp[x] = 0xFF;
			}
		}
		srcppp += src_pitch;
		srcpp += src_pitch;
		srcp += src_pitch;
		srcpn += src_pitch;
		srcpnn += src_pitch;
		cmkp += cmk_pitch;
		for (x=0; x<Width; ++x)
		{
			sFirst = srcp[x] - srcpp[x];
			sSecond = srcp[x] - srcpn[x];
			if ((sFirst > cthresh && sSecond > cthresh) || (sFirst < -cthresh && sSecond < -cthresh))
			{
				sFirst = abs(srcpnn[x]+(srcp[x]<<2)+srcpnn[x]-(3*(srcpp[x]+srcpn[x])));
				if (sFirst > cthresh6) cmkp[x] = 0xFF;
			}
		}
		srcppp += src_pitch;
		srcpp += src_pitch;
		srcp += src_pitch;
		srcpn += src_pitch;
		srcpnn += src_pitch;
		cmkp += cmk_pitch;
		for (y=2; y<Height-2; ++y)
		{
			for (x=0; x<Width; ++x)
			{
				sFirst = srcp[x] - srcpp[x];
				sSecond = srcp[x] - srcpn[x];
				if ((sFirst > cthresh && sSecond > cthresh) || (sFirst < -cthresh && sSecond < -cthresh))
				{
					sFirst = abs(srcppp[x]+(srcp[x]<<2)+srcpnn[x]-(3*(srcpp[x]+srcpn[x])));
					if (sFirst > cthresh6) cmkp[x] = 0xFF;
				}	
			}
			srcppp += src_pitch;
			srcpp += src_pitch;
			srcp += src_pitch;
			srcpn += src_pitch;
			srcpnn += src_pitch;
			cmkp += cmk_pitch;
		}
		for (x=0; x<Width; ++x)
		{
			sFirst = srcp[x] - srcpp[x];
			sSecond = srcp[x] - srcpn[x];
			if ((sFirst > cthresh && sSecond > cthresh) || (sFirst < -cthresh && sSecond < -cthresh))
			{
				sFirst = abs(srcppp[x]+(srcp[x]<<2)+srcppp[x]-(3*(srcpp[x]+srcpn[x])));
				if (sFirst > cthresh6) cmkp[x] = 0xFF;
			}
		}
		srcppp += src_pitch;
		srcpp += src_pitch;
		srcp += src_pitch;
		srcpn += src_pitch;
		srcpnn += src_pitch;
		cmkp += cmk_pitch;
		for (x=0; x<Width; ++x)
		{
			sFirst = srcp[x] - srcpp[x];
			if (sFirst > cthresh || sFirst < -cthresh)
			{
				sFirst = abs(srcppp[x]+(srcp[x]<<2)+srcppp[x]-(3*(srcpp[x]+srcpp[x])));
				if (sFirst > cthresh6) cmkp[x] = 0xFF;
			}
		}
	}
	if (chroma) 
	{
		cmkp = cmask->GetWritePtr(PLANAR_Y);
		cmkpU = cmask->GetWritePtr(PLANAR_U);
		cmkpV = cmask->GetWritePtr(PLANAR_V);
		Width = cmask->GetRowSize(PLANAR_V);
		Height = cmask->GetHeight(PLANAR_V);
		cmk_pitch = cmask->GetPitch(PLANAR_Y);
		cmk_pitchUV = cmask->GetPitch(PLANAR_V);
		cmkpp = cmkp - cmk_pitch;
		cmkpn = cmkp + cmk_pitch;
		cmkpnn = cmkpn + cmk_pitch;
		cmkppU = cmkpU - cmk_pitchUV;
		cmkpnU = cmkpU + cmk_pitchUV;
		cmkppV = cmkpV - cmk_pitchUV;
		cmkpnV = cmkpV + cmk_pitchUV;
		cmk_pitch <<= 1;
		for (y=1; y<Height-1; ++y)
		{
			cmkpp += cmk_pitch;
			cmkp += cmk_pitch;
			cmkpn += cmk_pitch;
			cmkpnn += cmk_pitch;
			cmkppV += cmk_pitchUV;
			cmkpV += cmk_pitchUV;
			cmkpnV += cmk_pitchUV;
			cmkppU += cmk_pitchUV;
			cmkpU += cmk_pitchUV;
			cmkpnU += cmk_pitchUV;
			for (x=1; x<Width-1; ++x)
			{
				if ((cmkpV[x] == 0xFF && (cmkpV[x-1] == 0xFF || cmkpV[x+1] == 0xFF ||
					 cmkppV[x-1] == 0xFF || cmkppV[x] == 0xFF || cmkppV[x+1] == 0xFF ||
					 cmkpnV[x-1] == 0xFF || cmkpnV[x] == 0xFF || cmkpnV[x+1] == 0xFF)) || 
					(cmkpU[x] == 0xFF && (cmkpU[x-1] == 0xFF || cmkpU[x+1] == 0xFF ||
					 cmkppU[x-1] == 0xFF || cmkppU[x] == 0xFF || cmkppU[x+1] == 0xFF ||
					 cmkpnU[x-1] == 0xFF || cmkpnU[x] == 0xFF || cmkpnU[x+1] == 0xFF)))
				{
					((unsigned short*)cmkp)[x] = (unsigned short) 0xFFFF;
					((unsigned short*)cmkpn)[x] = (unsigned short) 0xFFFF;
					if (y&1) ((unsigned short*)cmkpp)[x] = (unsigned short) 0xFFFF;
					else ((unsigned short*)cmkpnn)[x] = (unsigned short) 0xFFFF;
				}
			}
		}
	}
	cmk_pitch = cmask->GetPitch(PLANAR_Y);
	cmkp = cmask->GetWritePtr(PLANAR_Y) + cmk_pitch;
	cmkpp = cmkp - cmk_pitch;
	cmkpn = cmkp + cmk_pitch;
	Width = cmask->GetRowSize(PLANAR_Y);
	Height = cmask->GetHeight(PLANAR_Y);
	xblocks = ((Width+xhalf)>>xshift) + 1;
	xblocks4 = xblocks<<2;
	yblocks = ((Height+yhalf)>>yshift) + 1;
	arraysize = (xblocks*yblocks)<<2;
	memset(cArray,0,arraysize*sizeof(int));
	for (y=1; y<Height-1; ++y)
	{
		temp1 = (y>>yshift)*xblocks4;
		temp2 = ((y+yhalf)>>yshift)*xblocks4;
		for (x=0; x<Width; ++x)
		{
			if (cmkpp[x] == 0xFF && cmkp[x] == 0xFF && cmkpn[x] == 0xFF)
			{
				box1 = (x>>xshift)<<2;
				box2 = ((x+xhalf)>>xshift)<<2;
				++cArray[temp1+box1+0];
				++cArray[temp1+box2+1];
				++cArray[temp2+box1+2];
				++cArray[temp2+box2+3];
			}
		}
		cmkpp += cmk_pitch;
		cmkp += cmk_pitch;
		cmkpn += cmk_pitch;
	}
	MICount = -1;
	for (x=0; x<arraysize; ++x)
	{
		if (cArray[x] > MICount) 
		{
			MICount = cArray[x];
			blockN = x;
			xblocksi = xblocks4;
		}
	}
	if (MICount > MI)
	{
		if (debug)
		{
			sprintf(buf,"TFM:  match %c  - Detected As Combed! (%d > %d)\n", MTC(match), MICount, MI);
			OutputDebugString(buf);
		}
		return true;
	}
	if (debug)
	{
		sprintf(buf,"TFM:  match %c  - Detected As NOT Combed! (%d <= %d)\n", MTC(match), MICount, MI);
		OutputDebugString(buf);
	}
	blockN = xblocks = -20;
	return false;
}

void TFM::buildDiffMapPlaneYV12(const unsigned char *prvp, const unsigned char *nxtp, 
		unsigned char *dstp, int prv_pitch, int nxt_pitch, int dst_pitch, int Height, int Width)
{
	const unsigned char *prvpp = prvp - prv_pitch;
	const unsigned char *prvpn = prvp + prv_pitch;
	const unsigned char *nxtpp = nxtp - nxt_pitch;
	const unsigned char *nxtpn = nxtp + nxt_pitch;
	int x, y, u, diff, count;
	for (y=2; y<Height-2; y+=2)
	{
		for (x=1; x<Width-1; ++x)
		{
			diff = abs(prvp[x]-nxtp[x]);
			if (diff > 3)
			{
				for (count=0,u=x-1; u<x+2 && count<2; ++u)
				{
					if (abs(prvpp[u]-nxtpp[u]) > 3) ++count;
					if (abs(prvp[u]-nxtp[u]) > 3) ++count;
					if (abs(prvpn[u]-nxtpn[u]) > 3) ++count;
				}
				if (count > 1) ++dstp[x];
				if (diff > 19)
				{
					for (count=0,u=x-1; u<x+2 && count<3; ++u)
					{
						if (abs(prvpp[u]-nxtpp[u]) > 19) ++count;
						if (abs(prvp[u]-nxtp[u]) > 19) ++count;
						if (abs(prvpn[u]-nxtpn[u]) > 19) ++count;
					}
					if (count > 2) dstp[x] += 2;
				}
			}
		}
		prvpp += prv_pitch;
		prvp += prv_pitch;
		prvpn += prv_pitch;
		nxtpp += nxt_pitch;
		nxtp += nxt_pitch;
		nxtpn += nxt_pitch;
		dstp += dst_pitch;
	}
}

void TFM::DrawYV12(PVideoFrame &dst, int x1, int y1, const char *s) 
{
	int x, y = y1 * 20, num, tx, ty;
	int pitchY = dst->GetPitch(PLANAR_Y), pitchUV = dst->GetPitch(PLANAR_V);
	unsigned char *dpY, *dpU, *dpV;
	unsigned int width = dst->GetRowSize(PLANAR_Y);
	int height = dst->GetHeight(PLANAR_Y);
	if (y+20 >= height) return;
	for (int xx = 0; *s; ++s, ++xx) 
	{
		x = (x1 + xx) * 10;
		if (x+10 >= (int)(width)) return;
		num = *s - ' ';
		for (tx = 0; tx < 10; tx++)
		{
			for (ty = 0; ty < 20; ty++)
			{
				dpY = &dst->GetWritePtr(PLANAR_Y)[(x + tx) + (y + ty) * pitchY];
				if (font[num][ty] & (1 << (15 - tx))) *dpY = 235;
				else *dpY = (unsigned char) (*dpY >> 1);
			}
		}
		for (tx = 0; tx < 10; tx++)
		{
			for (ty = 0; ty < 20; ty++)
			{
				dpU = &dst->GetWritePtr(PLANAR_U)[((x + tx)/2) + ((y + ty)/2) * pitchUV];
				dpV = &dst->GetWritePtr(PLANAR_V)[((x + tx)/2) + ((y + ty)/2) * pitchUV];
				if (font[num][ty] & (1 << (15 - tx)))
				{
					*dpU = 128;
					*dpV = 128;
				} 
				else
				{
					*dpU = (unsigned char) ((*dpU + 128) >> 1);
					*dpV = (unsigned char) ((*dpV + 128) >> 1);
				}
			}
		}
	}
}

void TFM::drawBoxYV12(PVideoFrame &dst, int blockN, int xblocks)
{
	unsigned char *dstp = dst->GetWritePtr(PLANAR_Y);
	int width = dst->GetRowSize(PLANAR_Y);
	int height = dst->GetHeight(PLANAR_Y);
	int pitch = dst->GetPitch(PLANAR_Y);
	int cordy, cordx, x, y, temp, xlim, ylim;
	cordy = blockN / xblocks;
	cordx = blockN - (cordy*xblocks);
	temp = cordx%4;
	cordx = (cordx>>2);
	cordy *= blocky;
	cordx *= blockx;
	if (temp == 1) cordx -= (blockx>>1);
	else if (temp == 2) cordy -= (blocky>>1);
	else if (temp == 3) { cordx -= (blockx>>1); cordy -= (blocky>>1); }
	xlim = cordx + blockx;
	if (xlim > width) xlim = width;
	ylim = cordy + blocky;
	if (ylim > height) ylim = height;
	for (y=max(cordy,0), temp=cordx+blockx-1; y<ylim; ++y)
	{
		(dstp+y*pitch)[cordx] = 255;
		if (temp < width) (dstp+y*pitch)[temp] = 255;
	}
	for (x=max(cordx,0), temp=cordy+blocky-1; x<xlim; ++x)
	{
		(dstp+cordy*pitch)[x] = 255;
		if (temp < height) (dstp+temp*pitch)[x] = 255;
	}
}