// Copyright (c) 2002 Tom Barry.  All rights reserved.
//      trbarry@trbarry.com
// Requires Avisynth source code to compile for Avisynth
// Avisynth Copyright 2000 Ben Rudiak-Gould.
//      http://www.math.berkeley.edu/~benrg/avisynth.html
/////////////////////////////////////////////////////////////////////////////
//
//	This file is subject to the terms of the GNU General Public License as
//	published by the Free Software Foundation.  A copy of this license is
//	included with this software distribution in the file COPYING.  If you
//	do not have a copy, you may obtain a copy by writing to the Free
//	Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
//	This software is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details
//  
//  Also, this program is "Philanthropy-Ware".  That is, if you like it and 
//  feel the need to reward or inspire the author then please feel free (but
//  not obligated) to consider joining or donating to the Electronic Frontier
//  Foundation. This will help keep cyber space free of barbed wire and bullsh*t.  
//
/////////////////////////////////////////////////////////////////////////////
// Change Log
//
// Date				Version			Developer       Changes
//
// 2003/05/01		V 0.1.0.0		Tom Barry		Initial test release 

// CVS Log
//
/////////////////////////////////////////////////////////////////////////////
#include "internal.h"

//#define BOOL unsigned int
//#define BYTE unsigned char


class UnComb : public GenericVideoFilter {
    int  Max_Frames;
	bool SSEMMXenabled;
	int __stdcall Fieldcopy(void *dest, const void *src, size_t count, 
				int rows, int dst_pitch, int src_pitch);

public:
    UnComb(PClip _child,  IScriptEnvironment* env) : GenericVideoFilter(_child)
    {
        Max_Frames = vi.num_frames;
		SSEMMXenabled = (env->GetCPUFlags() & CPUF_INTEGER_SSE) !=0;

	    if  (! vi.IsYV12())
		{
			env->ThrowError("UnComb: Supports YV12 color format only");
		}

	    if (!SSEMMXenabled)
		{
			env->ThrowError("UnComb: Only SSEMMX CPU's currently supported");
		}

    }

    PVideoFrame __stdcall GetFrame(int inFrame, IScriptEnvironment* env);

};


PVideoFrame __stdcall UnComb::GetFrame(int inFrame, IScriptEnvironment* env) 
{
PVideoFrame src;
PVideoFrame dst;
PVideoFrame PrevSrc;

const BYTE* pSrc;
const BYTE* pPrev;
BYTE* dstp;
const BYTE* pSrcW;
const BYTE* pPrevW;

int src_pitch;
int prev_pitch;
int dst_pitch;

int rowsize;
int height;
int ctH;
int ctV;
int offsetH;
int offsetV;
int DiffTop;		// How well did Top field match prev bottom?
int DiffBottom;		// How well did Bottom field match prev top
int DiffSame;		// How well did Top & Bottom of this field match?
	
	src = child->GetFrame(inFrame, env);
	pSrc = src->GetReadPtr(PLANAR_Y);

	PrevSrc = inFrame
		? child->GetFrame(inFrame-1, env)
		: child->GetFrame(inFrame+1, env);

	pPrev = PrevSrc->GetReadPtr(PLANAR_Y);

	src_pitch = src->GetPitch(PLANAR_Y);
	prev_pitch = PrevSrc->GetPitch(PLANAR_Y);
	rowsize = src->GetRowSize(PLANAR_Y);
	height = src->GetHeight();

	offsetH = (rowsize >> 8) << 5;		// ignore 1/8 of width on both ends, in 32 byte chunks
	offsetV = (height >> 5) << 2;		// ignore 1/8 of height on both ends, in 4 line chunks
	ctH = (rowsize - 2 * offsetH) >> 5;	// do 32 bytes at a time
	ctV = (height - 2 * offsetV) >> 2;	// do 4 lines at a time
	pSrcW = pSrc + offsetH + offsetV * src_pitch;
	pPrevW = pPrev + offsetH + offsetV * prev_pitch;

	__asm
	{
		mov		esi, pSrcW			// curr source
		mov		edi, pPrevW			// prev source
		mov		eax, src_pitch
		mov		ebx, prev_pitch
		pxor	mm5, mm5			// DiffTop
		pxor	mm6, mm6			// DiffBottom
		pxor	mm7, mm7			// DiffSame

LoopV:
		mov		ecx, ctH			// horizontal count of 32 byte chuncks

			align   16
	LoopH:
			movq	mm0, qword ptr[esi]			// curr top even line
			movq	mm1, qword ptr[esi+eax]		// curr top odd line
			movq	mm2, qword ptr[edi]			// prev top even line
			pavgb	mm0, qword ptr[esi+2*eax]	// interp possible curr top odd line
			pavgb   mm2, qword ptr[edi+2*ebx]	// interp possible prev top odd line

			psadbw  mm2, mm1					// sum abs diff, calc DiffBottom
			psadbw  mm1, mm0					// sum abs diff, calc DiffSame
			psadbw  mm0, qword ptr[edi+ebx]		// sum abs diff, calc DiffTop

			paddd	mm5, mm0					// accum DiffTop
			paddd	mm7, mm1					// accum DiffSame
			paddd	mm6, mm2					// accum DiffBottom

			// unroll loop once and do it all again
			movq	mm0, qword ptr[esi+16]			// curr top even line
			movq	mm1, qword ptr[esi+eax+16]		// curr top odd line
			movq	mm2, qword ptr[edi+16]			// prev top even line
			pavgb	mm0, qword ptr[esi+2*eax+16]	// interp possible curr top odd line
			pavgb   mm2, qword ptr[edi+2*ebx+16]	// interp possible prev top odd line

			psadbw  mm2, mm1					// sum abs diff, calc DiffBottom
			psadbw  mm1, mm0					// sum abs diff, calc DiffSame
			psadbw  mm0, qword ptr[edi+ebx+16]		// sum abs diff, calc DiffTop

			paddd	mm5, mm0					// accum DiffTop
			paddd	mm7, mm1					// accum DiffSame
			paddd	mm6, mm2					// accum DiffBottom

			add		esi, 32
			add		edi, 32
			loop	LoopH						

		mov		esi, pSrcW					// curr source line begin
		lea		esi, [esi+4*eax]			// bump 4 lines
		mov		pSrcW, esi

		mov		edi, pPrevW					// prev source begin
		lea		edi, [edi+4*eax]			// bump 4 lines
		mov		pPrevW, edi

		dec		ctV
		jnz		LoopV						// go do another line
	
		movd	DiffTop, mm5				// done, save results
		movd	DiffBottom, mm6
		movd    DiffSame, mm7

		emms
	}

	// if curr frame already properly matched then return it, exit
	if (DiffSame <= DiffTop && DiffSame <= DiffBottom)
	{
		return src;
	}

	dst = env->NewVideoFrame(vi);
	dstp = dst->GetWritePtr(PLANAR_Y);
	dst_pitch = dst->GetPitch(PLANAR_Y);

	// else copy fields from two different frames - first Luma
	if (DiffTop <= DiffBottom)			// use Top from curr, bottom from Prev? 
	{
		env->BitBlt(dstp, 2 * dst_pitch, 					// copy top lines
			pSrc, 2 * src_pitch,			
			rowsize, height / 2);

		env->BitBlt(dstp + dst_pitch, 2 * dst_pitch,			// copy bottom lines
			pPrev + prev_pitch, 2 * prev_pitch,	
			rowsize, height / 2);
	}
	else
	{
		env->BitBlt(dstp, 2 * dst_pitch, 					// copy top lines
			pPrev, 2 * prev_pitch,			
			rowsize, height / 2);

		env->BitBlt(dstp + dst_pitch, 2 * dst_pitch,			// copy bottom lines
			pSrc + src_pitch, 2 * src_pitch,	
			rowsize, height / 2);
	}
		

//  void __stdcall BitBlt(BYTE* dstp, int dst_pitch, const BYTE* srcp, 
//		int src_pitch, int row_size, int height);

	// copy croma planes - U
	pSrc = src->GetReadPtr(PLANAR_U);
	pPrev = PrevSrc->GetReadPtr(PLANAR_U);
	dstp = dst->GetWritePtr(PLANAR_U);
	src_pitch = src->GetPitch(PLANAR_U);
	prev_pitch = PrevSrc->GetPitch(PLANAR_U);
	dst_pitch = dst->GetPitch(PLANAR_U);
	rowsize = dst->GetRowSize(PLANAR_U);
	height = dst->GetHeight(PLANAR_U);

	if (DiffTop <= DiffBottom)			// use Top from curr, bottom from Prev? 
	{
		env->BitBlt(dstp, 2 * dst_pitch, 					// copy top lines
			pSrc, 2 * src_pitch,			
			rowsize, height / 2);

		env->BitBlt(dstp + dst_pitch, 2 * dst_pitch,			// copy bottom lines
			pPrev + prev_pitch, 2 * prev_pitch,	
			rowsize, height / 2);
	}
	else
	{
		env->BitBlt(dstp, 2 * dst_pitch, 					// copy top lines
			pPrev, 2 * prev_pitch,			
			rowsize, height / 2);

		env->BitBlt(dstp + dst_pitch, 2 * dst_pitch,			// copy bottom lines
			pSrc + src_pitch, 2 * src_pitch,	
			rowsize, height / 2);
	}

	pSrc = src->GetReadPtr(PLANAR_V);
	pPrev = PrevSrc->GetReadPtr(PLANAR_V);
	dstp = dst->GetWritePtr(PLANAR_V);
	src_pitch = src->GetPitch(PLANAR_V);
	dst_pitch = dst->GetPitch(PLANAR_V);
	prev_pitch = PrevSrc->GetPitch(PLANAR_V);
	rowsize = dst->GetRowSize(PLANAR_V);
	height = dst->GetHeight(PLANAR_V);

	if (DiffTop <= DiffBottom)			// use Top from curr, bottom from Prev? 
	{
		env->BitBlt(dstp, 2 * dst_pitch, 					// copy top lines
			pSrc, 2 * src_pitch,			
			rowsize, height / 2);

		env->BitBlt(dstp + dst_pitch, 2 * dst_pitch,			// copy bottom lines
			pPrev + prev_pitch, 2 * prev_pitch,	
			rowsize, height / 2);
	}
	else
	{
		env->BitBlt(dstp, 2 * dst_pitch, 					// copy top lines
			pPrev, 2 * prev_pitch,			
			rowsize, height / 2);

		env->BitBlt(dstp + dst_pitch, 2 * dst_pitch,			// copy bottom lines
			pSrc + src_pitch, 2 * src_pitch,	
			rowsize, height / 2);
	}

	__asm emms;

    return dst;
}


int UnComb::Fieldcopy(void *dest, const void *src, size_t count, 
				int rows, int dst_pitch, int src_pitch)
{
BYTE* pDest = (BYTE*) dest;
BYTE* pSrc = (BYTE*) src;
	
	for (int i=0; i < rows; i++)
	{
		memcpy(pDest, pSrc, count);
		pSrc += src_pitch;
		pDest += dst_pitch;
	}
	return 0;
}


AVSValue __cdecl Create_UnComb(AVSValue args, void* user_data,
IScriptEnvironment* env) 
{
    return new UnComb(args[0].AsClip(), env);
}


extern "C" __declspec(dllexport) const char* __stdcall
AvisynthPluginInit2(IScriptEnvironment* env) {
    env->AddFunction("UnComb", "c", Create_UnComb, 0);
    return "`UnComb' Brain dead but fast IVTC";
}

