//
//	asharp (version 0.95) - adaptive sharpenning filter.
//
//	asharp engine implementation (C/MMX/ISSE)
//
//	Copyright (C) 2002 Marc Fauconneau
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  Please contact me for any bugs or questions.
//  marc.fd@libertysurf.fr

//  Change log :
//         27 nov 2002 - ver 0.95  - First GPL release.

#include "asharp.h"
#include "stdlib.h"
#include "memory.h"

// mmx/isse switch stuff
//	pminsw
#define _pminsw_mmx(mmr1,mmr2,mmrw) \
__asm pcmpeqw mmrw,mmrw		\
__asm psubusw mmrw,mmr2		\
__asm paddusw mmr1, mmrw	\
__asm psubusw mmr1, mmrw
#define _pminsw_isse(mmr1,mmr2,mmrw) \
__asm pminsw mmr1,mmr2
#define _pminsw(mmr1,mmr2,mmrw) _pminsw2(mmr1,mmr2,mmrw,cpu)
#define _pminsw2(mmr1,mmr2,mmrw,cpu) _pminsw3(mmr1,mmr2,mmrw,cpu) // link
#define _pminsw3(mmr1,mmr2,mmrw,cpu) _pminsw_##cpu##(mmr1,mmr2,mmrw)
//	pmaxub
#define _pmaxub_mmx(mmr1,mmr2) \
__asm psubusb mmr1,mmr2 \
__asm paddusb mmr1,mmr2
#define _pmaxub_isse(mmr1,mmr2) \
__asm pmaxub mmr1,mmr2
#define _pmaxub(mmr1,mmr2) _pmaxub2(mmr1,mmr2,cpu)
#define _pmaxub2(mmr1,mmr2,cpu) _pmaxub3(mmr1,mmr2,cpu) // link
#define _pmaxub3(mmr1,mmr2,cpu) _pmaxub_##cpu##(mmr1,mmr2)

// SIMD stuff

#define acc8b(srch,srcl,acch,accl,zero)		\
__asm movq			srcl,srch	\
__asm punpckhbw		srch,zero	\
__asm punpcklbw		srcl,zero	\
__asm paddusw		accl,srcl	\
__asm paddusw		acch,srch

#define check8b(cur,tmp,ref,max)	\
__asm movq			tmp,ref		\
__asm psubusb		tmp,cur		\
__asm psubusb		cur,ref		\
__asm por			tmp,cur		\
	 _pmaxub(max,tmp)

#define check8bm(cur,tmp,ref,max,msk)	\
__asm movq			tmp,ref		\
__asm psubusb		tmp,cur		\
__asm psubusb		cur,ref		\
__asm por			tmp,cur		\
__asm pand			tmp,msk		\
	 _pmaxub(max,tmp)

#define LoadAcc(offset,srch,srcl,acch,accl,zero)	\
__asm movq		srch,offset	\
	acc8b(srch,srcl,acch,accl,zero)

#define LoadCheck(offset,cur,tmp,ref,max)	\
__asm movq		cur,offset		\
	check8b(cur,tmp,ref,max)

#define LoadCheck2(offset,cur,tmp,ref,max,msk)	\
__asm movq		cur,offset		\
	check8bm(cur,tmp,ref,max,msk)

// basic diameter 3 unsharp masking
#define run_a(srcp,tmpp,last,stride,srch,srcl,acch,accl,zero)	\
__asm pxor	acch,acch		\
__asm pxor	accl,accl	\
	LoadAcc([tmpp-1],srch,srcl,acch,accl,zero)		\
	LoadAcc([tmpp  ],srch,srcl,acch,accl,zero)		\
	LoadAcc([tmpp+1],srch,srcl,acch,accl,zero)		\
	LoadAcc( last   ,srch,srcl,acch,accl,zero)		\
	LoadAcc([srcp  ],srch,srcl,acch,accl,zero)		\
	LoadAcc([srcp+1],srch,srcl,acch,accl,zero)		\
__asm add	srcp,stride	\
	LoadAcc([srcp-1],srch,srcl,acch,accl,zero)		\
	LoadAcc([srcp  ],srch,srcl,acch,accl,zero)		\
	LoadAcc([srcp+1],srch,srcl,acch,accl,zero)		\
__asm sub	srcp,stride	\

// scan for adaptive thresholds
#define run_c(srcp,tmpp,last,stride,cur,tmp,ref,max)		\
__asm pxor	max,max		\
__asm movq	ref,[srcp]	\
	LoadCheck([tmpp-1],cur,tmp,ref,max)		\
	LoadCheck([tmpp  ],cur,tmp,ref,max)		\
	LoadCheck([tmpp+1],cur,tmp,ref,max)		\
	LoadCheck( last  ,cur,tmp,ref,max)		\
	LoadCheck([srcp+1],cur,tmp,ref,max)		\
__asm add	srcp,stride	\
	LoadCheck([srcp-1],cur,tmp,ref,max)		\
	LoadCheck([srcp  ],cur,tmp,ref,max)		\
	LoadCheck([srcp+1],cur,tmp,ref,max)		\
__asm sub	srcp,stride	\

#define run_c2(srcp,tmpp,last,stride,cur,tmp,ref,max,c8b_00r,c8b_00l)		\
__asm pxor	max,max		\
__asm movq	ref,[srcp]	\
	LoadCheck2([tmpp-1],cur,tmp,ref,max,c8b_00l)		\
	LoadCheck ([tmpp  ],cur,tmp,ref,max)				\
	LoadCheck2([tmpp+1],cur,tmp,ref,max,c8b_00r)		\
	LoadCheck2( last   ,cur,tmp,ref,max,c8b_00l)		\
	LoadCheck2([srcp+1],cur,tmp,ref,max,c8b_00r)		\
__asm add	srcp,stride	\
	LoadCheck2([srcp-1],cur,tmp,ref,max,c8b_00l)		\
	LoadCheck ([srcp  ],cur,tmp,ref,max)				\
	LoadCheck2([srcp+1],cur,tmp,ref,max,c8b_00r)		\
__asm sub	srcp,stride	\

#define run_c2u(srcp,tmpp,last,stride,cur,tmp,ref,max,c8b_00r,c8b_00l)		\
__asm pxor	max,max		\
__asm movq	ref,[srcp]	\
	LoadCheck2([tmpp-1],cur,tmp,ref,max,c8b_00l)		\
	LoadCheck ([tmpp  ],cur,tmp,ref,max)				\
	LoadCheck2([tmpp+1],cur,tmp,ref,max,c8b_00r)		\
	LoadCheck2( last   ,cur,tmp,ref,max,c8b_00l)		\
	LoadCheck2([srcp+1],cur,tmp,ref,max,c8b_00r)		\

#define run_c2d(srcp,tmpp,last,stride,cur,tmp,ref,max,c8b_00r,c8b_00l)		\
__asm pxor	max,max		\
__asm movq	ref,[srcp]	\
	LoadCheck2( last   ,cur,tmp,ref,max,c8b_00l)		\
	LoadCheck2([srcp+1],cur,tmp,ref,max,c8b_00r)		\
__asm add	srcp,stride	\
	LoadCheck2([srcp-1],cur,tmp,ref,max,c8b_00l)		\
	LoadCheck ([srcp  ],cur,tmp,ref,max)				\
	LoadCheck2([srcp+1],cur,tmp,ref,max,c8b_00r)		\
__asm sub	srcp,stride	\

// compute adaptive thresholds
#define atresh_isse(thrh,thrl,c4w_D1,c4w_D2,c4w_Td4,zero,tmp)	\
__asm movq		thrl,thrh		\
__asm punpckhbw	thrh,zero		\
__asm punpcklbw	thrl,zero		\
__asm psllw		thrh,7			\
__asm psllw		thrl,7			\
__asm pmulhw	thrh,c4w_D1		\
__asm pmulhw	thrl,c4w_D2		\
	 _pminsw	(thrh,c4w_Td4,tmp)	\
	 _pminsw	(thrl,c4w_Td4,tmp)	\
__asm paddsw	thrh,c4w_Da		\
__asm paddsw	thrl,c4w_Da		\
__asm psllw		thrh,4			\
__asm psllw		thrl,4			\

// div9/diff
#define diff(srch,srcl,difl,difh,acch,accl,c4w_inv9,zero)	\
__asm pmulhw	acch,c4w_inv9	\
__asm pmulhw	accl,c4w_inv9	\
__asm movq		srcl,srch		\
__asm punpckhbw	srch,zero		\
__asm punpcklbw	srcl,zero		\
__asm movq		difh,srch		\
__asm movq		difl,srcl		\
__asm psubsw	difh,acch		\
__asm psubsw	difl,accl		\

// apply thresholded diff
#define apply(srch,srcl,difl,difh,thrh,thrl)	\
__asm psllw		difh,7			\
__asm psllw		difl,7			\
__asm pmulhw	difh,thrh		\
__asm pmulhw	difl,thrl		\
__asm paddw		srch,difh		\
__asm paddw		srcl,difl		\

// D=0 mode
#define run1(ptr,tmp)	\
__asm mov	ptr,[cfp]	\
__asm add	ptr,x	\
__asm mov	tmp,[lineptr]	\
__asm add	tmp,x	\
__asm pxor	mm0,mm0	\
	run_a(ptr,tmp,mm7,stride,mm2,mm3,mm4,mm5,mm0)	\
__asm movq	[tmp-1],mm7	\
__asm pxor	mm0,mm0	\
__asm movq	mm1,[ptr]	\
	diff(mm1,mm2,mm6,mm7,mm4,mm5,c4w_inv9,mm0)	\
__asm movq		mm3,c4w_T	\
__asm movq		mm4,mm3		\
	apply(mm1,mm2,mm6,mm7,mm3,mm4)	\
__asm packuswb	mm2,mm1	\
__asm movq		mm7,[ptr+7]	\
__asm movq		[ptr],mm2

// D>0 mode
#define run2(ptr,tmp,run_x)	\
__asm mov	ptr,[cfp]	\
__asm add	ptr,x	\
__asm mov	tmp,[lineptr]	\
__asm add	tmp,x	\
__asm pxor	mm0,mm0	\
	run_a(ptr,tmp,mm7,stride,mm2,mm3,mm4,mm5,mm0)	\
	run_x	\
__asm movq	[tmp-1],mm7	\
__asm pxor	mm0,mm0	\
__asm movq	mm1,[ptr]	\
	diff(mm1,mm2,mm6,mm7,mm4,mm5,c4w_inv9,mm0)	\
	atresh_isse(mm3,mm4,c4w_D1,c4w_D2,c4w_Td4,mm0,mm0)	\
	apply(mm1,mm2,mm6,mm7,mm3,mm4)	\
__asm packuswb	mm2,mm1	\
__asm movq		mm7,[ptr+7]	\
__asm movq		[ptr],mm2

void asharp_run_c(	uc* planeptr, int pitch,
					int height, int width, 
					int	T,int D, int B, int B2, bool bf )
	{
		uc* lineptr = new uc[width];
		int p = pitch;
		int y,x;
		uc* cfp = planeptr+pitch;
		uc* lp = lineptr;
		memcpy(lp,planeptr,width);
		for (y=1;y<height-1;y++) {
			int last = cfp[0];
			for (x=0;x<width-1;x++) {

				int avg = 0;
				int dev = 0;
				int T2;
				int diff;
				int tmp;

				avg += lp[x-1];
				avg += lp[x  ];
				avg += lp[x+1];
				avg += last;
				avg += cfp[x    ];
				avg += cfp[x  +1];
				avg += cfp[x+p-1];
				avg += cfp[x+p  ];
				avg += cfp[x+p+1];

				avg *= (65535/9);
				avg >>= 16;

				#define CHECK(A) \
				if (abs(A-cfp[x])>dev) dev = abs(A-cfp[x]);
				
				if (bf) {

				if (y%8>0) {
					if (x%8>0) CHECK(lp[x-1])
					CHECK(lp[x  ])
					if (x%8<7) CHECK(lp[x+1])
				}
				if (x%8>0) CHECK(last)
				if (x%8<7) CHECK(cfp[x  +1])
				if (y%8<7) {
					if (x%8>0) CHECK(cfp[x+p-1])
					CHECK(cfp[x+p  ])
					if (x%8<7) CHECK(cfp[x+p+1])
				}

				} else {
					CHECK(lp[x-p-1])
					CHECK(lp[x-p  ])
					CHECK(lp[x-p+1])
					CHECK(last)
					CHECK(cfp[x  +1])
					CHECK(cfp[x+p-1])
					CHECK(cfp[x+p  ])
					CHECK(cfp[x+p+1])
				}

				T2 = T;
				diff = cfp[x]-avg;
				int D2 = D;

				if (x%8==6) D2=(D2*B2)>>8;
				if (x%8==7) D2=(D2*B)>>8;
				if (x%8==0) D2=(D2*B)>>8;
				if (x%8==1) D2=(D2*B2)>>8;
				if (y%8==6) D2=(D2*B2)>>8;
				if (y%8==7) D2=(D2*B)>>8;
				if (y%8==0) D2=(D2*B)>>8;
				if (y%8==1) D2=(D2*B2)>>8;

				int Da = -32+(D>>7);
				if (D>0) T2 = ((((dev<<7)*D2)>>16)+Da)<<4;

				if (T2>T) T2=T;
				if (T2<-32) T2=-32;

				tmp = (((diff<<7)*T2)>>16)+cfp[x];

				if (tmp < 0) tmp = 0;
				if (tmp > 255) tmp = 255;
				lp[x-1] = last;
				last = cfp[x];
				cfp[x] = tmp;
			}
			lp[x] = cfp[x];
			cfp += pitch;
		}
	delete[] lineptr;
	} 

void asharp_run_mmx(	uc* planeptr, int pitch,
						int height, int width, 
						int T,int D, int B, int B2, bool bf )
	{
		uc* lineptrna = new uc[width+8];
		uc* lineptr = (uc*)(((int)lineptrna)+8-(((int)lineptrna)&7));

		memcpy(lineptr,planeptr,width);
		int p = pitch;
		int y,x;
		uc* cfp = planeptr+pitch;

		__declspec(align(8)) unsigned __int16 c4w_Dtab[3*8];
		__declspec(align(8)) unsigned __int16 *c4w_Dtabp = c4w_Dtab;
		__declspec(align(8)) unsigned __int64 c4w_D1,c4w_D2;

		for (int i=0;i<3;i++) {
			int D2=D;
			int D3=D;
			int D4=D;
			if (i==0) D2=(D2*B)>>8;
			if (i==1) D2=(D2*B2)>>8;
			D3=D2;
			D4=D2;
			D3=(D2*B2)>>8;
			D4=(D2*B)>>8;
			c4w_Dtab[i*8+0] = D4;
			c4w_Dtab[i*8+1] = D3;
			c4w_Dtab[i*8+2] = D2;
			c4w_Dtab[i*8+3] = D2;
			c4w_Dtab[i*8+4] = D2;
			c4w_Dtab[i*8+5] = D2;
			c4w_Dtab[i*8+6] = D3;
			c4w_Dtab[i*8+7] = D4;
		}

		__declspec(align(8)) unsigned __int16 c4w_inv9[4]	= {65535/9,65535/9,65535/9,65535/9};
		__declspec(align(8))   signed __int16 c4w_Da[4]		= {-32+(D>>7),-32+(D>>7),-32+(D>>7),-32+(D>>7)};
		__declspec(align(8)) unsigned __int8  c8b_00r[8]	= {0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0x00};
		__declspec(align(8)) unsigned __int8  c8b_00l[8]	= {0x00,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF};
		__declspec(align(8)) unsigned __int16 c4w_T[4]		= {T,T,T,T};
		__declspec(align(8)) unsigned __int16 c4w_Td4[4]	= {T>>4,T>>4,T>>4,T>>4};
		__declspec(align(8)) unsigned __int8  v8b_first[8];
		int stride = pitch;

		for (y=1;y<height-1;y++) {
			if ((y&7)==0 || (y&7)==7) {
				__asm {
					mov		eax,[c4w_Dtabp]
					movq	mm1,[eax]
					movq	mm2,[eax+8]
					movq	[c4w_D1],mm1
					movq	[c4w_D2],mm2
				}
			} else if ((y&7)==1 || (y&7)==6) {
				__asm {
					mov		eax,[c4w_Dtabp]
					movq	mm1,[eax+16]
					movq	mm2,[eax+24]
					movq	[c4w_D1],mm1
					movq	[c4w_D2],mm2
				}
			} else {
				__asm {
					mov		eax,[c4w_Dtabp]
					movq	mm1,[eax+32]
					movq	mm2,[eax+40]
					movq	[c4w_D1],mm1
					movq	[c4w_D2],mm2
				}
			}

			__asm {
					mov		eax,[cfp]
					movq	mm7,[eax]
					psllq	mm7,8
					movq	v8b_first,mm7
			}
			v8b_first[0] = v8b_first[1];
			__asm {
					movq	mm7,v8b_first
			}

			#define cpu mmx

			if (bf && D>0)  {
				if (y&7==0) { 
					for (x=0;x<width;x+=8)
						run2(esi,edi,run_c2(esi,edi,mm7,stride,mm2,mm0,mm6,mm3,c8b_00r,c8b_00l))
				} else if (y&7==7) {
					for (x=0;x<width;x+=8)
						run2(esi,edi,run_c2u(esi,edi,mm7,stride,mm2,mm0,mm6,mm3,c8b_00r,c8b_00l))
				} else 
					for (x=0;x<width;x+=8)
						run2(esi,edi,run_c2d(esi,edi,mm7,stride,mm2,mm0,mm6,mm3,c8b_00r,c8b_00l))
			} else if (D>0) {
				for (x=0;x<width;x+=8)
					run2(esi,edi,run_c(esi,edi,mm7,stride,mm2,mm0,mm6,mm3))
			} else 
				for (x=0;x<width;x+=8)
					run1(esi,edi)

			#undef cpu

			cfp += pitch;
		}
		delete[] lineptrna;

		__asm emms;
	}


// this implementation is only a copypaste of asharp_run_mmx. (it's not the reference)
void asharp_run_isse(	uc* planeptr, int pitch,
						int height, int width, 
						int T,int D, int B, int B2, bool bf )
	{
		uc* lineptrna = new uc[width+8];
		uc* lineptr = (uc*)(((int)lineptrna)+8-(((int)lineptrna)&7));

		memcpy(lineptr,planeptr,width);
		int p = pitch;
		int y,x;
		uc* cfp = planeptr+pitch;

		__declspec(align(8)) unsigned __int16 c4w_Dtab[3*8];
		__declspec(align(8)) unsigned __int16 *c4w_Dtabp = c4w_Dtab;
		__declspec(align(8)) __int64 c4w_D1,c4w_D2;

		for (int i=0;i<3;i++) {
			int D2=D;
			int D3=D;
			int D4=D;
			if (B<=128) {
				if (i==0) D2=(D2*B)>>8;
				if (i==1) D2=(D2*B2)>>8;
				D3=D2;
				D4=D2;
				D3=(D2*B2)>>8;
				D4=(D2*B)>>8;
			}
			c4w_Dtab[i*8+0] = D4;
			c4w_Dtab[i*8+1] = D3;
			c4w_Dtab[i*8+2] = D2;
			c4w_Dtab[i*8+3] = D2;
			c4w_Dtab[i*8+4] = D2;
			c4w_Dtab[i*8+5] = D2;
			c4w_Dtab[i*8+6] = D3;
			c4w_Dtab[i*8+7] = D4;
		}

		__declspec(align(8)) unsigned __int16 c4w_inv9[4]	= {65535/9,65535/9,65535/9,65535/9};
		__declspec(align(8))   signed __int16 c4w_Da[4]		= {-32+(D>>7),-32+(D>>7),-32+(D>>7),-32+(D>>7)};
		__declspec(align(8)) unsigned __int8  c8b_00r[8]	= {0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0x00};
		__declspec(align(8)) unsigned __int8  c8b_00l[8]	= {0x00,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF};
		__declspec(align(8)) unsigned __int16 c4w_T[4]		= {T,T,T,T};
		__declspec(align(8)) unsigned __int16 c4w_Td4[4]	= {T>>4,T>>4,T>>4,T>>4};
		__declspec(align(8)) unsigned __int8  v8b_first[8];
		int stride = pitch;

		for (y=1;y<height-1;y++) {
			if ((y&7)==0 || (y&7)==7) {
				__asm {
					mov		eax,[c4w_Dtabp]
					movq	mm1,[eax]
					movq	mm2,[eax+8]
					movq	[c4w_D1],mm1
					movq	[c4w_D2],mm2
				}
			} else if ((y&7)==1 || (y&7)==6) {
				__asm {
					mov		eax,[c4w_Dtabp]
					movq	mm1,[eax+16]
					movq	mm2,[eax+24]
					movq	[c4w_D1],mm1
					movq	[c4w_D2],mm2
				}
			} else {
				__asm {
					mov		eax,[c4w_Dtabp]
					movq	mm1,[eax+32]
					movq	mm2,[eax+40]
					movq	[c4w_D1],mm1
					movq	[c4w_D2],mm2
				}
			}

			__asm {
					mov		eax,[cfp]
					movq	mm7,[eax]
					psllq	mm7,8
					movq	v8b_first,mm7
			}
			v8b_first[0] = v8b_first[1];
			__asm {
					movq	mm7,v8b_first
			}

			#define cpu isse

			if (bf && D>0)  {
				if (y&7==0) { 
					for (x=0;x<width;x+=8)
						run2(esi,edi,run_c2(esi,edi,mm7,stride,mm2,mm0,mm6,mm3,c8b_00r,c8b_00l))
				} else if (y&7==7) {
					for (x=0;x<width;x+=8)
						run2(esi,edi,run_c2u(esi,edi,mm7,stride,mm2,mm0,mm6,mm3,c8b_00r,c8b_00l))
				} else 
					for (x=0;x<width;x+=8)
						run2(esi,edi,run_c2d(esi,edi,mm7,stride,mm2,mm0,mm6,mm3,c8b_00r,c8b_00l))
			} else if (D>0) {
				for (x=0;x<width;x+=8)
					run2(esi,edi,run_c(esi,edi,mm7,stride,mm2,mm0,mm6,mm3))
			} else 
				for (x=0;x<width;x+=8)
					run1(esi,edi)

			#undef cpu

			cfp += pitch;
		}
		delete[] lineptrna;

		__asm emms;
	} 
