//
//	asharp (version 0.95) - adaptive sharpenning filter.
//
//	asharp Avisynth 2.5 API
//
//	Copyright (C) 2002 Marc Fauconneau
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  Please contact me for any bugs or questions.
//  marc.fd@libertysurf.fr

//  Change log :
//         27 nov 2002 - ver 0.95  - First GPL release.

#include <windows.h>
#include "avisynth.h"
#include "asharp.h"
#include "asharp4avisynth.h"

PVideoFrame __stdcall PostProc::GetFrame(int n, IScriptEnvironment* env) 
{
	PVideoFrame cf = child->GetFrame(n, env);
	env->MakeWritable(&cf);

//#define SELFCHECK
#ifdef SELFCHECK

/********************************************************
 *	- Selfcheck results									*
 *	MMX and ISSE are matching 100%						*
 *	C vs SIMD in D=0 mode : only left border mismatch	*
 *	C vs SIMD in D>0 mode : bad. but not noticable.		*
 ********************************************************/

	int h = vi.height;
	int w = vi.width;
	int p = cf->GetPitch();

	uc* t1 = new uc[h*p];
	uc* t2 = new uc[h*p];

	memcpy(t1,cf->GetReadPtr(),h*p);
	memcpy(t2,cf->GetReadPtr(),h*p);

	asharp_run_c	(t1, p, vi.height,vi.width,T,D,B,B2,bf);
	asharp_run_mmx	(t2, p, vi.height,vi.width,T,D,B,B2,bf);

#define k 1 // tolerance (1 is bitwise)

	for(int y=0;y<h;y++)
		for(int x=0;x<w;x++)
			if (t1[x+p*y]>t2[x+p*y]-k && t1[x+p*y]<t2[x+p*y]+k)
				cf->GetWritePtr()[x+p*y]=0x00;
			else
				cf->GetWritePtr()[x+p*y]=0xFF;

	delete[] t1;
	delete[] t2;

#else

	asharp_run_fct* asharp_run;
	asharp_run = asharp_run_c;
	if (env->GetCPUFlags()&CPUF_MMX) asharp_run = asharp_run_mmx;
	if (env->GetCPUFlags()&CPUF_INTEGER_SSE) asharp_run = asharp_run_isse;
	asharp_run(cf->GetWritePtr(PLANAR_Y), cf->GetPitch(PLANAR_Y), vi.height,vi.width,T,D,B,B2,bf);

#endif

	return cf;
}