//
//	asharp (version 0.95) - adaptive sharpenning filter.
//
//	asharp Avisynth 2.5 API
//
//	Copyright (C) 2002 Marc Fauconneau
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  Please contact me for any bugs or questions.
//  marc.fd@libertysurf.fr

//  Change log :
//         27 nov 2002 - ver 0.95  - First GPL release.

#include "avisynth.h"

#define uc unsigned char

class PostProc : public GenericVideoFilter {
	int T,D,B,B2;
	bool bf;
public:
	PostProc(AVSValue args, IScriptEnvironment* env) : GenericVideoFilter(args[0].AsClip()) 
	{
		double t = args[1].AsFloat(2);
		double d = args[2].AsFloat(4);
		double b = args[3].AsFloat(-1);
		bf = args[4].AsBool(false);

		// parameters floating point to fixed point convertion
		T = (int)(t*(4<<7));
		D = (int)(d*(4<<7));
		B = (int)(256-b*64);
		B2= (int)(256-b*48);

		// clipping (recommended for SIMD code)

		if (T<-(4<<7)) T = -(4<<7); // yes, negatives values are accepted
		if (D<0) D = 0;
		if (B<0) B = 0;
		if (B2<0) B2 = 0;

		if (T>(32*(4<<7))) T = (32*(4<<7));
		if (D>(16*(4<<7))) D = (16*(4<<7));
		if (B>256) B = 256;
		if (B2>256) B2 = 256;
	}
    PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};


AVSValue __cdecl Create_PostProc(AVSValue args, void* user_data, IScriptEnvironment* env) {
	return new PostProc(args,env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit(IScriptEnvironment* env) {
    env->AddFunction("asharp", "c[T]f[D]f[B]f[hqbf]b", Create_PostProc, 0);
    return "";
}
