#include "internal.h"
#include "avisynth.h"
#include "windows.h"
#include "stdio.h"

#define MCACHE_LINE 64

 struct
  {
   unsigned char *blur,*work;
  } m;




class WarpSharp : public GenericVideoFilter
{


public:
 
    
    WarpSharp(PClip _child, int _wsdepth, int _wsthreshold, IScriptEnvironment* env) :
	    GenericVideoFilter(_child), wsdepth(_wsdepth),wsthreshold(_wsthreshold),warpcubic(-0.6,true)
	{
		if (!vi.IsYV12())
			env->ThrowError("Warpsharp: requires YV12 source; use ConvertToYV12()");
    m.blur=m.work=NULL;
   }
    PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
    virtual ~WarpSharp() {
      if (m.blur) _aligned_free(m.blur);m.blur=0;
      if (m.work) _aligned_free(m.work);m.work=0;
      m.blur=m.work=NULL;
    }
 

private:
  void warpsharpbump(const unsigned char *src,unsigned char *dst,int wsthresh);
  void warpsharpblur(const unsigned char * const src0,unsigned char * const dst0);
    int dyY;
    int dxY;
    int strideY;
    int strideYdst;
    int wsdepth;
    int wsthreshold;
   struct Cubic4Table {
      int table[1024];
      Cubic4Table(double A,bool mmx);
      int interpolate(const unsigned char *src, int stride, int x, int y);
    } warpcubic;
};

class XSharpen : public GenericVideoFilter
{


public:
 
   
    XSharpen(PClip _child, int _xstrength, int _xthreshold, IScriptEnvironment* env) :
	    GenericVideoFilter(_child), xstrength(_xstrength),xthreshold(_xthreshold)
	{
		if (!vi.IsYV12())
			env->ThrowError("XSharpen: requires YV12 source; use ConvertToYV12()");
    Ymin=Ymax=NULL;
   }
    PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
    virtual ~XSharpen() {
      if (Ymin) _aligned_free(Ymin);Ymin=0;
      if (Ymax) _aligned_free(Ymax);Ymax=0;
      if (aline) free(aline); aline=0;
    }


private:
    unsigned char *Ymin,*Ymax; 
    unsigned char *aline;
    int dyY;
    int dxY;
    int strideY;
    int strideYdst;
    int xstrength;
    int xthreshold;
};


void WarpSharp::warpsharpbump(const unsigned char *src,unsigned char *dst,int wsthresh) { //pitch ok
 memset(dst,0,dxY);
 src+=strideY;dst+=strideYdst;
 int stride_=strideY;
 static __declspec(align(8)) unsigned __int64 wsthresh64;
 wsthresh64=0xffff-wsthresh;wsthresh64=(wsthresh64<<48)|(wsthresh64<<32)|(wsthresh64<<16)|wsthresh64;
 for (const unsigned char *srcEnd=src+(dyY-2)*strideY;src<srcEnd;src+=strideY,dst+=strideYdst)
  {
   dst[0]=0;dst[dxY-1]=0;
   const unsigned char *srcLnEnd=src+dxY-1;
   __asm {
     mov  esi,[src]
     inc  esi
     mov  edi,[dst]
     inc  edi
     mov  ecx,stride_ //ecx -  stride
     mov  edx,ecx
     neg  edx         //edx - -stride
     pxor mm7,mm7
     movq mm5,[wsthresh64]
     mov  eax,[srcLnEnd]
    bump1:
     movd mm0,[esi+edx-1]
     punpcklbw mm0,mm7
     movd mm1,[esi-1]
     punpcklbw mm1,mm7
     paddw mm0,mm1
     movd mm1,[esi+ecx-1]
     punpcklbw mm1,mm7
     paddw mm0,mm1
     movd mm1,[esi+edx+1]
     punpcklbw mm1,mm7
     psubw mm0,mm1
     movd mm1,[esi+1]
     punpcklbw mm1,mm7
     psubw mm0,mm1
     movd mm1,[esi+ecx+1]
     punpcklbw mm1,mm7
     psubw mm0,mm1          // mm0 - dx
     
     movq  mm6,mm0
     psraw mm6,15
     pxor  mm0,mm6
     psubw mm0,mm6          // mm0 - abs(dx)
     
     movd mm1,[esi+edx-1]
     punpcklbw mm1,mm7
     movd mm2,[esi+edx]
     punpcklbw mm2,mm7
     paddw mm1,mm2
     movd mm2,[esi+edx+1]
     punpcklbw mm2,mm7
     paddw mm1,mm2
     movd mm2,[esi+ecx-1]
     punpcklbw mm2,mm7
     psubw mm1,mm2
     movd mm2,[esi+ecx]
     punpcklbw mm2,mm7
     psubw mm1,mm2
     movd mm2,[esi+ecx+1]
     punpcklbw mm2,mm7
     psubw mm1,mm2          // mm1 - dy
     
     movq  mm6,mm1
     psraw mm6,15
     pxor  mm1,mm6
     psubw mm1,mm6          // mm1 - abs(dy)
     
     movq  mm2,mm0
     paddw mm2,mm1
     movq  mm3,mm2
     paddw mm2,mm2
     paddw mm2,mm3          // mm2 - 3*(dx+dy)
     
     //abs(mm1-mm0)
     movq mm3,mm0
     psubusb mm0,mm1
     psubusb mm1,mm3
     por mm1,mm0            // mm1 - abs(dx-dy)
     
     paddw   mm1,mm2        // mm1 - level
     
     paddusw mm1,mm5
     psubusw mm1,mm5        // mm1 - min(level,wsthresh)

     packuswb mm1,mm7
     movd [edi],mm1         
     add esi,4
     add edi,4
     cmp esi,eax
     jl  bump1 
    }
  }
 memset(dst,0,dxY);
}

void WarpSharp::warpsharpblur(const unsigned char * const src0,unsigned char * const dst0) // src0 = dst (image) dst0 = m.blur
{
 const unsigned char *src1,*src1end;short *buf1;
 for (src1=src0,src1end=src1+dyY*strideYdst,buf1=(short*)m.work;src1<src1end;src1+=strideYdst,buf1+=strideY)
  {
   buf1[0]=src1[0]*16;buf1[1]=src1[1]*16;
   buf1[dxY-2]=src1[dxY-2]*16;buf1[dxY-1]=src1[dxY-1]*16;
   const unsigned char *srcLnEnd=src1+dxY-2;
   __asm
    {
     mov esi,[src1]
     add esi,2
     mov edi,[buf1]
     add edi,4
     mov eax,[srcLnEnd]
     pxor mm7,mm7
    blur1:
     movd      mm0,[esi-2]
     punpcklbw mm0,mm7
     movd      mm1,[esi-1]
     punpcklbw mm1,mm7
     paddw     mm1,mm1
     paddw     mm1,mm1
     movd      mm2,[esi]
     punpcklbw mm2,mm7
     paddw     mm2,mm2
     movq      mm6,mm2
     paddw     mm2,mm2
     paddw     mm2,mm6
     movd      mm3,[esi+1]
     punpcklbw mm3,mm7
     paddw     mm3,mm3
     paddw     mm3,mm3
     movd      mm4,[esi+2]
     punpcklbw mm4,mm7
     
     paddusw mm0,mm1
     paddusw mm0,mm2
     paddusw mm0,mm3
     paddusw mm0,mm4

     movq [edi],mm0
     add esi,4
     add edi,8
     cmp esi,eax
     jl  blur1
    }
  }
 const short *buf1end; 
 unsigned char *dst1;
 int stride_=strideY;
 static __declspec(align(8)) const unsigned __int64 m128=0x0080008000800080;
 for (buf1=(short*)m.work+strideY*2,buf1end=(short*)m.work+(dyY-2)*strideY,dst1=dst0+strideY*2;buf1<buf1end;buf1+=strideY,dst1+=strideY)
  {
   const short *bufLnEnd=buf1+dxY;
   __asm
    {
     mov ecx,stride_
     add ecx,ecx
     mov edx,ecx
     neg edx
     mov esi,[buf1]
     mov edi,[dst1]
     mov eax,[bufLnEnd]
     pxor mm6,mm6
     movq mm7,[m128]
    blur2:
     movq    mm0,[esi+edx*2]
     movq    mm1,[esi+edx]
     paddw   mm1,mm1
     paddw   mm1,mm1
     paddw   mm0,mm1
     movq    mm1,[esi]
     paddw   mm1,mm1
     movq    mm2,mm1
     paddw   mm1,mm1
     paddw   mm0,mm1
     paddw   mm0,mm2
     movq    mm1,[esi+ecx]
     paddw   mm1,mm1
     paddw   mm1,mm1
     paddw   mm0,mm1
     paddw   mm0,[esi+ecx*2]
     
     paddw   mm0,mm7
     psrlw   mm0,8

     packuswb mm0,mm6
     movd [edi],mm0

     add esi,8
     add edi,4
     cmp esi,eax
     jl  blur2
    }
  }
  memcpy(dst0,src0,dxY);    
  memcpy(dst0+strideY,src0+strideY,dxY);    
  memcpy(dst0+(dyY-1)*strideY,src0+(dyY-1)*strideY,dxY);
  memcpy(dst0+(dyY-2)*strideY,src0+(dyY-2)*strideY,dxY); 
}


PVideoFrame __stdcall WarpSharp::GetFrame(int n, IScriptEnvironment* env)
{

//  int get;

	/* Get the current frame. */
  PVideoFrame srcf = child->GetFrame(n, env);

    /* Get destination frame. */
	PVideoFrame dstf = env->NewVideoFrame(vi);

  const unsigned char *src = (unsigned char*)srcf->GetReadPtr();
  unsigned char *dst = dstf->GetWritePtr();


  strideY = srcf->GetPitch(PLANAR_Y);
  strideYdst = dstf->GetPitch(PLANAR_Y);
  dxY = srcf->GetRowSize(PLANAR_Y_ALIGNED);
  dyY = srcf->GetHeight(PLANAR_Y);

 if (!m.work) m.work=(unsigned char*)_aligned_malloc(strideY*dyY*2,MCACHE_LINE);
 warpsharpbump(src,dst,wsthreshold);
 if (!m.blur) m.blur=(unsigned char*)_aligned_malloc(strideY*dyY,MCACHE_LINE);
 warpsharpblur(dst,m.blur);

 memcpy(dst,src,dxY);
 dst+=strideYdst;
 src+=strideY;
 memcpy(dst,src,dxY);
 dst+=strideYdst;
 src+=strideY;
 memcpy(dst,src,dxY);
 dst+=strideYdst;
 src+=strideY;
 const BYTE *srcEnd=src+(dyY-5)*strideY;
 const BYTE *bum=m.blur+strideY*2;
 for (;src<srcEnd;src+=strideY,dst+=strideYdst,bum+=strideY)
  {
   dst[0]=src[0];dst[1]=src[1];dst[2]=src[2];
   dst[dxY-2]=src[dxY-2];dst[dxY-1]=src[dxY-1];dst[dxY-2]=src[dxY-2];
   int x=3;
   const BYTE *bumLn=bum+3;
   for (;x<dxY-2;bumLn++,x++)
    {
     int dispx=(int)bumLn[-1]-(int)bumLn[1];
     int dispy=(int)bumLn[-strideY]-(int)bumLn[strideY];
     dispx=dispx*wsdepth+512;dispy=dispy*wsdepth+512;

     if (dispx<-2048) dispx=-2048; 
     else if (dispx>(2*1024)) dispx=2*1024;
     if (dispy<-2048) dispy=-2048; 
     else if (dispy>(2*1024)) dispy=2*1024;
     
     int qx=(dispx>>2)&255,qy=(dispy>>2)&255;
     dispx>>=10;dispy>>=10;
     const BYTE* passme=src+(strideY*(dispy-1))+x+(dispx-1);
     dst[x]=warpcubic.interpolate(passme,strideY,qx,qy);
    }
  }
 memcpy(dst,src,dxY);src+=strideY;dst+=strideYdst;
 memcpy(dst,src,dxY);src+=strideY;dst+=strideYdst;
 memcpy(dst,src,dxY);
 __asm{emms}
 env->BitBlt(dstf->GetWritePtr(PLANAR_U),dstf->GetPitch(PLANAR_U),srcf->GetReadPtr(PLANAR_U),srcf->GetPitch(PLANAR_U),srcf->GetRowSize(PLANAR_U),srcf->GetHeight(PLANAR_U));
 env->BitBlt(dstf->GetWritePtr(PLANAR_V),dstf->GetPitch(PLANAR_V),srcf->GetReadPtr(PLANAR_V),srcf->GetPitch(PLANAR_V),srcf->GetRowSize(PLANAR_V),srcf->GetHeight(PLANAR_V));

  return dstf;
}
 
WarpSharp::Cubic4Table::Cubic4Table(double A,bool mmx) {
 int i;
 for (i=0; i<256; i++) 
  {
   double d = (double)i / 256.0;
   int y1, y2, y3, y4, ydiff;

   // Coefficients for all four pixels *must* add up to 1.0 for
   // consistent unity gain.
   //
   // Two good values for A are -1.0 (original VirtualDub bicubic filter)
   // and -0.75 (closely matches Photoshop).

   y1 = (int)floor(0.5 + (        +     A*d -       2.0*A*d*d +       A*d*d*d) * 16384.0);
   y2 = (int)floor(0.5 + (+ 1.0             -     (A+3.0)*d*d + (A+2.0)*d*d*d) * 16384.0);
   y3 = (int)floor(0.5 + (        -     A*d + (2.0*A+3.0)*d*d - (A+2.0)*d*d*d) * 16384.0);
   y4 = (int)floor(0.5 + (                  +           A*d*d -       A*d*d*d) * 16384.0);

   // Normalize y's so they add up to 16384.

   ydiff = (16384 - y1 - y2 - y3 - y4)/4;
//   assert(ydiff > -16 && ydiff < 16);

   y1 += ydiff;
   y2 += ydiff;
   y3 += ydiff;
   y4 += ydiff;

			if (mmx) 
			 {
 				table[i*4 + 0] = (y2<<16) | (y1 & 0xffff);
	 			table[i*4 + 1] = (y4<<16) | (y3 & 0xffff);
		 		table[i*4 + 2] = 0;
		 		table[i*4 + 3] = 0;
 			}
 	 else
 	  {
	 			table[i*4 + 0] = y1;
		 		table[i*4 + 1] = y2;
			 	table[i*4 + 2] = y3;
				 table[i*4 + 3] = y4;
			 }
  }
}
int WarpSharp::Cubic4Table::interpolate(const unsigned char *src, int stride, int x, int y)
{
	const unsigned char *p0 = src;
	const unsigned char *p1 = p0 + stride;
	const unsigned char *p2 = p1 + stride;
	const unsigned char *p3 = p2 + stride;
 const int *qx=table+x*4,*qy=table+y*4;
#if 0
	int l;
	l  = ((p0[0] * qx[0] + p0[1] * qx[1] + p0[2] * qx[2] + p0[3] * qx[3] + 128) >> 8) * qy[0];
	l += ((p1[0] * qx[0] + p1[1] * qx[1] + p1[2] * qx[2] + p1[3] * qx[3] + 128) >> 8) * qy[1];
	l += ((p2[0] * qx[0] + p2[1] * qx[1] + p2[2] * qx[2] + p2[3] * qx[3] + 128) >> 8) * qy[2];
	l += ((p3[0] * qx[0] + p3[1] * qx[1] + p3[2] * qx[2] + p3[3] * qx[3] + 128) >> 8) * qy[3];

	l = (l + (1 << 19)) >> 20;
	if (l<0)	l=0;	else if (l>255) l=255;
	return l;
#else
		static __declspec(align(8)) const __int64
			_00ff00ff00ff00ff = 0x00ff00ff00ff00ffi64,
			_0000ffff0000ffff = 0x0000ffff0000ffffi64,
			_0000008000000080 = 0x0000008000000080i64,
			_0008000000000000 = 0x0008000000000000i64;

		int luma;

		__asm {
			mov			eax,src
			mov			edx,stride

   pxor  mm7,mm7
			movd		mm0,[eax]
			punpcklbw mm0,mm7
			movd		mm1,[eax+edx]
			punpcklbw mm1,mm7
			movd		mm2,[eax+edx*2]
			punpcklbw mm2,mm7
			add			eax,edx
			movd		mm3,[eax+edx*2]
			punpcklbw mm3,mm7

			mov			eax,qx
			movq		mm5,[eax]

			//pand		mm0,mm4
			//pand		mm1,mm4
			//pand		mm2,mm4
			//pand		mm3,mm4

			pmaddwd		mm0,mm5
			movq		mm4,_00ff00ff00ff00ff
			pmaddwd		mm1,mm5
			punpckldq	mm4,mm0
			pmaddwd		mm2,mm5
			paddd		mm0,mm4
			pmaddwd		mm3,mm5
			psrlq		mm0,32
			punpckldq	mm5,mm1
			punpckldq	mm6,mm2
			paddd		mm1,mm5
			paddd		mm2,mm6
			psrlq		mm1,32
			punpckldq	mm7,mm3
			psrlq		mm2,32
			paddd		mm3,mm7

			psrlq		mm3,32

			movq		mm4,_0000008000000080



			punpckldq	mm0,mm2
			punpckldq	mm1,mm3

			paddd		mm0,mm4
			paddd		mm1,mm4

			movq		mm4,_0000ffff0000ffff

			psrlq		mm0,8
			psllq		mm1,8

			pand		mm0,mm4
			pandn		mm4,mm1

			mov			eax,qy
			por			mm0,mm4

			pmaddwd		mm0,[eax]
			punpckldq	mm4,mm0
			paddd		mm0,mm4

			pxor		mm4,mm4
			paddd		mm0,_0008000000000000
			psrad		mm0,4

			pcmpgtw		mm4,mm0
			pandn		mm4,mm0
			packuswb	mm4,mm4
			psrlq		mm4,56
			movd		luma,mm4
			
		}

		return luma;
#endif	
}


AVSValue __cdecl Create_WarpSharp(AVSValue args, void* user_data, IScriptEnvironment* env)
{

    return new WarpSharp(args[0].AsClip(),
		args[1].AsInt(30),		// depth
		args[2].AsInt(40),		// threshold
		env);
}


AVSValue __cdecl Create_XSharpen(AVSValue args, void* user_data, IScriptEnvironment* env)
{

    return new XSharpen(args[0].AsClip(),
		args[1].AsInt(30),		// stength
		args[2].AsInt(40),		// threshold
		env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
    env->AddFunction("WarpSharp", "c[depth]i[threshold]i", Create_WarpSharp, 0);
    env->AddFunction("XSharpen", "c[strength]i[threshold]i", Create_XSharpen, 0);
    return 0;
}




/*
XSharpen(PClip _child, int _wsdepth, int _wsthreshold, IScriptEnvironment* env) :
  GenericVideoFilter(_child)
	{
		if (!vi.IsYV12())
			env->ThrowError("XSharpen: requires YV12 source; use ConvertToYV12()");
   Ymin=Ymax=NULL;
   }

//  PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
    virtual ~XSharpen() {
      if (m.blur) _aligned_free(m.blur);m.blur=0;
      if (m.work) _aligned_free(m.work);m.work=0;
      m.blur=m.work=NULL;
    }
*/

PVideoFrame __stdcall XSharpen::GetFrame(int n, IScriptEnvironment* env)
{

	/* Get the current frame. */
  PVideoFrame srcf = child->GetFrame(n, env);

    /* Get destination frame. */
	PVideoFrame dstf = env->NewVideoFrame(vi);

  const unsigned char *src = (unsigned char*)srcf->GetReadPtr();
  unsigned char *dst = dstf->GetWritePtr();


  strideY = srcf->GetPitch(PLANAR_Y);
  strideYdst = dstf->GetPitch(PLANAR_Y);
  dxY = srcf->GetRowSize(PLANAR_Y_ALIGNED);
  dyY = srcf->GetHeight(PLANAR_Y);

  int dx=dxY,stride=strideY,dy=dyY;

 const unsigned char *srcLnEnd=src+stride*dy;
 const unsigned char *srcLn;unsigned char *YminLn,*YmaxLn,*dstLn;
 srcLn=src;
 if (!Ymin)
  {
   Ymin=(unsigned char*)_aligned_malloc(strideY*dyY*2,MCACHE_LINE);
   Ymax=(unsigned char*)_aligned_malloc(strideY*dyY,MCACHE_LINE);
   aline=(unsigned char*)malloc(dxY*2);
  }

 YminLn=Ymin;
 YmaxLn=Ymax;


 for (;srcLn<srcLnEnd;srcLn+=stride,YminLn+=stride,YmaxLn+=stride)
  {
   const unsigned char *srcPend=srcLn+(dx-1);
   __asm
    {
     mov esi,[srcLn]
     mov ecx,[YminLn]
     mov edx,[YmaxLn]
     inc esi
     inc ecx
     inc edx
    lineLoop1:
     movq mm1,[esi-1]
     movq mm2,[esi]
     movq mm5,mm2
     movq mm3,[esi+1]

     PMINUB mm2,mm1
     PMINUB mm2,mm3 //mm2 = min

     PMAXUB mm5,mm1
     PMAXUB mm5,mm3 //mm5 = max

     movq [ecx],mm2
     movq [edx],mm5

     add esi,8
     add ecx,8
     add edx,8
     cmp esi,[srcPend]
     jl lineLoop1
    }
  }
 
 srcLnEnd=src+stride*(dy-1);
 memcpy(dst,src,dx);
 memcpy(dst+stride*(dy-1),src+stride*(dy-1),dx);

 static __declspec(align(8)) const __int64 ones=0xffffffffffffffff;
 __int64 mfd_strength=xstrength; // 0-127
 __int64 mfd_strengthInv=127-mfd_strength;
 __int64 mfd_threshold=xthreshold; // 0-255
 static __declspec(align(8)) __int64 mtf_strength64;
 static __declspec(align(8)) __int64 mtf_strengthInv64;
 static __declspec(align(8)) __int64 mtf_thresh64;
 mtf_strength64=mfd_strength +(mfd_strength <<16)+(mfd_strength <<32)+(mfd_strength <<48);
 mtf_strengthInv64=mfd_strengthInv +(mfd_strengthInv <<16)+(mfd_strengthInv <<32)+(mfd_strengthInv <<48);
 mtf_thresh64=mfd_threshold+(mfd_threshold<<16)+(mfd_threshold<<32)+(mfd_threshold<<48);
 for (srcLn=src+stride,dstLn=dst+stride,YminLn=Ymin+stride,YmaxLn=Ymax+stride;srcLn<srcLnEnd;srcLn+=stride,dstLn+=stride,YminLn+=stride,YmaxLn+=stride)
  {
   const unsigned char *srcPend=srcLn+dx-1;
   __asm
    {
     mov eax,[stride]
     mov ebx,eax
     neg ebx
     
     mov ecx,[YminLn]
     mov edx,[YmaxLn]
     mov esi,[srcLn]
     mov edi,[dstLn]
    lineLoop2:
     pxor mm7,mm7
     
     movd mm1,[ecx]
     punpcklbw mm1,mm7;        //mm0-0,YminLn[0]    ,0,YminLn[1]    ,0,YminLn[2]    ,0,YminLn[3]      
       
     movd mm2,[ecx+ebx]
     punpcklbw mm2,mm7        //mm1-0,YminLn[-dx+0],0,YminLn[-dx+1],0,YminLn[-dx+2],0,YminLn[-dx+3]
       
     movd mm3,[ecx+eax]  
     punpcklbw mm3,mm7        //mm2-0,YminLn[+dx+0],0,YminLn[+dx+1],0,YminLn[+dx+2],0,YminLn[+dx+3]
       
     PMINSW mm1,mm2
     PMINSW mm1,mm3           //mm1-minimum z okolia (lumamin)
       
     movd mm3,[edx]
     punpcklbw mm3,mm7        //mm0-0,YmaxLn[0]    ,0,YmaxLn[1]    ,0,YmaxLn[2]    ,0,YmaxLn[3] 

     movd mm0,[edx+ebx]  
     punpcklbw mm0,mm7        //mm3-0,YmaxLn[-dx+0],0,YmaxLn[-dx+1],0,YmaxLn[-dx+2],0,YmaxLn[-dx+3]
       
     movd mm2,[edx+eax]  
     punpcklbw mm2,mm7        //mm4-0,YmaxLn[+dx+0],0,YmaxLn[+dx+1],0,YmaxLn[+dx+2],0,YmaxLn[+dx+3]

     PMAXSW mm3,mm0
     PMAXSW mm3,mm2           //mm3-maximum z okolia (lumamax)
       
     movd mm0,[esi]  
     punpcklbw mm0,mm7        //mm0-0,lumac[0],0,lumac[1],0,lumac[2],0,lumca[3]
       
     movq mm7,[ones]//pxor mm7,mm7      //mm7-0
     movq mm2,mm0      //  mm1-lumamin
     psubw mm2,mm1     //mm2-mindiff=lumac-lumamin
     movq mm4,mm3      //  mm3-lumamax
     psubw mm4,mm0     //mm4-maxdiff=lumamax-lumac
     movq mm5,mm2
     pcmpgtw mm5,mm4   //mm5-mindiff>maxdiff
     movq mm6,[mtf_thresh64] 
     pcmpgtw mm6,mm4   //mm6-mfd_threshold>maxdiff
     pand mm6,mm5      //mm6-mindiff>maxdiff AND mfd_threshold>maxdiff 
     pand mm7,mm6
     pand mm7,mm3//por  mm7,mm3
     movq mm4,[mtf_thresh64]
     pcmpgtw mm4,mm2   //mm4-mfd_threshol>mindiff
     pxor mm5,[ones]
     pand mm4,mm5      //mm4-mindiff<maxdiff AND mfd_threshold>maxdiff    
     pand mm1,mm4
     por  mm7,mm1      //mm7-p, ale len to, co treba zmenit
     por  mm4,mm6      //mm4- =y (mindiff>maxdiff AND mfd_threshold>maxdiff) OR (mindiff<maxdiff AND mfd_threshold>maxdiff)
     pmullw mm7,[mtf_strength64]
     movq mm6,mm0      //mm6-Y
     pmullw mm6,[mtf_strengthInv64]
     paddw mm7,mm6
     psrlw mm7,7
     pand mm7,mm4
     pxor mm4,[ones]
     pand mm0,mm4
     por  mm0,mm7
     pxor mm1,mm1
     packuswb mm0,mm1
     movd [edi],mm0
     add edi,4
     add ecx,4
     add edx,4
     add esi,4
     cmp esi,[srcPend]
     jl lineLoop2
    }
  }  
 const unsigned char *srcL=src+stride,*srcR=srcL+dx-1;unsigned char *dstL=dst+stride,*dstR=dstL+dx-1;
 for (int y=1;y<dy-1;srcL+=stride,srcR+=stride,dstL+=stride,dstR+=stride,y++)
  {
   *dstL=*srcL;
   *dstR=*srcR;
  }
 __asm{emms}
 env->BitBlt(dstf->GetWritePtr(PLANAR_U),dstf->GetPitch(PLANAR_U),srcf->GetReadPtr(PLANAR_U),srcf->GetPitch(PLANAR_U),srcf->GetRowSize(PLANAR_U),srcf->GetHeight(PLANAR_U));
 env->BitBlt(dstf->GetWritePtr(PLANAR_V),dstf->GetPitch(PLANAR_V),srcf->GetReadPtr(PLANAR_V),srcf->GetPitch(PLANAR_V),srcf->GetRowSize(PLANAR_V),srcf->GetHeight(PLANAR_V));
 return dstf;
}
